/*
   md.c : Multiple Devices driver for Linux
	  Copyright (C) 1998, 1999, 2000 Ingo Molnar

     completely rewritten, based on the MD driver code from Marc Zyngier

   Changes:

   - RAID-1/RAID-5 extensions by Miguel de Icaza, Gadi Oxman, Ingo Molnar
   - RAID-6 extensions by H. Peter Anvin <hpa@zytor.com>
   - boot support for linear and striped mode by Harald Hoyer <HarryH@Royal.Net>
   - kerneld support by Boris Tobotras <boris@xtalk.msk.su>
   - kmod support by: Cyrus Durgin
   - RAID0 bugfixes: Mark Anthony Lisher <markal@iname.com>
   - Devfs support by Richard Gooch <rgooch@atnf.csiro.au>

   - lots of fixes and improvements to the RAID1/RAID5 and generic
     RAID code (such as request based resynchronization):

     Neil Brown <neilb@cse.unsw.edu.au>.

     
     - Added pluggable metadata support.
     Copyright (C) 2004 Adaptec Inc.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   You should have received a copy of the GNU General Public License
   (for example /usr/src/linux/COPYING); if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <linux/module.h>
#include <linux/config.h>
#include <linux/linkage.h>
#include <linux/raid/emd/md.h>
#include <linux/sysctl.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/miscdevice.h>

#include <linux/fs.h>
#include <linux/init.h>

#include <linux/major.h>
#include <linux/locks.h>
#ifdef CONFIG_KMOD
#include <linux/kmod.h>
#endif

#include <asm/unaligned.h>
#include <linux/blk.h>

#ifdef MD_DEBUG
#define MD_DBG_PRINT(A)	printk A
#else
#define MD_DBG_PRINT(A)
#endif

static int autodetect_running;
static int autodetect_allowed;
static int unit_number_assignment_allowed;

#define DEBUG 0
#define dprintk(x...) ((void)(DEBUG && printk(x)))

#define set_capacity(disk, size) disk->part->nr_sects

/*
 *  * default readahead
 *  * vm_max_readahead was substituted for MAX_READAHEAD in 2.4.16.  Red Hat
 *  * includes this change in 2.4.7-10 and TurboLinux also includes this change
 *  * in their 2.4.9-3 kernel.  So instead of making this depend on a specific
 *  * kernel version, just depend on the existence of the MAX_READAHEAD macro.
 *   */
#ifdef MAX_READAHEAD
#define MD_READAHEAD    MAX_READAHEAD
#else
#define MD_READAHEAD    vm_max_readahead
#endif

#if (MINORBITS != 8)
#error MD doesnt handle bigger kdev yet
#endif

#define get_seconds()       CURRENT_TIME
#define MAX_MDP_DEVS		(1 << (MINORBITS - EMD_MINOR_SHIFT))

static int md_open(struct inode *, struct file *);
static int md_release(struct inode *, struct file *);
static int md_ioctl(struct inode *, struct file *, u_int, unsigned long);
static int md_revalidate(kdev_t dev);
static int do_md_stop(mddev_t * mddev, int ro);

/*
 * these have to be allocated separately because external
 * subsystems want to have a pre-defined structure
 */

struct hd_struct mdp_hd_struct[MAX_MDP_DEVS<<EMD_MINOR_SHIFT];
static int mdp_blocksizes[MAX_MDP_DEVS<<EMD_MINOR_SHIFT];
static int mdp_hardsect_sizes[MAX_MDP_DEVS<<EMD_MINOR_SHIFT];
static int mdp_maxreadahead[MAX_MDP_DEVS<<EMD_MINOR_SHIFT];
int mdp_size[MAX_MDP_DEVS<<EMD_MINOR_SHIFT];

static struct block_device_operations md_fops =
{
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,9)
	.owner		= THIS_MODULE,
#endif
	.open		= md_open,
	.release	= md_release,
	.ioctl		= md_ioctl,
	.revalidate	= md_revalidate
};

static struct gendisk mdp_gendisk=
{
	major_name: "emd",
	minor_shift: EMD_MINOR_SHIFT,
	max_p: 1<<EMD_MINOR_SHIFT,
	part: mdp_hd_struct,
	sizes: mdp_size,
	nr_real: MAX_MDP_DEVS,
	real_devices: NULL,
	next: NULL,
	fops: &md_fops,
};

static inline int invalidate_partition(mddev_t *mddev)
{
	return invalidate_device(MKDEV(EMD_MAJOR, EMDMINOR(mddev)),
				 mddev->unit);
}

typedef struct dev_mapping_s {
	mddev_t *mddev;
	void *data;
} dev_mapping_t;

static dev_mapping_t mddev_map[MAX_MD_DEVS];

static void add_mddev_mapping(mddev_t *mddev, int unit, void *data)
{
	if (mddev_map[unit].mddev) {
		MD_BUG();
		return;
	}
	mddev_map[unit].mddev = mddev;
	mddev_map[unit].data = data;
}

static void del_mddev_mapping(mddev_t *mddev, int unit)
{
	if (mddev->member.name[0] == 0) 
		return;

	if (mddev_map[unit].mddev != mddev) {
		MD_BUG();
		return;
	}
	mddev_map[unit].mddev = NULL;
	mddev_map[unit].data = NULL;
}

static mddev_t * kdev_to_mddev(kdev_t dev)
{
	mddev_t *mddev;
	u_long flags;
        int unit;

	if (MAJOR(dev) == EMD_MAJOR) {
        	int minor;

		minor = MINOR(dev);
		if (minor >= (MAX_MDP_DEVS << EMD_MINOR_SHIFT))
			return NULL;
		unit = minor >> EMD_MINOR_SHIFT;
		spin_lock_irqsave(&toplevel_arrays.member_lock, flags);
		mddev = mddev_map[unit].mddev;
		if (mddev) {
			member_get(&mddev->member);
			if (atomic_read(&mddev->member.refcount) == 1)
				mddev = NULL;
		}
		spin_unlock_irqrestore(&toplevel_arrays.member_lock, flags);
		return mddev;
	}
	return NULL;
}

static int
mdp_make_request(request_queue_t *q, int rw, struct buffer_head *bh)
{
	mddev_t *mddev;
	int minor;
	int retval = 0;

	mddev = kdev_to_mddev(bh->b_rdev);
	minor = MINOR(bh->b_rdev);

	if (mddev == NULL || mddev->member.pers == NULL) {
		buffer_IO_error(bh);
		MD_DBG_PRINT(("returning EINVAL from mdp_make_request\n"));
		if (mddev != NULL)
			member_put(&mddev->member);
		return (-EINVAL);
	}

	if (MAJOR(bh->b_rdev) == EMD_MAJOR)
		/* map the sector for partitions */
		bh->b_rsector += mdp_hd_struct[minor].start_sect;
	retval = mddev->member.pers->make_request(to_member(mddev), rw, bh);
	member_put(&mddev->member);
	return (retval);
}
#define bdev_get_queue(blkdev) blk_get_queue(blkdev->bd_inode->i_dev);

static void autostart_arrays (void);
static mdk_personality_t *pers[MAX_PERSONALITY];
static spinlock_t pers_lock = SPIN_LOCK_UNLOCKED;

void
md_get_array_stats(mddev_t *array, int *optimal, int *degraded,
		   int *failed, int *build_targets)
{
	struct list_head *list_item;
	u_int optimal_cnt;
	u_int degraded_cnt;
	u_int failed_cnt;
	u_int build_target_cnt;

	optimal_cnt = 0;
	degraded_cnt = 0;
	failed_cnt = 0;
	build_target_cnt = 0;

	/*
	 * Walk our member list and record the
	 * component state that effects our array
	 * state.
	 */
	list_item = array->member.child_set.next;
	while (list_item != &array->member.child_set) {
		mdk_member_t *member;

		member = list_entry(list_item, mdk_member_t, peer_links);
		list_item = list_item->next;
		switch (member->state & ~MDU_MSTATE_DIRTY) {
		case MDU_MSTATE_OPTIMAL:
			optimal_cnt++;
			break;
		case MDU_MSTATE_OPTIMAL|MDU_MSTATE_REBUILD_TARGET:
		case MDU_MSTATE_DEGRADED|MDU_MSTATE_REBUILD_TARGET:
			build_target_cnt++;
			break;
		case MDU_MSTATE_DEGRADED:
			degraded_cnt++;
			break;
		case MDU_MSTATE_FAILED:
		default:
			failed_cnt++;
			break;
		}
	}

	if (optimal != NULL)
		*optimal = optimal_cnt;
	if (degraded != NULL)
		*degraded = degraded_cnt;
	if (failed != NULL)
		*failed = failed_cnt;
	if (build_targets != NULL)
		*build_targets = build_target_cnt;
}

/*
 * Read num_sectors 512 byte sectors from the given device, starting at
 * the given sector offset.  Copy the result into the buffer supplied by
 * the user.  Return NULL if there is a failure, otherwise return the
 * user-supplied buffer.
 *
 * Hopefully there is a better way to do this.
 */

int md_read_blocks(kdev_t dev, sector_t sector_offset, 
		int num_sectors, uint8_t *dest_buf)
{
	struct buffer_head *bh = NULL;
	int i = 0;

	fsync_dev(dev);
	set_blocksize(dev, 512);

	for (i = 0; i < num_sectors; i++) {
		bh = bread(dev, sector_offset + i, 512);
		if (bh == NULL) {
			MD_DBG_PRINT(("md_read_blocks: bread failed\n"));
			return(-EIO);
		}
		memcpy(dest_buf + (i * 512), bh->b_data, 512);
		brelse(bh);
	}

	return 0;
}

int md_write_blocks(kdev_t dev, sector_t sector_offset,
		    int num_sectors, uint8_t *src_buf)
{
	struct buffer_head *bh;
	int error, i;

	error = 0;

	fsync_dev(dev);
	set_blocksize(dev, 512);

	for (i = 0; i < num_sectors; i++) {
		bh = bread(dev, sector_offset + i, 512);
		if (bh == NULL) {
			MD_DBG_PRINT(("md: bread failed for device %d/%d\n",
			       MAJOR(dev), MINOR(dev)));
			return (-EIO);
		}
		memcpy(bh->b_data, src_buf + (i * 512), 512);
		mark_buffer_dirty(bh);
		ll_rw_block(WRITE, 1, &bh);
		wait_on_buffer(bh);
		if (buffer_uptodate(bh) == 0) {
			MD_DBG_PRINT(("md_write_buffers: write failed\n"));
			error = -EIO;
		}
		brelse(bh);
		if (error != 0)
			break;
	}
	return (error);
}

/*
 * Current RAID-1,4,5 parallel reconstruction 'guaranteed speed limit'
 * is 1000 KB/sec, so the extra system load does not show up that much.
 * Increase it if you want to have more _guaranteed_ speed. Note that
 * the RAID driver will use the maximum available bandwith if the IO
 * subsystem is idle. There is also an 'absolute maximum' reconstruction
 * speed limit - in case reconstruction slows down your system despite
 * idle IO detection.
 *
 * you can change it via /proc/sys/dev/raid/speed_limit_min and _max.
 */

static int sysctl_speed_limit_min = 1000;
static int sysctl_speed_limit_max = 200000;

/*
 * Add a child to a member.  Returns the number of entries with
 * the same priority that were found during the priority insertion
 * sort.
 */
int md_member_add_child(mdk_member_t *member, mdk_member_t *child)
{
	struct list_head *list_item;
	u_long flags;
	int same_cnt;

	same_cnt = 0;
	child->parent = member;
	member_get(member);
	spin_lock_irqsave(&member->member_lock, flags);
	list_item = member->child_set.next;
	while (list_item != &member->child_set) {
		mdk_member_t *list_member;

		list_member = list_entry(list_item, mdk_member_t, peer_links);
		if (list_member->priority > child->priority)
			break;
		if (list_member->priority == child->priority)
			same_cnt++;
		list_item = list_item->next;
	}
	if (list_item != &member->child_set)
		list_add_tail(&child->peer_links, list_item);
	else
		list_add_tail(&child->peer_links, &member->child_set);
	member->child_set_gencount++;
	spin_unlock_irqrestore(&member->member_lock, flags);

	return (same_cnt);
}

void md_member_orphan(mdk_member_t *member)
{
	u_long flags;

	if (member->parent != NULL) {
		spin_lock_irqsave(&member->parent->member_lock, flags);
		list_del(&member->peer_links);
		member->parent->child_set_gencount++;
		spin_unlock_irqrestore(&member->parent->member_lock, flags);
		member_put(member->parent);
	}
}

static void member_init(mdk_member_t *member, mdk_member_t *parent,
			mdk_mtype_t type, mdk_msubtype_t subtype,
			int priority, void (*destructor)(mdk_member_t *))
{
	spin_lock_init(&member->member_lock);
	member->parent = parent;
	member->type = type;
	member->subtype = subtype;
	INIT_LIST_HEAD(&member->child_set);
	INIT_LIST_HEAD(&member->peer_links);
	member->priority = priority;
	member->destructor = destructor;
	atomic_set(&member->refcount, 1);

	member->state = MDU_MSTATE_UNKNOWN;

	/*
	 * Sort this new member into the
	 * child list of the parent.
	 */
	if (parent != NULL)
		md_member_add_child(parent, member);
}

static void member_destructor(mdk_member_t *member)
{
	mdk_member_t *parent;

	parent = member->parent;
	if (parent != NULL
	 && parent->pers != NULL)
		parent->pers->raid_op(parent, MDK_RAID_OP_RELEASE_MEMBER,
				      member);

	if (member->pers != NULL)
		member->pers->raid_op(member, MDK_RAID_OP_RELEASE, NULL);

	if (member->metapers != NULL)
		member->metapers->member_op(member, MDK_MEMBER_OP_RELEASE,
					    NULL);

	md_member_orphan(member);
}

enum md_bcast_e
{
	MD_BCAST_PERS,
	MD_BCAST_METAPERS
};
typedef enum md_bcast_e md_bcast_t;

/*
 * Perform a depth first traversal of the member
 * tree rooted in "member" calling the member personality
 * specified by "who" with the opcode "op" and the argument
 * "arg".
 */
static void md_member_bcast(mdk_member_t *member, int op, void *arg,
			    md_bcast_t who)
{
	struct list_head *tmp;
	mdk_member_t *child;
	u_long flags;

	ITERATE_CHILDREN(member, child, tmp, flags) {
		md_member_bcast(child, op, arg, who);
	}
	switch (who) {
	case MD_BCAST_PERS:
		if (member->pers != NULL)
			member->pers->raid_op(member, op, arg);
		break;
	case MD_BCAST_METAPERS:
		if (member->metapers != NULL)
			member->metapers->member_op(member, op, arg);
		break;
	default:
		break;
	}
}

static struct ctl_table_header *raid_table_header;
static LIST_HEAD(md_metapersonality_list);
static spinlock_t metapersonality_list_lock = SPIN_LOCK_UNLOCKED;
static int metapers_change_count;

/*
 * Iterates through all registered metadata personalities.
 * We take care to grab the personality list lock whenever
 * navigating the list, and to always hold a refcount when
 * unlocked.  Any code which breaks out of this loop will
 * own a reference to the personality pointed to by metapers
 * that must evenutally be released.
 */
#define ITERATE_METAPERS(metapers,tmp)					    \
									    \
for (({ spin_lock(&metapersonality_list_lock);				    \
	tmp = md_metapersonality_list.next;				    \
	metapers = NULL;});						    \
     ({ if (tmp != &md_metapersonality_list)				    \
		metapers_get(list_entry(tmp, mdk_metapersonality_t, links));\
	spin_unlock(&metapersonality_list_lock);			    \
	if (metapers) metapers_put(metapers);				    \
	metapers = list_entry(tmp, mdk_metapersonality_t, links);	    \
	tmp != &md_metapersonality_list;});				    \
     ({ spin_lock(&metapersonality_list_lock);				    \
	tmp = tmp->next;})						    \
	)

static inline mdk_metapersonality_t *
metapers_get(mdk_metapersonality_t *metapers)
{
	atomic_inc(&metapers->refcount);
	return metapers;
}

static void metapers_put(mdk_metapersonality_t *metapers)
{
	if (!atomic_dec_and_lock(&metapers->refcount,
				 &metapersonality_list_lock)) 
		return;
	list_del(&metapers->links);
	/* XXX Handle metapers release... Decrement module count? */
	spin_unlock(&metapersonality_list_lock);
}

static ctl_table raid_table[] = {
	{
		.ctl_name	= DEV_RAID_SPEED_LIMIT_MIN,
		.procname	= "speed_limit_min",
		.data		= &sysctl_speed_limit_min,
		.maxlen		= sizeof(int),
		.mode		= 0644,
		.proc_handler	= &proc_dointvec,
	},
	{
		.ctl_name	= DEV_RAID_SPEED_LIMIT_MAX,
		.procname	= "speed_limit_max",
		.data		= &sysctl_speed_limit_max,
		.maxlen		= sizeof(int),
		.mode		= 0644,
		.proc_handler	= &proc_dointvec,
	},
	{ .ctl_name = 0 }
};

static ctl_table raid_dir_table[] = {
	{
		.ctl_name	= DEV_RAID,
		.procname	= "raid",
		.maxlen		= 0,
		.mode		= 0555,
		.child		= raid_table,
	},
	{ .ctl_name = 0 }
};

static ctl_table raid_root_table[] = {
	{
		.ctl_name	= CTL_DEV,
		.procname	= "dev",
		.maxlen		= 0,
		.mode		= 0555,
		.child		= raid_dir_table,
	},
	{ .ctl_name = 0 }
};

static int md_status_read_proc(char *page, char **start, off_t off,
			int count, int *eof, void *data)
{
	mdk_rdev_t *rdev;
	mddev_t *mddev;
	mdk_member_t *array, *disk;
	struct list_head *tmp1, *tmp2;
	mdk_member_t *member;
	mdk_metapersonality_t *metapers;
	char *noun;
	mdk_checkpoint_arg_t checkpoint_arg;
	mdk_syncstatsop_arg_t syncstats;
	u_long flags1, flags2;
	u_long resync, max_blocks, res, rt;
	int error;
	int sz, j;
	int pool_spare_available;

	sz = 0;
	sz += sprintf(page + sz, "Personalities : ");
	for (j = 0; j < MAX_PERSONALITY; j++)
		if (pers[j])
			sz += sprintf(page + sz, "[%s] ", pers[j]->name);

	sz += sprintf(page + sz, "\n");

	sz += sprintf(page + sz, "read_ahead ");
	if (read_ahead[EMD_MAJOR] == INT_MAX)
		sz += sprintf(page + sz, "not set\n");
	else
		sz += sprintf(page + sz, "%d sectors\n", read_ahead[EMD_MAJOR]);
	/*
	 * Populate Pool spare details
	 */	
	spin_lock_irqsave(&spare_arrays.member_lock, flags1);
	if (list_empty(&spare_arrays.child_set)) 
		pool_spare_available = 0;
	else	
		pool_spare_available = 1;
	spin_unlock_irqrestore(&spare_arrays.member_lock, flags1);
	if (pool_spare_available)
	{
		sz += sprintf(page + sz , "Pool spares  ");
		ITERATE_CHILDREN(&spare_arrays, array, tmp1, flags1)  
			ITERATE_CHILDREN(array, disk, tmp2, flags2) { 
				sz += sprintf(page + sz, " (%d,%d)[%s]", 
			        	      MAJOR(to_rdev(disk)->kdev),
					      MINOR(to_rdev(disk)->kdev),
				     	      disk->metapers->name);
			}
		sz += sprintf(page + sz , "\n");
	}
	/*
	 * Polpulate array details
	 */
	ITERATE_CHILDREN(&toplevel_arrays, array, tmp1, flags1) {
		int priority;
 
		mddev = to_mddev(array);
		sz += sprintf(page + sz, "%s : %sactive", 
			      mdname(array),
			      array->pers ? "" : "in");
		if (array->pers) {
			if (mddev->ro)
				sz += sprintf(page + sz, " (read-only)");
			sz += sprintf(page + sz, " %s", array->pers->name);
		}
		sz += sprintf(page + sz, " state %#x", array->state);

		ITERATE_CHILDREN(array, disk, tmp2, flags2) {
			rdev = to_rdev(disk);
			sz += sprintf(page + sz, " (%d,%d)", 
			              MAJOR(to_rdev(disk)->kdev),
				      MINOR(to_rdev(disk)->kdev));
			if (disk->subtype == MDK_MSUBTYPE_DEDICATED_SPARE)
				sz += sprintf(page + sz, "[SPARE]");
		}

		sz += sprintf(page + sz, "\n      %lu blocks ", 
			     (to_member(mddev))->data_size >> 1);

		sz += sprintf(page + sz, " [");
		priority = 0;
		ITERATE_CHILDREN(array, disk, tmp2, flags2) {
			while (priority < disk->priority) {
				sz += sprintf(page + sz, "_");
				priority++;
			}
			if(disk->state == MDU_MSTATE_OPTIMAL)
				sz += sprintf(page + sz, "U");
			else 
				sz += sprintf(page + sz, "_");
			priority = disk->priority + 1;
		}
		while (priority < to_mddev(array)->max_disks) {
			sz += sprintf(page + sz, "_");
			priority++;
		}
		sz += sprintf(page + sz, "]\n");

		if (!array->pers) {
			sz += sprintf(page+sz, "\n");
			continue;
		}
		/*
		 * Populate rebuild or init progress details
		 */
		member = to_member(mddev);
		metapers = member->metapers;

		error = metapers->member_op(member,
					    MDK_MEMBER_OP_GET_CHECKPOINT,
					    &checkpoint_arg);
		if (error != 0) { 
			MD_DBG_PRINT(("Error in get checkpoint\n"));
			sz += sprintf(page+sz, "\n");
			continue;
		}
		error = member->pers->raid_op(member, MDK_RAID_OP_SYNC_STATS, 
					      &syncstats);
		if (error != 0) { 
			sz += sprintf(page+sz, "\n");
			continue;
		}

		switch (syncstats.syncop) {
		case MD_SYNCOP_REBUILD:
			noun = "rebuild";
			break;
		case MD_SYNCOP_INIT:
			noun = "init";
			break;
		case MD_SYNCOP_VERIFY:
			noun = "verify";
			break;
		case MD_SYNCOP_VERIFY_FIX:
			noun = "verify w/ fix";
			break;
		default:
			continue;
		}

		resync = checkpoint_arg.checkpoint>>1;
		max_blocks = (to_member(mddev))->data_size>>1;

		/*
	 	* Should not happen.
	 	*/
		if (!max_blocks) {
			MD_BUG();
			continue;
		}

		res = (resync/1024)*1000/(max_blocks/1024 + 1);
		{
			int i, x = res/50, y = 20-x;
			sz += sprintf(page + sz, "      [");
			for (i = 0; i < x; i++)
				sz += sprintf(page + sz, "=");
			sz += sprintf(page + sz, ">");
			for (i = 0; i < y; i++)
				sz += sprintf(page + sz, ".");
			sz += sprintf(page + sz, "] ");
		}

		sz += sprintf(page + sz , "%s", noun);
			
		sz += sprintf(page + sz, " = %3lu.%lu%% (%lu/%lu)",
		      res/10, res % 10, resync, max_blocks);

		/*
	 	* We do not want to overflow, so the order of operands and
	 	* the * 100 / 100 trick are important. We do a +1 to be
	 	* safe against division by zero. We only estimate anyway.
	 	*
	 	* rt: remaining time
	 	*/
		if (syncstats.stat_sectors == 0 || syncstats.stat_time == 0) {
			sz += sprintf(page + sz, "\n");
			continue;
		}
	
		/* sectors to 1K blocks */
		syncstats.stat_sectors >>= 1;
		rt = (syncstats.stat_time
		    * ((max_blocks-resync) / (syncstats.stat_sectors)))/HZ;

		sz += sprintf(page + sz, " finish=%lu.%lumin",
			      rt / 60, (rt % 60)/6);

		sz += sprintf(page + sz, " speed=%ldK/sec\n",
			      (syncstats.stat_sectors / 
			       syncstats.stat_time) * HZ); 
	}
	sz += sprintf(page + sz, "\n");
	return sz;
}

/*
 * All top-level arrays are children of the
 * "toplevel_arrays".  This simplifies code since
 * all "members" have a parent and we can use
 * the same "member" rules and semantics to
 * manage not only the top level array list,
 * but array children too.
 */
mdk_member_t toplevel_arrays = {
	.type		= MDK_MTYPE_MASTER,
	.subtype	= MDK_MSUBTYPE_NONE,
	.member_lock	= SPIN_LOCK_UNLOCKED,
	.child_set	= LIST_HEAD_INIT(toplevel_arrays.child_set),
	.refcount	= ATOMIC_INIT(1)
};

/* This is the top-level spare pool where all spare drives are
 * added. This is the "toplevel_arrays" equivalent for spares.
 * This holds both "global" as well as "dedicated" spares.
 * The array that intends to use the spare drive must check
 * the metadata type the spare drive has been associated
 * with before using it.
 */
mdk_member_t spare_arrays = {
	.type		= MDK_MTYPE_MASTER,
	.subtype	= MDK_MSUBTYPE_NONE,
	.member_lock	= SPIN_LOCK_UNLOCKED,
	.child_set	= LIST_HEAD_INIT(spare_arrays.child_set),
	.refcount	= ATOMIC_INIT(1)
};

static void mddev_destructor(mdk_member_t *member)
{
	mddev_t *mddev;

	mddev = to_mddev(member);
	del_mddev_mapping(mddev, mddev->unit);
	member_destructor(member);
	kfree(mddev);
}

/* 
 * md_get_unit returns the lowest free index of mddev_map array, if
 * the requested_unit is -1. For requested_unit is >= 0, it returns
 * the same requested_unit, if the index is free else returns any 
 * other free index. 
 */
static int md_get_unit(int requested_unit)
{
	int unit;
	
	if ((requested_unit >= MAX_MD_DEVS) || (requested_unit < -1))
		return -1;
	if ((requested_unit != -1) && 
	    (mddev_map[requested_unit].mddev == NULL))
		return requested_unit;
	
	for (unit = 0; unit < MAX_MD_DEVS; unit++) {
		if (mddev_map[unit].mddev == NULL)
			return unit;
	}
	return -1;
}

mddev_t *md_mddev_alloc(mdk_member_t *parent, mdk_msubtype_t mddev_subtype,
			int priority)
{
	mddev_t *mddev;

	mddev = (mddev_t *) kmalloc(sizeof(*mddev), GFP_KERNEL);
	if (!mddev)
		return NULL;

	memset(mddev, 0, sizeof(*mddev));

	init_MUTEX(&mddev->reconfig_sem);
	
	member_init(&mddev->member, parent, MDK_MTYPE_MDDEV,
		    mddev_subtype, priority, mddev_destructor);

	return mddev;
}

static void md_force_autodetect_dev(void);

void md_teardown_member(mdk_member_t *member)
{
	mdk_set_state_arg_t sstate;

	sstate.oldstate = MDU_MSTATE_TEARDOWN;
	sstate.newstate = MDU_MSTATE_TEARDOWN;

	md_member_bcast(member, MDK_MEMBER_OP_SET_STATE,
			&sstate, MD_BCAST_METAPERS);
}

const char *md_member_type_string(mdk_member_t *member)
{
	switch (member->type) {
	case MDK_MTYPE_NONE:
		return "NONE";
	case MDK_MTYPE_RDEV:
		return "RDEV";
	case MDK_MTYPE_MDDEV:
		return "MDDEV";
	case MDK_MTYPE_MASTER:
		return "MASTER";
	default:
		return "UNKNOWN";
	}
}

static void print_rdev(mdk_rdev_t *rdev)
{
	mdk_member_t *member;

	member = to_member(rdev);
	MD_DBG_PRINT(("<%s>", mdname(member)));
	MD_DBG_PRINT((KERN_INFO "md: rdev %s, SZ:%08llu S:%x DN:%u\n",
		mdname(member), (unsigned long long)member->data_size,
	       	member->state, member->priority));
}

static void md_print_member(mdk_member_t *member)
{
	struct list_head *tmp;
	mdk_member_t *child;
	u_long flags;

	if (member->type == MDK_MTYPE_RDEV) {
		/* Leaf node */
		print_rdev(to_rdev(member));
		return;
	}

	/* Print out the array and all children of the array. */
	MD_DBG_PRINT(("%s: ", mdname(member)));
	ITERATE_CHILDREN(member, child, tmp, flags) {
		md_print_member(child);
	}
	MD_DBG_PRINT(("\n"));
	
}

void md_print_devices(void)
{
	struct list_head *tmp;
	mdk_member_t *child;
	u_long flags;

	MD_DBG_PRINT(("\n"));
	MD_DBG_PRINT(("md:	**********************************\n"));
	MD_DBG_PRINT(("md:	* <COMPLETE RAID STATE PRINTOUT> *\n"));
	MD_DBG_PRINT(("md:	**********************************\n"));
	
	ITERATE_CHILDREN(&toplevel_arrays, child, tmp, flags) {
		md_print_member(child);
	}
	MD_DBG_PRINT(("md:	**********************************\n"));
	MD_DBG_PRINT(("\n"));
}

static void rdev_destructor(mdk_member_t *member)
{
	mdk_rdev_t *rdev;

	rdev = to_rdev(member);
	member_destructor(member);
	kfree(rdev);
}

mdk_rdev_t *md_rdev_alloc(mdk_member_t *parent, mdk_msubtype_t subtype,
			  kdev_t newdev, int priority)
{
	mdk_rdev_t *rdev;

	rdev = (mdk_rdev_t *) kmalloc(sizeof(*rdev), GFP_KERNEL);
	if (!rdev) {
		MD_DBG_PRINT((KERN_ERR "md: could not alloc mem for %s!\n",
		       kdevname(newdev)));
		return (NULL);
	}
	memset(rdev, 0, sizeof(*rdev));

	rdev->member.data_offset = 0;
	rdev->kdev = newdev;

	sprintf(rdev->member.name, "%s", kdevname(rdev->kdev));
	member_init(&rdev->member, parent, MDK_MTYPE_RDEV,
		    subtype, priority, rdev_destructor);

	return rdev;
}

void md_wakeup_thread(mdk_thread_t *thread);

/*
 * Write idle timer.  Queue a sync handler workitem to
 * to do the possibly blocking work of updating the
 * meta-data to place us in safemode.
 */
static void md_safemode_timeout(unsigned long data)
{
	mdk_sync_support_t *syncinfo;

	syncinfo = (mdk_sync_support_t *)data;
	spin_lock_irq(&syncinfo->sync_lock);
	if (syncinfo->safemode_workitem.work_flags == MD_WF_NONE) {
		md_queue_workitem(&syncinfo->work_state,
				  &syncinfo->safemode_workitem,
				  /*head_of_queue*/0);
	}
	spin_unlock_irq(&syncinfo->sync_lock);
}

int md_check_chunksize(mddev_t * mddev)
{
	struct list_head *tmp;
	mdk_member_t *child;
	mdk_member_t *member;
	mdk_rdev_t *rdev;
	u_long flags;
	int pnum, chunk_size, chunk_sects;

	chunk_size = mddev->chunk_size;
	chunk_sects = chunk_size >> 9;
	pnum = level_to_pers(mddev->level);

	if ((pnum != MULTIPATH) && (pnum != RAID1)) {
		if (!chunk_size) {
			/*
			 * 'default chunksize' in the old md code used to
			 * be PAGE_SIZE, baaad.
			 * we abort here to be on the safe side. We don't
			 * want to continue the bad practice.
			 */
			MD_DBG_PRINT((KERN_ERR 
				"no chunksize specified, see 'man raidtab'\n"));
			return -EINVAL;
		}
		if (chunk_size > MAX_CHUNK_SIZE) {
			MD_DBG_PRINT((KERN_ERR "too big chunk_size: %d > %d\n",
				chunk_size, MAX_CHUNK_SIZE));
			return -EINVAL;
		}
		/*
		 * chunk-size has to be a power of 2 and multiples of PAGE_SIZE
		 */
		if ( (1 << ffz(~chunk_size)) != chunk_size) {
			MD_BUG();
			return -EINVAL;
		}
		if (chunk_size < PAGE_SIZE) {
			MD_DBG_PRINT((KERN_ERR "too small chunk_size: %d < %ld\n",
				chunk_size, PAGE_SIZE));
			return -EINVAL;
		}

		member = (mdk_member_t *)mddev;
		/* devices must have minimum size of one chunk */
		ITERATE_CHILDREN(member, child, tmp, flags) {
			if (child->type == MDK_MTYPE_RDEV) {
				rdev = (mdk_rdev_t *) child;
				if (rdev->member.state != MDU_MSTATE_OPTIMAL)
					continue;
				if (rdev->member.data_size < chunk_sects) {
					MD_DBG_PRINT((KERN_WARNING
					"md: Dev %s smaller than chunk_size:"
					" %lluk < %dk\n",
					mdname(to_member(rdev)),
					(unsigned long long)rdev->member.data_size,
					chunk_sects));
					return -EINVAL;
				}
			}
		}
	}
	return 0;
}

int md_assign_unit_number(mddev_t *mddev)
{
	mdk_personality_t *pers;
	mdk_member_t *member;
	static DECLARE_MUTEX(unit_sem);
	u_long flags;
	kdev_t kdev;
	int unit;
	
	if (!mddev)
		MD_BUG();
	
	if (unit_number_assignment_allowed == 0)
		return 0;

	pers = mddev->member.pers;
	member = to_member(mddev);

	down(&unit_sem);
	spin_lock_irqsave(&member->member_lock, flags);
	if ((mddev->flags & MD_MDDEV_UNIT_NUMBER_ASSIGNED) != 0
	 || mddev->member.subtype != MDK_MSUBTYPE_TOPLEVEL) {
		spin_unlock_irqrestore(&member->member_lock, flags);
		up(&unit_sem);
		return 0;
	}
	spin_unlock_irqrestore(&member->member_lock, flags);

	if ((unit = md_get_unit(mddev->preferred_unit)) == -1){
		MD_DBG_PRINT(("md_assign_unit_number: no unit available, " 
		       "requested %d\n", mddev->preferred_unit));
		up(&unit_sem);
		return -1;
	}
	if ((mddev->preferred_unit != -1) && (unit != mddev->preferred_unit))
		MD_DBG_PRINT(("md_assign_unit_number: preferred unit %d unavailable. "
		       "Assigned %d instead\n", mddev->preferred_unit, unit));
	mddev->unit = unit;
	MD_DBG_PRINT(("Assigning unit %d to array with priority %#x\n", mddev->unit,
	       mddev->member.priority));
	add_mddev_mapping(mddev, unit, 0);

	kdev = MKDEV(EMD_MAJOR, mddev->unit);

	sprintf(member->name, "emd%c", 'a' + mddev->unit);
	mdp_size[EMDMINOR(mddev)] = mddev->member.data_size >> 1;
	spin_lock_irqsave(&member->member_lock, flags);
	mddev->flags |= MD_MDDEV_UNIT_NUMBER_ASSIGNED;
	spin_unlock_irqrestore(&member->member_lock, flags);
	up(&unit_sem);

	return 0;
}

int md_export_array(mddev_t *mddev)
{
	static DECLARE_MUTEX(disks_sem);
	mdk_member_t *member;
	mdk_personality_t *pers;
	u_long flags;
	
	member = to_member(mddev);
	mddev_lock(mddev);
	spin_lock_irqsave(&member->member_lock, flags);
	if ((mddev->flags & MD_MDDEV_ALLOWED_TO_RUN) == 0
	 || (mddev->flags & MD_MDDEV_UNIT_NUMBER_ASSIGNED) == 0
	 || (mddev->flags & MD_MDDEV_EXPORTED_TO_OS) != 0) {
		spin_unlock_irqrestore(&member->member_lock, flags);
		mddev_unlock(mddev);
		return -EINVAL;
	}
	
	mddev->flags |= MD_MDDEV_EXPORTED_TO_OS;
	spin_unlock_irqrestore(&member->member_lock, flags);
	mddev_unlock(mddev);
	
	pers = mddev->member.pers;

	down(&disks_sem);
	register_disk(&mdp_gendisk, MKDEV(EMD_MAJOR, 
		      mddev->unit << EMD_MINOR_SHIFT), 
		      1 << EMD_MINOR_SHIFT, &md_fops,
		      mddev->member.data_size);
	up(&disks_sem);
	return 0;
}

static int restart_array(mddev_t *mddev)
{
	int err;

	/*
	 * Complain if it has no devices
	 */
	err = -ENXIO;
	if (list_empty(&mddev->member.child_set))
		goto out;

	if (mddev->member.pers) {
		err = -EBUSY;
		if (!mddev->ro)
			goto out;

		mddev->ro = 0;
		set_device_ro(MKDEV(EMD_MAJOR, EMDMINOR(mddev)), 0);

		MD_DBG_PRINT((KERN_INFO "%s: switched to read-write mode.\n",
		       mdname(to_member(mddev))));
#if 0
		/*
		 * Kick recovery or resync if necessary
		 */
		set_bit(MD_RECOVERY_NEEDED, &mddev->recovery);
		md_wakeup_thread(mddev->thread);
#endif
		err = 0;
	} else {
		MD_DBG_PRINT((KERN_ERR "%s: has no personality assigned.\n",
		       mdname(to_member(mddev))));
		err = -EINVAL;
	}

out:
	return err;
}

static int do_md_ro(mddev_t *mddev, int ro)
{
	int i, first_minor = EMDMINOR(mddev);
	for (i = first_minor ; i < (first_minor + (1 << EMD_MINOR_SHIFT)); i++)
		set_device_ro(MKDEV(EMD_MAJOR, i), ro);
	mddev->ro = ro;
	return 0;
}


static int get_version(void * arg)
{
	mdu_version_t ver;

	ver.major = MD_MAJOR_VERSION;
	ver.minor = MD_MINOR_VERSION;
	ver.patchlevel = MD_PATCHLEVEL_VERSION;

	if (copy_to_user(arg, &ver, sizeof(ver)))
		return -EFAULT;

	return 0;
}

static int get_array_info(mddev_t * mddev, void * arg)
{
	mdu_array_info_t info;
	int nr,working,spare;
	mdk_member_t *member;
	u_int active;
	u_int degraded_cnt;
	u_int failed;
	u_int build_targets;

	nr=working=active=failed=spare=0;

	member = (mdk_member_t *)mddev;
	md_get_array_stats(mddev, &active, &degraded_cnt,
			   &failed, &build_targets);

	nr = active + degraded_cnt + failed + build_targets;
	working = active + build_targets;

	info.major_version = mddev->major_version;
	info.minor_version = mddev->minor_version;
	info.patch_version = 1;
	info.ctime         = mddev->ctime;
	info.level         = mddev->level;
	info.size          = mddev->member.data_size;
	info.nr_disks      = nr;
	info.raid_disks    = mddev->max_disks;
	info.md_minor      = EMDMINOR(mddev);
	info.not_persistent= !mddev->persistent;

	info.utime         = mddev->utime;
	info.state         = mddev->member.state;
	info.active_disks  = active;
	info.working_disks = working;
	info.failed_disks  = failed;
	info.spare_disks   = spare;

	info.layout        = mddev->layout;
	info.chunk_size    = mddev->chunk_size;
	strcpy( info.metaname, mddev->member.metapers->name );
	info.parent_cookie = (unsigned long) mddev->member.parent;
	info.cookie = (unsigned long) mddev;

	if (copy_to_user(arg, &info, sizeof(info)))
		return -EFAULT;

	return 0;
}

static int get_disk_info(mdk_member_t *member, void * arg)
{
	mdu_disk_info_t info;
	mdk_rdev_t *rdev;

	if ( member->type != MDK_MTYPE_RDEV ) {
		MD_DBG_PRINT((KERN_WARNING "ERROR: get_disk_info() Struct is not an RDEV\n"));
		return -EFAULT;
	}

	if (copy_from_user(&info, arg, sizeof(info)))
		return -EFAULT;

	rdev = to_rdev(member);

	if (rdev) {
		info.major = MAJOR(rdev->kdev);
		info.minor = MINOR(rdev->kdev);
		info.raid_disk = rdev->member.priority;
		info.number = rdev->member.priority;
		info.state = rdev->member.state;
	} else {
		info.major = info.minor = 0;
		info.raid_disk = -1;
		info.state = MDU_MSTATE_FAILED;
	}
	
	info.device_type = rdev->member.type;
	info.parent_cookie = (unsigned long) rdev->member.parent;
	info.cookie = (unsigned long) rdev;

	if (copy_to_user(arg, &info, sizeof(info)))
		return -EFAULT;

	return 0;
}

static int hot_generate_error(mddev_t * mddev, dev_t dev)
{
	return 0;
}

static int hot_remove_disk(mddev_t * mddev, dev_t dev)
{
	return -EBUSY;
}

static int hot_add_disk(mddev_t * mddev, dev_t dev)
{
	return -ENODEV;
}

static mdk_member_t *get_member_from_cookie(mdk_member_t *member, u_long cookie)
{
	mdk_member_t *child;
	struct list_head *tmp;
	u_long flags;

	member_get(member);
	if (member == (mdk_member_t *)cookie)
		return (member);

	ITERATE_CHILDREN(member, child, tmp, flags) {
		mdk_member_t *grandchild;

		grandchild = get_member_from_cookie(child, cookie);
		if (grandchild != NULL) {
			member_put(child);
			member_put(member);
			return (grandchild); 
		}
	}
	member_put(member);
	return (NULL);
}

static mdk_member_t *get_child_from_index(u_long parent_cookie, int index )
{
	struct list_head *tmp;
	mdk_member_t *parent;
	mdk_member_t *child;
	u_long flags;
	
	parent = get_member_from_cookie(&toplevel_arrays, parent_cookie);
	
	if (parent == NULL)
		parent = get_member_from_cookie(&spare_arrays, parent_cookie);

	if (parent == NULL)
		return (NULL);
		
	ITERATE_CHILDREN(parent, child, tmp, flags) {
		if (child->priority == index) {
			member_put(parent);
			return (child);
		}
	}
	
	member_put(parent);
	return (NULL);
}

static mdk_member_t *get_member_from_device(mdk_member_t *member, kdev_t device)
{
	mdk_member_t *child;
	struct list_head *tmp;
	u_long flags;

	member_get(member);
	if (member->type == MDK_MTYPE_RDEV) {
		mdk_rdev_t *rdev;

		rdev = to_rdev(member);
		if (rdev->kdev == device)
			return (member); 
	} else if (member->type == MDK_MTYPE_MDDEV) {
		mddev_t *mddev;
		dev_t  tempdev;

		mddev = to_mddev(member);
		tempdev = MKDEV(EMD_MAJOR, EMDMINOR(mddev));
		if (tempdev == device)
			return (member);
	}

	ITERATE_CHILDREN(member, child, tmp, flags) {
		mdk_member_t *grandchild;
	
		grandchild = get_member_from_device(child, device);
		if (grandchild != NULL) {
			member_put(child);
			member_put(member);
			return (grandchild); 
		}
	}

	member_put(member);
	return (NULL);
}

static int get_cookie_info(mdk_member_t *member, void *arg)
{
	mdu_cookie_info_t *cookie_info;
	mdk_rdev_t *rdev;

	cookie_info = (mdu_cookie_info_t *) arg;
	
	switch (member->type)
	{
	case MDK_MTYPE_RDEV:
		rdev = to_rdev(member);
		cookie_info->major_minor.major = MAJOR(rdev->kdev);
		cookie_info->major_minor.minor = MINOR(rdev->kdev);
			
		break;
	
	case MDK_MTYPE_MDDEV: 
		cookie_info->major_minor.major = EMD_MAJOR;
		cookie_info->major_minor.minor = EMDMINOR(to_mddev(member));
		break;
		
	default:
		cookie_info->major_minor.major = 0;
		cookie_info->major_minor.minor = 0;
		break;
	}	
	
	if (member->subtype != MDK_MSUBTYPE_TOPLEVEL)
	   	cookie_info->parent_cookie = (unsigned long)member->parent;
	else
		cookie_info->parent_cookie = 0;

	cookie_info->cookie = (unsigned long) member;                  
	cookie_info->type = member->type;
	cookie_info->subtype = member->subtype; 
	cookie_info->gencount = member->child_set_gencount;
	
	return 0;
}

static int md_reread_partition(mddev_t *mddev)
{
	/* copied from ide_revalidate_disk */
	int minor = EMDMINOR(mddev);
	int part = 0;

	if (mddev->unit >= MAX_MDP_DEVS)
		return 0;
	foreach_part(part, mddev) {
		if (mdp_hd_struct[part].nr_sects > 0) {
			kdev_t dev = MKDEV(EMD_MAJOR, part);

			invalidate_device(dev, 1);

			set_blocksize(dev, mdp_hardsect_sizes[part]);
		}
		mdp_hd_struct[part].start_sect= 0;
		mdp_hd_struct[part].nr_sects = 0;
	}
	/* and from revalidate_scsidisk.. */
	grok_partitions(&mdp_gendisk, minor >> EMD_MINOR_SHIFT,
			1<<EMD_MINOR_SHIFT, mdp_size[minor]<<1);
	return 0;
}

static int md_ioctl(struct inode *inode, struct file *file,
			unsigned int cmd, unsigned long arg)
{
	mddev_t *mddev;
	mdk_member_t *member;
	struct hd_struct *hdp;
	kdev_t dev;
	int err;

	if (!capable(CAP_SYS_ADMIN))
		return -EACCES;

	dev = inode->i_rdev;
	mddev = kdev_to_mddev(dev);

	if (mddev == NULL)
		return -ENODEV;
	
	member = to_member(mddev);
	hdp = &mdp_hd_struct[MINOR(dev)];

	/*
	 * Partitioning and other blkdev ioctls.
	 */
	switch (cmd) {
	case HDIO_GETGEO:
	{
		struct hd_geometry *loc;
		u_int heads, sectors;

		loc = (struct hd_geometry *)arg;
		if (!loc) {
			err = -EINVAL;
			break;
		}
		heads = 255;
		sectors = 63;
		err = put_user (heads, (char *) &loc->heads);
		if (err)
			break;
		err = put_user (sectors, (char *) &loc->sectors);
		if (err)
			break;
		err = put_user(hdp->nr_sects / (heads * sectors), 
			(short *) &loc->cylinders);
		if (err)
			break;
		err = put_user(hdp->start_sect, (long *)&loc->start);
		break;
	}
	case BLKRRPART: /* Re-read partition tables */
		err = md_reread_partition(mddev);
		break;
	case BLKGETSIZE:
		err = put_user(hdp->nr_sects, (unsigned long *)arg);
		break;
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,9)
	case BLKGETSIZE64:
#endif
#if defined(CONFIG_IA64)
	case BLKGETLASTSECT:
	case BLKSETLASTSECT:
#endif
	case BLKROSET:
	case BLKROGET:
	case BLKRASET:
	case BLKRAGET:
	case BLKFLSBUF:
	case BLKSSZGET:
	case BLKPG:
	case BLKELVGET:
	case BLKELVSET:
	case BLKBSZGET:
	case BLKBSZSET:
		err = blk_ioctl(dev, cmd, arg);
		break;
	default:
		err = -EINVAL;
	}

	if (member)
		member_put(member);

	return err;
}

static int md_open(struct inode *inode, struct file *file)
{
	/*
	 * Succeed if we can lock the mddev, which confirms that
	 * it isn't being stopped right now.
	 */
	mddev_t *mddev = kdev_to_mddev(inode->i_rdev);
	int err;

	if (mddev == NULL)
		return -ENODEV;
	
	if ((mddev->flags & MD_MDDEV_EXPORTED_TO_OS) == 0) {
		member_put(&mddev->member);
		return -ENXIO;
	}
	
	if ((err = mddev_lock(mddev)))
		goto out;

 	atomic_inc(&mddev->open_count);
	mddev_unlock(mddev);

	err = 0;
	check_disk_change(inode->i_dev);
 out:
	if (err != 0)
		member_put(&mddev->member);

	return err;
}

static int md_release(struct inode *inode, struct file *file)
{
	mddev_t *mddev = kdev_to_mddev(inode->i_rdev);
	int err;

	if (mddev == NULL)
		return 0;

	if ((err = mddev_lock(mddev))) {
		member_put(&mddev->member);
		return err;
	}

	if (atomic_read(&mddev->open_count)) { 
 		atomic_dec(&mddev->open_count);
		/* for md_open */
		member_put(&mddev->member);
	}
	
	mddev_unlock(mddev);
	
	/* for kdev_to_mddev */
	member_put(&mddev->member);

	return 0;
}

static int md_revalidate(kdev_t dev)
{
	mddev_t *mddev;
	int err;

	mddev = kdev_to_mddev(dev);

	if (mddev == NULL)
		return -ENODEV;

	err = md_reread_partition(mddev);
	member_put(&mddev->member);

	return err;
}

void md_update_thread_string(mdk_thread_t *thread, const char *fmt)
{
	if (thread->mddev != NULL) {
		sprintf(thread->tsk->comm, fmt,
			mdname(to_member(thread->mddev)));
	} else {
		sprintf(thread->tsk->comm, fmt);
	}
	thread->name = thread->tsk->comm;
}
	
int md_thread(void * arg)
{
	mdk_thread_t *thread = arg;

	lock_kernel();

	/*
	 * Detach thread
	 */
	daemonize();

	current->exit_signal = SIGCHLD;
	siginitsetinv(&current->blocked, sigmask(SIGKILL));

	md_flush_signals();

	thread->tsk = current;

	md_update_thread_string(thread, thread->name);

	/*
	 * md_thread is a 'system-thread', it's priority should be very
	 * high. We avoid resource deadlocks individually in each
	 * raid personality. (RAID5 does preallocation) We also use RR and
	 * the very same RT priority as kswapd, thus we will never get
	 * into a priority inversion deadlock.
	 *
	 * we definitely have to have equal or higher priority than
	 * bdflush, otherwise bdflush will deadlock if there are too
	 * many dirty RAID5 blocks.
	 */
	unlock_kernel();

	if (thread->event && thread->run) {
		complete(thread->event);
		thread->event = NULL;
	}
	while (thread->run) {
		mdk_thread_handler_t *run;
		DECLARE_WAITQUEUE(wait, current);

		add_wait_queue(&thread->wqueue, &wait);
		set_task_state(current, TASK_INTERRUPTIBLE);
		if (!test_bit(THREAD_WAKEUP, &thread->flags)) {
			dprintk("md: thread %p went to sleep.\n", thread);
			schedule();
			dprintk("md: thread %p woke up.\n", thread);
		}
		current->state = TASK_RUNNING;
		remove_wait_queue(&thread->wqueue, &wait);
		clear_bit(THREAD_WAKEUP, &thread->flags);

		run = thread->run;
		if (run) {
			run(thread->mddev, thread->arg);
			run_task_queue(&tq_disk);
		}
		if (signal_pending(current)){
			md_flush_signals();
		}
	}
	if (thread->event) {
		complete(thread->event);
		thread->event = NULL;
	}
	return 0;
}

void md_wakeup_thread(mdk_thread_t *thread)
{
	if (thread) {
		dprintk("md: waking up MD thread %s.\n", thread->tsk->comm);
		set_bit(THREAD_WAKEUP, &thread->flags);
		wake_up(&thread->wqueue);
	}
}

mdk_thread_t *md_register_thread(mdk_thread_handler_t *run, mddev_t *mddev,
				 void *arg, const char *name)
{
	mdk_thread_t *thread;
	int ret;
	struct completion event;

	thread = (mdk_thread_t *) kmalloc
				(sizeof(mdk_thread_t), GFP_KERNEL);
	if (!thread)
		return NULL;

	memset(thread, 0, sizeof(mdk_thread_t));
	init_waitqueue_head(&thread->wqueue);

	init_completion(&event);
	thread->event = &event;
	thread->run = run;
	thread->arg = arg;
	thread->mddev = mddev;
	thread->name = name;
	ret = kernel_thread(md_thread, thread, 0);
	if (ret < 0) {
		kfree(thread);
		return NULL;
	}
	wait_for_completion(&event);
	return thread;
}

void md_interrupt_thread(mdk_thread_t *thread)
{
	if (!thread->tsk) {
		MD_BUG();
		return;
	}
	dprintk("interrupting MD-thread pid %d\n", thread->tsk->pid);
	send_sig(SIGKILL, thread->tsk, 1);
}

void md_unregister_thread(mdk_thread_t *thread)
{
	struct completion event;

	init_completion(&event);

	thread->event = &event;
	thread->run = NULL;
	thread->arg = NULL;
	thread->name = NULL;
	md_interrupt_thread(thread);
	wait_for_completion(&event);
	kfree(thread);
}

mdk_personality_t *md_find_pers(int level)
{
	mdk_personality_t *mdpers;
	int pnum;

	pnum = level_to_pers(level);
	if (pnum >= MAX_PERSONALITY) {
		MD_BUG();
		return NULL;
	}

	spin_lock(&pers_lock);
#ifdef CONFIG_KMOD
	if (!pers[pnum]) {
		spin_unlock(&pers_lock);
		{
			char mod_name[32];
			sprintf(mod_name, "md-personality-%d", pnum);
			request_module(mod_name);
		}
		spin_lock(&pers_lock);
	}
#endif

	mdpers = NULL;
	if (pers[pnum])
		mdpers = md_pers_get(pers[pnum]);
	spin_unlock(&pers_lock);

	if (mdpers == NULL)
		MD_DBG_PRINT((KERN_ERR "md: personality %d is not loaded!\n", pnum));

	return (mdpers);
}

mdk_personality_t *md_pers_get(mdk_personality_t *pers)
{
#if 0
	if (inter_module_get(pers->owner->name) == NULL)
		return NULL;
#endif
	return pers;
}

void md_pers_put(mdk_personality_t *pers)
{
#if 0
	inter_module_put(pers->owner->name);
#endif
}

mdk_metapersonality_t *md_metapers_get(mdk_metapersonality_t *metapers)
{
#if 0
	if (inter_module_get(metapers->owner->name) == NULL)
		return NULL;
#endif
	return metapers;
}

void md_metapers_put(mdk_metapersonality_t *metapers)
{
#if 0
	inter_module_put(metapers->owner->name);
#endif
}

int register_md_personality(int pnum, mdk_personality_t *p)
{
	if (pnum >= MAX_PERSONALITY) {
		MD_DBG_PRINT((KERN_ERR
		       "md: tried to install personality %s as nr %d, "
		       "but max is %lu\n",
		       p->name, pnum, MAX_PERSONALITY-1));
		return -EINVAL;
	}

	spin_lock(&pers_lock);
	if (pers[pnum]) {
		spin_unlock(&pers_lock);
		MD_BUG();
		return -EBUSY;
	}

	pers[pnum] = p;
	printk(KERN_INFO "md: %s personality registered as nr %d\n",
	       p->name, pnum);
	spin_unlock(&pers_lock);
	return 0;
}

int unregister_md_personality(int pnum)
{
	if (pnum >= MAX_PERSONALITY) {
		MD_BUG();
		return -EINVAL;
	}

	printk(KERN_INFO "md: %s personality unregistered\n", pers[pnum]->name);
	spin_lock(&pers_lock);
	pers[pnum] = NULL;
	spin_unlock(&pers_lock);
	return 0;
}

int register_md_metapersonality(mdk_metapersonality_t *p)
{
	struct list_head *ph;
	mdk_metapersonality_t *pers;

	printk("md: registering metapers %s\n", p->name);
	spin_lock(&metapersonality_list_lock);
	list_for_each(ph, &md_metapersonality_list) {
		pers = list_entry(ph, mdk_metapersonality_t, links);
		if (strncmp(p->name, pers->name, MD_METAPERS_NAME_LENGTH) == 0){
			spin_unlock(&metapersonality_list_lock);
			MD_DBG_PRINT(("md: metapers already registered, ignoring\n"));
			return -1;
		}
	}
	list_add(&p->links, &md_metapersonality_list);
	metapers_change_count++;
	spin_unlock(&metapersonality_list_lock);

	/*
	 * See if any unclaimed devices can be
	 * claimed by this personality.
	 */
	autostart_arrays();
	return 0;
}

int unregister_md_metapersonality(mdk_metapersonality_t *p)
{
	printk("md: unregistering metapers %s\n", p->name);

	list_del(&p->links);
	return 0;
}

void md_sync_acct(mdk_member_t *member, unsigned long nr_sectors)
{
}

int md_notify_reboot(struct notifier_block *this,
		     unsigned long code, void *x)
{
	struct list_head *tmp;
	mdk_member_t *member;
	mddev_t *mddev;

	printk("md_notify_reboot called with code %#lx\n", code);

	/*
	 * Don't allow any newly found devices to attach.
	 */
	autodetect_allowed = 0;
	unit_number_assignment_allowed = 0;
	if ((code == SYS_DOWN) || (code == SYS_HALT) || (code == SYS_POWER_OFF)) {
		u_long flags;

		MD_DBG_PRINT((KERN_INFO "md: stopping all md devices.\n"));

		ITERATE_CHILDREN(&toplevel_arrays, member, tmp, flags) {
			mddev = to_mddev(member);
			if (mddev_trylock(mddev) == 0)
				do_md_stop (mddev, 1);
		}
		/*
		 * certain more exotic SCSI devices are known to be
		 * volatile wrt too early system reboots. While the
		 * right place to handle this issue is the given
		 * driver, we do want to have a safe RAID driver ...
		 */
		mdelay(1000*1);
	}
	return NOTIFY_DONE;
}

struct notifier_block md_notifier = {
	.notifier_call	= md_notify_reboot,
	.next		= NULL,
	.priority	= INT_MAX, /* before any real devices */
};

static void md_geninit(void)
{
	int i;

	dprintk("md: sizeof(mdp_super_t) = %d\n", (int)sizeof(mdp_super_t));

	for (i = 0; i < MAX_MDP_DEVS; i++) {
       		mdp_blocksizes[i] = 512;
        	mdp_size[i] = 0;
        	mdp_hardsect_sizes[i] = 512;
        	mdp_maxreadahead[i] = MD_READAHEAD;
	}
	blksize_size[EMD_MAJOR] = mdp_blocksizes;
	blk_size[EMD_MAJOR] = mdp_size;
	max_readahead[EMD_MAJOR] = mdp_maxreadahead;
	hardsect_size[EMD_MAJOR] = mdp_hardsect_sizes;
#ifdef CONFIG_PROC_FS
	create_proc_read_entry("emdstat", 0, NULL, md_status_read_proc, NULL);
#endif
}

/*
 * Searches all registered partitions for autorun RAID arrays
 * at boot time.
 */
struct unclaimed_dev_s
{
	struct	list_head links;
	kdev_t	kdev;
};

typedef struct unclaimed_dev_s unclaimed_dev_t;

static LIST_HEAD(unclaimed_dev_list);
static spinlock_t unclaimed_dev_list_lock = SPIN_LOCK_UNLOCKED;

void md_autodetect_dev(kdev_t dev)
{
	unclaimed_dev_t *new_dev;

	new_dev = (unclaimed_dev_t *)kmalloc(sizeof(*new_dev), GFP_KERNEL);
	if (new_dev == NULL) {
		MD_DBG_PRINT((KERN_WARNING
		       "md: Failed to create setup item for dev %s.\n",
		       kdevname(dev)));
		return;
	}
	new_dev->kdev = dev;
	spin_lock(&unclaimed_dev_list_lock);
	list_add_tail(&new_dev->links, &unclaimed_dev_list);
	spin_unlock(&unclaimed_dev_list_lock);

	/*
	 * Whenever a new device arrives, probe for metadata.
	 */
	autostart_arrays();
}

static inline void md_genlist(int major_num)
{
	struct gendisk *gp;
	int i=0;

	if ((gp = get_gendisk(MKDEV(major_num, 0))) == NULL)
		return;

	for (i = 0; i < (gp->nr_real << gp->minor_shift);
	     i += (1 << gp->minor_shift)) {
		if (gp->part[i].nr_sects == 0)
			continue;
		md_autodetect_dev(MKDEV(major_num, i));
	}
}

static void md_force_autodetect_dev(void)
{
	md_genlist(SCSI_DISK0_MAJOR);
	md_genlist(IDE0_MAJOR);
	md_genlist(IDE1_MAJOR);
	md_genlist(IDE2_MAJOR);
	md_genlist(IDE3_MAJOR);
}

static void autostart_arrays(void)
{
	LIST_HEAD(still_unclaimed_list);
	struct list_head *tmp;
	unclaimed_dev_t *cur_dev;
	int last_metapers_change_count;

	/*
	 * Only allow one scan process at a time.
	 * Also honor the core's request to defer
	 * the autodetection process until we are
	 * ready to handle it.
	 */
	spin_lock(&unclaimed_dev_list_lock);
	if (autodetect_running || !autodetect_allowed) {
		spin_unlock(&unclaimed_dev_list_lock);
		return;
	}

	MD_DBG_PRINT((KERN_INFO "md: Autodetecting RAID arrays.\n"));
	autodetect_running = 1;

	/*
	 * Record the metadata personality change count so
	 * that we can rescan should the list change mid-scan.
	 */
	spin_lock(&metapersonality_list_lock);
rescan:
	last_metapers_change_count = metapers_change_count;
	spin_unlock(&metapersonality_list_lock);
	while ((tmp = unclaimed_dev_list.next) != &unclaimed_dev_list) {
		mdk_metapersonality_t *pers;
		mdk_metapersonality_t *high_bidder;
		int		       high_bid;
		int		       bid;
		int		       error;

		list_del(tmp);
		spin_unlock(&unclaimed_dev_list_lock);
		cur_dev = list_entry(tmp, unclaimed_dev_t, links);

		/*
		 * Auction off the device to all
		 * registered meta-data modules.
		 */
		high_bidder = NULL;
		high_bid = 0;
		ITERATE_METAPERS(pers, tmp) {

			bid = 0;
			error = pers->dev_op(cur_dev->kdev,
					     MDK_DEVOP_CLAIM,
					     &bid);
			if (error != 0) {
				/*
				 * XXX need to bail out early from bidding
				 *     process for some errors?
				 */
				continue;
			}
			if (bid > high_bid) {
				/*
				 * Keep a reference count on the
				 * highest bidder so that this
				 * personality is still available
				 * when we exit our loop.
				 */
				if (high_bidder != NULL)
					metapers_put(high_bidder);
				high_bidder = pers;
				high_bid = bid;
				metapers_get(high_bidder);
			}
		}
		if (high_bidder != NULL) {
			error = high_bidder->dev_op(cur_dev->kdev,
						    MDK_DEVOP_ATTACH, NULL);
			metapers_put(high_bidder);
			if (error == 0) {
				kfree(cur_dev);
				spin_lock(&unclaimed_dev_list_lock);
				continue;
			}
		}
		list_add_tail(&cur_dev->links, &still_unclaimed_list);
		spin_lock(&unclaimed_dev_list_lock);
	}
	/*
	 * Return any devices that we were unable to
	 * associate into arrays to the unclaimed_dev_list.
	 * A new metadata personality may be registered later
	 * that will claim them.
	 */
	list_splice(&still_unclaimed_list, &unclaimed_dev_list);

	spin_lock(&metapersonality_list_lock);
	if (last_metapers_change_count != metapers_change_count)
		goto rescan;
	spin_unlock(&metapersonality_list_lock);
	autodetect_running = 0;
	spin_unlock(&unclaimed_dev_list_lock);
}

static int do_md_stop(mddev_t * mddev, int ro)
{
	mdk_member_t *member;
	int err;
	
	err = 0;
	member = to_member(mddev);
	if (mddev->member.pers) {

		if (mddev->member.name[0] != 0)
			invalidate_partition(mddev);

		if (ro) {
			err  = -ENXIO;
			if (mddev->ro)
				goto out;
			do_md_ro(mddev, 1);
			member->pers->raid_op(member, MDK_RAID_OP_STOP,
					      &mddev->ro);
		} else {
			mddev->ro = 0;
			member->pers->raid_op(member, MDK_RAID_OP_STOP,
					      &mddev->ro);
		}
	}
	/*
	 * Free resources if final stop
	 */
	if (!ro) {
		MD_DBG_PRINT((KERN_INFO "%s: stopped.\n", mdname(member)));

		mddev->member.data_size = 0;
		mdp_size[EMDMINOR(mddev)] = 0;
		md_teardown_member(member);
	} else
		MD_DBG_PRINT((KERN_INFO "%s: switched to read-only mode.\n",
		       mdname(member)));
	err = 0;
out:
	return err;
}

int md_init_workitem(mdk_workitem_t *work, u_int opcode,
		     mdk_member_callback_t *control,
		     mdk_work_handler_t *handler, void *data)
{
	memset(work, 0, sizeof(*work));
	work->work_opcode = opcode;
	work->work_control = control;
	work->work_handler = handler;
	work->work_data = data;
	INIT_LIST_HEAD(&work->links);

	return 0;
}

void
md_cancel_workitems(mdk_workitem_support_t *workinfo)
{
	wait_queue_t wait;

	init_waitqueue_entry(&wait, current);

	spin_lock_irq(&workinfo->work_state_lock);
	add_wait_queue(&workinfo->work_wait, &wait);
	set_current_state(TASK_UNINTERRUPTIBLE);
	while (!list_empty(&workinfo->work_items)) {

		spin_unlock_irq(&workinfo->work_state_lock);
		md_interrupt_thread(workinfo->work_thread);
		md_wakeup_thread(workinfo->work_thread);
		run_task_queue(&tq_disk);
		schedule();
		spin_lock_irq(&workinfo->work_state_lock);
	}
	current->state = TASK_RUNNING;
	remove_wait_queue(&workinfo->work_wait, &wait);
	spin_unlock_irq(&workinfo->work_state_lock);
}

/*
 * Wait for a workitem to complete.  Passing in a NULL workitem
 * indicates a desire to wait until the work queue is empty.
 */
void
md_wait_workitem(mdk_workitem_support_t *workinfo, mdk_workitem_t *searchitem)
{
	wait_queue_t wait;

	init_waitqueue_entry(&wait, current);

	spin_lock_irq(&workinfo->work_state_lock);
	add_wait_queue(&workinfo->work_wait, &wait);
	for (;;) {
		struct list_head *list_item;
		int found;
		
		set_current_state(TASK_UNINTERRUPTIBLE);
		found = 0;
		list_item = workinfo->work_items.next;
		while (list_item != &workinfo->work_items) {
			mdk_workitem_t *work;

			work = list_entry(list_item, mdk_workitem_t, links);
			if (work == searchitem || searchitem == NULL) {
				found++;
				break;
			}
			list_item = list_item->next;
		}

		if (!found)
			break;

		spin_unlock_irq(&workinfo->work_state_lock);
		run_task_queue(&tq_disk);
		schedule();
		spin_lock_irq(&workinfo->work_state_lock);
	}
	current->state = TASK_RUNNING;
	remove_wait_queue(&workinfo->work_wait, &wait);
	spin_unlock_irq(&workinfo->work_state_lock);
}

mdk_workitem_flags_t md_find_workitem(mdk_workitem_support_t *workinfo,
				      u_int opcode, int *pfound)
{
	struct list_head *list_item;
	mdk_workitem_flags_t flags;
	int found;

	found = 0;
	flags = MD_WF_NONE;
	spin_lock_irq(&workinfo->work_state_lock);
	list_item = workinfo->work_items.next;
	while (list_item != &workinfo->work_items) {
		mdk_workitem_t *work;

		work = list_entry(list_item, mdk_workitem_t, links);
		if (work->work_opcode == opcode) {
			found++;
			if ((work->work_flags & MD_WF_RUNNING) != 0)
				flags |= MD_WF_RUNNING;
			else if ((work->work_flags & MD_WF_QUEUED) != 0
			      && (flags & MD_WF_RUNNING) == 0)
				flags |= MD_WF_QUEUED;
		}
		list_item = list_item->next;
	}
	spin_unlock_irq(&workinfo->work_state_lock);
	if (pfound != NULL)
		*pfound = found;
	return flags;
}

int md_queue_workitem(mdk_workitem_support_t *workinfo,
		      mdk_workitem_t *work, int head_of_queue)
{
	int error;
	mdk_workitem_t *first_work;

	error = 0;

	if (workinfo->work_thread == NULL) {
		error = -EINVAL;
		MD_DBG_PRINT(("md_queue_workitem: no work thread, exiting\n"));
		goto bailout;
	}

	work->work_flags = MD_WF_QUEUED;
	spin_lock_irq(&workinfo->work_state_lock);

	/*
	 * If the user wants this item to go to the head of the queue,
	 * we insert this on the list *after* currently running item, if
	 * any.  Then we interrupt the thread to kill the current task.
	 * The new work state will run once the current task has been
	 * terminated.
	 */
	if (head_of_queue != 0
	 &&!list_empty(&workinfo->work_items)) {
		first_work = list_entry(workinfo->work_items.next,
					mdk_workitem_t, links);
		if ((first_work->work_flags & MD_WF_RUNNING) != 0) {
			list_add_tail(&work->links, &first_work->links);
			md_interrupt_thread(workinfo->work_thread);
		} else {
			list_add(&work->links, &workinfo->work_items);
		}
		md_wakeup_thread(workinfo->work_thread);
	} else {
		list_add(&work->links, &workinfo->work_items);

		/*
		 * If a task is running, don't wake up the thread, since
		 * this task will get run after the current one finishes.
		 */

		first_work = list_entry(workinfo->work_items.next,
					mdk_workitem_t, links);
		if ((first_work->work_flags & MD_WF_RUNNING) == 0)
			md_wakeup_thread(workinfo->work_thread);
	}

	spin_unlock_irq(&workinfo->work_state_lock);

bailout:
	return (error);
}

/*
 * Thread handler for running array workitems.
 */
static void
md_handle_workitems(mddev_t *mddev, void *arg)
{
	mdk_workitem_support_t *workinfo;
	mdk_workitem_t *workitem;
	mdk_member_t *member;
	int error;

	if (mddev != NULL)
		member = to_member(mddev);
	else
		member = NULL;
	workinfo = (mdk_workitem_support_t *)arg;

	if (workinfo == NULL) {
		MD_DBG_PRINT(("%s: md_handle_workitems(): NULL workitem_suppott!\n",
		       current->comm));
		return;
	}

	MD_DBG_PRINT((KERN_INFO "%s: work thread woken up\n", current->comm));

	spin_lock_irq(&workinfo->work_state_lock);

	while (!list_empty(&workinfo->work_items)) {

		/*
		 * Indicate that we are actively servicing
		 * this workitem.
		 */
		workitem = list_entry(workinfo->work_items.next,
				      mdk_workitem_t, links);
		workitem->work_flags |= MD_WF_RUNNING;
		spin_unlock_irq(&workinfo->work_state_lock);
		error = workitem->work_control(member, MDK_WORKITEM_OP_OPEN,
					       workitem);
		if (error != 0) {
			if (error != -ESTALE) {
				MD_DBG_PRINT((KERN_ERR
				       "%s: cannot open work item\n",
					current->comm));
			}
			goto dequeue;
		}

		error = workitem->work_handler(mddev, workitem);

		/*
		 * Dequeue and free the workitem.
		 */
dequeue:
		spin_lock_irq(&workinfo->work_state_lock);
		list_del(&workitem->links);
		workitem->work_control(member, MDK_WORKITEM_OP_CLOSE, workitem);
		spin_unlock_irq(&workinfo->work_state_lock);
		workitem->work_control(member, MDK_WORKITEM_OP_FREE, workitem);

		/*
		 * Indicate to waiters that our queue has changed.
		 */
		wake_up_all(&workinfo->work_wait);

		/*
		 * Lock for next loop iteration.
		 */
		spin_lock_irq(&workinfo->work_state_lock);

		if (error == -EINTR || error == -ENOMEM)
			break;
	}
	spin_unlock_irq(&workinfo->work_state_lock);
	MD_DBG_PRINT((KERN_INFO "%s: work thread done\n", current->comm));
}

/*
 * Initialize a workitem support structure.  These
 * are usually embedded in the private data area of
 * a RAID personality.
 */
int md_init_workitem_support(mddev_t *mddev, mdk_workitem_support_t *workinfo,
			     const char *thread_string)
{
	memset(workinfo, 0, sizeof(*workinfo));
	spin_lock_init(&workinfo->work_state_lock);
	INIT_LIST_HEAD(&workinfo->work_items);
	init_MUTEX(&workinfo->work_config_sem);
	init_waitqueue_head(&workinfo->work_wait);

	if (thread_string == NULL)
		thread_string = "%s_work";
	workinfo->work_thread = md_register_thread(md_handle_workitems,
						   mddev, workinfo,
						   thread_string);
	if (workinfo->work_thread == NULL)
		return -ENOMEM;
	return 0;
}

void md_teardown_workitem_support(mdk_workitem_support_t *workinfo)
{
	if (workinfo->work_thread != NULL)
		md_unregister_thread(workinfo->work_thread);
}

static int md_sync_safemode_control(mdk_member_t *member, int op, void *arg)
{
	mdk_workitem_t *work;
	mdk_sync_support_t *syncinfo;

	work = (mdk_workitem_t *)arg;
	syncinfo = (mdk_sync_support_t *)work->work_data;
	switch (op) {
	case MDK_WORKITEM_OP_OPEN:
	{
		mddev_t *mddev;

		/*
		 * If it is safe to enter safe mode, mark us
		 * as in safemode and allow the workitem to
		 * put us into safemode.
		 */
		spin_lock_irq(&syncinfo->sync_lock);
		if ((syncinfo->safemode != MDK_SAFEMODE_IDLE_WAIT
		  && syncinfo->safemode != MDK_SAFEMODE_OFF)
		 || syncinfo->writes_pending[0] != 0
		 || syncinfo->writes_pending[1] != 0
		 || syncinfo->sync_pending[0] != 0
		 || syncinfo->sync_pending[1] != 0) {
			spin_unlock_irq(&syncinfo->sync_lock);
			return -EBUSY;
		}
		syncinfo->safemode = MDK_SAFEMODE_ACTIVE;
		mddev = syncinfo->work_state.work_thread->mddev;
		syncinfo->lock_range.start = 0;
		syncinfo->lock_range.end = to_member(mddev)->data_size;
		spin_unlock_irq(&syncinfo->sync_lock);
		break;
	}
	case MDK_WORKITEM_OP_CLOSE:
		/* NOOP */
		break;
	case MDK_WORKITEM_OP_FREE:
		work->work_flags = MD_WF_NONE;
		break;
	default:
		return (-EINVAL);
	}
	return 0;
}

int
md_sync_safemode_handler(mddev_t *mddev, mdk_workitem_t *workitem)
{
	mdk_sync_support_t *syncinfo;
	mdk_member_t *member;

	syncinfo = (mdk_sync_support_t *)workitem->work_data;
	del_timer_sync(&syncinfo->safemode_timer);
	member = to_member(mddev);
	member->pers->raid_op(member, MDK_RAID_OP_ENTER_SAFEMODE, NULL);
	return 0;
}

/*
 * Initialize a sync support structure.  These
 * are usually embedded in the private data area of
 * a RAID personality.
 */
int md_init_sync_support(mddev_t *mddev, mdk_sync_support_t *syncinfo,
			 const char *thread_string)
{
	int error;

	memset(syncinfo, 0, sizeof(*syncinfo));
	error = md_init_workitem_support(mddev, &syncinfo->work_state,
					 thread_string);
	if (error != 0)
		return error;

	atomic_set(&syncinfo->reads_pending, 0);
	spin_lock_init(&syncinfo->sync_lock);

	/*
	 * The lock range is 0 length at the start of
	 * the array when no sync operation is in progress.
	 * This ensures that no newly added sync target
	 * will be trusted until the window opens or is
	 * moved up the volume.  The only exception to
	 * this rule is in safemode.  There we open the
	 * lock range to cover the entire disk so that
	 * all writes block until we exit safemode.  Start
	 * the disk out in safemode.
	 */
	syncinfo->lock_range.start = 0;
	syncinfo->lock_range.end = to_member(mddev)->data_size;
	syncinfo->safemode = MDK_SAFEMODE_ACTIVE;
	init_waitqueue_head(&syncinfo->io_wait);
	init_waitqueue_head(&syncinfo->sync_wait);

	md_init_workitem(&syncinfo->safemode_workitem, MD_SYNCOP_ENTER_SAFEMODE,
			 md_sync_safemode_control, md_sync_safemode_handler, 
			 syncinfo);
	init_timer(&syncinfo->safemode_timer);
	syncinfo->safemode_timer.function = md_safemode_timeout;
	syncinfo->safemode_timer.data = (unsigned long)syncinfo;
	syncinfo->safemode_delay = 20 * HZ;

	/*
	 * A zero window size, io_limit_min and max
	 * indicate use the defaults.
	 */
	return 0;
}

void md_teardown_sync_support(mddev_t *mddev, mdk_sync_support_t *syncinfo)
{
	del_timer_sync(&syncinfo->safemode_timer);
	md_teardown_workitem_support(&syncinfo->work_state);
}

int md_get_sync_op_stat(mdk_sync_support_t *syncinfo,
	           mdk_syncstatsop_arg_t *syncstat)
{
	mdk_workitem_support_t *workinfo = &syncinfo->work_state;
	mdk_workitem_flags_t flags;
	int syncops[] = { MD_SYNCOP_REBUILD, MD_SYNCOP_INIT, MD_SYNCOP_VERIFY,  
			  MD_SYNCOP_VERIFY_FIX };
	int i;

	for (i = 0 ; i < sizeof(syncops); i++){
		flags = md_find_workitem(workinfo, syncops[i], NULL);
		if (flags & MD_WF_RUNNING){
			syncstat->stat_sectors = syncinfo->stat_sectors;
			syncstat->stat_time = syncinfo->stat_time;
			syncstat->syncop = syncops[i];
			return 0;
		}
	}
	memset(syncstat, 0, sizeof(*syncstat));
	return -1;
}
/*
 * Generic workitem handler that "walks" a volume
 * issuing sync (resync, initialize, verify, verify/fix)
 * operations.  The actual nature of the sync operation
 * is unknown to this routine but it does guarantee that
 * oustanding sync operations have exclusive write access
 * to the range on which they operate.
 */
#define SYNC_MARKS	32
#define	SYNC_MARK_STEP	(HZ/2)
int md_sync_handler(mddev_t *mddev, mdk_workitem_t *workitem)
{
	mdk_checkpoint_arg_t checkpoint_arg;
	mdk_sync_support_t *syncinfo;
	mdk_member_t *member;
	mdk_metapersonality_t *metapers;
	mdk_personality_t *pers;
	char *verb, *noun;
	u_long stat_sectors[SYNC_MARKS];
	u_long stat_time[SYNC_MARKS];
	u_long currspeed;
	sector_t sync_start;
	int error;
	int stat_bucket;

	syncinfo = (mdk_sync_support_t *)workitem->work_data;
	member = to_member(mddev);
	metapers = member->metapers;
	pers = member->pers;
 
	switch (workitem->work_opcode) {
	case MD_SYNCOP_REBUILD:
		verb = "reconstructing";
		noun = "reconstruction";
		break;
	case MD_SYNCOP_INIT:
		verb = "initializing";
		noun = "initialization";
		break;
	case MD_SYNCOP_VERIFY:
	case MD_SYNCOP_VERIFY_FIX:
		verb = "verifying";
		noun = "verification";
		break;
	default:
		/* silence warnings */
		verb = NULL;
		noun = NULL;
		MD_BUG();
		break;
	}

	MD_DBG_PRINT((KERN_INFO "md: %s RAID array %s\n", verb, mdname(member)));
	MD_DBG_PRINT((KERN_INFO "md: minimum _guaranteed_ %s speed: %d KB/sec/disc.\n",
	       noun, sysctl_speed_limit_min));
	MD_DBG_PRINT((KERN_INFO "md: using maximum available idle IO bandwith\n"));
	MD_DBG_PRINT((KERN_INFO "md: (but not more than %d KB/sec) for %s.\n",
	       sysctl_speed_limit_max, noun));

	/*
	 * Initialize our statistics counters.
	 */
	for (stat_bucket = 0; stat_bucket < SYNC_MARKS; stat_bucket++) {
		stat_sectors[stat_bucket] = 0;
		stat_time[stat_bucket] = jiffies;
	}
	stat_bucket = 0;
	syncinfo->stat_sectors = 0;

	/*
	 * We start the sync operation at the last checkpointed
	 * location that is applicable to this type of workitem.
	 */
	checkpoint_arg.checkpoint_type = workitem->work_opcode;
	error = metapers->member_op(member,
				    MDK_MEMBER_OP_GET_CHECKPOINT,
				    &checkpoint_arg);
	if (error != 0) {
		/*
		 * XXX record workitem failures somewhere?
		 */
		return error;
	}

	/*
	 * Round the starting sector down to a window boundary.  The
	 * RAID personalities should specify a window that is an even
	 * multiple of any minimum I/O size required to do a sync.
	 * We assume that window_size is a power of 2.
	 */
	sync_start = checkpoint_arg.checkpoint & ~(syncinfo->window_size - 1);
	syncinfo->sync_pending[0] = 0;
	syncinfo->sync_pending[1] = 0;
	syncinfo->sync_range[syncinfo->phase].start = sync_start;
	syncinfo->sync_range[syncinfo->phase].end = sync_start;
	syncinfo->sync_range[syncinfo->phase ^ 1].start = sync_start;
	syncinfo->sync_range[syncinfo->phase ^ 1].end = sync_start;

	spin_lock_irq(&syncinfo->sync_lock);

	/*
	 * Make sure that we are not in safemode.  We will be
	 * actively updating both our members and their meta-data.
	 * This also avoids having safemode operations clobber our
	 * lock range.
	 */
	md_exit_safemode(mddev, syncinfo, /*disable*/1);

	/*
	 * Initialize our lock_range.  The main loop opens
	 * the range, so don't do that here.
	 */
	syncinfo->lock_range.start = sync_start;
	syncinfo->lock_range.end = sync_start;

	/*
	 * Main sync loop;
	 * - if more syncs to issue (e.g. lock_range.end < array-size), loop.
	 * - else, wait for lock_range to close
	 * - return.
	 */
	while (syncinfo->lock_range.start < member->data_size) {
		int old_phase;
		int new_phase;

		old_phase = syncinfo->phase;
		new_phase = syncinfo->phase ^ 1;

		/*
		 * Wait for sync I/O in oldest phase to cease.  Once
		 * we toggle the phase, this oldest sync window will
		 * become the newest one, so we must be careful not
		 * to collide with any pending operations.
		 */
		wait_event_lock_irq(syncinfo->io_wait,
				    (syncinfo->sync_pending[new_phase] == 0),
				    syncinfo->sync_lock);
		/*
		 * Update the checkpoint of this operation.  The checkpoint
		 * is extended to cover all transactions in the oldest phase
		 * that have completed thus far.
		 */
		checkpoint_arg.checkpoint = syncinfo->sync_range[new_phase].end;
		spin_unlock_irq(&syncinfo->sync_lock);
		error = metapers->member_op(member,
					    MDK_MEMBER_OP_SET_CHECKPOINT,
					    &checkpoint_arg);
		if (error != 0) {
			/*
			 * XXX record workitem failures somewhere?
			 *     close sync??
			 */
			return error;
		}
		spin_lock_irq(&syncinfo->sync_lock);

		/*
		 * Toggle the phase, extend the lock range, and create a
		 * new sync range to cover the new extention of the lock
		 * range.
		 */
		if (syncinfo->sync_range[old_phase].end
		 != syncinfo->lock_range.end) {
			MD_DBG_PRINT(("md: Unexepcted sync range discrepancy.\n"
			       "md: lock 0x[%lx,%lx], sync_old 0x[%lx,%lx], "
			       "sync_new 0x[%lx,%lx].\n",
			       syncinfo->lock_range.start,
			       syncinfo->lock_range.end,
			       syncinfo->sync_range[old_phase].start,
			       syncinfo->sync_range[old_phase].end,
			       syncinfo->sync_range[new_phase].start,
			       syncinfo->sync_range[new_phase].end));
			MD_BUG();
		}
		syncinfo->phase = new_phase;
		syncinfo->sync_range[new_phase].start =
		    syncinfo->lock_range.end;
		syncinfo->lock_range.end += syncinfo->window_size;
		if (syncinfo->lock_range.end > member->data_size)
			syncinfo->lock_range.end = member->data_size;
		syncinfo->sync_range[new_phase].end = syncinfo->lock_range.end;

		/*
		 * Before starting any I/O in the new sync range, wait
		 * for outstanding I/O in the old phase to complete.
		 * This guarantees that only I/O operations that have
		 * seen the new lock range are outstanding.
		 */
		wait_event_lock_irq(syncinfo->sync_wait,
				    (syncinfo->writes_pending[old_phase] == 0),
				    syncinfo->sync_lock);
		spin_unlock_irq(&syncinfo->sync_lock);

		/*
		 * Issue syncs in the new range closing the new sync
		 * range as we go.
		 */
		while (syncinfo->sync_range[new_phase].start
		     < syncinfo->sync_range[new_phase].end) {
			sector_t issued;
	
			issued = syncinfo->sync_range[new_phase].start;
			error = pers->raid_op(member, MDK_RAID_OP_SYNC_REQUEST,
					      workitem);
			if (error) {
				MD_DBG_PRINT((KERN_INFO "%s: md_sync_handler() "
				       "error on sync I/O... exiting\n",
				       mdname(member)));
				spin_lock_irq(&syncinfo->sync_lock);
				goto out;
			}
			issued = syncinfo->sync_range[new_phase].start - issued;
			stat_sectors[stat_bucket] += issued;
test_throttle:
			if (jiffies >= stat_time[stat_bucket]) {
				int next_bucket;

				/*
				 * Rotate statistics and update
				 * utility visible statistics counters.
				 * Record actual time of switch
				 * for proper accounting.
				 */
				stat_time[stat_bucket] = jiffies;
				next_bucket = (stat_bucket + 1) % SYNC_MARKS;
				syncinfo->stat_sectors +=
				    stat_sectors[stat_bucket];
				syncinfo->stat_sectors -=
				    stat_sectors[next_bucket];
				syncinfo->stat_time =
				    jiffies - stat_time[next_bucket];
				stat_sectors[next_bucket] = 0;
				stat_time[next_bucket] =
				    jiffies + SYNC_MARK_STEP;
				stat_bucket = next_bucket;
			}

			if (signal_pending(current)) {
				/*
				 * got a signal, exit.
				 */
				MD_DBG_PRINT((KERN_INFO "%s: md_sync_handler() "
				       "got signal ... exiting\n",
				       mdname(member)));
				flush_signals(current);
				error = -EINTR;
				spin_lock_irq(&syncinfo->sync_lock);
				goto out;
			}

			/*
			 * Throttle our resync based on calculated I/O
			 * load so long as no normal I/O is blocked waiting
			 * on our resync window.  Throttling kicks in any
			 * time we exceed the maximum resync rate as well
			 * as anytime normal I/O is in progress as we are
			 * exceeding the guaranteed resync rate.
			 */
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,9)
			cond_resched();
#endif
			if (syncinfo->stat_sectors == 0
			 || syncinfo->stat_time < 1)
				currspeed = 0;
			else
				currspeed = ((syncinfo->stat_sectors >> 1) * HZ)
					  / syncinfo->stat_time;
			if (!waitqueue_active(&syncinfo->io_wait)
			 && (currspeed > sysctl_speed_limit_max
			  || (currspeed > sysctl_speed_limit_min
			   && (syncinfo->writes_pending[syncinfo->phase]
			    || atomic_read(&syncinfo->reads_pending))))) {
				current->state = TASK_INTERRUPTIBLE;
				schedule_timeout(HZ/4);
				goto test_throttle;
			}
		}
		spin_lock_irq(&syncinfo->sync_lock);
	}

out:
	/*
	 * Now that all sync requests have been issued, wait for
	 * the syncs to complete, write a final checkpoint, release
	 * the lock_range, and exit.  We forst the end of the sync
	 * ranges to be the same as the start position just in
	 * case we exited due to a signal or error.  Otherwise
	 * the completion handlers will not know to signal us
	 * when the pending sync count drops to 0.
	 */
 	syncinfo->sync_range[0].end = syncinfo->sync_range[0].start;
 	syncinfo->sync_range[1].end = syncinfo->sync_range[1].start;
	wait_event_lock_irq(syncinfo->io_wait,
			    (syncinfo->sync_pending[0] == 0
			  && syncinfo->sync_pending[1] == 0),
			    syncinfo->sync_lock);
	syncinfo->lock_range.start = 0;
	syncinfo->lock_range.end = 0;
	md_enable_safemode(mddev, syncinfo);
	spin_unlock_irq(&syncinfo->sync_lock);
	checkpoint_arg.checkpoint = syncinfo->sync_range[syncinfo->phase].end;
	metapers->member_op(member, MDK_MEMBER_OP_SET_CHECKPOINT,
			    &checkpoint_arg);
	return error;
}

static mdk_metapersonality_t *md_check_metadata_support(char *metaname)
{
	struct list_head *ph;
	mdk_metapersonality_t *pers;

	spin_lock(&metapersonality_list_lock);
	list_for_each(ph, &md_metapersonality_list) {
		pers = list_entry(ph, mdk_metapersonality_t, links);
		if (strncmp(metaname, pers->name, 
			    MD_METAPERS_NAME_LENGTH) == 0){
			spin_unlock(&metapersonality_list_lock);
			return pers;
		}
	}
	spin_unlock(&metapersonality_list_lock);
	
	return NULL;
}

/*
 * Mark meta-data as open, release sync_lock range, and update
 * safemode state to reflect that safemode is off.  Called with
 * the syncinfo lock held.
 */
void md_exit_safemode(mddev_t *mddev, mdk_sync_support_t *syncinfo, int disable)
{
	mdk_member_t *member;

	/*
	 * Prevent other I/Os from attempting to end
	 * safe-mode by indicating that we are leaving
	 * safemode.  Future writes will still end up
	 * blocking on the lock-range that still covers
	 * the entire volume.
	 */
	syncinfo->safemode = MDK_SAFEMODE_EXIT;
	spin_unlock_irq(&syncinfo->sync_lock);

	/*
	 * We may be called from the sync handler.  In
	 * this case, our oldstate may well have been
	 * MDK_SAFEMODE_IDLE_WAIT.  Ensure that the idle
	 * timer is canceled.
	 */
	del_timer_sync(&syncinfo->safemode_timer);

	member = to_member(mddev);

	member->pers->raid_op(member, MDK_RAID_OP_EXIT_SAFEMODE, NULL);

	spin_lock_irq(&syncinfo->sync_lock);
	if (disable)
		syncinfo->safemode = MDK_SAFEMODE_DISABLED;
	else
		syncinfo->safemode = MDK_SAFEMODE_OFF;
	syncinfo->lock_range.start = 0;
	syncinfo->lock_range.end = 0;

	/*
	 * Wakeup any I/O blocked on the exit from safemode.
	 */
	wake_up_all(&syncinfo->io_wait);
}

void md_enable_safemode(mddev_t *mddev, mdk_sync_support_t *syncinfo)
{
	if (syncinfo->safemode != MDK_SAFEMODE_DISABLED)
		return;

	/*
	 * Re-enable safemode.  If no writes are pending,
	 * kick our safemode timer so that we will eventually
	 * go into safemode.
	 */
	syncinfo->safemode = MDK_SAFEMODE_OFF;
	if (syncinfo->writes_pending[syncinfo->phase] == 0)
		md_update_safemode_timer(syncinfo);
}

static int set_disk_faulty(dev_t dev)
{
	mdk_member_t *member;

	member = get_member_from_device(&toplevel_arrays, dev);
	if (member == NULL
	 || member->type != MDK_MTYPE_RDEV)
		return -ENODEV;

	md_set_member_state(member, MDU_MSTATE_FAILED);

	member_put(member);

	return 0;
}

/*
 * md_remove_from_unclaimed list removes the kdev from unclaimed
 * list.
 */
static int md_remove_from_unclaimed_list(kdev_t kdev)
{
	unclaimed_dev_t *cur_dev;
	int error;
	
	error = -ENODEV;
	spin_lock(&unclaimed_dev_list_lock);
	list_for_each_entry(cur_dev, &unclaimed_dev_list, links) {
		if (cur_dev->kdev == kdev) {
			list_del(&cur_dev->links);
			kfree(cur_dev);
			error = 0;
			break;
		}
	}
	spin_unlock(&unclaimed_dev_list_lock);
	return (error);
}

static int md_run_array(mddev_t *mddev)
{
	mdk_member_t *member;
	mdk_personality_t *pers;
	u_long flags;

	member = to_member(mddev);
	pers = member->pers;
	spin_lock_irqsave(&member->member_lock, flags);
	mddev->flags |= MD_MDDEV_ALLOWED_TO_RUN;
	spin_unlock_irqrestore(&member->member_lock, flags);
	return pers->raid_op(member, MDK_RAID_OP_RUN, 0);
}

static int md_ctl_open(struct inode *inode, struct file *file)
{
	/* XXX Should do a refcount here and limit the number of openers. */
	return (0);
}

static int md_ctl_ioctl(struct inode *inode, struct file *file,
			unsigned int cmd, unsigned long arg)
{
	kdev_t dev;
	int minor, err = 0;

	dev = inode->i_rdev;
	minor = MINOR(dev);
	if (!capable(CAP_SYS_ADMIN))
		return -EACCES;

	if (minor >= MAX_MDP_DEVS << EMD_MINOR_SHIFT) 
		return -EINVAL;

	switch (cmd) {
	/*
	 * Commands dealing with the RAID driver but not any
	 * particular array:
	 */
	case RAID_VERSION:
	{
		err = get_version((void *)arg);
		return 0;
	}
#if 0
	case PRINT_RAID_DEBUG:
	{
		err = 0;
		md_print_devices();
		return 0;
	}
	case RAID_AUTORUN:
	{
		err = 0;
		autostart_arrays();
		return 0;
	}
#endif
	
	case EMD_GET_DEVICE_INFO:
	{
		unsigned int major, minor;
		mdu_cookie_info_t *pcookie = (mdu_cookie_info_t *)arg;
		mdu_cookie_info_t cookie;
		mdk_member_t *member;
		
		if (pcookie == NULL)
			return -EINVAL;

		if ((err = copy_from_user(&cookie, pcookie, sizeof(cookie))))
			return err;
		
		major = cookie.major_minor.major;
		minor = cookie.major_minor.minor;
		
		member = get_member_from_device(&toplevel_arrays,
			       			MKDEV(major, minor));
		
		if (member == NULL) {
			if ((member = get_member_from_device(&spare_arrays,
				    	MKDEV(major, minor))) == NULL)
				return -ENODEV;
		}
		
		err = get_cookie_info(member, &cookie);
		member_put(member);

		if (!err) {
			if (copy_to_user(pcookie, &cookie, sizeof(cookie)))
				return -EFAULT;
		}
		return err;
	}
	/* This IOCTL passes an array structure as the arg */
	case EMD_GET_ARRAY_INFO:
	{
		mdu_array_info_t *parray_info;
		mdu_array_info_t array_info;
		mdk_member_t *member;
		mddev_t *mddev;
		
		parray_info = (mdu_array_info_t *)arg;
		if (parray_info == NULL)
			return -EINVAL;
		
		if ((err = copy_from_user(&array_info, parray_info, 
					  sizeof(array_info))))
			return err;

		member = get_member_from_cookie(&toplevel_arrays,
						array_info.cookie);
		
		if (member == NULL) {
			if ((member = get_member_from_cookie(&spare_arrays, 
						array_info.cookie)) == NULL)
				return -ENODEV;
		}
		
		/* If this is an rdev, you really want */
		/* the array info from the parent. */
		if (member && (member->type == MDK_MTYPE_RDEV)) {
			mdk_member_t *parent;

			parent = member->parent;
			if (parent)
				member_get(parent);
			member_put(member);
			member = parent;
		}
		
		if (member == NULL)
			return -ENODEV;
		
		mddev = to_mddev(member);
		err = get_array_info(mddev, parray_info);
		member_put(member);

		return err;
	}
	case EMD_STOP_ARRAY_RO:
	case EMD_RESTART_ARRAY_RW:
	{
		mddev_t *mddev;
		mdu_major_minor_t major_minor;
		
		err = copy_from_user(&major_minor, (void *)arg,
			      	     sizeof(mdu_major_minor_t));
		if (err)
			return err;
		mddev = kdev_to_mddev(MKDEV(major_minor.major,
					    major_minor.minor));
		if (!mddev)
			return -EINVAL;
		
		if (atomic_read(&mddev->open_count)) {
			member_put(&mddev->member);
			return -EBUSY;
		}

		if (cmd == EMD_STOP_ARRAY_RO)
			err = do_md_ro(mddev, 1);
		else
			err = do_md_ro(mddev, 0);
		member_put(&mddev->member);
		return err;
	}
	case EMD_CANCEL_VERIFY:
	case EMD_VERIFY_ARRAY:
	case EMD_VERIFY_WITH_FIX_ARRAY:
	{
		mddev_t *mddev;
		mdu_major_minor_t major_minor;
		mdk_member_t *member;
		int fix;
		
		err = copy_from_user(&major_minor, (void *)arg,
			      	     sizeof(mdu_major_minor_t));
		if (err)
			return err;
		mddev = kdev_to_mddev(MKDEV(major_minor.major,
					    major_minor.minor));
		if (!mddev)
			return -ENODEV;
		member = to_member(mddev);
		if (!member->pers) {
			member_put(member);
			return -EFAULT;
		}
		if (cmd == EMD_CANCEL_VERIFY) {
			err = member->pers->raid_op(member,
						    MDK_RAID_OP_CANCEL_VERIFY,
						    NULL);
			member_put(member);
			return err;

		}
		fix = (cmd == EMD_VERIFY_ARRAY) ? 0 : 1;
		err = member->pers->raid_op(member, MDK_RAID_OP_REQUEST_VERIFY,
					    (void *)fix);
		member_put(member);
		return err;
	}
	case EMD_KILL_META_DATA:
	{
		mdu_major_minor_t major_minor;
		kdev_t kdev;
		mdk_member_t *member;
		mdk_metapersonality_t *pers;
		struct list_head *tmp;
			
		err = copy_from_user(&major_minor, (void *)arg,
			      	     sizeof(mdu_major_minor_t));
		if (err)
			return err;
		kdev = MKDEV(major_minor.major, major_minor.minor);
		member = get_member_from_device(&toplevel_arrays, kdev);
		if (member != NULL) {
			/*
			 * kdev is part of an array
			 */
			member_put(member);
			return -EBUSY;
		}

		/*
		 * Checking if valid metadata is present before zeroing.
		 */
		err = -ENODATA;
		ITERATE_METAPERS(pers, tmp) {
			int bid;
			int error;

			bid = 0;
			error = pers->dev_op(kdev, MDK_DEVOP_CLAIM, &bid);
			if (error != 0)
				continue;
			if (bid > 0) {
				pers->dev_op(kdev, MDK_DEVOP_WIPE_METADATA,
					     NULL);
				err = 0;
			}
		}
		
		return err;
	}
	case EMD_SET_ARRAY_INFO:
	{
		mdu_array_info_t array_k;
		mdu_array_info_t *parray_info = (mdu_array_info_t *)arg;
		mddev_t *pmddev;
		mdk_member_t *member;
		int device;
		
		if (parray_info == NULL) 
			return -EINVAL;
		
		if ((err = copy_from_user(&array_k, parray_info, 
					  sizeof(array_k))))
			return err;

		device = MKDEV(EMD_MAJOR, array_k.md_minor);
		
		member = get_member_from_device(&toplevel_arrays, device);
		pmddev = to_mddev(member);
		
		if (pmddev == NULL) {
			mdk_metapersonality_t *metapers;

			if ((metapers = md_check_metadata_support(
					array_k.metaname)) == NULL) {
				MD_DBG_PRINT(("Metadata:[%s] not registered\n",
			       		array_k.metaname));
				return -EFAULT;
			}
			err = metapers->member_op(member, 
					    MDK_MEMBER_OP_SET_ARRAY_INFO, 
					    &array_k);

		} else {
			member_put(member);
			MD_DBG_PRINT(("md: Error! Array already exists.\n"));
			return -EINVAL;
		}
	
		/* copy back cookie info to user */
		if (!err) {
			if (copy_to_user(parray_info, &array_k, 
					 sizeof(array_k))) {
				return -EFAULT;
			}
		}
		
		return err;
	}
	case EMD_ADD_NEW_DISK:
	{	
		mdu_disk_info_t *pdisk = (mdu_disk_info_t *)arg;
		mdu_disk_info_t disk_info;
		mdk_member_t *parent;
		kdev_t device_num;

		if (pdisk == NULL) 
			return -EINVAL;

		err = copy_from_user(&disk_info, pdisk, sizeof(disk_info));
		if (err) 
			return err;

		parent = get_member_from_cookie(&toplevel_arrays,
						disk_info.parent_cookie);
		if (parent == NULL)
			return -ENODEV; 

		if (parent->type != MDK_MTYPE_MDDEV) 
			return -ENODEV; 
		

		/*
		 * Only allow devices to be added to an array if
		 * they are not already incorperated.
		 */
		device_num = MKDEV(disk_info.major, disk_info.minor);
		if (get_member_from_device(&toplevel_arrays, device_num)
		 || get_member_from_device(&spare_arrays, device_num)) 
			return -EBUSY;

		/*
		 * If the device is in the unclaimed list, it
		 * is no longer unclaimed.
		 */
		md_remove_from_unclaimed_list(device_num);

		/* 
		 * The device has been removed from the unclaimed 
		 * device list, now call the appropriate 
		 * member_op routine to stamp metadata on the disk.
		 */
		if (disk_info.device_type == MDU_PHYSICAL_DEVICE) {
			
			if (parent == NULL) 
				return -ENODEV;
			/* We assume the caller has the proper mddev received
			 * prior to issuing this call using a cookie.
			 */
			
			err = parent->metapers->member_op(parent, 
						  MDK_MEMBER_OP_ADD_NEW_DISK,
						  &disk_info);
		} 
		
		if (disk_info.device_type == MDU_HOT_SPARE_DEVICE) {

			if (parent == NULL) {
				mdk_metapersonality_t *metapers;

				if ((metapers = md_check_metadata_support(
					disk_info.metaname)) == NULL) {
					MD_DBG_PRINT(("Metadata:[%s] not registered\n",
					disk_info.metaname));
					member_put(parent);
					return -EINVAL;
				}
	
				err = metapers->member_op( NULL,
					 	  MDK_MEMBER_OP_CREATE_SPARE,
					          &disk_info);
			} else {
				err = parent->metapers->member_op(parent, 
					 	  MDK_MEMBER_OP_CREATE_SPARE,
					          &disk_info);

			}
		}
		member_put(parent);
		
		return err;
	}
	case EMD_GET_SPARES:
	{
		mdk_metapersonality_t *metapers;
		mdu_spare_disk_info_t  spare_disk;
		mdu_disk_info_t	       *user_disk_info;
		mdk_member_t	      *parent;
		size_t		       disk_info_size;

		err = copy_from_user(&spare_disk, (void *)arg,
				     sizeof(spare_disk));
		if (err)
			return err;
		
		parent = get_member_from_cookie(&toplevel_arrays,
				spare_disk.parent_cookie);
		metapers = md_check_metadata_support(
				spare_disk.metaname);
		if (metapers == NULL) {
			MD_DBG_PRINT(("Metadata:[%s] not registered\n",
				spare_disk.metaname));
			kfree(spare_disk.spare_info);
			if (parent)
				member_put(parent);
			return -EINVAL;
		}
		if (parent != NULL && parent->metapers != metapers) {
			MD_DBG_PRINT(("EMD_GET_SPARES: Metadata format (%s) "
			       "!= array's format (%s).\n",
			       spare_disk.metaname,
			       parent->metapers == NULL
			     ? "None" : parent->metapers->name));
			member_put(parent);
			return -EINVAL;
		}
	
		user_disk_info = NULL;
		disk_info_size = 0;	
		if (spare_disk.count != 0) {
			user_disk_info = spare_disk.spare_info;
			disk_info_size = sizeof(mdu_disk_info_t) 
					 * spare_disk.count;
			spare_disk.spare_info = 
					kmalloc(GFP_KERNEL, disk_info_size);
			if (spare_disk.spare_info == NULL) {
				if (parent)
					member_put(parent);
				return -ENOMEM;
			}
			memset(spare_disk.spare_info, 0, disk_info_size);
			err = copy_from_user(spare_disk.spare_info, 
					     user_disk_info,
					     disk_info_size);
			if (err) {
				if (parent)
					member_put(parent);
				kfree(spare_disk.spare_info);
				return -EINVAL;
			}
		}
		err = metapers->member_op(parent,  MDK_MEMBER_OP_GET_SPARES,
					  &spare_disk);
		if (!err) {

			if (spare_disk.spare_info != NULL) {
				err = copy_to_user(user_disk_info,
					spare_disk.spare_info, 
					disk_info_size);
				kfree(spare_disk.spare_info);
			} else {
				err = copy_to_user((void *)arg, &spare_disk,
					   sizeof(spare_disk));
			}
		}
		if (parent)
			member_put(parent);
		return err;
	}
	case EMD_GET_FREESIZE:
	{
		mdu_device_metapers_t mpers_dev;
		kdev_t kdev;
		mdk_metapersonality_t *metapers;
		sector_t freesize;
		
		err = copy_from_user(&mpers_dev, (void *)arg, 
				     sizeof(mdu_device_metapers_t));
		if (err)
			return err;
		
		kdev = MKDEV(mpers_dev.major_minor.major, 
			     mpers_dev.major_minor.minor);
		
		metapers = md_check_metadata_support(mpers_dev.metapers_name);
		if (metapers == NULL) {
			MD_DBG_PRINT(("Metadata:[%s] not registered\n",
			       mpers_dev.metapers_name));
			return -EFAULT;
		}
		err = metapers->dev_op(kdev, MDK_DEVOP_GET_FREESIZE, 
					&freesize);
		if (err)
			return err;
		err = copy_to_user((void *)arg, &freesize, 
					   sizeof(sector_t));
		return err;
	}
	case EMD_GET_NUM_OF_METAPERS:
	{
		int count = 0;
		mdk_metapersonality_t *pers;
		struct list_head *tmp;
		
		ITERATE_METAPERS(pers, tmp) {
			count++;
		}
		err = copy_to_user((void *)arg, &count, sizeof(count));
		
		return err;
	}
	case EMD_GET_METAPERS:
	{
		int count = 0;
		int namelen = MD_METAPERS_NAME_LENGTH;
		char name[MD_METAPERS_NAME_LENGTH];
		mdk_metapersonality_t *pers;
		struct list_head *tmp;
		
		ITERATE_METAPERS(pers, tmp) {
			memset(name, 0, namelen);
			strcpy(name, pers->name);
			err = copy_to_user((void *)(arg + (namelen * count)), 
					   name, namelen); 
			if(err)
				return err;
			count++;
		}
		return 0;
	}
	case EMD_SET_DISK_FAULTY:
	{
		mdu_disk_info_t *pdisk = (mdu_disk_info_t *)arg;
		mdu_disk_info_t disk;
		kdev_t device_num;
		
		if (pdisk == NULL)
			return -ENODEV;

		err = copy_from_user(&disk, pdisk, 
				     sizeof(disk));
		if (err)
			return err;

		device_num = MKDEV(disk.major, disk.minor);
		err = set_disk_faulty(device_num);
		return err;
	}
	case EMD_STOP_ARRAY:
	{
		mdk_member_t *pmember;
		mdk_member_t *member;
		mddev_t *mddev;

		pmember = (mdk_member_t *)arg;
		if (pmember == NULL)
			return -EINVAL;

		member = get_member_from_cookie(&toplevel_arrays, arg);

		if (member == NULL) {
			if ((member = 
			 get_member_from_cookie(&spare_arrays, arg)) == NULL) {
				return -ENODEV;
			} else {
				err = member->metapers->member_op(member, 
					 	  MDK_MEMBER_OP_DELETE_SPARE,
					          NULL);
				/* Drop all refcount and destroy spare */
				member_put(member);
				member_put(member->parent);
				return 0;
			}
		}

		if (member->type != MDK_MTYPE_MDDEV)
			return -ENODEV;
		
		mddev = to_mddev(member);
		
		if (mddev == NULL)
			return -ENODEV;
		
		if (atomic_read(&mddev->open_count)) {
			member_put(member);
			return -EBUSY;
		}
		err = do_md_stop(mddev, 0);
		member_put(member);

		return err;
	}
	case EMD_GET_DISK_INFO_FROM_PARENT:
	{
		mdu_disk_info_t *pdisk;
		mdu_disk_info_t disk_info;
		mdk_member_t *member;
		
		pdisk = (mdu_disk_info_t *) arg;
		
		if ((err = copy_from_user(&disk_info, pdisk, 
					  sizeof(disk_info))))
			return err;

		if ((member = get_child_from_index(disk_info.parent_cookie,
					      disk_info.number)) == NULL)
			return -ENODEV;
		
		if (member->type != MDK_MTYPE_RDEV)
			return -ENODEV;
		
		err = get_disk_info(member, (void *)arg);
		member_put(member);
	
		return err;
	}
	case EMD_GET_MEMBER_INFO:
	{
		mdu_cookie_info_t *pcookie = (mdu_cookie_info_t *)arg;
		mdu_cookie_info_t cookie;
		mdk_member_t *member = NULL;
		
		if (pcookie == NULL)
			return -EINVAL;

		if ((err = copy_from_user(&cookie, pcookie, sizeof(cookie))))
			return err;
	
		member = get_member_from_cookie(&toplevel_arrays,
						cookie.cookie);
		if (member == NULL)
			return -ENODEV;
		
		err = get_cookie_info(member, &cookie);		
		member_put(member);

		if (!err) {
			if (copy_to_user(pcookie, &cookie, sizeof(cookie))) {
				return -EFAULT;
			}
		}
	
		return err;
	}
	case EMD_RUN_ARRAY:
	{
		mdk_member_t *member;
		mddev_t *mddev;

		member = get_member_from_cookie(&toplevel_arrays, arg);

		if (member == NULL)
			return -ENODEV;

		mddev = to_mddev(member);
		md_run_array(mddev);
		member_put(member);
		return err;
	}	
	default:
		if (_IOC_TYPE(cmd) == MD_MAJOR)
			MD_DBG_PRINT((KERN_WARNING "md: %s(pid %d) used"
				" obsolete MD ioctl, upgrade your"
				" software to use new ioctls.\n",
				current->comm, current->pid));
		//return -EINVAL;
	}


/* The below ioctls are yet to be tested for functionality */	

	switch (cmd) {
	/* These IOCTLs pass a disk structure as the arg */	
	case EMD_GET_DISK_INFO:
	//case EMD_HOT_REMOVE_DISK:
	{
		mdu_disk_info_t *disk;
		mdk_member_t *member;
		mddev_t *mddev;
		int device;

		disk = (mdu_disk_info_t *)arg;
		if (disk == NULL)
			return -ENODEV;

		device = MKDEV(disk->major, disk->minor);
		if ((member = get_member_from_device(&toplevel_arrays, 
						     device)) == NULL)
			return -ENODEV;
		
		if (member->type != MDK_MTYPE_RDEV)
			return -ENODEV;

		if (cmd == EMD_GET_DISK_INFO)
			err = get_disk_info(member, (void *)arg);
		else
			err = hot_remove_disk(mddev, device);

		member_put(member);
		
		return err;
	}
	/* These IOCTLs pass a device number as the arg */	
	case HOT_GENERATE_ERROR:
	//case EMD_HOT_ADD_DISK:
	{
		mdu_disk_info_t *disk;
		mdk_member_t *member;
		mddev_t *mddev;
		int dev;
	
		disk = (mdu_disk_info_t *) arg;
		if (disk == NULL)
			return -ENODEV;

		dev = MKDEV(disk->major, disk->minor);
		member = (mdk_member_t *) disk->parent_cookie;
		mddev = to_mddev(member);

		if (cmd == EMD_HOT_ADD_DISK)
			err = hot_add_disk(mddev, dev);
		else
			err = hot_generate_error(mddev, dev);
		
		member_put(member);
		
		return err;
	}
	default:
		MD_DBG_PRINT(("Unsupported Ioctl\n"));
		return -ENXIO;
	}
	
	return err;
}

static struct file_operations md_ctl_fops = {
	.owner	= THIS_MODULE,
	.open	= md_ctl_open,
	.ioctl	= md_ctl_ioctl
};

static struct miscdevice md_ctl_miscdev = {
	.minor = EMDCTL_MINOR,
	.name = "emdctl",
	.fops = &md_ctl_fops
};

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
int __init raid0_init (void);
void raid0_exit (void);
int __init raid1_init (void);
void raid1_exit (void);
int __init ddf_init(void);
void ddf_exit(void);
int __init md_asr_init(void);
void md_asr_exit(void);
#endif

int __init md_init(void)
{
	mdk_member_t *member;
	mddev_t *mddev;
	struct list_head *tmp;
	u_long flags;

	printk(KERN_INFO "emd: Enhanced md driver %d.%d.%d MAX_MD_DEVS=%d,"
			" MD_SB_DISKS=%d\n",
			MD_MAJOR_VERSION, MD_MINOR_VERSION,
			MD_PATCHLEVEL_VERSION, MAX_MD_DEVS, MD_SB_DISKS);

	if (devfs_register_blkdev (EMD_MAJOR, "emd", &md_fops)) {
		MD_DBG_PRINT((KERN_ALERT "emd: Unable to get major %d for md\n",
		       EMD_MAJOR));
		return -1;
	}

	if (misc_register(&md_ctl_miscdev) != 0)
		MD_DBG_PRINT(("md_init: warning, unable to register emdctl node\n"));

	mdp_gendisk.major = EMD_MAJOR;

	/* forward all md request to md_make_request */
	blk_queue_make_request(BLK_DEFAULT_QUEUE(EMD_MAJOR), mdp_make_request);

	read_ahead[EMD_MAJOR] = INT_MAX;

	/* Register with the block layer */
	add_gendisk(&mdp_gendisk);
	md_geninit();

	register_reboot_notifier(&md_notifier);
	raid_table_header = register_sysctl_table(raid_root_table, 1);

	raid0_init();
	raid1_init();
	ddf_init();
	md_asr_init();

	md_force_autodetect_dev();
	autodetect_allowed = 1;
	unit_number_assignment_allowed = 1;
	autostart_arrays();

	ITERATE_CHILDREN(&toplevel_arrays, member, tmp, flags) {
		mddev = to_mddev(member);
		md_run_array(mddev);
	}
	return (0);
}

static __exit void md_exit(void)
{
	mdk_member_t *member;
	mddev_t *mddev;
	struct list_head *tmp;
	u_long flags;

	unregister_sysctl_table(raid_table_header);
	misc_deregister(&md_ctl_miscdev);
	devfs_unregister_blkdev(EMD_MAJOR, "emd");
	unregister_reboot_notifier(&md_notifier);
#ifdef CONFIG_PROC_FS
	remove_proc_entry("emdstat", NULL);
#endif
	ITERATE_CHILDREN(&toplevel_arrays, member, tmp, flags) {

		mddev = to_mddev(member);
		do_md_stop(mddev, 0);
	}
	del_gendisk(&mdp_gendisk);
	ddf_exit();
	md_asr_exit();
	raid0_exit();
	raid1_exit();
}


module_init(md_init)
module_exit(md_exit)

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
EXPORT_SYMBOL(register_md_personality);
EXPORT_SYMBOL(unregister_md_personality);
EXPORT_SYMBOL(register_md_metapersonality);
EXPORT_SYMBOL(unregister_md_metapersonality);
EXPORT_SYMBOL(md_sync_acct);
EXPORT_SYMBOL(md_done_sync);
EXPORT_SYMBOL(md_write_start);
EXPORT_SYMBOL(md_write_end);
EXPORT_SYMBOL(md_handle_safemode);
EXPORT_SYMBOL(md_register_thread);
EXPORT_SYMBOL(md_unregister_thread);
EXPORT_SYMBOL(md_update_thread_string);
EXPORT_SYMBOL(md_wakeup_thread);
EXPORT_SYMBOL(md_print_devices);
EXPORT_SYMBOL(md_interrupt_thread);
EXPORT_SYMBOL(md_get_array_stats);
EXPORT_SYMBOL(md_read_blocks);
EXPORT_SYMBOL(md_write_blocks);
EXPORT_SYMBOL(md_member_add_child);
EXPORT_SYMBOL(md_member_orphan);
EXPORT_SYMBOL(md_rdev_alloc);
EXPORT_SYMBOL(md_mddev_alloc);
EXPORT_SYMBOL(md_find_pers);
EXPORT_SYMBOL(md_pers_get);
EXPORT_SYMBOL(md_pers_put);
EXPORT_SYMBOL(md_metapers_get);
EXPORT_SYMBOL(md_metapers_put);
EXPORT_SYMBOL(md_assign_unit_number);
EXPORT_SYMBOL(md_teardown_member);
EXPORT_SYMBOL(md_export_array);
EXPORT_SYMBOL(md_check_chunksize);
EXPORT_SYMBOL(toplevel_arrays);
EXPORT_SYMBOL(spare_arrays);
EXPORT_SYMBOL(md_init_sync_support);
EXPORT_SYMBOL(md_teardown_sync_support);
EXPORT_SYMBOL(md_cancel_workitems);
EXPORT_SYMBOL(md_wait_workitem);
EXPORT_SYMBOL(md_teardown_workitem_support);
EXPORT_SYMBOL(md_sync_handler);
EXPORT_SYMBOL(md_exit_safemode);
EXPORT_SYMBOL(md_init_workitem);
EXPORT_SYMBOL(md_queue_workitem);
EXPORT_SYMBOL(md_find_workitem);
EXPORT_SYMBOL(md_member_type_string);
#else
EXPORT_NO_SYMBOLS;
#endif

MODULE_LICENSE("GPL");
