/*
 * Copyright (c) 2001, 2002, 2004 Adaptec Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions, and the following disclaimer,
 *    without modification.
 * 2. Redistributions in binary form must reproduce at minimum a disclaimer
 *    substantially similar to the "NO WARRANTY" disclaimer below
 *    ("Disclaimer") and any redistribution must be conditioned upon
 *    including a substantially similar Disclaimer requirement for further
 *    binary redistribution.
 * 3. Neither the names of the above-listed copyright holders nor the names
 *    of any contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * Alternatively, this software may be distributed under the terms of the
 * GNU General Public License ("GPL") version 2 as published by the Free
 * Software Foundation.
 *
 * NO WARRANTY
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGES.
 */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/config.h>
#include <linux/raid/emd/md.h>
#include <asm/div64.h>
#include "md_metadata_adaptec.h"

/*
 * DPT RAID table is 8K, and the reserved block is 512 bytes.  These two
 * blocks are at the very end of the disk -- they take up the last 17 512
 * byte sectors of the disk.  The reserved block is the last block of the
 * disk, and contains a pointer to the start of the raid table, which
 * generally takes up the 8KB before the reserved block.
 */
#define ADAPTEC_ASR_RT_SIZE	8192
#define ADAPTEC_ASR_RB_SIZE	512

/*
 * The priority this module returns
 * when claiming a device.
 */
#define	MD_ASR_CLAIM_PRIORITY	1

/*
 * The "penalty" added to the priority of
 * all devices that are not marked bootable.
 */
#define	MD_ASR_NON_BOOT_PRIORITY 0x8000

/*
 * Modular arithmetic macros to deal with ASR RAID table sequence numbers,
 * which are 16 bit unsigned integers that can wrap.
 */
#define	ASR_SEQ_LT(a,b)		((int16_t)((a)-(b)) < 0)
#define	ASR_SEQ_LEQ(a,b)	((int16_t)((a)-(b)) <= 0)
#define	ASR_SEQ_GT(a,b)		((int16_t)((a)-(b)) > 0)
#define	ASR_SEQ_GEQ(a,b)	((int16_t)((a)-(b)) >= 0)

#ifdef MD_DEBUG
#define MD_ASR_DEBUG(A)	printk A
#else
#define MD_ASR_DEBUG(A)
#endif

struct md_asr_meta_info {
	struct asr_reservedblock *asr_rb;
	struct asr_raidtable	 *asr_rt;
	kdev_t			  kdev;
};

typedef enum {
	MD_ASR_MD_DIRTY		= 0x1,	/*
					 * Meta-data should be written
					 * at shutdown.  Used for things
					 * like checkpoints that we don't
					 * bother to commit to media on
					 * every change.
					 */
	MD_ASR_MD_NEED_SYNC	= 0x2,	/*
					 * Meta-data should be written
					 * immediately.
					 */
	MD_ASR_MD_SYNC_INPROG	= 0x4,	/*
					 * A Meta-data update request is
					 * being processed.
					 */
	MD_ASR_MD_SYNC_ALLOWED	= 0x8	/*
					 * We defer performing any meta-data
					 * sync operations until the array
					 * has been opened for write operation
					 * or we are degraded and start a
					 * rebuild operation or allocate a
					 * spare.  This avoids incrementing
					 * the sequence number and invalidating
					 * optimal members until our meta-data
					 * state changes.  This flag is the
					 * control for this behavior.
					 */
} md_asr_md_flags;

struct md_asr_top_level_array;

/*
 * Information held by all members of an array
 * object hierarchy. 
 */
struct md_asr_member {
	struct md_asr_top_level_array	*tl;
	struct asr_raid_configline	*rcl;
	int				flags;
#define MD_ASR_MFLAGS_PUT_AFTER_SYNC	0x1
};

/*
 * Configuration data held by leaf device objects.
 */
struct md_asr_device {
	struct md_asr_member		member;
	struct asr_reservedblock	asr_rb;
};

/*
 * Configuration data held by top-level array objects.
 */
struct md_asr_top_level_array {
	struct md_asr_member		 member;
	md_asr_md_flags			 flags;
	int				 num_levels;
	mddev_t				*spare_pool;
	mdk_workitem_t			 sync_md_workitem;
	spinlock_t			 rt_lock;
	struct asr_raidtable		*asr_rt;
	struct asr_raidtable		*sync_rt;
};

#define MD_ASR_WORKOP_SYNC 0x1

#define	member_to_asr_member(member) \
	((struct md_asr_member *)member->metadata_private)

#define	mddev_to_top_level_array(mddev) \
	((struct md_asr_top_level_array *)mddev->member.metadata_private)

#define	mddev_to_asr_member(mddev) \
	((struct md_asr_member *)mddev->member.metadata_private)

#define	rdev_to_device(rdev) \
	((struct md_asr_device *)rdev->member.metadata_private)

#define get_rb(rdev) \
	(&rdev_to_device(rdev)->asr_rb)

mdk_workitem_support_t	md_asr_syncer;

static int md_asr_work_control(mdk_member_t *member, int op, void *arg);
static int md_asr_sync_minfo(struct md_asr_meta_info *minfo);
static int md_asr_sync_handler(mddev_t *mddev, mdk_workitem_t *workitem);
static void md_asr_schedule_sync(struct md_asr_top_level_array *tl_array,
				 int wait, int enable_sync);
static mddev_t *md_asr_find_mddev(mdk_member_t *parent, uint32_t raidmagic);
static int md_asr_check_for_metadata(struct md_asr_meta_info *);
static int md_asr_dev_op_attach(dev_t dev);
static int md_asr_add_member(mddev_t *array, struct md_asr_meta_info *minfo, 
			     int allow_collisions);
static int md_asr_add_spare(mddev_t *array, struct md_asr_meta_info *minfo);
static int md_asr_find_toplevel(struct asr_raidtable *asr_rt);
static int md_asr_find_member(struct asr_raidtable *asr_rt, uint32_t magic,
			      int type, int *parent_index,
			      int *component_index);
static sector_t md_asr_get_member_disk_capacity(mdk_member_t *member); 
static mdk_rdev_t *md_asr_find_spare_by_magic(uint32_t sparemagic);
static mddev_t *md_asr_mddev_alloc(mdk_member_t *parent,
				   struct asr_raidtable *asr_rt,
				   struct asr_raid_configline *asr_rcl,
				   mdk_msubtype_t subtype, int priority);
static mdk_rdev_t *md_asr_rdev_alloc(mddev_t *array,
				     struct asr_reservedblock *asr_rb,
				     struct asr_raid_configline *asr_rcl,
				     kdev_t dev, mdk_msubtype_t subtype,
				     int priority, int allow_collisions);
static int md_asr_update_array_state(mddev_t *array);
static void md_asr_mark_missing_members(mddev_t *mddev);
static int md_asr_validate_member(mdk_member_t *member,
				  struct asr_raidtable *asr_rt);
static int md_asr_apply_rt(mddev_t *toplevel_array, mdk_member_t *member,
			   struct asr_raidtable *asr_rt);
static void md_asr_change_magic(mddev_t *toplevel_array, mdk_member_t *member);
static int md_asr_validate_member(mdk_member_t *member,
				  struct asr_raidtable *asr_rt);
static int md_asr_member_to_array_state(struct md_asr_member *asr_member,
					mdu_mstate_t md_state,
					uint8_t *adaptec_state);
static int md_asr_array_to_member_state(int adaptec_state,
					mdu_mstate_t *md_state);
static int md_asr_member_to_disk_state(mdu_mstate_t md_state,
				       uint8_t *adaptec_state);
static int md_asr_disk_to_member_state(mddev_t *mddev, int adaptec_state,
				       mdu_mstate_t *md_state);
static int md_asr_raidtype_to_pers(uint8_t asr_raidtype);
static int md_asr_raidtype_to_level(uint8_t asr_raidtype);
static int md_asr_level_to_asr_raidtype(int level, uint8_t *asr_raidtype);
static const char *md_asr_raidtype_string(int raidtype);
static void md_asr_dump_rb(struct asr_reservedblock *asr_rb);
static void md_asr_dump_rcl(struct asr_raid_configline *rcl, int index);
static void md_asr_dump_rt(struct asr_raidtable *asr_rt);
static void md_asr_dump_minfo( struct md_asr_meta_info *minfo);

static int md_asr_dev_op(kdev_t, mdk_dev_op_t, void *);
static int md_asr_member_op(mdk_member_t *, int, void *);

static int md_asr_member_op_enter_safemode(mddev_t *mddev);
static int md_asr_member_op_exit_safemode(mddev_t *mddev);
static int md_asr_member_op_get_checkpoint(mddev_t *mddev,
					   mdk_checkpoint_arg_t *);
static int md_asr_member_op_set_checkpoint(mddev_t *mddev,
					   mdk_checkpoint_arg_t *);
static int md_asr_member_op_set_state(mdk_member_t *member,
				      mdk_set_state_arg_t *sstate);
static uint32_t md_asr_next_target_id(void);
static int
md_asr_locate_spare(mdk_member_t *member, int op, 
		    mdu_disk_info_t *spare_disk_info, mddev_t **spare_mddev, 
		    mdk_rdev_t **spare_rdev, int *spare_count);
static int md_asr_member_op_get_spares(mdk_member_t *member, 
		            mdu_spare_disk_info_t *spare_info);
static int md_asr_member_op_create_spare(mdk_member_t *member, 
				      mdu_disk_info_t *spare_info);
static int md_asr_member_op_activate_spare(mdk_member_t *member);
static int md_asr_member_op_delete_spare(mdk_member_t *member);
static int md_asr_member_op_set_array_info(mdk_member_t *member, 
					   mdu_array_info_t *array_k);
static int md_asr_member_op_add_new_disk(mddev_t *mddev, 
					 mdu_disk_info_t *disk_info);
static int md_asr_member_op_release(mdk_member_t *member);

static mdk_metapersonality_t md_asr_pers = {
 	.name		= "Adaptec ASR",
	.owner		= THIS_MODULE,
	.dev_op		= md_asr_dev_op,
	.member_op	= md_asr_member_op,
	.refcount	= ATOMIC_INIT(1)
};

static int
md_asr_minfo_alloc(kdev_t kdev, struct md_asr_meta_info **pminfo)
{
	struct md_asr_meta_info *minfo;

	minfo = kmalloc(sizeof(*minfo), GFP_KERNEL);
	if (minfo == NULL)
		return (-ENOMEM);
	minfo->asr_rt = kmalloc(sizeof(*minfo->asr_rt), GFP_KERNEL);
	if (minfo->asr_rt == NULL) {
		kfree(minfo);
		return (-ENOMEM);
	}
	memset(minfo->asr_rt, 0, sizeof(*minfo->asr_rt));

	minfo->asr_rb = kmalloc(sizeof(*minfo->asr_rb), GFP_KERNEL);
	if (minfo->asr_rb == NULL) {
		kfree(minfo->asr_rt);
		kfree(minfo);
		return (-ENOMEM);
	}
	memset(minfo->asr_rb, 0, sizeof(*minfo->asr_rb));
	minfo->kdev = kdev;
	*pminfo = minfo;
	return (0);
}

static void
md_asr_minfo_free(struct md_asr_meta_info *minfo)
{
	if (minfo->asr_rt != NULL)
		kfree(minfo->asr_rt);
	if (minfo->asr_rb != NULL)
		kfree(minfo->asr_rb);
	kfree(minfo);
}

static int
md_asr_dev_op(kdev_t dev, mdk_dev_op_t op, void *arg)
{
	MD_ASR_DEBUG(("md_asr_dev_op called\n"));
	switch (op) {
	case MDK_DEVOP_CLAIM:
	{
		struct md_asr_meta_info *minfo;
		int retval;

		MD_ASR_DEBUG(("md_asr_dev_op: MDK_DEVOP_CLAIM\n"));
		retval = md_asr_minfo_alloc(dev, &minfo);
		if (retval != 0)
			return (retval);
		retval = md_asr_check_for_metadata(minfo);
		if (retval == 0){
			*(int *)arg = MD_ASR_CLAIM_PRIORITY;
		} else {
			*(int *)arg = -1;
		}
		md_asr_minfo_free(minfo);
		return (retval);
	}
	case MDK_DEVOP_ATTACH:
		return (md_asr_dev_op_attach(dev));
	case MDK_DEVOP_WIPE_METADATA:
	{
		struct md_asr_meta_info *minfo;
		struct gendisk *gd;
		sector_t drive_capacity;
		int retval;


		MD_ASR_DEBUG(("md_asr_dev_op: MDK_DEVOP_WIPE_METADATA\n"));
		retval = md_asr_minfo_alloc(dev, &minfo);
		if (retval != 0)
			return (retval);
		
		gd = get_gendisk(minfo->kdev);
		if (gd == NULL){
			MD_ASR_DEBUG(("md_asr_dev_op: gendisk is NULL\n"));
			md_asr_minfo_free(minfo);
			return (-ENODEV);
		}
		drive_capacity = gd->part[MINOR(minfo->kdev)].nr_sects;
		minfo->asr_rb->raidtbl = htonl(drive_capacity - 
					       (RTBLBLOCKS + 1));
		retval = md_asr_sync_minfo(minfo);
		md_asr_minfo_free(minfo);
		return retval;
	}
 	case MDK_DEVOP_GET_FREESIZE:
 	{
		uint64_t freesize;
		struct gendisk *gd;
 
 		gd = get_gendisk(dev);
 		if (gd == NULL){
 			MD_ASR_DEBUG(("md_asr_dev_op: gendisk is NULL\n"));
 			return (-ENODEV);
 		}
 		
 		/* last 17 sectors are reserved for Adaptec ASR metadata */
		freesize = (uint64_t)gd->part[MINOR(dev)].nr_sects * 512;
		do_div(freesize, 100000000);
		freesize *= 100000000;
		freesize /= 512;
		/* Reserve 17 sectors for meta-data. */
		freesize -= 17;
 		*((sector_t *)arg) = freesize;
 		return 0;
 	}
	case MDK_DEVOP_INITIALIZE:
	default:
		return (-EOPNOTSUPP);
	}
}

static int
md_asr_member_op(mdk_member_t *member, int op, void *arg)
{
	switch (op) {
	case MDK_MEMBER_OP_ENTER_SAFEMODE:
	case MDK_MEMBER_OP_EXIT_SAFEMODE:
	{
		mddev_t *mddev;

		/*
		 * We only care about safemode events on
		 * array objects.
		 */
		if (member->type != MDK_MTYPE_MDDEV)
			return (0);

		mddev = to_mddev(member);
		if (op == MDK_MEMBER_OP_ENTER_SAFEMODE)
			return md_asr_member_op_enter_safemode(mddev);
		else
			return md_asr_member_op_exit_safemode(mddev);
		/* NOTREACHED */
	}
	case MDK_MEMBER_OP_GET_CHECKPOINT:
	case MDK_MEMBER_OP_SET_CHECKPOINT:
	{
		mdk_checkpoint_arg_t *checkpoint;
		mddev_t *mddev;

		mddev = to_mddev(member);
		checkpoint = (mdk_checkpoint_arg_t *)arg;
		if (op == MDK_MEMBER_OP_GET_CHECKPOINT)
			return md_asr_member_op_get_checkpoint(mddev,
								checkpoint);
		else
			return md_asr_member_op_set_checkpoint(mddev,
								checkpoint);
	}
	case MDK_MEMBER_OP_SET_STATE:
		return md_asr_member_op_set_state(member,
						  (mdk_set_state_arg_t *)arg);
	case MDK_MEMBER_OP_SET_ARRAY_INFO:
		return md_asr_member_op_set_array_info(member, arg);
	case MDK_MEMBER_OP_ADD_NEW_DISK:
		return md_asr_member_op_add_new_disk(to_mddev(member), arg);
	case MDK_MEMBER_OP_RELEASE:
		return md_asr_member_op_release(member);
	case MDK_MEMBER_OP_GET_SPARES:
		return md_asr_member_op_get_spares(member, arg);
	case MDK_MEMBER_OP_CREATE_SPARE:
		return md_asr_member_op_create_spare(member, arg);
	case MDK_MEMBER_OP_ACTIVATE_SPARE:
		return md_asr_member_op_activate_spare(member);
	case MDK_MEMBER_OP_DELETE_SPARE:
		return md_asr_member_op_delete_spare(member);
	
	case MDK_MEMBER_OP_SET_ARRAY_ACTION:
	case MDK_MEMBER_OP_SET_USER_ACTION:
	case MDK_MEMBER_OP_GET_ARRAY_ACTION:
	case MDK_MEMBER_OP_GET_USER_ACTION:
	default:
		return (-EOPNOTSUPP);
	}
}

static int
md_asr_work_control(mdk_member_t *member, int op, void *arg)
{
	mdk_workitem_t *work;
	mddev_t *mddev;
	struct md_asr_top_level_array *tl_array;
	struct asr_raidtable *asr_rt;
	int error;

	work = (mdk_workitem_t *)arg;
	mddev = (mddev_t *)work->work_data;
	member = to_member(mddev);
	tl_array = mddev_to_top_level_array(mddev);
	asr_rt = tl_array->asr_rt;
	error = 0;

	switch (op) {
	case MDK_WORKITEM_OP_OPEN:
	{
		uint8_t *pbyte;
		int i;
		uint16_t elmcnt;
		uint16_t new_seq;
		uint16_t new_cksum;

		spin_lock_irq(&tl_array->rt_lock);
		if ((tl_array->flags & MD_ASR_MD_NEED_SYNC) == 0
		 || (tl_array->flags & MD_ASR_MD_SYNC_ALLOWED) == 0) {
			spin_unlock_irq(&tl_array->rt_lock);
			error = -ESTALE;
			break;
		}
		new_seq = htons(ntohs(tl_array->member.rcl->raidseq) + 1);
		elmcnt = ntohs(asr_rt->elmcnt);
		for (i = 0; i < elmcnt; i++)
			asr_rt->ent[i].raidseq = new_seq;
		memcpy(tl_array->sync_rt, asr_rt, sizeof(*asr_rt));
		tl_array->flags &= ~(MD_ASR_MD_DIRTY|MD_ASR_MD_NEED_SYNC);
		spin_unlock_irq(&tl_array->rt_lock);

		/*
		 * Update checksum.  It only covers the configured elements.
		 */
		pbyte = (uint8_t *)asr_rt->ent;
		new_cksum = 0;
		for (i = 0; i < (sizeof(asr_rt->ent[0]) * elmcnt); i++)
			new_cksum += pbyte[i];
		/* 16-bit checksum */
		new_cksum &= 0xffff;
		tl_array->sync_rt->rchksum = ntohs(new_cksum);
		break;
	}
	case MDK_WORKITEM_OP_CLOSE:
	{
		/* NO-OP */
		break;
	}
	case MDK_WORKITEM_OP_FREE:
	{
		int requeue;

		spin_lock_irq(&tl_array->rt_lock);
		if ((tl_array->flags & MD_ASR_MD_NEED_SYNC) != 0
		 && (tl_array->flags & MD_ASR_MD_SYNC_ALLOWED) != 0) {
			requeue = 1;
		} else {
			tl_array->flags &= ~MD_ASR_MD_SYNC_INPROG;
			requeue = 0;
		}
		spin_unlock_irq(&tl_array->rt_lock);
		if (requeue)
			error = md_queue_workitem(&md_asr_syncer,
						  work, /*head_of_queue*/0);
		break;
	}
	default:
		error = -EINVAL;
	}
	return (error);
}

static int
md_asr_sync_minfo(struct md_asr_meta_info *minfo)
{
	struct gendisk *gd;
	sector_t rb_location;
	int retval;

	gd = get_gendisk(minfo->kdev);
	if (gd == NULL){
		MD_ASR_DEBUG(("md_asr_sync_minfo: gendisk is NULL\n"));
		return (-ENODEV);
	}

	rb_location = gd->part[MINOR(minfo->kdev)].nr_sects - 1;
	retval = md_write_blocks(minfo->kdev, rb_location,
				 ADAPTEC_ASR_RB_SIZE / 512,
				 (uint8_t *)minfo->asr_rb);
	if (retval != 0)
		return (retval);

	retval = md_write_blocks(minfo->kdev, ntohl(minfo->asr_rb->raidtbl),
			       	 ADAPTEC_ASR_RT_SIZE / 512,
				 (uint8_t *)minfo->asr_rt);
	return (retval);
}

static int
md_asr_sync_handler(mddev_t *mddev, mdk_workitem_t *workitem)
{
	struct md_asr_meta_info minfo;
	mdk_member_t *array;
	struct md_asr_top_level_array *tl_array;
	mdk_member_t *child;
	mdk_member_t *gchild;
	struct list_head *tmp;
	struct list_head *tmp2;
	u_long flags;
	u_long flags2;

	mddev = (mddev_t *)workitem->work_data;
	tl_array = mddev_to_top_level_array(mddev);
	minfo.asr_rt = tl_array->sync_rt;
	MD_ASR_DEBUG(("md_asr_sync_handler: %p syncing Checkpoint %d\n",
		      mddev, tl_array->member.rcl->curAppBlock));
	array = to_member(mddev);

	ITERATE_CHILDREN(array, child, tmp, flags) {
		struct md_asr_member *asr_member;
		mdk_rdev_t *rdev;

		if (child->type == MDK_MTYPE_RDEV) {

			rdev = to_rdev(child);
			minfo.kdev = rdev->kdev;
			minfo.asr_rb = get_rb(rdev);
			if (md_asr_sync_minfo(&minfo) == -EIO) {
				md_set_member_state(child, MDU_MSTATE_FAILED);
				MD_ASR_DEBUG(("md_asr_sync_handler: Unable to sync "
				       "metadata. So marking the drive (%d,%d) "
				       "as failed\n", MAJOR(rdev->kdev),
				       MINOR(rdev->kdev)));
			}
			asr_member = member_to_asr_member(child);
			spin_lock_irq(&child->member_lock);
			if (asr_member->flags & MD_ASR_MFLAGS_PUT_AFTER_SYNC) {
				asr_member->flags &=
				    ~MD_ASR_MFLAGS_PUT_AFTER_SYNC;
				member_put(child);
			}
			spin_unlock_irq(&child->member_lock);
			continue;
		}

		ITERATE_CHILDREN(child, gchild, tmp2, flags2) {

			if (gchild->type != MDK_MTYPE_RDEV)
				BUG();

			rdev = to_rdev(gchild);
			minfo.kdev = rdev->kdev;
			minfo.asr_rb = get_rb(rdev);
			if (md_asr_sync_minfo(&minfo) == -EIO) {
				md_set_member_state(gchild, MDU_MSTATE_FAILED);
				MD_ASR_DEBUG(("md_asr_sync_handler: Unable to sync "
				       "metadata. So marking the drive (%d,%d) "
				       "as failed\n", MAJOR(rdev->kdev),
				       MINOR(rdev->kdev)));
			}
			asr_member = member_to_asr_member(gchild);
			spin_lock_irq(&gchild->member_lock);
			if (asr_member->flags & MD_ASR_MFLAGS_PUT_AFTER_SYNC) {
				asr_member->flags &=
				    ~MD_ASR_MFLAGS_PUT_AFTER_SYNC;
				member_put(gchild);
			}
			spin_unlock_irq(&gchild->member_lock);
		}
	}
	return (0);
}

static void
md_asr_schedule_sync(struct md_asr_top_level_array *tl_array,
		     int wait, int enable_sync)
{
	u_long flags;
	int sched_work;

	sched_work = 0;
	spin_lock_irqsave(&tl_array->rt_lock, flags);
	tl_array->flags |= MD_ASR_MD_DIRTY|MD_ASR_MD_NEED_SYNC;
	if (enable_sync)
		tl_array->flags |= MD_ASR_MD_SYNC_ALLOWED;
	if ((tl_array->flags & MD_ASR_MD_SYNC_INPROG) == 0) {
		tl_array->flags |= MD_ASR_MD_SYNC_INPROG;
		sched_work = 1;
	}
	spin_unlock_irqrestore(&tl_array->rt_lock, flags);
	if (sched_work) {
		md_queue_workitem(&md_asr_syncer,
				  &tl_array->sync_md_workitem,
				  /*head_of_queue*/0);
	}
	if (!wait)
		return;

	for (;;) {
		int syncing;

		md_wait_workitem(&md_asr_syncer, &tl_array->sync_md_workitem);
		spin_lock_irqsave(&tl_array->rt_lock, flags);
		syncing = (tl_array->flags & MD_ASR_MD_SYNC_INPROG);
		spin_unlock_irqrestore(&tl_array->rt_lock, flags);
		if (!syncing)
			break;
	}
}

static mddev_t *
md_asr_find_mddev(mdk_member_t *parent, uint32_t raidmagic)
{
	mddev_t *mddev;
	mdk_member_t *child;
	struct md_asr_member *asr_member;
	struct list_head *tmp;
	u_long flags;

	mddev = NULL;
	ITERATE_CHILDREN(parent, child, tmp, flags) {

		if (child->metapers != &md_asr_pers) 
			continue;

		mddev = to_mddev(child);
		asr_member = mddev_to_asr_member(mddev);
		if (ntohl(asr_member->rcl->raidmagic) == raidmagic)
			return (mddev);
	}
	return (NULL);
}

/*
 * Read the reserved block and the RAID table off of the disk and check
 * both to make sure the metadata is okay.  Return 1 for success, 0 for
 * failure.
 */
static int
md_asr_check_for_metadata(struct md_asr_meta_info *minfo)
{
	struct gendisk *gd;
	struct asr_reservedblock *asr_rb;
	struct asr_raidtable *asr_rt;
	unsigned int rt_cksum;
	uint8_t *rclptr;
	sector_t rb_location;
	u_int elmcnt;
	int retval, i;

	retval = 0;

	MD_ASR_DEBUG(("md_check_for_sb_adaptec called for %s %d/%d\n",
		      kdevname(minfo->kdev), MAJOR(minfo->kdev),
		      MINOR(minfo->kdev)));

	asr_rb = minfo->asr_rb;
	asr_rt = minfo->asr_rt;

	/*
	 * The reserved block, if it is around, will be on the last sector
	 * of the disk.
	 */
	gd = get_gendisk(minfo->kdev);
	if (gd == NULL){
		MD_ASR_DEBUG(("md_asr_check_for_metadata: gendisk is NULL\n"));
		return (-ENODEV);
	}

	rb_location = gd->part[MINOR(minfo->kdev)].nr_sects - 1;

	retval = md_read_blocks(minfo->kdev, rb_location,
				ADAPTEC_ASR_RB_SIZE / 512, (uint8_t *)asr_rb);
	if (retval != 0) {		   
		/* print an error here? */
		MD_ASR_DEBUG(("md_asr_check_for_metadata: "
			      "md_read_blocks failed %d\n", retval));
		return (retval);
	}

	if ((ntohl(asr_rb->b0idcode) != B0RESRVD)
	 || (ntohl(asr_rb->smagic) != SVALID)) {
		MD_ASR_DEBUG(("md_asr_check_for_metadata: bailing out, "
			      "b0idcode= %x, smagic= %x\n",
			      ntohl(asr_rb->b0idcode), ntohl(asr_rb->smagic)));
		return (-ENODEV);
	}

	retval = md_read_blocks(minfo->kdev, ntohl(asr_rb->raidtbl),
			       ADAPTEC_ASR_RT_SIZE / 512, (uint8_t *)asr_rt);
	if (retval != 0) {
		/* print error? */
		MD_ASR_DEBUG(("md_check_for_sb_adaptec: "
			      "2nd md_read_blocks failed\n"));
		return (retval);
	}

	if (ntohl(asr_rt->ridcode) != RVALID2) {
		MD_ASR_DEBUG(("md_asr_check_for_metadata: bailing out, "
			      "ridcode= %d\n", asr_rt->ridcode));
		return (-ENODEV);
	}

	rclptr = (uint8_t *)asr_rt->ent;
	rt_cksum = 0;
	elmcnt = ntohs(asr_rt->elmcnt);

	if (elmcnt > RCTBL_MAX_ENTRIES) {
		MD_ASR_DEBUG(("md_asr_check_for_metadata: RAID table "
			      "invalid entry count %d\n", elmcnt));
		return (-ENODEV);
	}

	/*
	 * The checksum isn't calculated over the entire RAID table, only
	 * over the filled-in config line elements.
	 */
	for (i = 0; i < (sizeof(asr_rt->ent[0]) * elmcnt); i++)
		rt_cksum += rclptr[i];
	/* 16-bit checksum */
	rt_cksum &= 0xffff;

	if (ntohs(asr_rt->rchksum) != rt_cksum) {
		MD_ASR_DEBUG(("md_asr_check_for_metadata: RAID table "
			      "checksum %#x doesn't match calculated "
			      "checksum %#x\n",
			      ntohs(asr_rt->rchksum), rt_cksum));
		return (-ENODEV);
	}

	return (0);
}

/*
 * Add a disk that we have successfully bid for to our
 * array configuration.
 */
static int
md_asr_dev_op_attach(kdev_t dev)
{
	struct md_asr_meta_info *minfo;
	struct md_asr_top_level_array *tl_array;
	struct asr_raidtable *asr_rt;
	struct asr_raid_configline *toplevel_rcl;
	mddev_t *array;
	mdk_member_t *parent;
	uint32_t raidmagic;
	uint16_t raidseq;
	int toplevel;
	int retval;

	MD_ASR_DEBUG(("md_add_to_array_adaptec called\n"));

	retval = md_asr_minfo_alloc(dev, &minfo);
	if (retval != 0)
		return (retval);
	retval = md_asr_check_for_metadata(minfo);

	if (retval != 0) {
		md_asr_minfo_free(minfo);
		return (retval);
	}

	md_asr_dump_minfo(minfo);

	asr_rt = minfo->asr_rt;

	/*
	 * Determine the toplevel array to which this device,
	 * or the logical array parent of this device, belongs.
	 */
	toplevel = md_asr_find_toplevel(asr_rt);
	toplevel_rcl = &asr_rt->ent[toplevel];
	raidmagic = ntohl(toplevel_rcl->raidmagic);
	raidseq = ntohs(toplevel_rcl->raidseq);

	/*
	 * Spare handling differs between HostRAID and processor
	 * RAID controller generated meta-data.
	 * 
	 * HostRAID supports dedicated and pool spares.  The RAID array
	 * members indicate their dedicated spare via a non-zero
	 * sparedrivemagic in the raidtable of all of their members. The
	 * dedicated spare also points back to its associated raid array
	 * by setting its raidmagic field in the raidtable to the magic
	 * of the raid array.
	 *
	 * Processor based cards do not have the concept of dedicated
	 * spares.  The controller instead combines all drives into a single
	 * spare pool array.  In this configuration, the sparedrivemagic
	 * field and the raidmagic field in the raidtable will always be 0.
	 *
	 * In both cases, the spare drives are "pooled" in that a logical
	 * spare array object is listed in the metadata as the parent of
	 * the spare.  To simplify the handling of both cases, even
	 * dedicated spares are made part of a "pool" with the parent array
	 * object pointing to that pool.
	 * 
	 * RAID arrays are parented by the md core's toplevel_array object.
	 * Spare pools are parented by the spare_arrays object.  Modify the
	 * parent used to search for an existing array accordingly.
	 */
	if (toplevel_rcl->raidtype == ASR_RAIDSPR)
		parent = &spare_arrays;
	else
		parent = &toplevel_arrays;
	if (raidmagic == 0xFFFFFFFFUL)
		/* 
		 * new mddev has to be allocated for each and 
		 * every pool spare as they dont share a common
		 * raid table. 
		 */
		array = NULL;
	else
		array = md_asr_find_mddev(parent, raidmagic);
	if (array == NULL) {
		mdk_msubtype_t subtype;
		int priority;

allocate_array:
		/*
		 * This is the first member to arrive for an array.
		 * Create an array object to house the new array
		 * configuration.
		 */
		if (toplevel_rcl->raidtype == ASR_RAIDSPR) {
			priority = MDK_PRIORITY_UNKNOWN;
			subtype = MDK_MSUBTYPE_POOL_SPARE;
			if (asr_rt->raidmagic != 0
			 || asr_rt->raidmagic != 0xFFFFFFFFUL)
				subtype = MDK_MSUBTYPE_DEDICATED_SPARE;
		} else {
			subtype = MDK_MSUBTYPE_TOPLEVEL;
			priority = ntohl(toplevel_rcl->raidid);
			if ((asr_rt->irocFlags & ASR_IF_BOOTABLE) == 0)
				priority |= 0x8000;
		}
		array = md_asr_mddev_alloc(parent, asr_rt, toplevel_rcl,
					   subtype, priority);
		if (array == NULL) {
			md_asr_minfo_free(minfo);
			return (-ENOMEM);
		}

		tl_array = mddev_to_top_level_array(array);

		/*
		 * Hold an extra ref-count on the mddev so the
		 * refcounting scheme is the code can treat it
		 * the same as a pre-exisiting array.
		 */
		member_get(to_member(array));
	} else {
		mdk_member_t *member;
		uint16_t cur_raidseq;

		tl_array = mddev_to_top_level_array(array);
		cur_raidseq = ntohs(tl_array->member.rcl->raidseq);
		member = to_member(array);

		if (ASR_SEQ_GT(raidseq, cur_raidseq)
			&& toplevel_rcl->raidtype != ASR_RAIDSPR) {
			/* 
			 * If the newly arrived drive has a newer sequence
			 * number than that reported in an existing array, vet
			 * the already attached members against the newer
			 * meta-data.  Members that no longer belong in the
			 * array will be removed accordingly.
			 */
			retval = mddev_lock(array);
			if (retval != 0) {
				member_put(member);
				goto allocate_array;
			}
			retval = md_asr_apply_rt(array, member, asr_rt);
			mddev_unlock(array);
			if (retval != 0) {
				/*
				 * md_asr_apply_rt will fail if the
				 * already exising array has been exported
				 * to the OS.  In this case, md_asr_apply_rt
				 * will change the raidmagic of the "stale"
				 * array so as not to conflict with the
				 * current array information.  Create a
				 * new array object in this case.
				 */
				member_put(member);
				goto allocate_array;
			}

		} else if (ASR_SEQ_LT(raidseq, cur_raidseq)
			&& toplevel_rcl->raidtype != ASR_RAIDSPR) {
			uint32_t drivemagic;
			int index;

			/* 
			 * If this member has an older sequencer number,
			 * only allow it into the array if it still exists
			 * there.  Otherwise, we fail the attach so this
			 * drive can be accessed as a "single".
			 */
			drivemagic = ntohl(minfo->asr_rb->drivemagic);
			index = md_asr_find_member(tl_array->asr_rt,
						   drivemagic, FWP,
						   NULL, NULL);
			if (index < 0) {
				md_asr_minfo_free(minfo);
				member_put(&array->member);
				return (-EINVAL);
			}
		}
	}

	if (toplevel_rcl->raidtype == ASR_RAIDSPR)
		retval = md_asr_add_spare(array, minfo);
	else
		retval = md_asr_add_member(array, minfo, /*no collisions*/0);

	member_put(&array->member);
	md_asr_minfo_free(minfo);

	return (retval);
}

/*
 * Handle notification of our array transitioning into
 * Read/Write mode or, in MD parlance, "exiting safemode".
 */
static int
md_asr_member_op_exit_safemode(mddev_t *mddev)
{
	struct md_asr_member *asr_member;
	struct md_asr_top_level_array *tl_array;
	u_long flags;
	int sched_write;

	/*
	 * Update the sequence number and mark the
	 * array meta-data as open.  We use this
	 * status to detect improper shutdown of
	 * the array.  Incrementing the sequence number
	 * automaticallly prevents "missing" members
	 * that arrive late from being accepted into
	 * the array.
	 */
	MD_ASR_DEBUG(("ASR Exit Safemode\n"));
	asr_member = mddev_to_asr_member(mddev);
	tl_array = asr_member->tl;
	sched_write = 0;
	spin_lock_irqsave(&tl_array->rt_lock, flags);
	if (tl_array->asr_rt->dirty != ARRAY_STATE_DIRTY) {
		tl_array->asr_rt->dirty = ARRAY_STATE_DIRTY;
		sched_write = 1;
	}
	spin_unlock_irqrestore(&tl_array->rt_lock, flags);
	if (sched_write)
		md_asr_schedule_sync(tl_array, /*wait4complete*/1, /*enable*/1);
	return (0);
}

/*
 * Handle notification of our array transitioning to Read-Only
 * mode or, in MD parlance, "entering safemode".
 */
static int
md_asr_member_op_enter_safemode(mddev_t *mddev)
{
	struct md_asr_member *asr_member;
	struct md_asr_top_level_array *tl_array;
	u_long flags;
	int sched_write;

	/*
	 * Update the sequence number and mark the
	 * array meta-data as closed.  We increment
	 * the sequence number so that we are not
	 * confused by any members that for some
	 * reason are not written before we lose
	 * power or are incorrectly shutdown.
	 */
	MD_ASR_DEBUG(("ASR Enter Safemode\n"));
	asr_member = mddev_to_asr_member(mddev);
	tl_array = asr_member->tl;
	sched_write = 0;
	spin_lock_irqsave(&tl_array->rt_lock, flags);
	if (tl_array->asr_rt->dirty == ARRAY_STATE_DIRTY) {
		tl_array->asr_rt->dirty = ARRAY_STATE_OK;
		sched_write = 1;
	}
	spin_unlock_irqrestore(&tl_array->rt_lock, flags);
	if (sched_write)
		md_asr_schedule_sync(tl_array, /*wait*/1, /*enable*/1);
	return (0);
}

static int
md_asr_member_op_get_checkpoint(mddev_t *mddev, mdk_checkpoint_arg_t *chkpt_arg)
{
	struct md_asr_member *asr_member;
	struct md_asr_top_level_array *tl_array;
	mdk_member_t *array;
	u_long flags;

	array = to_member(mddev);
	asr_member = mddev_to_asr_member(mddev);
	tl_array = asr_member->tl;
	spin_lock_irqsave(&tl_array->rt_lock, flags);
	chkpt_arg->checkpoint = ntohl(asr_member->rcl->curAppBlock);
	spin_unlock_irqrestore(&tl_array->rt_lock, flags);
#if 0	
	MD_ASR_DEBUG(("md_asr_member_op_get_checkpoint: %p RCL %d "
		      "Chkpoint %ld\n", mddev,
		      asr_member->rcl - tl_array->asr_rt->ent,
		      chkpt_arg->checkpoint));
#endif	
	return (0);
}

static sector_t
md_asr_get_member_disk_capacity(mdk_member_t *member)
{
	mdk_member_t *disk;
	sector_t size;
	u_long flags;

	size = 0;
	spin_lock_irqsave(&member->member_lock, flags);
	if (list_empty(&member->child_set) == 0) {
		disk = list_entry(member->child_set.next,
				  mdk_member_t, peer_links);
		size = disk->data_size;
	}
	spin_unlock_irqrestore(&member->member_lock, flags);
	return (size);
}

static int
md_asr_locate_spare(mdk_member_t *member, int op, 
		    mdu_disk_info_t *spare_disk_info, mddev_t **spare_mddev, 
		    mdk_rdev_t **spare_rdev, int *spare_count)
{
	struct list_head *tmp;
	struct md_asr_member *asr_array_member, *asr_spare_member;
	struct md_asr_top_level_array *tl_array;
	struct asr_raidtable *asr_spare_rt, *asr_array_rt;
	mddev_t	*array_mddev, *mddev; 
	mdk_member_t *child;
	mdk_rdev_t *rdev;
	u_long flags;
	int spare_index;
	int found = 0;

		
	tl_array = NULL;
	asr_array_rt = NULL;
	rdev = NULL;
	spare_index = 0;
	array_mddev = NULL;
	mddev = NULL;
	if (member
	 && member->type == MDK_MTYPE_MDDEV
	 && member->metapers == &md_asr_pers) {
		array_mddev = to_mddev(member);
		asr_array_member = mddev_to_asr_member(to_mddev(member));
		if (asr_array_member == NULL)
			return (-ENODEV);
		tl_array = asr_array_member->tl;
		if (tl_array == NULL)
			return -ENODEV;
		asr_array_rt = tl_array->asr_rt;
		if (asr_array_rt->ent[0].raidtype != ASR_RAID1) {
			MD_ASR_DEBUG(("emd:No spares available:"
				      "Invalid RAID Level\n"));
			return (-EINVAL);
		}
	}
	ITERATE_CHILDREN(&spare_arrays, child, tmp, flags) {
		struct list_head *tmp2;
		mdk_member_t *child2;
		u_long flags2;

		if (child->metapers != &md_asr_pers) 
			continue;
		if (child->type != MDK_MTYPE_MDDEV)
			continue;
		/*
		 * If an array has a dedicated spare pool,
		 * only list spares from it.
		 */
		if (tl_array != NULL
		 && tl_array->spare_pool != NULL
		 && tl_array->spare_pool != to_mddev(child))
			continue;

		mddev = to_mddev(child);
		asr_spare_member = mddev_to_asr_member(mddev);
		asr_spare_rt = asr_spare_member->tl->asr_rt;

		ITERATE_CHILDREN(child, child2, tmp2, flags2) {
			if (child2->type != MDK_MTYPE_RDEV)
				continue;
			rdev = to_rdev(child2);
			if (op == MDK_MEMBER_OP_DELETE_SPARE 
			 && member && member->type == MDK_MTYPE_RDEV
			 && member == child2) {
				down_interruptible(&mddev->reconfig_sem);
				*spare_mddev = mddev;
				*spare_rdev = rdev;
				return 0;
			}
				
			if (member && member->type == MDK_MTYPE_MDDEV) {
				if (tl_array != NULL
				 && tl_array->spare_pool != NULL
				 && tl_array->spare_pool == mddev) {
					found = 1;
				} else {
					if (ntohl(asr_spare_rt->ent[0].lcapcty) 
					>= ntohl(asr_array_rt->ent[0].lcapcty))
						found = 1;
				}
			}

			if (op == MDK_MEMBER_OP_ACTIVATE_SPARE 
			 && found == 1) {
				if (mddev->dspare_tag != NULL 
				  && mddev->dspare_tag != array_mddev) {
					found = 0;
					continue;
				} else {
					down_interruptible(
							&mddev->reconfig_sem);
					*spare_mddev = mddev;
					*spare_rdev = rdev;
					return 0;
				}
			}
		
			if (member && member->type == MDK_MTYPE_MDDEV 
			 && found != 1) 
				continue;

			if (spare_disk_info) {
				spare_disk_info[spare_index].major = 
						MAJOR(rdev->kdev);
				spare_disk_info[spare_index].minor = 
						MINOR(rdev->kdev);
				spare_disk_info[spare_index].number =
				       		spare_index;
				spare_disk_info[spare_index].device_type =
				       		MDU_HOT_SPARE_DEVICE;
			}
			spare_index++;
			found = 0;
		}
	}
	if (spare_count)
		*spare_count = spare_index;
	return 0;
}
/*
 * md_asr_member_op_create_spare() 
 * Assigns a drive as either a Global or Dedicated spare. 
 * The routine supports both global as well as dedicated spares. 
 * Dedicated spares are supported by HostRAID BIOS upto v4.10 and
 * not above v4.30 ??. The code for adding a dedicated spare is #ifdef'ed
 * out right now and it needs to be discussed if this support has to be
 * given. Moreover there are limitations in implementing dedicated spares
 * as the exact target id of the spare device must be stamped in the metadata
 * and the EMD driver doesnt have a way of finding the target id.
 *
 * Input:
 *  member represents logical array to which the spare is to be attached.
 *   'member' is non-NULL implies 'Add as Dedicated Spare'. 
 *   'member' is NULL implies 'Add as Global Spare'.
 *   
 * Return:
 *  Status
 */

static int
md_asr_member_op_create_spare(mdk_member_t *member, 
		           mdu_disk_info_t *spare_info)
{
	struct asr_raidtable *asr_spare_rt, *asr_array_rt;
	struct asr_raid_configline *toplevel_rcl, *sublevel_rcl;
	struct asr_reservedblock *asr_rb;
	struct md_asr_meta_info *minfo;
	struct md_asr_member *asr_array_member = NULL, *asr_spare_member;
	struct timeval md_timestamp;
	struct gendisk *gd;
	struct list_head *tmp;
	mddev_t	*array_mddev = NULL, *spare_mddev, *mddev;
	mdk_member_t *spare_member;
	mdk_member_t *child;
	sector_t drive_capacity;
	kdev_t kdev;
	u_long flags;
	uint32_t raidmagic;
	int retval = 0;

	kdev = MKDEV(spare_info->major, spare_info->minor);
	
	retval = md_asr_minfo_alloc(kdev, &minfo);
	if (retval != 0) {
		return (retval);
	}

	gd = get_gendisk(minfo->kdev);
	if (gd == NULL) {
		md_asr_minfo_free(minfo);
		return (-ENODEV);
	}
	
	drive_capacity = gd->part[MINOR(kdev)].nr_sects;

	if (member != NULL) {
		/*
		 * dedicated spare or activating spool or dedicated spare.
		 */
		if (md_asr_get_member_disk_capacity(member) > drive_capacity){
			MD_ASR_DEBUG(("Spare Creation Fail!"
			       "Device (%d,%d) too small for this array ", 
						spare_info->major, 
						spare_info->minor));
			md_asr_minfo_free(minfo);
			return -EINVAL;
		}
		
		array_mddev = to_mddev(member);
		asr_array_member = mddev_to_asr_member(array_mddev);
		asr_array_rt = asr_array_member->tl->asr_rt;

		if (asr_array_rt == NULL) {
			md_asr_minfo_free(minfo);
			return (-ENODEV);
		}
	}
	MD_ASR_DEBUG(("Spare Info: (%d,%d)\n",
		      spare_info->major, spare_info->minor));
	
	asr_rb = minfo->asr_rb;
	memset(asr_rb, 0, sizeof(*asr_rb));
	
	asr_spare_rt = minfo->asr_rt;
	memset(asr_spare_rt, 0, sizeof(*asr_spare_rt));

	/* Populate device specific reserved block */
	asr_rb->b0idcode = htonl(B0RESRVD);
	asr_rb->smagic = htonl(SVALID);
	asr_rb->resver = 0x08;
	asr_rb->raidtbl = htonl(drive_capacity - (RTBLBLOCKS + 1));
	asr_rb->fstrsvrb = asr_rb->raidtbl;
	
	/*
	 * Populate RAID table. 
	 */
	asr_spare_rt->ridcode = htonl((uint32_t)RVALID2);
	asr_spare_rt->rversion = htonl(0x2);
	asr_spare_rt->maxelm = htons(RCTBL_MAX_ENTRIES);
	asr_spare_rt->elmcnt = htons(0x2);
	asr_spare_rt->elmsize = htons(sizeof(struct asr_raid_configline));
	do_gettimeofday(&md_timestamp);
	asr_spare_rt->timestamp = htonl(md_timestamp.tv_sec);

#if MD_ASR_DEDICATED_SPARE_SUPPORT
	if (member == NULL) {
#endif		
		/* Global spare */
		asr_spare_rt->spareid = 0x0;
		asr_spare_rt->actionPriority = 0x0;
		asr_spare_rt->sparedrivemagic = htonl(0x0);
		asr_spare_rt->raidmagic = htonl(0xFFFFFFFFUL);
#if MD_ASR_DEDICATED_SPARE_SUPPORT
	} else {
		/* Dedicated spare */
		asr_spare_rt->spareid = targetid;//MINOR(kdev) | 0x80;TBD
		asr_spare_rt->actionPriority = 0x8;//8/32/64
		
		/*
		 * should use this for FWP also as well as update 
		 * back in asr_array_rt 
		 */
		get_random_bytes(&raidmagic, sizeof(raidmagic));
		asr_spare_rt->sparedrivemagic = htonl(raidmagic);
		
		asr_spare_rt->raidmagic = asr_array_rt->ent[0].raidmagic;
	}
#endif		
	/*
	 * Populate RAID CLine entry for
	 * the toplevel logical array.
	 */	
	toplevel_rcl = &asr_spare_rt->ent[0];
	toplevel_rcl->raidcnt = htons(0x1);

#if MD_ASR_DEDICATED_SPARE_SUPPORT
	if (member == NULL){
#endif		
		/* Global spare */
		toplevel_rcl->raidseq = htons(0x1);
		toplevel_rcl->raidmagic = htonl(0xFFFFFFFFUL);
#if MD_ASR_DEDICATED_SPARE_SUPPORT
	} else {
		/* Dedicated spare */
		toplevel_rcl->raidseq = asr_array_rt->ent[0].raidseq;
		get_random_bytes(&raidmagic, sizeof(raidmagic));
		toplevel_rcl->raidmagic = htonl(raidmagic);
	}
#endif		
	
	toplevel_rcl->raidlevel = FWL;
	toplevel_rcl->raidtype = ASR_RAIDSPR; 

	toplevel_rcl->raidstate = LSU_ARRAY_STATE_OPTIMAL;
		
	/* XXX Not really correct. */
	toplevel_rcl->raidid = htonl(MINOR(kdev));
	toplevel_rcl->lcapcty  = htonl(drive_capacity - (RTBLBLOCKS + 1));
	toplevel_rcl->strpsize = htons(0x10);

	/*
	 * Populate RAID CLine entry for
	 * the physical disk.
	 */	
	sublevel_rcl = toplevel_rcl + 1;
	sublevel_rcl->raidseq = toplevel_rcl->raidseq;
#if MD_ASR_DEDICATED_SPARE_SUPPORT
	if (member == NULL) {
#endif		
		get_random_bytes(&raidmagic, sizeof(raidmagic));
		sublevel_rcl->raidmagic = htonl(raidmagic);
		/*asr_array_rt->spareid = 0;
		asr_array_rt->actionPriority = 0x10;
		asr_array_rt->sparedrivemagic = htonl(0x0); */
#if MD_ASR_DEDICATED_SPARE_SUPPORT
	} else {
		sublevel_rcl->raidmagic = asr_spare_rt->sparedrivemagic;
		/* Update metadata in the array */
		asr_array_rt->spareid = asr_spare_rt->spareid;
		asr_array_rt->actionPriority = asr_spare_rt->actionPriority;
		asr_array_rt->sparedrivemagic = asr_spare_rt->sparedrivemagic;
		asr_array_rt->raidmagic = htonl(0x0);
	}
#endif		
	sublevel_rcl->raidlevel = FWP;
	sublevel_rcl->raidtype = ASR_RAIDSPR;
	sublevel_rcl->raidstate = LSU_COMPONENT_STATE_OPTIMAL;

	/* XXX Not really correct. */
	sublevel_rcl->raidid = htonl(MINOR(kdev));
	
	sublevel_rcl->lcapcty = toplevel_rcl->lcapcty;
	sublevel_rcl->strpsize = toplevel_rcl->strpsize; 
	
	asr_rb->drivemagic = sublevel_rcl->raidmagic; 

	/* Allocate an mddev for the Spare and put it in spare_arrays */
	spare_mddev = md_asr_mddev_alloc(&spare_arrays, asr_spare_rt,
					 toplevel_rcl,
					 MDK_MSUBTYPE_POOL_SPARE,
					 MDK_PRIORITY_UNKNOWN);
	if (spare_mddev == NULL) {
		md_asr_minfo_free(minfo);
		return(-ENOMEM);
	}

	MD_ASR_DEBUG(("Spare mddev allocated:0x%p\n", spare_mddev));
	spare_member = to_member(spare_mddev);

	/* Mark the spare as dedicated to the 'member' */
	if (member != NULL)
		spare_mddev->dspare_tag	= to_mddev(member);
	
	retval = md_asr_add_spare(spare_mddev, minfo);
	if (retval == 0) {
		asr_spare_member = mddev_to_asr_member(spare_mddev);
		asr_spare_rt = asr_spare_member->tl->asr_rt;

		if (asr_spare_rt == NULL) {
			md_asr_minfo_free(minfo);
			return (-ENODEV);
		}
		
		md_asr_schedule_sync(asr_spare_member->tl, /*wait*/1, /*enable*/1);

		if (member != NULL) {
#if MD_ASR_DEDICATED_SPARE_SUPPORT
		/* Update metadata in array in case of dedicated spare */
			md_asr_schedule_sync(asr_array_member->tl, /*wait*/1, 
					     /*enable*/1);

#endif	
			if (member->pers != NULL) {
				member->pers->raid_op(member, 
					     MDK_RAID_OP_SPARE_ADDED, 
					     NULL);
				/* Remove dedicated spare tag */

				if (spare_member) {
					spin_lock_irq (
						&spare_member->member_lock);
					spare_mddev->dspare_tag	= NULL; 
					spin_unlock_irq (
						&spare_member->member_lock);
				}
			}
		} else {
			mddev = NULL;
			ITERATE_CHILDREN(&toplevel_arrays, child, tmp, flags) {

				if (child->metapers != &md_asr_pers) 
					continue;

				if (child->type != MDK_MTYPE_MDDEV)
					continue;

				mddev = to_mddev(child);
			
				/* TBD:
				 * should break once all the available 
				 * spares are used up
				 */
				child->pers->raid_op(child, 
					     MDK_RAID_OP_SPARE_ADDED, 
					     NULL);
				/* TBD: check refcount */
			}
	
		}	
	}

	/* destruct array */
	//member_put(to_member(array_mddev));
	md_asr_minfo_free(minfo);
	return retval;
}

static int
md_asr_member_op_activate_spare(mdk_member_t *member)
{
	struct asr_raidtable *asr_spare_rt, *asr_array_rt;
	struct asr_raid_configline *asr_rcl;
	struct asr_reservedblock *asr_rb;
	struct md_asr_meta_info *minfo;
	struct md_asr_member *asr_array_member, *asr_spare_member;
	mdk_member_t *spare_member, *spare_rdev_member;
	mdk_rdev_t *rdev;
	mddev_t	*array_mddev, *spare_mddev, *mddev; 
	int retval;
	int i;
	
	if (member == NULL)
		return -ENODEV;

	if ((member->state & MDU_MSTATE_DEGRADED) == 0) {
		MD_ASR_DEBUG(("md_asr_activate_spare:" 
			"Invalid array RAID state 0x%x\n", member->state));
		return (-EINVAL);
	}

	array_mddev = to_mddev(member);
	asr_array_member = mddev_to_asr_member(array_mddev);
	if (asr_array_member == NULL)
		return (-ENODEV);

	/*
	 * If we need to activate a spare, we should
	 * first update the metadata to account for
	 * missing devices.  Otherwise we may not be
	 * able to find a valid slot in the VDCR to
	 * place the spare.
	 */
	md_asr_mark_missing_members(array_mddev);

	asr_array_rt = asr_array_member->tl->asr_rt;
	if (asr_array_rt->ent[0].raidtype != ASR_RAID1) {
		MD_ASR_DEBUG(("emd: Spare activation fail!"
			      "Invalid RAID Level\n"));
		return (-EINVAL);
	}

	spare_mddev = NULL;
	asr_spare_rt = NULL;
	rdev = NULL;
	mddev = NULL;
	md_asr_locate_spare(member, MDK_MEMBER_OP_ACTIVATE_SPARE, 
			    NULL, &spare_mddev, &rdev, NULL);

	if (spare_mddev == NULL || rdev == NULL) {
		MD_ASR_DEBUG(("emd: Spare activation fail!" 
			      "No available spares\n"));
		return (-ENODEV);
	}
	spare_member = to_member(spare_mddev);
	spare_rdev_member = to_member(rdev);
	spin_lock_irq(&spare_member->member_lock);	
	spare_mddev->dspare_tag = NULL;
	spin_unlock_irq(&spare_member->member_lock);
	
	asr_spare_member = mddev_to_asr_member(spare_mddev);
	if (asr_spare_member == NULL)
		return -ENODEV;
	asr_spare_rt = asr_spare_member->tl->asr_rt;
	if (asr_spare_rt == NULL)
		return -ENODEV;
	retval = md_asr_minfo_alloc(rdev->kdev, &minfo);
	if (retval != 0) {
		up(&spare_mddev->reconfig_sem);
		member_put(spare_member);
		return (retval);
	}
	asr_rb = minfo->asr_rb;
	/* Populate device specific reserved block */
	memset(asr_rb, 0, sizeof(*asr_rb));
	asr_rb->b0idcode = htonl(B0RESRVD);
	asr_rb->smagic = htonl(SVALID);
	asr_rb->resver = 0x08;
	asr_rb->raidtbl = asr_spare_rt->ent[0].lcapcty; 
	asr_rb->fstrsvrb = asr_rb->raidtbl;

	/* Find the RAID configuration line of the FAILED physical disk */
	spin_lock_irq(&asr_array_member->tl->rt_lock);
	for (asr_rcl = &asr_array_rt->ent[1], i = 1;
     	     i < asr_array_rt->ent[0].raidcnt; 
	     asr_rcl++, i++) {
		if (asr_rcl->raidstate == LSU_COMPONENT_STATE_FAILED)
			break;
	}
	
	/*
	 * Update existing RAID CLine. 
	 */
	asr_rcl->raidstate = LSU_COMPONENT_STATE_REPLACED;
	asr_rcl->raidid = asr_spare_rt->ent[1].raidid;
	asr_rcl->raidmagic =  asr_spare_rt->ent[1].raidmagic;

	/*
	 * Update existing RAID Table. 
	 */
	asr_array_rt->spareid = 0x0;
	asr_array_rt->actionPriority = 0x0;
	asr_array_rt->sparedrivemagic = htonl(0x0);
	asr_array_rt->raidmagic = htonl(0x0);
	asr_rb->drivemagic = asr_rcl->raidmagic;
	spin_unlock_irq(&asr_array_member->tl->rt_lock);
	
	retval = md_asr_add_member(array_mddev, minfo, /*allow collisions*/1);
	
	up(&spare_mddev->reconfig_sem);
	
	/* drop recently acquired ref count first */
	member_put(spare_rdev_member);
	member_put(spare_member);

	if (retval == 0) {
		struct md_asr_top_level_array *tl_array;

		MD_ASR_DEBUG(("Spare activated successfully:" 
			      "Updating Metadata\n"));
		md_asr_schedule_sync(asr_array_member->tl, /*wait*/1, 
				     /*enable*/1); 
		tl_array = mddev_to_top_level_array(array_mddev);

		/* if its a dedicated spare then drop reference on it */
		if (tl_array->spare_pool != NULL)
			member_put(to_member(tl_array->spare_pool));
		/* destroy spare from spare pool */
		member_put(spare_rdev_member);
		member_put(spare_member);
	}
	md_asr_minfo_free(minfo);
	
	return retval;
}

static int
md_asr_member_op_delete_spare(mdk_member_t *member)
{
	struct asr_raidtable *asr_spare_rt, *asr_array_rt;
	struct md_asr_member *asr_array_member, *asr_spare_member;
	struct list_head *tmp;
	mdk_member_t *child, *spare_member, *spare_rdev_member;
	mdk_rdev_t *rdev;
	mddev_t	*array_mddev, *spare_mddev, *mddev; 
	u_long flags;
	
	if (member == NULL)
		return -ENODEV;
	
	spare_mddev = NULL;
	asr_spare_rt = NULL;
	asr_array_rt = NULL;
	asr_array_member = NULL;
	rdev = NULL;
	mddev = NULL;
	
	md_asr_locate_spare(member, MDK_MEMBER_OP_DELETE_SPARE, 
			    NULL, &spare_mddev, &rdev, NULL);
	if (spare_mddev == NULL || rdev == NULL) {
		MD_ASR_DEBUG(("emd: Spare deletion fail! Spare not found\n"));
		return (-ENODEV);
	}
	spare_member = to_member(spare_mddev);
	spare_rdev_member = to_member(rdev);

	array_mddev = NULL;
	asr_spare_member = mddev_to_asr_member(spare_mddev);
	asr_spare_rt = asr_spare_member->tl->asr_rt;
	if (asr_spare_rt->sparedrivemagic != 0) {
		/* This is a dedicated spare */
		mddev = NULL;
		ITERATE_CHILDREN(&toplevel_arrays, child, tmp, flags) {
			if (child->metapers != &md_asr_pers) 
				continue;
			if (child->type != MDK_MTYPE_MDDEV)
				continue;
			mddev = to_mddev(child);
			asr_array_member = mddev_to_asr_member(mddev);
			asr_array_rt = asr_array_member->tl->asr_rt;
			if (asr_array_rt->ent[0].raidtype != ASR_RAID1)
				continue;
			if (ntohl(asr_spare_rt->ent[1].raidmagic) 
			 == ntohl(asr_array_rt->sparedrivemagic)) {
				if (ntohl(asr_spare_rt->raidmagic) 
				 == ntohl(asr_array_rt->ent[0].raidmagic)) {
					array_mddev = mddev;
					break;
				}
			}
		}
		if (array_mddev != NULL) {
			/* Update metadata on the array to which the spare was
			 * dedicated
			 */
			struct md_asr_top_level_array *tl_array;
			tl_array = mddev_to_top_level_array(array_mddev);

			asr_array_rt->spareid = 0x0;
			asr_array_rt->actionPriority = 0x10;
			asr_array_rt->sparedrivemagic = htonl(0x0);
			asr_array_rt->raidmagic = htonl(0x0);
			
			if (tl_array->spare_pool != NULL) {
				member_put(to_member(tl_array->spare_pool));
			}

			md_asr_schedule_sync(asr_array_member->tl, /*wait*/1, 
						     /*enable*/1);

			member_put(to_member(array_mddev));
		}
	}

	/* drop recently acquired refcount first */
	member_put(spare_rdev_member);
	member_put(spare_member);

	/* drop refcount acquired from the ioctl path, the 
	 * ioctl path code will drop last refcount and eventually 
	 * the member would get destroyed
	 */
	member_put(spare_rdev_member);
	member_put(spare_member);

	up(&spare_mddev->reconfig_sem);
	
	MD_ASR_DEBUG(("emd: Spare deleted successfully\n"));

	return 0;

}

static int
md_asr_member_op_get_spares(mdk_member_t *member, 
		            mdu_spare_disk_info_t *user_spare_info)
{
	mdu_disk_info_t *spare_disk_info;

	if (user_spare_info->count == 0) {
		return md_asr_locate_spare(member,  MDK_MEMBER_OP_GET_SPARES,
				NULL, NULL, NULL, &user_spare_info->count);
	}
	spare_disk_info = user_spare_info->spare_info;
	return md_asr_locate_spare(member,  MDK_MEMBER_OP_GET_SPARES, 
				   spare_disk_info, NULL, NULL, NULL);
}

static int
md_asr_member_op_set_checkpoint(mddev_t *mddev, mdk_checkpoint_arg_t *chkpt_arg)
{
	struct md_asr_member *asr_member;
	struct md_asr_top_level_array *tl_array;
	mdk_member_t *array;
	int build_complete;
	int sync_md;

	array = to_member(mddev);
	asr_member = mddev_to_asr_member(mddev);
	tl_array = asr_member->tl;
	sync_md = 0;
	spin_lock_irq(&tl_array->rt_lock);
	if (tl_array->asr_rt->dirty != ARRAY_STATE_DIRTY) {
		tl_array->asr_rt->dirty = ARRAY_STATE_DIRTY;
		sync_md = 1;
	}
	if ((chkpt_arg->checkpoint_type & MD_SYNCOP_VERIFY) != 0) {
		uint8_t irocFlags;
		
		irocFlags = tl_array->asr_rt->irocFlags;
		irocFlags &= ~ASR_IF_VERIFY_WITH_AUTOFIX;
		if (chkpt_arg->checkpoint_type == MD_SYNCOP_VERIFY_FIX)
			irocFlags |= ASR_IF_VERIFY_WITH_AUTOFIX;
		if (irocFlags != tl_array->asr_rt->irocFlags) {
			tl_array->asr_rt->irocFlags = irocFlags;
			sync_md = 1;
		}
	}
	if (chkpt_arg->checkpoint < array->data_size) {
		asr_member->rcl->curAppBlock = htonl(chkpt_arg->checkpoint);
		build_complete = 0;
	} else {
		asr_member->rcl->curAppBlock = 0;
		build_complete = 1;
		sync_md = 1;
	}
	spin_unlock_irq(&tl_array->rt_lock);

	if (sync_md)
		md_asr_schedule_sync(tl_array, /*wait*/0, /*enable*/1);

	if (build_complete) {
		mdk_member_t *child;
		struct list_head *tmp;
		u_long cflags;

		/*
		 * Mark any build target children as optimal.
		 * Do this without holding any locks in case
		 * this update result in updates in our own
		 * state.
		 */
		ITERATE_CHILDREN(array, child, tmp, cflags) {
			u_long cflags;

			spin_lock_irqsave(&child->member_lock, cflags);
			if (MDK_WRITABLE_MEMBER(child)
			 && (child->state & MDU_MSTATE_REBUILD_TARGET) != 0) {
				mdk_set_state_arg_t sstate;
				mdk_metapersonality_t *mpers;

				sstate.oldstate = child->state;
				spin_unlock_irqrestore(&child->member_lock,
						       cflags);
				sstate.newstate = child->state
						& ~MDU_MSTATE_REBUILD_TARGET;
				mpers = child->metapers;
				mpers->member_op(child,
						 MDK_MEMBER_OP_SET_STATE,
						 &sstate);
			} else {
				spin_unlock_irqrestore(&child->member_lock,
						       cflags);
			}
		}

		/*
		 * When verifies complete, clear the dirty flag.
		 */
		if ((chkpt_arg->checkpoint_type & MD_SYNCOP_VERIFY) != 0) {
			mdk_set_state_arg_t sstate;

			spin_lock_irqsave(&array->member_lock, cflags);
			sstate.oldstate = array->state;
			sstate.newstate = array->state & ~MDU_MSTATE_DIRTY;
			spin_unlock_irqrestore(&array->member_lock, cflags);
			md_asr_member_op_set_state(array, &sstate);
		}
	}
	return (0);
}

/*
 * Handle state change notifications on objects.  In general
 * state changes are only allowed if the old presented to us
 * matches the current state of the object.  The teardown state
 * is special.  We don't bother to check the previous state
 * of the member.  We just change its externally visible state
 * to the teardown state and decrement its reference so it goes
 * away.
 */
static int
md_asr_member_op_set_state(mdk_member_t *member, mdk_set_state_arg_t *sstate)
{
	struct md_asr_member *asr_member;
	struct md_asr_top_level_array *tl_array;
	u_long rtflags;
	u_long mflags;
	int error;
	int i;
	int put_member;
	int put_member_after_sync;
	uint8_t asr_state;

	asr_member = member_to_asr_member(member);
	tl_array = asr_member->tl;
	put_member = 0;
	put_member_after_sync = 0;


	switch (member->type) {
	case MDK_MTYPE_RDEV:
		error = md_asr_member_to_disk_state(sstate->newstate,
						    &asr_state);
		/*
		 * When RDEVs fail, schedule them for destruction
		 * by dropping their sentinal reference.
		 */
		if (sstate->newstate == MDU_MSTATE_FAILED
		 && sstate->oldstate != MDU_MSTATE_FAILED)
			put_member_after_sync = 1;
		break;
	case MDK_MTYPE_MDDEV:

		error = md_asr_member_to_array_state(asr_member,
						     sstate->newstate,
						     &asr_state);
		break;
	default:
		BUG();
		return (-EINVAL);
	}

	if (error != 0) {
		MD_ASR_DEBUG(("member state %#x can't be mapped to ASR MD!\n",
		       sstate->newstate));
		return (-EINVAL);
	}

	spin_lock_irqsave(&tl_array->rt_lock, rtflags);
	spin_lock_irqsave(&member->member_lock, mflags);
	if (sstate->newstate != MDU_MSTATE_TEARDOWN) {
		uint32_t raidmagic;

		if (sstate->oldstate != member->state) {
			spin_unlock_irqrestore(&member->member_lock, mflags);
			spin_unlock_irqrestore(&tl_array->rt_lock, rtflags);
			return (-EINVAL);
		}

		/*
		 * Traverse the entire RAID table as in the multi-level
		 * case, the same second level array is listed twice.
		 */
		raidmagic = asr_member->rcl->raidmagic;
		for (i = 0; i < ntohs(tl_array->asr_rt->elmcnt); i++) {
			if (tl_array->asr_rt->ent[i].raidmagic != raidmagic)
				continue;
			tl_array->asr_rt->ent[i].raidstate = asr_state;
		}
	} else if (member->state != MDU_MSTATE_TEARDOWN) {
		if (member->type != MDK_MTYPE_RDEV
		 || member->state != MDU_MSTATE_FAILED)
			put_member = 1;
		MD_ASR_DEBUG(("ASR_TEARDOWN %p - type %s: Refcount %d\n",
			      member, md_member_type_string(member),
			      (int)atomic_read(&member->refcount)));
		/* If I'm a dedicated spare pool, remove me from my
		 * parent object.
		 */
		if (tl_array->spare_pool != NULL)
			member_put(to_member(tl_array->spare_pool));
	}
	member->state = sstate->newstate;
	if (put_member_after_sync)
		asr_member->flags |= MD_ASR_MFLAGS_PUT_AFTER_SYNC;	
	spin_unlock_irqrestore(&member->member_lock, mflags);
	spin_unlock_irqrestore(&tl_array->rt_lock, rtflags);

	if (sstate->newstate != MDU_MSTATE_TEARDOWN) {
		int enable_sync;

		enable_sync = 0;
		if (member->type == MDK_MTYPE_RDEV
		 && sstate->newstate & (MDU_MSTATE_FAILED|MDU_MSTATE_DEGRADED))
			enable_sync = 1;

		if (member->subtype == MDK_MSUBTYPE_TOPLEVEL
		 || md_asr_update_array_state(to_mddev(member->parent)) == 0
		 || enable_sync)
			md_asr_schedule_sync(tl_array, /*wait*/0, enable_sync);

		if (member->pers != NULL) {
			member->pers->raid_op(member,
					      MDK_RAID_OP_MSTATE_CHANGED,
					      member);
		}
	}

	if (put_member) {
		mdk_member_t *parent;

		/*
		 * The rules for "sentinel puts" require us to
		 * guarantee that all other threads have seen
		 * a state change to either TEARDOWN or FAILED
		 * prior to the last reference count being released.
		 * This is achieved by briefly acquiring and
		 * releasing the parent's lock so that all threads
		 * iterating on the parent's child set have exited
		 * their loop since the state change and prior to
		 * our put.
		 */
		parent = member->parent;
		if (parent) {
			spin_lock_irqsave(&parent->member_lock, mflags);
			spin_unlock_irqrestore(&parent->member_lock, mflags);
		}

		member_put(member);
	}

	return (0);
}

/*
 * Determine a "reasonable" target_id value for the ASR
 * meta-data for a new array.  We do this by traversing
 * all exiting ASR arrays coming up with a new "maximum"
 * ID.  This algorithm is sufficient for simple
 * configuraitons, but would certainly benefit from having
 * more topology information about the devices we are
 * managing.
 */
static uint32_t
md_asr_next_target_id()
{
	struct list_head *tmp;
	mdk_member_t *child;
	u_long flags;
	uint32_t max;

	max = 0;
	ITERATE_CHILDREN(&toplevel_arrays, child, tmp, flags) {
		mddev_t *mddev;

		struct md_asr_top_level_array *tl_array;
		uint32_t array_raidid;

		mddev = to_mddev(child);
		if (mddev->member.metapers != &md_asr_pers)
			continue;
		tl_array = mddev_to_top_level_array(mddev);
		array_raidid = ntohl(tl_array->member.rcl->raidid);
		if (array_raidid >= max)
			max = array_raidid + 1;
	}
	
	return (max);
}

static int
md_asr_member_op_set_array_info(mdk_member_t *member, 
				mdu_array_info_t *user_array_info)
{
	struct timeval md_timestamp;
	struct asr_raidtable *asr_rt;
	struct asr_raid_configline *toplevel_rcl;
	struct asr_raid_configline *sublevel_rcl;
	mddev_t	*array;
	int retval;
	int i;
	uint32_t raidmagic;
	uint32_t targetid;
	uint32_t array_sectors;
	uint32_t stripe_sectors;
	uint32_t child_stripe_sectors;

	asr_rt = kmalloc(sizeof(*asr_rt), GFP_KERNEL);
	if (asr_rt == NULL)
		return (-ENOMEM);

	/*
	 * Populate RAID table based on passed-in
	 * configuration data.
	 *
	 * XXX Here we assume a single level array where
	 * this set_array is creating the top-level array
	 * object.  To support multi-level arrays, we need
	 * some method for passing in the parent array,
	 * so that child arrays can be properly supported.
	 */
	memset(asr_rt, 0, sizeof(*asr_rt));
	asr_rt->ridcode = htonl((uint32_t)RVALID2);
	asr_rt->rversion = htonl(0x2);
	asr_rt->maxelm = htons(RCTBL_MAX_ENTRIES);
	asr_rt->elmcnt = htons(user_array_info->raid_disks + 1);
	asr_rt->elmsize = htons(sizeof(struct asr_raid_configline));
	do_gettimeofday(&md_timestamp);
	asr_rt->timestamp = htonl(md_timestamp.tv_sec);
	get_random_bytes(&raidmagic, sizeof(raidmagic));
	asr_rt->raidmagic = htonl(raidmagic);

	/*
	 * Populate RAID CLine entry for
	 * the toplevel logical array.
	 */	
	asr_rt->ent[0].raidlevel = FWL;
	toplevel_rcl = &asr_rt->ent[0];
	toplevel_rcl->raidcnt = htons(user_array_info->raid_disks);
	get_random_bytes(&raidmagic, sizeof(raidmagic));
	toplevel_rcl->raidmagic = htonl(raidmagic);
	retval = md_asr_level_to_asr_raidtype(user_array_info->level,
					      &toplevel_rcl->raidtype);
	if (retval != 0) {
		kfree(asr_rt);
		return (retval);
	}
	
	retval = md_asr_member_to_array_state(NULL,
					      user_array_info->state,
					      &toplevel_rcl->raidstate);
	if (retval != 0) {
		kfree(asr_rt);
		return (retval);
	}
	
	/*
	 * XXX We don't have a way for the utility to
	 *     mark this array as a bootable array.
	 */
	targetid = md_asr_next_target_id() | MD_ASR_NON_BOOT_PRIORITY;
	toplevel_rcl->raidid = htonl(targetid);

	/*
	 * Calculate the individual member "configured size"
	 * based on the total array size in user_array_info->size.
	 */
	array_sectors = user_array_info->size;
	stripe_sectors = user_array_info->chunk_size >> 9;
	child_stripe_sectors = stripe_sectors;
	
	switch (toplevel_rcl->raidtype) {
	case ASR_RAID0:

		if (user_array_info->raid_disks < 2) {
			MD_ASR_DEBUG((KERN_WARNING "md_asr_set_array_info:"
				"Attempt to create RAID0 with less than 2"
				"members : Failing creation\n"));
			kfree(asr_rt);
			return (-EINVAL);
		}
		array_sectors *= user_array_info->raid_disks;

		/*
		 * The ASR metadata format's array entry
		 * "stripe size" is the sum of all member's
		 * "chunk size".
		 */
		stripe_sectors *= user_array_info->raid_disks;
		break;
	case ASR_RAID1:

		if (user_array_info->raid_disks < 2) {
			MD_ASR_DEBUG((KERN_WARNING "md_asr_set_array_info:" 
				"Attempt to create RAID1 with less than 2"
				"members : Failing creation\n"));
			kfree(asr_rt);
			return (-EINVAL);
		}
		stripe_sectors = 128;
		child_stripe_sectors = 128;
		break;
	case ASR_RAID5:

		if (user_array_info->raid_disks < 3) {
			MD_ASR_DEBUG((KERN_WARNING "md_asr_set_array_info:"
				"Attempt to create RAID5 with less than 3"
			        "members : Failing creation\n"));
			kfree(asr_rt);
			return (-EINVAL);
		}
		array_sectors *= (user_array_info->raid_disks - 1);

		/*
		 * The ASR metadata format's array entry
		 * "stripe size" is the sum of all member's
		 * "chunk size".
		 */
		stripe_sectors *= (user_array_info->raid_disks - 1);
		break;
	default:
		kfree(asr_rt);
		return (-EINVAL);
	}

	if (child_stripe_sectors != 128
	 && child_stripe_sectors != 64
	 && child_stripe_sectors != 32) {
		MD_ASR_DEBUG((KERN_WARNING "md_asr_set_array_info:"
			"Attempt to create array with invalid chunk size\n"
		       "md_asr_set_array_info: No 64k, 32K, or 16K\n"
		       "md_asr_set_array_info: Failing creation\n"));
		kfree(asr_rt);
		return (-EINVAL);
	}
	
	toplevel_rcl->lcapcty = htonl(array_sectors);
	toplevel_rcl->strpsize = htons(stripe_sectors);
	memcpy(toplevel_rcl->name, user_array_info->array_name, 
	       MAX_ARRAY_NAME_LENGTH);

	for (i = 0, sublevel_rcl = toplevel_rcl + 1;
	     i < user_array_info->raid_disks; i++, sublevel_rcl++) {
		sublevel_rcl->strpsize = htons(child_stripe_sectors);
		get_random_bytes(&raidmagic, sizeof(raidmagic));
		sublevel_rcl->raidmagic = htonl(raidmagic);
	}

	array = md_asr_mddev_alloc(&toplevel_arrays, asr_rt,
				   toplevel_rcl,
				   MDK_MSUBTYPE_TOPLEVEL,
				   ntohl(toplevel_rcl->raidid));
	if (array == NULL) {
		retval = (-ENOMEM);
	} else {
		md_assign_unit_number(array);
		user_array_info->cookie = (unsigned long)array;
	}

	kfree(asr_rt);
	return retval;
}


static int
md_asr_member_op_add_new_disk(mddev_t *mddev, mdu_disk_info_t *disk_info)
{
	mdk_member_t *array;
	struct md_asr_meta_info *minfo;
	struct gendisk *gd;
	struct md_asr_member *asr_member;
	struct asr_reservedblock *asr_rb;
	struct asr_raid_configline *asr_rcl;
	kdev_t	dev;
	sector_t drive_capacity;
	uint32_t drive_size;
	int retval;

	/*
	 * Verify that this component is not already present.
	 */
	array = to_member(mddev);
	if (disk_info->raid_disk >= mddev->max_disks)
		return (-EINVAL);

	dev = MKDEV(disk_info->major, disk_info->minor);
	retval = md_asr_minfo_alloc(dev, &minfo);
	if (retval != 0)
		return (retval);

	gd = get_gendisk(minfo->kdev);
	if (gd == NULL) {
		md_asr_minfo_free(minfo);
		return (-ENODEV);
	}

	asr_rb = minfo->asr_rb;
	asr_member = mddev_to_asr_member(mddev);

	/* Populate device specific reserved block */
	memset(asr_rb, 0, sizeof(*asr_rb));
	asr_rb->b0idcode = htonl(B0RESRVD);
	asr_rb->smagic = htonl(SVALID);
	asr_rb->resver = 0x08;

	drive_capacity = gd->part[disk_info->minor].nr_sects;
	asr_rb->raidtbl = htonl(drive_capacity - (RTBLBLOCKS + 1));
	asr_rb->fstrsvrb = asr_rb->raidtbl;
		
	/* Find the RAID configuration line for this physical disk */
	asr_rcl = asr_member->rcl + disk_info->raid_disk + 1;
	asr_rcl->raidseq = asr_member->rcl->raidseq;
	asr_rcl->raidtype = asr_member->rcl->raidtype;
	asr_rcl->raidlevel = FWP;
	retval = md_asr_member_to_disk_state(disk_info->state,
				    	     &asr_rcl->raidstate);
	if (retval != 0) {
		md_asr_minfo_free(minfo);
		return retval;
	}

	/* XXX Not really correct. */
	asr_rcl->raidid = htonl(MINOR(dev));

	/* calculate drive size from array size */
	drive_size = ntohl(asr_member->rcl->lcapcty);
	switch (asr_member->rcl->raidtype) {
	case ASR_RAID0:
		drive_size /= mddev->max_disks;
		break;
	case ASR_RAID1:
		break;
	case ASR_RAID5:
		drive_size /= mddev->max_disks - 1;
		break;
	default:
		md_asr_minfo_free(minfo);
		return -EINVAL;
	}

	if ((drive_capacity - (RTBLBLOCKS + 1)) < drive_size){
		md_asr_minfo_free(minfo);
		return -ENODEV;
	}
	
	asr_rcl->lcapcty = htonl(drive_size);
	asr_rb->drivemagic = asr_rcl->raidmagic;
	
	retval = md_asr_add_member(mddev, minfo, /*no collisions*/0);
	if (retval != 0) {
		md_asr_minfo_free(minfo);
		return retval;
	}
	md_asr_schedule_sync(asr_member->tl, /*wait*/1, /*enable*/1);

	md_asr_minfo_free(minfo);
	return (retval);
}

/*
 * Free up meta-data module state associated with a member.
 */
static int
md_asr_member_op_release(mdk_member_t *member)
{
	switch (member->type) {
	case MDK_MTYPE_MDDEV:
	{

		if (member->subtype == MDK_MSUBTYPE_TOPLEVEL) {
			mddev_t *mddev;
			struct md_asr_top_level_array *tl_array;

			mddev = to_mddev(member);
			tl_array = mddev_to_top_level_array(mddev);
			if (tl_array->spare_pool != NULL)
				member_put(to_member(tl_array->spare_pool));
			if (tl_array->asr_rt != NULL)
				kfree(tl_array->asr_rt);
			if (tl_array->sync_rt != NULL)
				kfree(tl_array->sync_rt);
		}
		break;
	}
	default:
		break;
	}
	if (member->metadata_private != NULL)
		kfree(member->metadata_private);
	member->metadata_private = NULL;
	md_metapers_put(member->metapers);
	member->metapers = NULL;
	return (0);
}

/*
 * Add a device, indicated by kdev_t, with the metadata in minfo,
 * to the toplevel array tree rooted by "array".
 */
static int
md_asr_add_member(mddev_t *array, struct md_asr_meta_info *minfo, 
		  int allow_collisions)
{
	struct md_asr_top_level_array *tl_array;
	struct asr_reservedblock *asr_rb;
	struct asr_raidtable *asr_rt;
	struct asr_raid_configline *asr_rcl;
	mdk_rdev_t *rdev;
	int rt_index;
	int parent_index;
	int component_index;

	tl_array = mddev_to_top_level_array(array);
	asr_rb = minfo->asr_rb;
	asr_rt = tl_array->asr_rt;

	/*
	 * Find this member in the raid table.
	 */
	rt_index = md_asr_find_member(asr_rt, ntohl(asr_rb->drivemagic),
				      FWP, &parent_index, &component_index);
	MD_ASR_DEBUG(("md_asr_add_member called: rti %d, pi %d, ci %d\n",
		      rt_index, parent_index, component_index));
	if (rt_index < 0
	 || parent_index < 0
	 || component_index < 0) {
		MD_ASR_DEBUG(("Invalid Raid Table\n"));
		return (-EINVAL);
	}
	asr_rcl = &asr_rt->ent[rt_index];

	/*
	 * Array is the top-level array.  If it heads
	 * a multi-level array, find or allocate our
	 * direct parent.
	 */
	if (tl_array->num_levels == 2) {
		mddev_t *child;
		struct asr_raid_configline *parent_rcl;

		MD_ASR_DEBUG(("Member of Dual Level Array\n"));
		parent_rcl = &asr_rt->ent[parent_index];
		child = md_asr_find_mddev(&array->member,
					  ntohl(parent_rcl->raidmagic));
		if (child == NULL) {
			int parent_component_index;

			/*
			 * Determine parents component index
			 * in the array.
			 */
			md_asr_find_member(asr_rt,
					   ntohl(parent_rcl->raidmagic),
					   FWL, NULL, &parent_component_index);
			if (parent_component_index < 0) {
				MD_ASR_DEBUG(("Invalid Raid Table "
					      "Parent Offset\n"));
				return (-EINVAL);
			}
			array = md_asr_mddev_alloc(to_member(array), asr_rt,
						   parent_rcl,
						   MDK_MSUBTYPE_CHILD,
						   parent_component_index);
			if (array == NULL)
				return (-ENOMEM);
			member_get(to_member(array));
		}
	} else {
		member_get(to_member(array));
	}

	rdev = md_asr_rdev_alloc(array, asr_rb, asr_rcl,
				 minfo->kdev, MDK_MSUBTYPE_CHILD,
				 component_index, allow_collisions);
	member_put(to_member(array));
	if (rdev == NULL)
		return (-ENOMEM);
	return (0);
}

/*
 * Add a spare to the spare pool array "array".
 */
static int
md_asr_add_spare(mddev_t *array, struct md_asr_meta_info *minfo)
{
	struct md_asr_member *asr_member;
	struct asr_raidtable *asr_rt;
	mdk_msubtype_t subtype;
	mdk_rdev_t *rdev;
	int rt_index;

	asr_member = mddev_to_asr_member(array);
	asr_rt = asr_member->tl->asr_rt;

	subtype = MDK_MSUBTYPE_POOL_SPARE;
	if (asr_rt->raidmagic != 0)
		subtype = MDK_MSUBTYPE_DEDICATED_SPARE;
	rt_index = md_asr_find_member(asr_rt,
				      ntohl(minfo->asr_rb->drivemagic),
				      FWP, NULL, NULL);
	rdev = md_asr_rdev_alloc(array, minfo->asr_rb,
				 &asr_rt->ent[rt_index],
				 minfo->kdev, subtype, /*priority*/0, 
				 /*no collisions*/0);
	if (rdev == NULL)
		return (-ENOMEM);
	return (0);
}

/*
 * Find the index in the ASR RaidTable of the top
 * level array to which this member belongs.
 */
static int
md_asr_find_toplevel(struct asr_raidtable *asr_rt)
{
	int i, toplevel, physlevel, found;

	found = 0;
	toplevel = 0;
	physlevel = 0;

	for (i = 0; i < ntohs(asr_rt->elmcnt); i++) {
		if (asr_rt->ent[i].raidlevel == FWL_2) {
			toplevel = i;
			found = 1;
			break;
		} else if (asr_rt->ent[i].raidlevel == FWL) {
			physlevel = i;
		}
	}

	if (found == 0)
		toplevel = physlevel;

	MD_ASR_DEBUG(("md_asr_find_toplevel returning %d\n", toplevel));
	return (toplevel);
}

/*
 * Find a member of type "type" with magic "magic" in
 * the passed in raid-table.  Returns raid table index
 * on success, -1 on search failure.  Optionally returns
 * both the raid table index of the parent and the component
 * index relative to the parent of the found entry (e.g.
 * component 0 or 1 of the array parented by the entry at
 * parent index).  A top-level array with have a parent
 * and component index of -1.
 */
static int
md_asr_find_member(struct asr_raidtable *asr_rt, uint32_t magic, int type,
		   int *p_parent_index, int *p_component_index)
{
	int i;
	int member_index;
	int parent_index;
	int component_index;
	int toplevel;

	member_index = -1;
	parent_index = -1;
	component_index = -1;
	for (i = 0; i < ntohs(asr_rt->elmcnt); i++) {

		if (asr_rt->ent[i].raidlevel == type
		 && ntohl(asr_rt->ent[i].raidmagic) == magic) {
			member_index = i;
			break;
		}

		if (asr_rt->ent[i].raidlevel != FWP) {
			/*
			 * Anytime we encounter an array object,
			 * reset our array object relative component
			 * index.  We also set the parent index so
			 * that this new array is treated as the parent
			 * if we find the member we are looking for
			 * before seeing another array.
			 */
			parent_index = i;
			component_index = 0;
		} else
			component_index++;
	}

	toplevel = md_asr_find_toplevel(asr_rt);
	if (member_index == toplevel) {
		/*
		 * Toplevel Arrays have no parent or component index,
		 * but our initial loop may have set while passing
		 * through child arrays.  Clear them again now.
		 */
		parent_index = -1;
		component_index = -1;
	} else if (type == FWL
		&& member_index != -1) {
		int i;

		/*
		 * For child arrays of a toplevel array, we must
		 * determine the component index by finding the
		 * second FWL instance for this array under the
		 * top-level array entry.
		 */
		parent_index = toplevel;
		component_index = 0;
		for (i = toplevel + 1; i < ntohs(asr_rt->elmcnt); i++) {
			if (asr_rt->ent[i].raidlevel == FWL
			 && ntohl(asr_rt->ent[i].raidmagic) == magic)
				break;
			component_index++;
		}
	}

	if (p_parent_index != NULL)
		*p_parent_index = parent_index;

	if (p_component_index != NULL)
		*p_component_index = component_index;

	return (member_index);
}

static mdk_rdev_t *
md_asr_find_spare_by_magic(uint32_t sparemagic)
{
	mdk_member_t *spare_array;
	struct list_head *tmp;
	u_long aflags;

	ITERATE_CHILDREN(&spare_arrays, spare_array, tmp, aflags) {
		mdk_member_t *child;
		struct list_head *tmp2;
		u_long sflags;

		ITERATE_CHILDREN(spare_array, child, tmp2, sflags) {
			struct asr_reservedblock *rb;
			mdk_rdev_t *spare;

			spare = to_rdev(child);
			rb = get_rb(spare);
			if (sparemagic == ntohl(rb->drivemagic)) {
				/*
				 * Drop extra refcount on parent
				 * array object.  The child already
				 * holds a refcount on it.
				 */
				member_put(spare_array);
				return (spare);
			}
				
		}
	}
	return (NULL);
}

static mddev_t *
md_asr_mddev_alloc(mdk_member_t *parent, struct asr_raidtable *asr_rt,
		   struct asr_raid_configline *asr_rcl,
		   mdk_msubtype_t subtype, int priority)
{
	mddev_t *array;
	mdk_member_t *member;
	struct md_asr_member *asr_member;
	struct md_asr_top_level_array *tl_array;
	int num_levels;

	array = md_mddev_alloc(NULL, subtype, priority);
	if (array == NULL)
		return (NULL);
	member = to_member(array);
	if (subtype == MDK_MSUBTYPE_TOPLEVEL
	 || subtype == MDK_MSUBTYPE_POOL_SPARE
	 || subtype == MDK_MSUBTYPE_DEDICATED_SPARE) {

		/*
		 * Allocate top-level array structure for all
		 * top-level array objects including pool spare
		 * objects.
		 */
		tl_array = kmalloc(sizeof(*tl_array), GFP_KERNEL);
		if (tl_array == NULL) {
			member_put(member);
			return (NULL);
		}
		memset(tl_array, 0, sizeof(*tl_array));
		tl_array->asr_rt = kmalloc(sizeof(*asr_rt), GFP_KERNEL);
		if (tl_array->asr_rt == NULL) {
			member_put(member);
			kfree(tl_array);
			return (NULL);
		}
		memcpy(tl_array->asr_rt, asr_rt, sizeof(*asr_rt));
		tl_array->sync_rt = kmalloc(sizeof(*asr_rt), GFP_KERNEL);
		if (tl_array->sync_rt == NULL) {
			member_put(member);
			kfree(tl_array->asr_rt);
			kfree(tl_array);
			return (NULL);
		}
		memset(tl_array->sync_rt, 0, sizeof(*asr_rt));
		spin_lock_init(&tl_array->rt_lock);
		md_init_workitem(&tl_array->sync_md_workitem,
				 MD_ASR_WORKOP_SYNC, md_asr_work_control,
				 md_asr_sync_handler, array);
		if (asr_rcl->raidlevel == FWL_2)
			num_levels = 2;
		else
			num_levels = 1;
		tl_array->num_levels = num_levels;

		asr_member = &tl_array->member;
	} else {

		/*
		 * Child array object.  Allocate and configure
		 * a member object for it.
		 */
		tl_array = mddev_to_top_level_array(to_mddev(parent));
		asr_member = kmalloc(sizeof(*asr_member), GFP_KERNEL);
		if (asr_member == NULL) {
			member_put(member);
			return (NULL);
		}
		memset(asr_member, 0, sizeof(*asr_member));
		num_levels = 1;
	}

	/*
	 * Make our member objects raid configuration line point to
	 * our line in the top-level array's copy of the raid table.
	 * Each array hierarchy has only one copy of the raid table.
	 */
	asr_member->tl = tl_array;
	asr_member->rcl = &tl_array->asr_rt->ent[asr_rcl - asr_rt->ent];
	asr_rcl = asr_member->rcl;
	asr_rt = tl_array->asr_rt;
	member->metadata_private = asr_member;

	MD_ASR_DEBUG(("md_asr_mddev_alloc: %p Level %d priority %#x, "
		      "subtype %d\n", array, num_levels, priority, subtype));

	/*
	 * Chunksize is the portion of the stripe on each child, so
	 * use the stripe size on our first child to export this
	 * information to MD.  ASR metadata specifies a fixed chunk
	 * size for RAID1 members too, but this is nonsensical.  Force
	 * a value of 0 when exporting this value to the rest of MD.
	 */
	if (asr_rcl->raidtype == ASR_RAID1
	 || asr_rcl->raidtype == ASR_RAIDSPR)
		array->chunk_size = 0;
	else
		array->chunk_size = ntohs((asr_rcl+1)->strpsize) << 9;
	array->preferred_unit = -1;
	array->persistent = 1;
	member->metapers = md_metapers_get(&md_asr_pers);
	member->state = MDU_MSTATE_OPTIMAL;
	if (asr_rcl->raidtype != ASR_RAIDSPR) {
		mdk_personality_t *raid_pers;
		int error;

		/*
		 * An array without any members must be in the
		 * failed state unless it is a spare array or
		 * it is marked for initialization. This will be
		 * updated as children are added.
		 */
		error = md_asr_array_to_member_state(asr_rcl->raidstate,
						     &member->state);
/* XXX Check Error!!! */
		member->state &= ~MDU_MSTATE_OPTIMAL;
		member->state |= MDU_MSTATE_FAILED;

		/*
		 * If the array was shutdown unexpectedly, the dirty
		 * flag will still be set.  If set, clear any checkpoint,
		 * and mark our state as dirty.
		 */
		if (asr_rt->dirty == ARRAY_STATE_DIRTY) {
			member->state |= MDU_MSTATE_DIRTY;
			asr_rt->irocFlags |= ASR_IF_VERIFY_WITH_AUTOFIX;
			asr_rcl->curAppBlock = 0;
		}

		if ((asr_rt->irocFlags & ASR_IF_VERIFY_WITH_AUTOFIX) != 0)
			array->verify_op = MD_SYNCOP_VERIFY_FIX;
		else
			array->verify_op = MD_SYNCOP_VERIFY;

		member->total_size = ntohl(asr_rcl->lcapcty);
		member->data_size = member->total_size;
/* XXX Verify component count for multi-level arrays. */
		array->max_disks = ntohs(asr_rcl->raidcnt);
		array->level = md_asr_raidtype_to_level(asr_rcl->raidtype);

		/*
		 * If we are a toplevel array and our array info indicates
		 * a dedicated spare pool, try to find it now in case the
		 * spare pool was enumerated before this array.
		 */
		if (subtype == MDK_MSUBTYPE_TOPLEVEL
		 && asr_rt->sparedrivemagic != 0) {
			mdk_rdev_t *spare;
			uint32_t sparemagic;

			sparemagic = ntohl(asr_rt->sparedrivemagic);
			spare = md_asr_find_spare_by_magic(sparemagic);
			if (spare != NULL) {
				mddev_t *spare_pool;

				/*
				 * Hold a reference to the spare pool for
				 * this dedicated spare.  All spare lookups
				 * occur by traversing spare pools, not
				 * individual spares.
				 */
				spare_pool = to_mddev(spare->member.parent);
				member_get(to_member(spare_pool));
				tl_array->spare_pool = spare_pool;

				/*
				 * Drop refcount from the lookup of
				 * spare now that we hold a reference
				 * to its parent.
				 */
				member_put(to_member(spare));
				MD_ASR_DEBUG(("attached spare %#x to array "
					      "%#x\n", sparemagic,
					      ntohl(asr_rcl->raidmagic)));
			}
		}

		raid_pers = md_find_pers(array->level);
		if (raid_pers == NULL) {
			member_put(member);
			return (NULL);
		}
		raid_pers->raid_op(member, MDK_RAID_OP_INIT, NULL);
		/*
		 * Lose extra reference from md_find_pers.  The
		 * RAID personality owns the personality reference
		 * for this object from INIT through RELEASE.
		 */
		md_pers_put(raid_pers);
	} else if (asr_rcl->raidtype == ASR_RAIDSPR) {
		if (subtype == MDK_MSUBTYPE_DEDICATED_SPARE) {
			mddev_t *assoc_array;
		
			/*
			 * Associate this dedicated spare pool with
			 * the array it is dedicated to if it has
			 * already been discovered.  This associated
			 * array holds a reference on the spare pool.
			 */
			assoc_array = md_asr_find_mddev(&toplevel_arrays,
						ntohl(asr_rt->raidmagic));
			if (assoc_array != NULL) {
				struct md_asr_top_level_array *assoc_tl;

				member_get(member);
				assoc_tl = mddev_to_top_level_array(
								assoc_array);
				assoc_tl->spare_pool = array;
				MD_ASR_DEBUG(
				("attached spare pool%#x to array %#x\n",
				      ntohl(asr_member->rcl->raidmagic),
				      ntohl(assoc_tl->member.rcl->raidmagic)));
				member_put(to_member(assoc_array));
			}
		}
		array->max_disks = ntohs(asr_rcl->raidcnt);
		/* Do not assign unit number to spare. This fixes
		 * the bug where the driver was honoring requests to
		 * phantom devices 
		 */
		array->unit = -1;
		array->level = md_asr_raidtype_to_level(asr_rcl->raidtype);
		member->total_size = ntohl(asr_rcl->lcapcty);
		member->data_size = member->total_size;
	}

	if (parent != NULL) {
		int collision;

		if (parent->pers != NULL)
			parent->pers->raid_op(parent, MDK_RAID_OP_INIT_MEMBER,
					      member);
		MD_ASR_DEBUG(("md_asr_mddev_alloc: Adding array %p to %p\n",
			      array, parent));
		collision = md_member_add_child(parent, member);
		if (subtype == MDK_MSUBTYPE_CHILD) {

			if (collision != 0) {
				
				/*
				 * The parent array has a duplicate child
				 * array object.  Tear this array down
				 * and warn the user.
				 */
				md_set_member_state(member,
						    MDU_MSTATE_TEARDOWN);
				MD_ASR_DEBUG((KERN_WARNING "md_asr_mddev_alloc:"
				       "Duplicate child array discarded.\n"));
				return (NULL);
			}
			md_asr_update_array_state(to_mddev(parent));
			if (parent->pers != NULL) {
				parent->pers->raid_op(member,
						      MDK_RAID_OP_ADD_MEMBER,
						      NULL);
			}
		}
	}
	return (array);
}

static mdk_rdev_t *
md_asr_rdev_alloc(mddev_t *array, struct asr_reservedblock *asr_rb,
		  struct asr_raid_configline *asr_rcl, kdev_t dev,
		  mdk_msubtype_t subtype, int priority, int allow_collisions)
{
	struct md_asr_device *asr_dev;
	mdk_rdev_t *rdev;
	mdk_member_t *parent;
	mdk_member_t *member;
	int error;
	int collision;

	asr_dev = kmalloc(sizeof(*asr_dev), GFP_KERNEL);
	if (asr_dev == NULL)
		return (NULL);
	memset(asr_dev, 0, sizeof(*asr_dev));

	memcpy(&asr_dev->asr_rb, asr_rb, sizeof(*asr_rb));
	asr_dev->member.tl = mddev_to_asr_member(array)->tl;
	asr_dev->member.rcl = asr_rcl;

	rdev = md_rdev_alloc(NULL, subtype, dev, priority);
	if (rdev == NULL) {
		kfree(asr_dev);
		return (NULL);
	}

	member = to_member(rdev);
	member->metadata_private = asr_dev;
	member->metapers = md_metapers_get(&md_asr_pers);
	error = md_asr_disk_to_member_state(array, asr_rcl->raidstate,
					    &member->state);
	if (error) {
		member_put(member);
		return (NULL);
	}
	member->total_size = ntohl(asr_rcl->lcapcty);
	member->data_size = member->total_size;
	member->data_offset = 0;
	parent = to_member(array);
	if (parent->pers != NULL)
		parent->pers->raid_op(parent, MDK_RAID_OP_INIT_MEMBER, member);
	collision = md_member_add_child(parent, member);
	if (collision != 0 && !allow_collisions) {
		/*
		 * The parent array has a duplicate child
		 * device object.  Discard the duplicate.
		 */
		md_set_member_state(member, MDU_MSTATE_TEARDOWN);
		MD_ASR_DEBUG((KERN_WARNING "md_asr_rdev_alloc: "
		       "Duplicate child device discarded.\n"));
		return (NULL);
	}
	MD_ASR_DEBUG(("Added drive %#x to array %#x\n",
		      ntohl(asr_rb->drivemagic),
		      ntohl(mddev_to_asr_member(array)->rcl->raidmagic)));

	/*
	 * Update our parents state to reflect the
	 * addition of this member.
	 */
	md_asr_update_array_state(array);
	if (parent->pers != NULL)
		parent->pers->raid_op(parent, MDK_RAID_OP_ADD_MEMBER, NULL);

	return (rdev);
}

/*
 * Revalidate our array state.   This routine is called whenever
 * a member is added, removed, or changes state.
 */
static int
md_asr_update_array_state(mddev_t *array)
{
	mdk_set_state_arg_t sstate;
	struct md_asr_member *asr_member;
	u_long flags;
	u_int configured_cnt;
	u_int degraded_need_cnt;
	u_int optimal_cnt;
	u_int degraded_cnt;
	u_int failed_cnt;
	u_int build_targets;

	asr_member = mddev_to_asr_member(array);
	MD_ASR_DEBUG(("md_asr_update_array_state: %p curstate = %#x\n",
		      array, array->member.state));
	
	/*
	 * Prevent state from changing during this update.
	 */
	spin_lock_irqsave(&array->member.member_lock, flags);
	sstate.oldstate = array->member.state;
	configured_cnt = ntohs(asr_member->rcl->raidcnt);

	md_get_array_stats(array, &optimal_cnt, &degraded_cnt,
			   &failed_cnt, &build_targets);
	MD_ASR_DEBUG(("md_asr_update_array_state: o %d d %d f %d b %d\n",
		      optimal_cnt, degraded_cnt, failed_cnt, build_targets));

	/*
	 * All of array types save RAID0 can operate
	 * with one failed/missing member.
	 */
	degraded_need_cnt = configured_cnt - 1;
	switch (asr_member->rcl->raidtype) {
	case ASR_RAIDSPR:
		/*
		 * Spare arrays are always optimal.
		 */
		sstate.newstate = MDU_MSTATE_OPTIMAL;
		break;
	case ASR_RAID0:
		/*
		 * RAID0's are marked failed unless
		 * all members are present and optimal
		 * or degraded.
		 */
		degraded_need_cnt = configured_cnt;
		/* FALLTHROUGH */
	case ASR_RAID1:
	case ASR_RAID4:
	case ASR_RAID5:
		if (optimal_cnt == configured_cnt) 
			sstate.newstate = MDU_MSTATE_OPTIMAL;
		else if (optimal_cnt + degraded_cnt >= degraded_need_cnt)
			sstate.newstate = MDU_MSTATE_DEGRADED;
		else if (sstate.oldstate & MDU_MSTATE_INITIALIZING) {
			sstate.newstate = MDU_MSTATE_INITIALIZING;
			/*
			 * If we have sufficent build targets,
			 * update our state.
			 */
			if (optimal_cnt + build_targets == configured_cnt) 
				sstate.newstate |= MDU_MSTATE_OPTIMAL;
			else if (optimal_cnt + build_targets
			      >= degraded_need_cnt)
				sstate.newstate |= MDU_MSTATE_DEGRADED;
		} else
			 sstate.newstate = MDU_MSTATE_FAILED;

		if ((sstate.oldstate & MDU_MSTATE_DIRTY) != 0)
			sstate.newstate |= MDU_MSTATE_DIRTY;
		break;
	default:
		sstate.newstate = sstate.oldstate;
		break;
	}
	spin_unlock_irqrestore(&array->member.member_lock, flags);

	if (sstate.newstate != sstate.oldstate) {

		md_asr_member_op_set_state(&array->member, &sstate);
		MD_ASR_DEBUG(("Array %p transitions from %#x to %#x\n",
			      array, sstate.oldstate, sstate.newstate));

		return (1);
	}
	return (0);
}

static void
md_asr_mark_missing_members(mddev_t *mddev)
{
	mdk_member_t *array;
	mdk_member_t *child;
	struct list_head *tmp;
	struct md_asr_member *asr_member;
	struct md_asr_top_level_array *tl_array;
	struct asr_raidtable *asr_rt;
	u_long tlflags;
	u_long mflags;
	u_int i;
	int sched_sync;

	array = to_member(mddev);
	asr_member = mddev_to_asr_member(mddev);
	tl_array = asr_member->tl;
	asr_rt = tl_array->asr_rt;
	sched_sync = 0;
	spin_lock_irqsave(&tl_array->rt_lock, tlflags);
	spin_lock_irqsave(&array->member_lock, mflags);
	for (i = 0; i < mddev->max_disks; i++) {
		struct asr_raid_configline *asr_rcl;
		int found;

		asr_rcl = asr_member->rcl + i + 1;
		found = 0;
		ITERATE_CHILDREN_LOCKED(array, child, tmp) {
			struct md_asr_member *asr_child_member;

			asr_child_member = child->metadata_private;
			if (asr_rcl == asr_child_member->rcl) {
				found = 1;
				break;
			}
		}
		if (found == 0
		 && asr_rcl->raidstate != LSU_COMPONENT_STATE_FAILED) {
			asr_rcl->raidstate = LSU_COMPONENT_STATE_FAILED;
			sched_sync = 1;
			break;
		}
	}
	spin_unlock_irqrestore(&tl_array->rt_lock, tlflags);
	spin_unlock_irqrestore(&array->member_lock, mflags);
	if (sched_sync)
		md_asr_schedule_sync(tl_array, /*wait4complete*/0, /*enable*/1);
}

static int
md_asr_validate_member(mdk_member_t *member, struct asr_raidtable *asr_rt)
{
	struct md_asr_member *asr_member;
	struct md_asr_top_level_array *tl_array;
	struct asr_raid_configline *o_rcl;
	struct asr_raid_configline *rcl;
	uint32_t raidmagic;
	int o_parent_index;
	int o_component_index;
	int o_rt_index;
	int rt_index;
	int parent_index;
	int component_index;
	int error;
	mdu_mstate_t state;

	asr_member = member_to_asr_member(member);
	tl_array = asr_member->tl;
	o_rcl = asr_member->rcl;
	raidmagic = ntohl(o_rcl->raidmagic);

	/*
	 * Consider a member valid only if its position
	 * in the configuration is unchanged and it is
	 * not listed as FAILED in the new raid table.
	 */
	o_rt_index = md_asr_find_member(tl_array->asr_rt, raidmagic,
					o_rcl->raidlevel, &o_parent_index,
					&o_component_index);
	rt_index = md_asr_find_member(asr_rt, raidmagic, o_rcl->raidlevel,
				      &parent_index, &component_index);

	if (rt_index == -1
	 || o_rt_index != rt_index
	 || o_parent_index != parent_index
	 || o_component_index != component_index)
		return (-ENODEV);

	rcl = &asr_rt->ent[rt_index];
	if (rcl->raidlevel == FWP)
		error = md_asr_disk_to_member_state(NULL, rcl->raidstate,
						    &state);
	else
		error = md_asr_array_to_member_state(rcl->raidstate, &state);

	if (error != 0)
		return (-ENODEV);

	if ((state & MDU_MSTATE_FAILED) != 0)
		return (-ENODEV);

	return (0);
}

/*
 * Update an existing array hierarchy to match the raid-table
 * pointed to by asr_rt.  This routine is typically called when
 * a member with more recent meta-data arrives after we've seen
 * at least one member with an out-of-date raid-table.
 */
static int
md_asr_apply_rt(mddev_t *toplevel_array, mdk_member_t *member,
		struct asr_raidtable *asr_rt)
{
	mdk_member_t *child;
	struct list_head *tmp;
	u_long flags;
	int exported;
	int saw_error;
	int error;

	saw_error = 0;
	spin_lock_irqsave(&to_member(toplevel_array)->member_lock, flags);
	exported = toplevel_array->flags & MD_MDDEV_EXPORTED_TO_OS;
	spin_unlock_irqrestore(&to_member(toplevel_array)->member_lock, flags);
	error = md_asr_validate_member(member, asr_rt);
	if (error != 0) {

		if (exported) {
			/*
			 * Update RAID table and leave intact.
			 */
			md_asr_change_magic(toplevel_array, member);
		} else
			md_teardown_member(member);
		saw_error++;
	}

	ITERATE_CHILDREN(member, child, tmp, flags) {

		saw_error += md_asr_apply_rt(toplevel_array, child, asr_rt);
	}

	if (to_member(toplevel_array) == member
	 && saw_error == 0) {
		struct md_asr_top_level_array *tl_array;

		tl_array = mddev_to_top_level_array(toplevel_array);
		spin_lock_irq(&tl_array->rt_lock);
		memcpy(tl_array->asr_rt, asr_rt, sizeof(*asr_rt));
		spin_unlock_irq(&tl_array->rt_lock);
		md_asr_schedule_sync(tl_array, /*wait4complete*/0, /*enable*/1);
	}
	return (saw_error);
}

static void
md_asr_change_magic(mddev_t *toplevel_array, mdk_member_t *member)
{
	struct md_asr_top_level_array *tl_array;
	struct md_asr_member *asr_member;
	struct asr_raidtable *asr_rt;
	uint32_t oldmagic;
	uint32_t newmagic;
	int i;
	
	tl_array = mddev_to_top_level_array(toplevel_array);
	asr_member = member_to_asr_member(member);
	asr_rt = tl_array->asr_rt;
	oldmagic = asr_member->rcl->raidmagic;

retry:
	get_random_bytes(&newmagic, sizeof(newmagic));
	spin_lock_irq(&tl_array->rt_lock);
	/*
	 * Keep resetting the magic number until it
	 * is unique.
	 */
	for (i = 0; i < ntohs(asr_rt->elmcnt); i++) {

		if (asr_rt->ent[i].raidmagic == newmagic) {
			spin_unlock_irq(&tl_array->rt_lock);
			goto retry;
		}
	}

	/*
	 * Now replace fields with oldmagic with new magic.
	 */
	for (i = 0; i < ntohs(asr_rt->elmcnt); i++) {

		if (asr_rt->ent[i].raidmagic == oldmagic)
			asr_rt->ent[i].raidmagic = newmagic;
	}

	if (member->type == MDK_MTYPE_RDEV) {
		mdk_rdev_t *rdev;
		struct asr_reservedblock *rb;

		rdev = to_rdev(member);
		rb = get_rb(rdev);
		rb->drivemagic = newmagic;
	}
	spin_unlock_irq(&tl_array->rt_lock);
	md_asr_schedule_sync(tl_array, /*wait4complete*/0, /*enable*/1);
}

static int
md_asr_member_to_array_state(struct md_asr_member *asr_member,
			     mdu_mstate_t md_state, uint8_t *adaptec_state)
{
	int building;

	if (asr_member != NULL) {
		u_long flags;

		spin_lock_irqsave(&asr_member->tl->rt_lock, flags);
		building = asr_member->rcl->curAppBlock != 0;
		spin_unlock_irqrestore(&asr_member->tl->rt_lock, flags);
	} else
		building = 0;

	switch (md_state & ~MDU_MSTATE_DIRTY) {
	case MDU_MSTATE_OPTIMAL:
		if ((md_state & MDU_MSTATE_DIRTY) != 0)
			*adaptec_state = LS_VERIFYING_M;
		else
			*adaptec_state = LS_OPTIMAL_M;
		break;
	case MDU_MSTATE_OPTIMAL|MDU_MSTATE_INITIALIZING:
	case MDU_MSTATE_DEGRADED|MDU_MSTATE_INITIALIZING:
		*adaptec_state = LS_BUILDING_M;
		break;
	case MDU_MSTATE_DEGRADED:
		if (building != 0)
			*adaptec_state = LS_REBUILDING_M;
		else
			*adaptec_state = LS_DEGRADED_M;
		break;
	case MDU_MSTATE_FAILED:
	case MDU_MSTATE_TEARDOWN:
		*adaptec_state = LS_DEAD_M;
		break;
	default:
		return (1);
	}

	return (0);
}

static int
md_asr_array_to_member_state(int adaptec_state, mdu_mstate_t *md_state)
{

	switch (adaptec_state) {
	case LS_OPTIMAL_M:
		*md_state = MDU_MSTATE_OPTIMAL;
		break;
	case LS_VERIFYING_M:
		*md_state = MDU_MSTATE_OPTIMAL|MDU_MSTATE_DIRTY;
		break;
	case LS_BUILDING_M:
		*md_state = MDU_MSTATE_OPTIMAL|MDU_MSTATE_INITIALIZING;
		break;
	case LS_REBUILDING_M:
	case LS_DEGRADED_M:
		*md_state = MDU_MSTATE_DEGRADED;
		break;
	case LS_DEAD_M:
		*md_state = MDU_MSTATE_FAILED;
		break;
	default:
		return (1);
	}

	return (0);
}

static int
md_asr_member_to_disk_state(mdu_mstate_t md_state, uint8_t *adaptec_state)
{

	switch (md_state) {
	case MDU_MSTATE_OPTIMAL:
		*adaptec_state = LSU_COMPONENT_STATE_OPTIMAL;
		break;
	case MDU_MSTATE_OPTIMAL|MDU_MSTATE_INITIALIZING:
		*adaptec_state = LSU_COMPONENT_STATE_UNINITIALIZED
			       | LSU_COMPONENT_SUBSTATE_BUILDING;
		break;
	case MDU_MSTATE_OPTIMAL|MDU_MSTATE_REBUILD_TARGET:
		*adaptec_state = LSU_COMPONENT_STATE_REPLACED
			       | LSU_COMPONENT_SUBSTATE_REBUILDING;
		break;
	case MDU_MSTATE_FAILED:
	case MDU_MSTATE_TEARDOWN:
		*adaptec_state = LSU_COMPONENT_STATE_FAILED;
		break;
	default:
		return (1);
	}

	return (0);
}

static int
md_asr_disk_to_member_state(mddev_t *array, int adaptec_state,
			    mdu_mstate_t *md_state)
{
	MD_ASR_DEBUG(("adaptec_disk_state_to_md: adaptec_state = %#x\n",
		     adaptec_state));

	switch (adaptec_state & MAIN_STATE) {
	case LSU_COMPONENT_STATE_OPTIMAL:
		*md_state = MDU_MSTATE_OPTIMAL;
		if (array != NULL
		 && (to_member(array)->state & MDU_MSTATE_INITIALIZING) != 0)
			*md_state |= MDU_MSTATE_REBUILD_TARGET;
		break;
	case LSU_COMPONENT_STATE_DEGRADED:
	case LSU_COMPONENT_STATE_FAILED:
		*md_state = MDU_MSTATE_FAILED;
		break;
	case LSU_COMPONENT_STATE_REPLACED:
		*md_state = MDU_MSTATE_REBUILD_TARGET|MDU_MSTATE_OPTIMAL;
		break;
	case LSU_COMPONENT_STATE_UNINITIALIZED:
		*md_state = MDU_MSTATE_INITIALIZING|MDU_MSTATE_OPTIMAL;
		break;
	case LSU_ARRAY_SUBSTATE_AWAIT_FORMAT:
	case LSU_COMPONENT_STATE_UNCONFIGURED:
	default:
		/*
		 * In general these states don't come up,
		 */
		MD_ASR_DEBUG(("adaptec_disk_state_to_md: unexpected "
			      "adaptec_state = %#x\n", adaptec_state));
		*md_state = MDU_MSTATE_FAILED;
		return (1);
	}

	return (0);
}

static int
md_asr_raidtype_to_pers(uint8_t asr_raidtype)
{
	int raidpers;

	switch (asr_raidtype) {
	case ASR_RAID0:
		raidpers = RAID0;
		break;
	case ASR_RAID1:
		raidpers = RAID1;
		break;
	case ASR_RAID5:
		raidpers = RAID5;
		break;
	case ASR_RAIDSPR:
		raidpers = SPARE;
		break;
	default:
		raidpers = MD_RESERVED;
		break;
	}

	return (raidpers);
}

static int
md_asr_raidtype_to_level(uint8_t asr_raidtype)
{
	return (pers_to_level(md_asr_raidtype_to_pers(asr_raidtype)));
}

static int
md_asr_level_to_asr_raidtype(int level, uint8_t *praidtype)
{
	uint8_t raidtype;

	switch (level) {
	case 0:
		raidtype = ASR_RAID0;
		break;
	case 1:
		raidtype = ASR_RAID1;
		break;
	case 5:
		raidtype = ASR_RAID5;
		break;
	default:
		return (-EINVAL);
	}

	if (praidtype != NULL)
		*praidtype = raidtype;
	return (0);
}

static const char *
md_asr_raidlevel_string(int raidlevel)
{
	const char *string;

	switch (raidlevel) {
	case FWP:
		string = "FWP";
		break;
	case FWL:
		string = "FWL";
		break;
	case OSI:
		string = "OSI";
		break;
	case FWL_2:
		string = "FWL_2";
		break;
	default:
		string = "Level UNKNOWN";
		break;
	}

	return (string);
}

static const char *
md_asr_raidtype_string(int raidtype)
{
	const char *string;

	switch (raidtype) {
	case ASR_RAID0:
		string = "RAID0";
		break;
	case ASR_RAID1:
		string = "RAID1";
		break;
	case ASR_RAID5:
		string = "RAID5";
		break;
	case ASR_RAIDRED:
		string = "RAIDRED";
		break;
	case ASR_RAIDSPR:
		string = "RAIDSPR";
		break;
	default:
		string = "Type UNKNOWN";
		break;
	}

	return (string);
}

static void
md_asr_dump_rb(struct asr_reservedblock *asr_rb)
{
#if MD_DEBUG
	MD_ASR_DEBUG(("ASR Reserved Block\n"));
	MD_ASR_DEBUG(("  b0idcode = 0x%x, sb0flags = 0x%x, jbodEnable = 0x%x\n",
	       ntohl(asr_rb->b0idcode), asr_rb->sb0flags, asr_rb->jbodEnable));
	MD_ASR_DEBUG(("  biosInfo = 0x%x, fstrsvrb = 0x%x, "
	       "svBlockStorageTid = 0x%x\n",
	       ntohs(asr_rb->biosInfo), ntohl(asr_rb->fstrsvrb),
	       ntohs(asr_rb->svBlockStorageTid)));
	MD_ASR_DEBUG(("  svtid = 0x%x, drivemagic = 0x%x, raidtbl = 0x%x\n",
	       ntohs(asr_rb->svtid), ntohl(asr_rb->drivemagic),
	       ntohs(asr_rb->raidtbl)));
#endif
}

static void
md_asr_dump_rcl(struct asr_raid_configline *rcl, int index)
{
#if MD_DEBUG
	MD_ASR_DEBUG(("RCL Entry [%d] %s %s Magic %#x LSU %#x\n",
	       index,  md_asr_raidlevel_string(rcl->raidlevel),
	       md_asr_raidtype_string(rcl->raidtype),
	       ntohl(rcl->raidmagic), ntohl(rcl->lsu)));
	MD_ASR_DEBUG(("    raidcnt = %d, raidseq = %d\n",
	       ntohs(rcl->raidcnt), ntohs(rcl->raidseq)));
	MD_ASR_DEBUG(("    raidstate = %#x, flags = %#x, refcnt = %d\n",
	       rcl->raidstate, rcl->flags, rcl->refcnt));
	MD_ASR_DEBUG(("    hba = %d, channel = %d, TID = %d, LUN = %d\n",
	       rcl->raidhba, rcl->raidchnl, ntohl(rcl->raidid), rcl->raidlun));
	MD_ASR_DEBUG(("    loffset = %d, lcapcty = %d, strpsize = %d\n",
	       ntohl(rcl->loffset), ntohl(rcl->lcapcty),
	       ntohs(rcl->strpsize)));
#endif
}

static void
md_asr_dump_rt(struct asr_raidtable *asr_rt)
{
#if MD_DEBUG
	u_int elmcnt;
	u_int i;

	elmcnt = ntohs(asr_rt->elmcnt);
	MD_ASR_DEBUG(("Raid Table RID %#x Version %#x NumElements %d\n",
	       ntohl(asr_rt->ridcode), ntohl(asr_rt->rversion), elmcnt));
	MD_ASR_DEBUG(("    timestamp = %#x, rchksum = %#x, raidFlags = %#x\n",
	       ntohs(asr_rt->timestamp), ntohl(asr_rt->rchksum),
	       ntohl(asr_rt->raidFlags)));
	MD_ASR_DEBUG(("    irocFlags = %#x, dirty = %#x, actionPriority = %#x\n",
	       asr_rt->irocFlags, asr_rt->dirty, asr_rt->actionPriority));
	MD_ASR_DEBUG(("    raidmagic = %#x, sparemagic = %#x, spareid = %#x\n",
	       ntohl(asr_rt->raidmagic), ntohl(asr_rt->sparedrivemagic),
	       asr_rt->spareid));
	MD_ASR_DEBUG(("    verifyDate = %#x, recreateDate = %#x\n",
	       ntohl(asr_rt->verifyDate), ntohl(asr_rt->recreateDate)));
	for (i = 0; i < elmcnt; i++)
		md_asr_dump_rcl(&asr_rt->ent[i], i);
#endif
}

static void
md_asr_dump_minfo( struct md_asr_meta_info *minfo)
{
	md_asr_dump_rb(minfo->asr_rb);
	md_asr_dump_rt(minfo->asr_rt);
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
static int __init
#else
int __init
#endif
md_asr_init(void)
{
	int error;

	error = md_init_workitem_support(NULL, &md_asr_syncer, "asr_syncer");
	if (error != 0)
		return (error);
	error = register_md_metapersonality(&md_asr_pers);
	if (error != 0)
		md_teardown_workitem_support(&md_asr_syncer);
	return (error);
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
static void md_asr_exit (void)
#else
void md_asr_exit(void)
#endif
{
	unregister_md_metapersonality(&md_asr_pers);
	md_teardown_workitem_support(&md_asr_syncer);
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
module_init(md_asr_init);
#endif
MODULE_LICENSE("Dual BSD/GPL");
