/*
 * Copyright (c) 2003, 2004 Adaptec Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions, and the following disclaimer,
 *    without modification.
 * 2. Redistributions in binary form must reproduce at minimum a disclaimer
 *    substantially similar to the "NO WARRANTY" disclaimer below
 *    ("Disclaimer") and any redistribution must be conditioned upon
 *    including a substantially similar Disclaimer requirement for further
 *    binary redistribution.
 * 3. Neither the names of the above-listed copyright holders nor the names
 *    of any contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * Alternatively, this software may be distributed under the terms of the
 * GNU General Public License ("GPL") version 2 as published by the Free
 * Software Foundation.
 *
 * NO WARRANTY
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGES.
 */
#include <linux/module.h>
#include <linux/config.h>
#include <linux/genhd.h>
#include <linux/list.h>
#include <linux/raid/emd/md.h>
#include <asm/div64.h>
#include "md_metadata_ddf.h"
#include "crc32.h"

#ifdef MD_DEBUG
#define MD_DDF_DEBUG(A)	printk A
#else
#define MD_DDF_DEBUG(A)
#endif

#define DDF_REV		"02.00.00"
#define DDF_NORMAL_BID	10
#define MD_DDF_MAX_PDE 15
#define MD_DDF_MAX_VDE 4

struct md_ddf_domain;
struct md_ddf_array_info;

struct md_ddf_sync_data
{
	mdk_workitem_t		  sync_workitem;
	struct md_ddf_domain	 *sync_dm;
	struct md_ddf_array_info *sync_ai;
	spinlock_t		 *sync_lock;
	int			  sync_flags;
#define DDF_SYNC_FLAGS_SYNC_INPROG	0x01
	/*
	 * We defer performing any meta-data sync operations
	 * until the array has been opened for write operation
	 * or we are degraded and start a rebuild operation or
	 * allocate a spare.  This avoids incrementing the
	 * sequence number and invalidating optimal members
	 * until our meta-data state changes.  This flag is
	 * the control for this behavior.
	 */
#define DDF_SYNC_FLAGS_SYNC_ALLOWED	0x02
};

/*
 * A DDF domain represents a collection of disks and arrays that are aware
 * of each other and can be seen by the same 'controller'.  The definition
 * of a controller varies in scope depending on the ability of different
 * levels of hardware and software to see the devices.  It could be a single
 * SCSI channel, or multiple channels on the same PCI card, or every disk in
 * the system.
 */
struct md_ddf_domain {
	struct list_head	ddf_domain_links;
	uint8_t			Controller_GUID[DDF_GUID_BYTES];
	struct md_ddf_sync_data sync_data;
	int			diskqueue_gencount;
	int			array_gencount;
	struct list_head	diskqueue;
	struct list_head	arrayqueue;
	struct ddf_cd_record	*cd;
	struct ddf_vd_record	*vdr;
	struct ddf_pd_record	*pdr;
	struct ddf_cd_record	*sync_cd;
	struct ddf_vd_record	*sync_vdr;
	struct ddf_pd_record	*sync_pdr;
	uint32_t		cd_length;
	uint32_t		pdr_length;
	uint32_t		vdr_length;
	uint16_t		crl;
	spinlock_t		ddf_lock;
	atomic_t		refcount;
	int			ddf_dirty_mask;
#define DM_DIRTY_CD	0x01
#define DM_DIRTY_PDR	0x02
#define DM_DIRTY_VDR	0x04
#define DM_DIRTY_LAST	DM_DIRTY_VDR
};

static LIST_HEAD(ddf_domains);
static spinlock_t ddf_domains_lock = SPIN_LOCK_UNLOCKED;
static int ddf_domains_gencount;

typedef enum {
	AI_VENDOR_UNKNOWN,
	AI_VENDOR_ADPT
} ai_vendor_t;

/*
 * Note:
 * Protected by the member_lock of the mddev this object shaddows.
 */
struct md_ddf_array_info {
	struct list_head	ddf_array_links;
	struct md_ddf_domain	*domain;
	struct ddf_vd_entry	*vde;
	struct ddf_vdc_record	*vdcr;
	struct ddf_vendor_log	*vendor;
	uint32_t		vendor_length;
	ai_vendor_t		vendor_type;
	atomic_t		refcount;
	int			ddf_dirty_mask;
#define	AI_DIRTY_VENDOR	(DM_DIRTY_LAST << 1)
#define	AI_DIRTY_VDCR	(DM_DIRTY_LAST << 2)

	uint32_t		*pds_array;
	uint32_t		*ss_array;
	int			ss_len;
	int			pds_len;
	mddev_t			*mddev;
	int			needed_disks;
	struct md_ddf_sync_data sync_data;
	uint64_t		checkpoint;
};

/*
 * Note:
 * Projected by the member_lock of the rdev this object shaddows.
 */
struct md_ddf_disk_info {
	struct list_head	ddf_disk_links;
	uint64_t		dev_size;
	kdev_t			dev;
	mdk_rdev_t		*rdev;
	struct md_ddf_domain	*domain;
	struct md_ddf_array_info *parent;
	int			cr_consumers;

	/* Pointers to things in the metadata that are hard to find */
	struct ddf_pd_entry	*pde;
	int			ss_len;
	int			pds_len;
	uint16_t		crl;
	uint8_t			vend_id[8];
	uint32_t		cd_length;
	uint32_t		pdr_length;
	uint32_t		vdr_length;
	uint32_t		cr_length;
	uint32_t		pdd_length;
	uint32_t		vendor_length;

	/* Here we keep a copy of the DDF info */
	struct ddf_header	*hdr_anchor;
	struct ddf_header	*hdr_pri;
	struct ddf_cd_record	*cd;
	struct ddf_pd_record	*pdr;
	struct ddf_vd_record	*vdr;
	uint8_t			*cr;	/* This holds the storage
					 * for the following three
					 * pointers.
					 */
	struct ddf_vdc_record	*vdcr;
	struct ddf_vuc_record	*vuc;
	struct ddf_sa_record	*sa;
	struct ddf_pdd_record	*pdd;
	struct ddf_bbm_log	*bbm;
	struct ddf_vendor_log	*vendor;

	uint32_t		ddf_dirty_mask;
#define DI_DIRTY_CD		DM_DIRTY_CD
#define DI_DIRTY_PDR		DM_DIRTY_PDR
#define DI_DIRTY_VDR		DM_DIRTY_VDR
#define DI_DIRTY_VENDOR		AI_DIRTY_VENDOR
#define DI_DIRTY_VDCR		AI_DIRTY_VDCR
#define DI_DIRTY_HDR_ANCHOR	(AI_DIRTY_VDCR << 1)
#define DI_DIRTY_HDR_PRI	(AI_DIRTY_VDCR << 2)
#define DI_DIRTY_HDR_SEC	(AI_DIRTY_VDCR << 3)
#define DI_DIRTY_VUC		(AI_DIRTY_VDCR << 4)
#define DI_DIRTY_SA		(AI_DIRTY_VDCR << 5)
#define DI_DIRTY_PDD		(AI_DIRTY_VDCR << 6)
#define DI_DIRTY_BBM		(AI_DIRTY_VDCR << 7)

	uint32_t		flags;
#define DI_FLAGS_PUT_AFTER_SYNC	0x01
};

static void ddf_dumpblock(void *, int) __attribute__((unused));
static int md_dev_op_ddf(kdev_t dev, mdk_dev_op_t, void *arg);
static int md_member_op_ddf(mdk_member_t *member, int, void *arg);
static int ddf_claim(kdev_t dev, void *arg);
static int ddf_get_anchor(kdev_t dev, struct ddf_header **hdr,
			  uint64_t *dev_size);
static int ddf_attach(kdev_t dev);
static int ddf_add_spare(struct md_ddf_disk_info *di);
struct md_ddf_disk_info *ddf_find_disk_info(kdev_t dev);
static int ddf_get_freesize(kdev_t dev, void *arg);
static int ddf_get_array_info(struct md_ddf_disk_info *di,
			      struct md_ddf_array_info **ai);
static int ddf_get_disk_info(kdev_t dev, struct ddf_header *anchor_header,
			     uint64_t dev_size, struct md_ddf_disk_info **di);
static int ddf_get_domain(struct md_ddf_disk_info *di,
			  struct md_ddf_domain **dm);
static int ddf_attach_drive(struct md_ddf_disk_info *di,
			    struct md_ddf_array_info *ai);
static mddev_t *ddf_attach_array(struct md_ddf_array_info *ai);
static int ddf_update_array_state(mddev_t *array);
static int ddf_to_md_disk_state(mddev_t *array, uint8_t pd_state);
static int ddf_to_md_array_state(u_int vde_state, u_int init_state,
				 mdu_mstate_t *state);
static int ddf_set_disk_state(mdk_rdev_t *disk, mdk_set_state_arg_t *sstate);
static int ddf_set_array_state(mddev_t *mddev, mdk_set_state_arg_t *sstate);
static uint8_t md_to_ddf_disk_state(mdu_mstate_t state);
static int md_to_ddf_array_state(mdu_mstate_t state,
				 u_int *ddf_state, u_int *init_state);
static struct md_ddf_domain *ddf_alloc_domain(void);
static struct md_ddf_array_info *ddf_alloc_array_info(struct md_ddf_domain *dm);
static struct md_ddf_disk_info *ddf_alloc_disk_info(kdev_t dev, uint64_t size);
static void ddf_free_domain(struct md_ddf_domain *dm);
static void ddf_free_array_info(struct md_ddf_array_info *);
static void ddf_free_disk_info(struct md_ddf_disk_info *);
static int md_ddf_work_control(mdk_member_t *member, int op, void *arg);
static int md_ddf_sync_handler(mddev_t *mddev, mdk_workitem_t *workitem);
static int ddf_set_checkpoint(mddev_t *mddev, mdk_checkpoint_arg_t *cp); 
static uint32_t ddf_calc_cksum(void *data, int len);
static struct ddf_header * ddf_init_header(int header_type, sector_t size,
					   int crl, int cr_size);
static void *ddf_init_cr(uint32_t crl, u_int num_recs);
static int ddf_init_cd(struct md_ddf_domain *dm);
static int ddf_init_pdr(struct md_ddf_domain *dm, int populated_pdes);
static struct ddf_pd_entry *
	   ddf_init_pde(struct ddf_pd_record *pdr, int entry, int type,
			int state, uint64_t size, char *guid, uint32_t ref);
static int ddf_init_vdr(struct md_ddf_domain *dm, int populated_vdes);
static struct ddf_vd_entry *
	   ddf_init_vde(struct ddf_vd_record *vdr, int entry, int type,
			int state, int init_state, char *name, char *guid);
static struct ddf_vdc_record *
	   ddf_init_vdcr(int count, int ss, int level, int rlq,
			 uint64_t size, int crl);
int ddf_init_sa(struct md_ddf_domain *dm, struct ddf_sa_record *sa);
static struct ddf_pdd_record *ddf_init_pdd(void);
static struct ddf_vendor_log *ddf_init_vendor(struct md_ddf_domain *dm);
static void ddf_schedule_sync(struct md_ddf_sync_data *, int wait, int enabl3e);
static int ddf_find_domain(struct md_ddf_domain **dm);
static int ddf_wipe_metadata(kdev_t dev);
static int ddf_set_array_info(mdk_member_t *member, mdu_array_info_t *marray);
static int ddf_add_new_disk(mddev_t *mddev, mdu_disk_info_t *ui);
static int ddf_exit_safemode(mddev_t *mddev);
static int ddf_enter_safemode(mddev_t *mddev);
static int ddf_get_spares(mddev_t *mddev, mdu_spare_disk_info_t *usi);
static int ddf_delete_spare(mdk_member_t *member);
static int ddf_create_spare(mddev_t *mddev, mdu_disk_info_t *spare_info);
static int ddf_activate_spare(mddev_t *mddev);
static void ddf_mark_missing_members(mddev_t *mddev);

static __inline struct md_ddf_domain *
ddf_dm_get(struct md_ddf_domain *dm)
{
	int success;

	success = atomic_inc_if_nonzero(&dm->refcount);
	MD_DDF_DEBUG(("ddf_dm_get %p refcount %d\n", dm,
		      (int)atomic_read(&dm->refcount)));
	if (success)
		return (dm);
	return (NULL);
}

static __inline
void ddf_dm_put(struct md_ddf_domain *dm)
{
	int delete;

	delete = atomic_dec_and_test(&dm->refcount);
	MD_DDF_DEBUG(("ddf_dm_put %p refcount %d\n", dm,
		      (int)atomic_read(&dm->refcount)));
	if (delete)
		ddf_free_domain(dm);
}

static __inline struct md_ddf_array_info *
ddf_ai_get(struct md_ddf_array_info *ai)
{
	int success;

	success = atomic_inc_if_nonzero(&ai->refcount);
	MD_DDF_DEBUG(("ddf_ai_get %p refcount %d\n", ai,
		      (int)atomic_read(&ai->refcount)));
	if (success)
		return (ai);
	return (NULL);
}

static __inline void ddf_ai_put(struct md_ddf_array_info *ai)
{
	int delete;

	delete = atomic_dec_and_test(&ai->refcount);
	MD_DDF_DEBUG(("ddf_ai_put %p refcount %d\n", ai,
		      (int)atomic_read(&ai->refcount)));
	if (delete)
		ddf_free_array_info(ai);
}

static __inline void
ddf_sync_data_get(struct md_ddf_sync_data *sync_data)
{
	if (sync_data->sync_ai != NULL)
		ddf_ai_get(sync_data->sync_ai);
	else
		ddf_dm_get(sync_data->sync_dm);
}

static __inline void
ddf_sync_data_put(struct md_ddf_sync_data *sync_data)
{
	if (sync_data->sync_ai != NULL)
		ddf_ai_put(sync_data->sync_ai);
	else
		ddf_dm_put(sync_data->sync_dm);
}

static mdk_metapersonality_t md_ddf_pers = {
	.name =		"DDF-"DDF_REV,
	.owner =	THIS_MODULE,
	.dev_op =	md_dev_op_ddf,
	.member_op =	md_member_op_ddf,
	.refcount =	ATOMIC_INIT(1),
};

static mdk_workitem_support_t	md_ddf_syncer;
#define MD_DDF_WORKOP_SYNC	0x01

static __inline uint32_t
ddf_gettimeofday(void)
{
	struct timeval tv;
	int32_t sec;

	do_gettimeofday(&tv);

	/* DDF mandates that the Epoch is Jan 1, 1980 */
	sec = tv.tv_sec - 315532800;
	if (sec < 0)
		sec = 0;
	return (sec);
}
 
static __inline int
ddf_md_to_ddf_raid_type(int md_level, int *ddf_level, int *ddf_rlq)
{
	switch (md_level) {
	case 0:
		*ddf_level = 0;
		*ddf_rlq = 0;
		break;
	case 1:
		*ddf_level = 1;
		*ddf_rlq = 1;
		break;
	default:
		return (-EINVAL);
	}

	return (0);
}

/* Convienience routine to dump a block in hex */
static void
ddf_dumpblock(void *data, int len)
{
        int i, j;

        for (i = 0; i < len; i++) {
                printk("\noffset= %03d: ", i);
                for (j = 0; j < 8; j++) {
                        if (i + j == len)
                                break;
                        printk("0x%02x ", (uint8_t)((uint8_t*)(data))[i + j]);
                }
                i += j - 1;
        }
        printk("\n");
}

/* Entry points into the module */
static int
md_dev_op_ddf(kdev_t dev, mdk_dev_op_t op, void *arg)
{
	switch (op) {
	case MDK_DEVOP_CLAIM:
		return (ddf_claim(dev, arg));
	case MDK_DEVOP_ATTACH:
		return (ddf_attach(dev));
	case MDK_DEVOP_GET_FREESIZE:
		return (ddf_get_freesize(dev, arg));
	case MDK_DEVOP_WIPE_METADATA:
		return (ddf_wipe_metadata(dev));
	case MDK_DEVOP_INITIALIZE:
	default:
		return (-EINVAL);
	}

	return (0);
}

static int
md_member_op_ddf(mdk_member_t *member, int op, void *arg)
{
	switch (op) {
	case MDK_MEMBER_OP_GET_CHECKPOINT:
	{
		mdk_checkpoint_arg_t *cp;
		struct md_ddf_array_info *ai;

		cp = (mdk_checkpoint_arg_t *)arg;
		ai = member->metadata_private;
		cp->checkpoint = ai->checkpoint;

		return (0);
	}
	case MDK_MEMBER_OP_SET_CHECKPOINT:
		return (ddf_set_checkpoint(to_mddev(member),
					   (mdk_checkpoint_arg_t *)arg));
	case MDK_MEMBER_OP_SET_STATE:
		if (member->type == MDK_MTYPE_RDEV)
			return (ddf_set_disk_state(to_rdev(member),
						   (mdk_set_state_arg_t *)arg));
		else
			return (ddf_set_array_state(to_mddev(member),
						    (mdk_set_state_arg_t*)arg));
		break;
	case MDK_MEMBER_OP_SET_ARRAY_INFO:
		return (ddf_set_array_info(member, arg));
	case MDK_MEMBER_OP_ADD_NEW_DISK:
		return (ddf_add_new_disk(to_mddev(member), arg));
	case MDK_MEMBER_OP_RELEASE:
		if (member->type == MDK_MTYPE_RDEV) {
			struct md_ddf_disk_info *di;

			di = member->metadata_private;
			ddf_free_disk_info(di);
		} else if (member->subtype == MDK_MSUBTYPE_POOL_SPARE) {
			struct md_ddf_domain *pool_dm;

			pool_dm = member->metadata_private;
			ddf_dm_put(pool_dm);
		} else {
			struct md_ddf_array_info *ai;

			ai = member->metadata_private;
			ddf_ai_put(ai);
		}
		member->metadata_private = NULL;
		md_metapers_put(member->metapers);
		member->metapers = NULL;
		return (0);
	case MDK_MEMBER_OP_ENTER_SAFEMODE:
	case MDK_MEMBER_OP_EXIT_SAFEMODE:
	{
		mddev_t *mddev;

		/*
		 * We only care about safemode events on
		 * array objects.
		 */
		if (member->type != MDK_MTYPE_MDDEV)
			return (0);

		mddev = to_mddev(member);
		if (op == MDK_MEMBER_OP_ENTER_SAFEMODE)
			return ddf_enter_safemode(mddev);
		else
			return ddf_exit_safemode(mddev);
		/* NOTREACHED */
	}
	case MDK_MEMBER_OP_CREATE_SPARE:
		return ddf_create_spare(to_mddev(member),
					(mdu_disk_info_t *)arg);
	case MDK_MEMBER_OP_ACTIVATE_SPARE:
		return ddf_activate_spare(to_mddev(member));
	case MDK_MEMBER_OP_GET_SPARES:
		return ddf_get_spares(to_mddev(member),
				      (mdu_spare_disk_info_t *)arg);
	case MDK_MEMBER_OP_DELETE_SPARE:
		return ddf_delete_spare(member);
	case MDK_MEMBER_OP_SET_ARRAY_ACTION:
	default:
		return (-EOPNOTSUPP);
	}

	return (0);
}

/* All that we want to do here is see if a valid anchor exists.  No other
 * actions are needed, nor can we save away any state.
 */
static int
ddf_claim(kdev_t dev, void *arg)
{
	struct ddf_header *header;
	int error;
	int *bid;

	bid = (int *)arg;

	error = ddf_get_anchor(dev, &header, NULL);

	if (error != 0)
		*bid = 0;
	else
		*bid = DDF_NORMAL_BID;

	if (header != NULL)
		kfree(header);

	return (0);
}

/* Return the DDF anchor record from a given drive. This function allocates
 * the buffer, but the caller needs to free it.  Since the claim code needs
 * to know the difference between a failure to get the block and the block
 * not having the right DDF info, only worry about returning status for the
 * first case.  The callers will each have to examine the contents.
 */
static int
ddf_get_anchor(kdev_t dev, struct ddf_header **hdr, uint64_t *dev_size)
{
	struct gendisk *gd;
	struct ddf_header *header;
	uint64_t ddf_anchor_base;
	uint32_t crc;

	/* Find the end of the disk, which holds Anchor DDF block */
	gd = get_gendisk(dev);
	if (gd == NULL)
		return (0);

	if (dev_size)
		*dev_size = gd->part[MINOR(dev)].nr_sects;
	ddf_anchor_base = gd->part[MINOR(dev)].nr_sects - 1;

	*hdr = NULL;
	header = (struct ddf_header *)kmalloc(DDF_HEADER_BYTES, GFP_KERNEL);
	if (header == NULL) {
		printk("md_metadata_ddf: failed to allocate buffer\n");
		return (-ENOMEM);
	}

	/* Get the anchor block and validate it */
	if (md_read_blocks(dev, ddf_anchor_base, 1, (uint8_t *)header)!= 0) {
		printk("ddf_get_anchor: cannot read anchor block\n");
		kfree(header);
		return (-EIO);
	}

	crc = ddf_calc_cksum(header, DDF_HEADER_BYTES);
	if ((le32_to_cpu(header->Signature) != DDF_HEADER_SIGNATURE) ||
	    (le32_to_cpu(header->Header_Type) != DDF_HEADER_ANCHOR) ||
	    (header->CRC != crc) ||
	    (le64_to_cpu(header->Primary_Header_LBA) >= ddf_anchor_base)) {
		kfree(header);
		return (-EINVAL);
	}

	*hdr = header;
	return (0);
}

/*
 * Do the work of fetching all of the records and doing fun things
 * with them.
 */
static int
ddf_attach(kdev_t dev)
{
	struct ddf_header *header;
	struct md_ddf_disk_info *di, *tmp_info;
	struct md_ddf_array_info *ai;
	struct md_ddf_domain *dm;
	uint64_t dev_size;
	u_long dmflags;
	mddev_t *mddev;
	int error;

	di = NULL;
	header = NULL;

	/* Start by getting the archor header */
	if ((error = ddf_get_anchor(dev, &header, &dev_size)) != 0)
		return (error);

	/* Next get the DDF records for the disk_info structure */
	if ((error = ddf_get_disk_info(dev, header, dev_size, &di)) != 0)
		return (error);

	/* Put this disk into the proper domain */
	if ((error = ddf_get_domain(di, &dm)) != 0) {
		ddf_free_disk_info(di);
		return (error);
	}

	/* Check to see if we already know about this disk.  If not, add it
	 * to the domain.
	 */
	spin_lock_irqsave(&dm->ddf_lock, dmflags);
	list_for_each_entry(tmp_info, &dm->diskqueue, ddf_disk_links) {
		kdev_t tmp_dev;

		/*
		 * Skip all non-matching entries.
		 */
		if (memcmp(&di->pde->PD_GUID[0],
			   &tmp_info->pde->PD_GUID[0], DDF_GUID_BYTES) != 0)
			continue;

		/* Duplicate GUID found, but device numbers don't
		 * match.  Bail out.
		 */
		tmp_dev = tmp_info->dev;
		spin_unlock_irqrestore(&dm->ddf_lock, dmflags);
		if (tmp_info->dev != di->dev) {
			printk("DDF disk %x Duplicates %x! "
			       "Bailing out\n", di->dev, tmp_info->dev);
		} else {
			/* Duplicate GUID and same device numbers.
			 * Must be the same disk.
			 */
			printk("ddf_attach: Disk %x attached "
			       "more than once!\n", di->dev);
		}
		ddf_free_disk_info(di);
		return (-EINVAL);
	}
	list_add(&di->ddf_disk_links, &dm->diskqueue);
	dm->diskqueue_gencount++;
	spin_unlock_irqrestore(&dm->ddf_lock, dmflags);

	/* Is this drive a spare? */
	if ((le16_to_cpu(di->pde->PD_Type) &
	    (DDF_PDE_GLOBAL_SPARE | DDF_PDE_CONFIG_SPARE)) != 0)
		return (ddf_add_spare(di));

	/* Is the drive a valid part of an array? */
	if (((le16_to_cpu(di->pde->PD_Type) & DDF_PDE_PARTICIPATING) == 0) ||
	    ((le16_to_cpu(di->pde->PD_State)
	     & (DDF_PDE_FAILED|DDF_PDE_ONLINE
	      | DDF_PDE_REBUILD|DDF_PDE_TRANSITION)) == 0))
		return (0);

	/* Now get the array_info structure */
	if ((error = ddf_get_array_info(di, &ai)) != 0) {
		ddf_free_disk_info(di);
		return (error);
	}

	/*
	 * Check the sequence numbers to make sure that this member
	 * is in sync with the others.  We should really be more
	 * intelligent here....
	 */
	if (di->vdcr->Sequence_Number != ai->vdcr->Sequence_Number) {
		printk("DDF: Disk %d/%d is out of date, rejecting\n",
		       MAJOR(di->dev), MINOR(di->dev));
		ddf_free_disk_info(di);
		return (-EINVAL);
	}

	/* If the mddev is NULL, the this array is new and hadn't been
	 * initialized yet.
	 */
	if (ai->mddev == NULL)
		mddev = ddf_attach_array(ai);
	else
		mddev = ai->mddev;

	if (mddev == NULL) {
		printk("ddf_attach: ddf_attach_array returned NULL\n");
		ddf_free_disk_info(di);
		return (-EINVAL);
	}

	/*
	 * We discard failed drives after getting the
	 * array information since we want failed arrays
	 * to show up, just like in the BIOS.
	 */
	if ((le16_to_cpu(di->pde->PD_State) & DDF_PDE_FAILED) != 0) {
		ddf_free_disk_info(di);
		return (-ENXIO);
	}

	return (ddf_attach_drive(di, ai));
}

static int
ddf_attach_drive(struct md_ddf_disk_info *di, struct md_ddf_array_info *ai)
{
	struct ddf_pd_entry *pde;
	mdk_member_t *rdmember, *mdmember;
	mdk_rdev_t *rdev;
	mddev_t *mddev;
	uint32_t pdseq;
	int i;
	int pds_index;

	mddev = ai->mddev;
	mdmember = to_member(mddev);

	/* Create an rdev member object for the disk */
	if ((rdev = md_rdev_alloc(NULL, MDK_MSUBTYPE_CHILD, di->dev,
				  MDK_PRIORITY_UNKNOWN)) == NULL) {
		printk("ddf_attach_drive: cannot alloc md_disk\n");
		return (-EINVAL);
	}

	di->rdev = rdev;
	rdmember = to_member(rdev);
	rdmember->data_size = (sector_t)le64_to_cpu(di->pde->Configured_Size);
	rdmember->metadata_private = di;
	rdmember->state = ddf_to_md_disk_state(mddev, di->pde->PD_State);
	rdmember->metapers = md_metapers_get(&md_ddf_pers);

	/* Verify that this drive is actually in the array */
	pde = di->pde;
	pds_index = -1;
	for (i = 0; i < le16_to_cpu(di->vdcr->Primary_Element_Count); i++) {
		pdseq = ai->pds_array[i];
		if (pdseq == pde->PD_Reference) {
			pds_index = i;
			break;
		}
	}
	if (pds_index == -1) {
		printk("DDF: Cannot find entry for drive in VD\n");
		member_put(rdmember);
		return (-EINVAL);
	}

	/* XXX Adaptec BIOS doesn't always get this field right.
	 * Do this here since we need to look at both array and disk
	 * attributes.
	 */
	if ((rdmember->data_size == 0) && (mddev->level == RAID1))
		rdmember->data_size = mdmember->data_size;

	/* The only way to know the data offset of the member is to
	 * look at the VDCR.  Since we are this far along, we know
	 * that we have an array, and thus a VDCR.
	 */
	rdmember->data_offset = le64_to_cpu(ai->ss_array[pds_index]);
	if (rdmember->data_offset == 0xffffffffffffffffULL) {
		printk("DDF: member starting block invalid, assuming 0\n");
		rdmember->data_offset = 0;
	}
	
	/*
	 * Add the disk into the array.
	 */
	rdmember->priority = pds_index;
	mdmember->pers->raid_op(mdmember, MDK_RAID_OP_INIT_MEMBER,
				(void *)rdmember);
	md_member_add_child(mdmember, rdmember);

	/* See if the array state has changed.  It is initialized to a
	 * non-running state to make sure that it does get started early.
	 * If a state transition happens, then notify the raid module.
	 */
	ddf_update_array_state(mddev);

	mdmember->pers->raid_op(mdmember, MDK_RAID_OP_ADD_MEMBER,
				(void *)rdmember);
	return (0);
}

static mddev_t *
ddf_attach_array(struct md_ddf_array_info *ai)
{
	mddev_t *mddev;
	mdk_member_t *member;
	mdk_personality_t *raid_pers;
	int error;
	int priority;

	/* Can we deal with this array? */
	if ((ai->vdcr->Primary_RAID_Level != 0) &&
	    (ai->vdcr->Primary_RAID_Level != 1)) {
		printk("DDF raid level %d not supported\n",
		       ai->vdcr->Primary_RAID_Level);
		return (NULL);
	}
	if (ai->vdcr->Secondary_Element_Count != 0x01) {
		/* XXX Handle the Adaptec interpretation */
		if (ai->vdcr->Secondary_Element_Count != 0x00) {
			printk("Multilevel DDF arrays not supported\n");
			return (NULL);
		}
	}

	priority = DDF_DEFAULT_PRIORITY;
	if (ai->vendor_type == AI_VENDOR_ADPT) {
		struct adpt_vendor_data *adpt;

		adpt = (struct adpt_vendor_data *)ai->vendor->data;
		if ((adpt->irocFlags & DDF_ADPT_IF_BOOT_VOLUME) != 0)
			priority = DDF_BOOT_PRIORITY;
	}
	mddev = md_mddev_alloc(NULL, MDK_MSUBTYPE_TOPLEVEL, priority);
	if (mddev == NULL) {
		printk("Cannot allocate md_array\n");
		return (NULL);
	}
	member = to_member(mddev);
	ai->mddev = mddev;
	ai->sync_data.sync_lock = &member->member_lock;

	switch (ai->vdcr->Primary_RAID_Level) {
	case 0:
		mddev->level = 0;
		ai->needed_disks = le16_to_cpu(ai->vdcr->Primary_Element_Count);
		break;
	case 1:
		mddev->level = 1;
		ai->needed_disks = 1;
		break;
	default:
		/* This should not be reached */
		break;
	}

	mddev->preferred_unit = -1;
	mddev->chunk_size = (1 << ai->vdcr->Stripe_Size) * 512;
	mddev->max_disks =le16_to_cpu(ai->vdcr->Primary_Element_Count);
	member->data_size = (sector_t)le64_to_cpu(ai->vdcr->Size);
	member->data_offset = 0;	/* XXX need a real value here */
	/*
	 * An array without any members must be in the
	 * failed state unless it is a spare array or
	 * it is marked for initialization. This will be
	 * updated as children are added.
	 */
	error = ddf_to_md_array_state(ai->vde->VD_State, ai->vde->Init_State,
				      &member->state);
	if (error != 0) {
		member_put(member);
		return (NULL);
	}
	member->state &= ~MDU_MSTATE_MAIN_MASK;
	member->state |= MDU_MSTATE_FAILED;

	/*
	 * If the array was shutdown unexpectedly, the dirty
	 * flag will still be set.  If set, clear any checkpoint,
	 * and mark our state as dirty.
	 *
	 * Some Adaptec BIOSes store the dirty information
	 * inside the vendor log.
	 */
 	if (ai->vendor_type == AI_VENDOR_ADPT) {
		struct adpt_vendor_data *adpt;

		adpt = (struct adpt_vendor_data *)ai->vendor->data;
		if (adpt->Dirty != 0)
			ai->vde->VD_State |= DDF_VDE_DIRTY;
	}
	if ((ai->vde->VD_State & DDF_VDE_DIRTY) != 0) {
		MD_DDF_DEBUG(("VD_State found dirty!\n"));
		member->state |= MDU_MSTATE_DIRTY;
		ai->checkpoint = 0;
		mddev->verify_op = MD_SYNCOP_VERIFY_FIX;
	} else {

		mddev->verify_op = MD_SYNCOP_VERIFY;
 		if (ai->vendor_type == AI_VENDOR_ADPT) {
			struct adpt_vendor_data *adpt;
			
			adpt = (struct adpt_vendor_data *)ai->vendor->data;

		 	if (adpt->CheckpointActivity
			 == DDF_ADPT_CHKPNT_ACT_VERIFY) {
				MD_DDF_DEBUG(("Array is verifying\n"));
				ai->vde->VD_State |= DDF_VDE_DIRTY;
				member->state |= MDU_MSTATE_DIRTY;
			}

			if ((adpt->irocFlags
			  &  DDF_ADPT_IF_VERIFY_WITH_AUTOFIX) != 0)
				mddev->verify_op = MD_SYNCOP_VERIFY_FIX;
		}
	}

	member->metapers = md_metapers_get(&md_ddf_pers);
	member->metadata_private = ddf_ai_get(ai);
	raid_pers = md_find_pers(mddev->level);
	if (raid_pers == NULL) {
		/* No personality was found for this.  Should we tear down
		 * all of the objects and wait for later?
		 */
		member_put(member);
		return (NULL);
	}

	/* Call the raid module to do basic init on the mddev */
	error = raid_pers->raid_op(to_member(mddev), MDK_RAID_OP_INIT, NULL);
	if (error != 0) {
		member_put(member);
		return (NULL);
	}

	/*
	 * Add to toplevel arrays.
	 */
	md_member_add_child(&toplevel_arrays, member);

	/*
	 * Lose extra reference from md_find_pers.  The
	 * RAID personality owns the personality reference
	 * for this object from INIT through RELEASE.
	 */
	md_pers_put(raid_pers);

	return (mddev);
}

struct md_ddf_disk_info *
ddf_find_disk_info(kdev_t dev)
{
	struct md_ddf_domain *dm;
	struct md_ddf_disk_info *list_di;
	struct md_ddf_disk_info *di;
	u_long domains_flags;
	u_long dmflags;

	di = NULL;
	spin_lock_irqsave(&ddf_domains_lock, domains_flags);
	list_for_each_entry(dm, &ddf_domains, ddf_domain_links) {
		spin_lock_irqsave(&dm->ddf_lock, dmflags);
		list_for_each_entry(list_di, &dm->diskqueue, ddf_disk_links) {
			if (list_di->dev != dev)
				continue;
			di = list_di;
			break;
		}
		spin_unlock_irqrestore(&dm->ddf_lock, dmflags);
	}
	spin_unlock_irqrestore(&ddf_domains_lock, domains_flags);

	return (di);
}

static int
ddf_get_freesize(kdev_t dev, void *arg)
{
	uint64_t freesize;
	sector_t *psize;
	struct md_ddf_disk_info *di;
	struct gendisk *gd;
 
	/*
	 * Lookup disk in the domain and use the configured
	 * capacity if we have it.
	 * then fallback ....
	 */
	psize = arg;
	di = ddf_find_disk_info(dev);
	if (di != NULL) {
		*psize = le64_to_cpu(di->pde->Configured_Size);
		return (0);
	}
 	gd = get_gendisk(dev);
 	if (gd == NULL){
 		printk("ddf_get_freesize: gendisk is NULL\n");
 		return (-ENODEV);
 	}
 		
	/*
	 * Round to a 100MB boundary and then subtract the metadata area.
	 * Note that here, MB is a drive industry MB: 1,000,000 bytes.
	 */
	freesize = (uint64_t)gd->part[MINOR(dev)].nr_sects * 512;
	do_div(freesize, 100000000);
	freesize *= 100000000;
	freesize /= 512;
	/* Reserve 32MB for meta-data. */
	freesize -= (32 * 1024 * 1024) / 512;
 	*((sector_t *)arg) = freesize;
	MD_DDF_DEBUG(("ddf_get_freesize: 0x%lx\n", *(sector_t *)arg));
 	return (0);
 }

#define DDF_VALIDATE_OFFSET(offset, len, rec_name)			       \
do {									       \
	if ((offset == 0xffffffff) ||					       \
	    (len == 0xffffffff)) {					       \
		printk("ddf_get_disk_info: %s len/offset fields unitialized\n",\
		       rec_name);					       \
		ddf_free_disk_info(di);					       \
		return (-EINVAL);					       \
	}								       \
	offset += ddf_primary_base;					       \
	if (offset >= anchor_base) {					       \
		printk("ddf_get_disk_info: %s out of bounds\n", rec_name);     \
		ddf_free_disk_info(di);					       \
		return (-EINVAL);					       \
	}								       \
	MD_DDF_DEBUG(("ddf_get_disk_info: %s len %x, offset %llx \n",	       \
		       rec_name, len, offset));				       \
} while (0)

#define DDF_ALLOCATE_RECORD(ptr, sec_len, rec_name)			       \
do {									       \
	ptr = kmalloc(sec_len * 512, GFP_KERNEL);			       \
	if (ptr == NULL) {						       \
		printk("ddf_get_disk_info: failed to allocate %s record\n",    \
		       rec_name);					       \
		ddf_free_disk_info(di);					       \
		return (-ENOMEM);					       \
	}								       \
} while (0)

/* Probe a disk for the relevant disk-specific DDF records, and return
 * them in a disk_info struct.  The caller is responsible for freeing this.
 */
static int
ddf_get_disk_info(kdev_t dev, struct ddf_header *anchor_header,
		  uint64_t dev_size, struct md_ddf_disk_info **disk_info)
{
	struct md_ddf_disk_info *di;
	uint64_t anchor_base;
	uint64_t ddf_primary_base;
	uint64_t offset;
	uint32_t pdref;
	uint32_t crc;
	uint16_t maxpde;
	uint16_t max_parts;
	uint8_t *buffer;
	int retval;
	int i;

	*disk_info = NULL;

	di = ddf_alloc_disk_info(dev, dev_size);
	if (di == NULL) {
		printk("ddf_get_disk_info: cannot allocate disk_info\n");
		kfree(anchor_header);
		return (-ENOMEM);
	}
	di->hdr_anchor = anchor_header;

	/* Find and read the Primary DDF header block */
	anchor_base = dev_size - 1;
	ddf_primary_base = le64_to_cpu(di->hdr_anchor->Primary_Header_LBA);
	if (ddf_primary_base >= anchor_base) {
		printk("ddf_get_disk_info: PRI Header out of bounds\n");
		ddf_free_disk_info(di);
		return (-EINVAL);
	}

	di->hdr_pri = kmalloc(DDF_HEADER_BYTES, GFP_KERNEL);
	if (di->hdr_pri == NULL) {
		printk("ddf_get_disk_info: failed to allocate PRI Hdr buf\n");
		ddf_free_disk_info(di);
		return (-ENOMEM);
	}

	if (md_read_blocks(dev, ddf_primary_base, 1,
			   (uint8_t *)di->hdr_pri) != 0) {
		printk("ddf_get_disk_info: cannot read primary header\n");
		ddf_free_disk_info(di);
		return (-EIO);
	}

	/* Validate the primary header. */
	crc = ddf_calc_cksum(di->hdr_pri, DDF_HEADER_BYTES);
	if ((le32_to_cpu(di->hdr_pri->Signature) != DDF_HEADER_SIGNATURE) ||
	    (le32_to_cpu(di->hdr_pri->Header_Type) != DDF_HEADER_PRIMARY) ||
	    (di->hdr_pri->CRC != crc)) {
		printk("ddf primary header signature does not match\n");
		ddf_free_disk_info(di);
		return (-EINVAL);
	}

	/*
	 * Pull all global configuration records off the disk
	 * for validation against our domain.  Start with the
	 * controller record.
	 */
	offset = le32_to_cpu(di->hdr_pri->cd_section);
	di->cd_length = le32_to_cpu(di->hdr_pri->cd_length);
	DDF_VALIDATE_OFFSET(offset, di->cd_length, "CD");
	DDF_ALLOCATE_RECORD(di->cd, di->cd_length, "CD");
	retval = md_read_blocks(dev, offset, di->cd_length, (uint8_t *)di->cd);
	if (retval != 0) {
		printk("ddf_get_disk_info: cannot read CD\n");
		ddf_free_disk_info(di);
		return (-EIO);
	}

	crc = ddf_calc_cksum(di->cd, di->cd_length * 512);
	if (le32_to_cpu(di->cd->Signature) != DDF_CD_SIGNATURE ||
	    di->cd->CRC != crc) {
		printk("ddf_get_disk_info: cd signature does not match: %x\n",
			le32_to_cpu(di->cd->Signature));
		ddf_free_disk_info(di);
		return (-EINVAL);
	}

	memcpy(&di->vend_id[0], &di->cd->Controller_GUID[0], 8);

	/* Pull in the PD records. */
	offset = le32_to_cpu(di->hdr_pri->pdr_section);
	di->pdr_length = le32_to_cpu(di->hdr_pri->pdr_length);
	DDF_VALIDATE_OFFSET(offset, di->pdr_length, "PD");
	DDF_ALLOCATE_RECORD(di->pdr, di->pdr_length, "PD");
	if (md_read_blocks(dev, offset, di->pdr_length,
			   (uint8_t *)di->pdr) != 0) {
		printk("ddf_get_disk_info: cannot read pdr\n");
		ddf_free_disk_info(di);
		return (-EIO);
	}

	crc = ddf_calc_cksum(di->pdr, di->pdr_length * 512);
	if (le32_to_cpu(di->pdr->Signature) != DDF_PDR_SIGNATURE ||
	    di->pdr->CRC != crc) {
		printk("ddf pdr signature does not match\n");
		ddf_free_disk_info(di);
		return (-EINVAL);
	}

	/* Pull in the VD Records. */
	offset = le32_to_cpu(di->hdr_pri->vdr_section);
	di->vdr_length = le32_to_cpu(di->hdr_pri->vdr_length);
	DDF_VALIDATE_OFFSET(offset, di->vdr_length, "VD");
	DDF_ALLOCATE_RECORD(di->vdr, di->vdr_length, "VD");
	if (md_read_blocks(dev, offset, di->vdr_length,
			   (uint8_t *)di->vdr) != 0) {
		printk("ddf_get_disk_info: cannot read VDR\n");
		ddf_free_disk_info(di);
		return (-EIO);
	}
	crc = ddf_calc_cksum(di->vdr, di->vdr_length * 512);
	if (le32_to_cpu(di->vdr->Signature) != DDF_VD_RECORD_SIGNATURE ||
	    di->vdr->CRC != crc) {
		printk("ddf_get_disk_info: VDR signature does not match\n");
		ddf_free_disk_info(di);
		return (-EINVAL);
	}

	/* Now pull in records that are local to this drive.
	 * Start with the Physical Disk Data record for this
	 * drive so we know what to reference in the rest of
	 * the metadata.
	 */
	offset = le32_to_cpu(di->hdr_pri->pdd_section);
	di->pdd_length = le32_to_cpu(di->hdr_pri->pdd_length);
	DDF_VALIDATE_OFFSET(offset, di->pdd_length, "PDD");
	DDF_ALLOCATE_RECORD(di->pdd, di->pdd_length, "PDD");
	if (md_read_blocks(dev, offset, di->pdd_length,
			   (uint8_t *)di->pdd) != 0) {
		printk("ddf_get_disk_info: cannot read pdd\n");
		ddf_free_disk_info(di);
		return (-EIO);
	}

	crc = ddf_calc_cksum(di->pdd, di->pdd_length * 512);
	if (le32_to_cpu(di->pdd->Signature) != DDF_PDD_SIGNATURE ||
	    di->pdd->CRC != crc) {
		printk("ddf_get_disk_info: pdd signature does not match\n");
		ddf_free_disk_info(di);
		return (-EINVAL);
	}

	/* Find the correct PD Entry for our drive */
	di->pde = NULL;
	pdref = di->pdd->PD_Reference;
	for (i = 0; i < le16_to_cpu(di->pdr->Max_PD_Entries); i++) {
		di->pde = &di->pdr->entry[i];
		if (di->pde->PD_Reference == pdref)
			break;
		di->pde = NULL;
	}

	if (di->pde == NULL) {
		printk("ddf_get_disk_info: Cannot find pde for disk!\n");
		ddf_free_disk_info(di);
		return (-EINVAL);
	}

	MD_DDF_DEBUG(("ddf_get_disk_info: found disk ref= 0x%x, "
		      "type= 0x%x, state= 0x%x, size= %lld\n",
		      di->pde->PD_Reference,
		      le16_to_cpu(di->pde->PD_Type),
		      le16_to_cpu(di->pde->PD_State),
		      le64_to_cpu(di->pde->Configured_Size)));

	/*
	 * See if there is an interesting Vendor-Unique record.
	 */
	offset = le32_to_cpu(di->hdr_pri->vendor_section);
	di->vendor_length = le32_to_cpu(di->hdr_pri->vendor_length);
	if ((offset != 0xffffffff) && (di->vendor_length != 0xffffffff)
	 && (offset += ddf_primary_base) < anchor_base) {
		DDF_ALLOCATE_RECORD(di->vendor, di->vendor_length, "vendor");
		if (md_read_blocks(dev, offset, di->vendor_length,
				   (uint8_t *)di->vendor) != 0) {
			printk("ddf_get_disk_info: cannot read vendor data\n");
			ddf_free_disk_info(di);
			return (-EIO);
		}
	}

	/* Parse the configuration records.
	 * XXX The VDCR might be really big.
	 *     This needs to be handled.
	 */
	offset = le32_to_cpu(di->hdr_pri->cr_section);
	di->cr_length = le32_to_cpu(di->hdr_pri->cr_length);
	DDF_VALIDATE_OFFSET(offset, di->cr_length, "CFG");
	DDF_ALLOCATE_RECORD(di->cr, di->cr_length, "CFG");
	if (md_read_blocks(dev, offset, di->cr_length,
			   di->cr) != 0) {
		printk("ddf_get_disk_info: cannot read config records\n");
		ddf_free_disk_info(di);
		return (-EIO);
	}

	di->crl = le16_to_cpu(di->hdr_pri->Configuration_Record_Length);
	max_parts = le16_to_cpu(di->hdr_pri->Max_Partitions);
	for (i = 0, buffer = di->cr; i < max_parts;
	     i++, buffer += di->crl * 512) {
		struct ddf_header *hdr;
		uint32_t signature;

		hdr = (struct ddf_header *)buffer;
		signature = le32_to_cpu(hdr->Signature);
		crc = ddf_calc_cksum(buffer, di->crl * 512);
		if (hdr->CRC != crc) {
			printk("DDF: Invalid Configuration Record CRC\n");
			continue;
		}
		switch (signature) {
		case DDF_VDCR_SIGNATURE:

			/* Only assign this pointer once.  We will use
			 * pointer arithmitec to index to the right spot
			 * later on.
			 */
			if (di->vdcr == NULL)
				di->vdcr = (struct ddf_vdc_record *)buffer;
			else {
				printk("ddf_get_disk_info: multiple VDCR "
				       "records found on disk, ignoring\n");
				break;
			}
			
			/* Store the variable-sized arrays from the VDCR */
			/* XXX This is still in a bit of flux.  For now, we
			 * will assume that the size is based off of the
			 * Max_PD_entries field, and special case the one
			 * anomoly mentioned in the spec.  Hopefully this
			 * will be cleared up and clarified in future version
			 * of the spec.
			 */
			maxpde = le16_to_cpu(di->hdr_pri->Max_PD_Entries);
			if ((maxpde == 0x40) && (di->crl == 0x03)) {
				di->pds_len = 512;
				di->ss_len = 512;
			} else {
				di->pds_len = maxpde * 4;
				di->ss_len = maxpde * 8;
			}
			di->cr_consumers++;
			break;
		case DDF_SA_SIGNATURE:
			if (di->sa == NULL) {
				di->sa = (struct ddf_sa_record *)buffer;
				di->cr_consumers++;
				MD_DDF_DEBUG(("ddf_get_disk_info: Found SA "
					      "record: type 0x%x\n",
					      di->sa->Spare_Type));
			} else
				printk("ddf_get_disk_info: Multiple SA records "
				       "found on disk, ignoring\n");
			break;
		case DDF_VUCR_SIGNATURE:
			if (di->vuc == NULL) {
				di->vuc = (struct ddf_vuc_record *)buffer;
				di->cr_consumers++;
			} else
				printk("DDF: Multiple VUCR records found on"
				       "disk, ignoring\n");
			break;
		default:
			printk("DDF: Unknown configuration record\n");
			ddf_free_disk_info(di);
			return (-EINVAL);
		}
	}
	*disk_info = di;
	return (0);
}
#undef DDF_VALIDATE_OFFSET
#undef DDF_ALLOCATE_RECORD

/* Read and validate the global state of the passed in device.
 * If this device is part of an existing domain, reconsile the
 * device's records with the current domain records.  Otherwise
 * allocate a new domain and register the disk there.
 */
static int
ddf_get_domain(struct md_ddf_disk_info *di, struct md_ddf_domain **domain)
{
	static DECLARE_MUTEX(ddf_domain_cfg_sem);
	struct md_ddf_domain *dm;
	u_long flags;
	int retval;

	down(&ddf_domain_cfg_sem);

	/* Search for a pre-exising domain that matches this 
	 * drive's controller GUID.
	 */
	spin_lock_irqsave(&ddf_domains_lock, flags);
	list_for_each_entry(dm, &ddf_domains, ddf_domain_links) {
		
		retval = memcmp(di->cd->Controller_GUID,
				dm->cd->Controller_GUID, DDF_GUID_BYTES);
		if (retval == 0)
			break;
	}

	if (&dm->ddf_domain_links == &ddf_domains) {
		/*
		 * No match.  Allocate new domain.
		 */
		spin_unlock_irqrestore(&ddf_domains_lock, flags);
		MD_DDF_DEBUG(("Allocating new Domain\n"));
		dm = ddf_alloc_domain();
		if (dm == NULL) {
			printk("ddf_get_domain: cannot create domain\n");
			up(&ddf_domain_cfg_sem);
			return (-ENOMEM);
		}

		/* This is the first disk in the domain, so use
		 * its metadata to initialized the domain.
		 */
		dm->sync_cd = kmalloc(di->cd_length * 512, GFP_KERNEL);
		dm->sync_pdr = kmalloc(di->pdr_length * 512, GFP_KERNEL);
		dm->sync_vdr = kmalloc(di->vdr_length * 512, GFP_KERNEL);
		if (dm->sync_cd == NULL || dm->sync_pdr == NULL ||
		    dm->sync_vdr == NULL) {
			printk("ddf_get_domain: cannot allocate records\n");
			ddf_free_domain(dm);
			up(&ddf_domain_cfg_sem);
			return (-ENOMEM);
		}
		dm->cd = di->cd;
		dm->pdr = di->pdr;
		dm->vdr = di->vdr;
		dm->cd_length = di->cd_length;
		dm->pdr_length = di->pdr_length;
		dm->vdr_length = di->vdr_length;
		memcpy(dm->Controller_GUID, di->cd->Controller_GUID,
		       DDF_GUID_BYTES);

		spin_lock_irqsave(&ddf_domains_lock, flags);
		list_add_tail(&dm->ddf_domain_links, &ddf_domains);
		ddf_domains_gencount++;
	} else {
		/* XXX Validate this drive's records agains the
		 * pre-existing records and then...
		 */
		di->pde = &dm->pdr->entry[di->pde - di->pdr->entry];
		kfree(di->cd);
		kfree(di->pdr);
		kfree(di->vdr);
		di->cd = dm->cd;
		di->pdr = dm->pdr;
		di->vdr = dm->vdr;
		dm = ddf_dm_get(dm);
	}
	di->domain = dm;
	*domain = dm;
	spin_unlock_irqrestore(&ddf_domains_lock, flags);

	up(&ddf_domain_cfg_sem);
	return (0);
}

static int
ddf_get_array_info(struct md_ddf_disk_info *di,
		   struct md_ddf_array_info **array_info)
{
	struct md_ddf_domain *dm;
	struct md_ddf_array_info *ai;
	u_long flags;
	int i;

	dm = di->domain;

	/*
	 * A disk with no configuration record
	 * cannot be part of an array.
	 */
	if (di->vdcr == NULL)
		return (-EINVAL);

	/* Do we already know about this array? */
	spin_lock_irqsave(&dm->ddf_lock, flags);
	list_for_each_entry(ai, &di->domain->arrayqueue, ddf_array_links) {
		if (ai->vdcr != NULL
		 && memcmp(&ai->vdcr->VD_GUID[0], &di->vdcr->VD_GUID[0],
			   DDF_GUID_BYTES) == 0) {
			di->parent = ddf_ai_get(ai);
			*array_info = ai;
			spin_unlock_irqrestore(&dm->ddf_lock, flags);
			return (0);
		}
	}	
	spin_unlock_irqrestore(&dm->ddf_lock, flags);

	/* This is a new array.  Discover its details */
	ai = ddf_alloc_array_info(dm);
	if (ai == NULL) {
		printk("md_add_to_array_ddf: cannot allocate array_info\n");
		return (-ENOMEM);
	}

	/* Make a copy of the VDR and VDCR so we don't have to share it with
	 * the disk object.
	 */
	ai->vdcr = kmalloc(di->crl * 512, GFP_KERNEL);
	if (ai->vdcr == NULL) {
		printk("ddf_get_array_info: cannot allocate records\n");
		ddf_ai_put(ai);
		return (-ENOMEM);
	}
	memcpy(ai->vdcr, di->vdcr, di->crl * 512);

	if (di->vendor != NULL) {
		ai->vendor = kmalloc(di->vendor_length * 512, GFP_KERNEL);
		if (ai->vendor == NULL) {
			printk("ddf_get_array_info: cannot allocate records\n");
			ddf_ai_put(ai);
			return (-ENOMEM);
		}
		memcpy(ai->vendor, di->vendor, di->vendor_length * 512);
		ai->vendor_length = di->vendor_length;
	}

	ai->checkpoint = 0;
	ai->vendor_type = AI_VENDOR_UNKNOWN;
	if (di->vendor != NULL) {
		uint32_t crc;
#ifdef MD_DEBUG
		int j;
#endif

		crc = ddf_calc_cksum(di->vendor, di->vendor_length * 512);
		if ((le32_to_cpu(di->vendor->Signature)
		  != DDF_VENDOR_LOG_SIGNATURE)
		 || di->vendor->CRC != crc) {
			printk("DDF: vendor signature does not match\n");
			goto unrecognized_vendor_record;
		}
#ifdef MD_DEBUG
		printk("Controller GUID:");
		for (j = 0; j < DDF_GUID_BYTES; j++)
			printk("%02x", di->cd->Controller_GUID[j]);
		printk("\n");
#endif

		if (memcmp(di->cd->Controller_GUID, "ADPT", 4) == 0
		 || memcmp(&di->cd->Controller_GUID[16], "ADPT", 4) == 0) {
			struct adpt_vendor_data *adpt;

			adpt = (struct adpt_vendor_data *)di->vendor->data;
			ai->checkpoint = adpt->CheckpointLBA;
			ai->vendor_type = AI_VENDOR_ADPT;
			MD_DDF_DEBUG(("DDF: checkpoint= %d\n",
				     adpt->CheckpointLBA));
		}
	}
unrecognized_vendor_record:

	/* Find the VD Entry in the record */
	ai->vde = NULL;
	for (i = 0; i < le16_to_cpu(dm->vdr->Max_VD_Entries); i++) {
#ifdef MD_DEBUG
		int j;
#endif

		ai->vde = &dm->vdr->entry[i];
#ifdef MD_DEBUG
		printk("Comparing ");
		for (j = 0; j < DDF_GUID_BYTES; j++)
			printk("%02x", ai->vde->VD_GUID[j]);
		printk(" with ");
		for (j = 0; j < DDF_GUID_BYTES; j++)
			printk("%02x", ai->vdcr->VD_GUID[j]);
		printk("\n");
#endif
		if (memcmp(&ai->vde->VD_GUID[0], &ai->vdcr->VD_GUID[0],
		    DDF_GUID_BYTES) == 0)
			break;
		ai->vde = NULL;
	}

	if (ai->vde == NULL) {
		printk("ddf_get_array_info: cannot find VDE [%d]\n",
		       le16_to_cpu(dm->vdr->Max_VD_Entries));
		ddf_ai_put(ai);
		*array_info = NULL;
		return (-EINVAL);
	}

	ai->pds_array =
	    (uint32_t *)&((uint8_t *)(ai->vdcr))[DDF_VDCR_PDS_OFFSET];
	ai->ss_array =
	    (uint32_t *)&((uint8_t *)(ai->vdcr))[DDF_VDCR_PDS_OFFSET +
						 di->pds_len];

#ifdef MD_DEBUG
	printk("FOUND NEW DDF ARRAY: PRL= %d RLQ= %d, State= %d, "
	       "Size= %llu, Stripe Size= %d\n",
	       ai->vdcr->Primary_RAID_Level, ai->vdcr->RLQ,
	       ai->vde->VD_State, ai->vdcr->Size,
	       (1 << ai->vdcr->Stripe_Size) * 512);
#endif

	spin_lock_irqsave(&dm->ddf_lock, flags);
	list_add(&ai->ddf_array_links, &ai->domain->arrayqueue);
	ai->domain->array_gencount++;
	spin_unlock_irqrestore(&dm->ddf_lock, flags);

	di->parent = ai;
	*array_info = ai;

	return (0);
}

/*
 * Find the spare pool for a domain.
 */
static mddev_t *
ddf_find_spare_pool(struct md_ddf_domain *dm)
{
	mdk_member_t *spare_pool;
	struct list_head *tmp;
	u_long flags;

	/*
	 * Search through the spare arrays to find
	 * the array that is associated with this disk's
	 * domain.
	 */
	ITERATE_CHILDREN(&spare_arrays, spare_pool, tmp, flags) {
		struct md_ddf_domain *pool_dm;

		if (spare_pool->metapers != &md_ddf_pers)
			continue;

		pool_dm = spare_pool->metadata_private;
		if (pool_dm == dm)
			return to_mddev(spare_pool);
	}
	return (NULL);
}

static int
ddf_add_spare(struct md_ddf_disk_info *di)
{
	mddev_t *mddev;
	mdk_member_t *spare_pool;
	mdk_member_t *disk;
	mdk_rdev_t *rdev;
	mdk_member_t *array;
	struct list_head *tmp;
	u_long flags;

	/* Create an rdev member object for the disk */
	rdev = md_rdev_alloc(NULL, MDK_MSUBTYPE_POOL_SPARE, di->dev,
			     MDK_PRIORITY_UNKNOWN);
	if (rdev == NULL) {
		printk("ddf_add_spare: cannot alloc rdev\n");
		return (-ENOMEM);
	}
	disk = to_member(rdev);

	mddev = ddf_find_spare_pool(di->domain);
	spare_pool = to_member(mddev);
	if (spare_pool == NULL) {

		/*
		 * This is the first spare for this domain.
		 * Allocate a new spare pool.
		 */
		mddev = md_mddev_alloc(NULL, MDK_MSUBTYPE_POOL_SPARE,
					    MDK_PRIORITY_UNKNOWN);
		if (mddev == NULL) {
			printk("ddf_add_spare: Cannot allocate md_array\n");
			member_put(disk);
			return (-ENOMEM);
		}
		spare_pool = to_member(mddev);
		spare_pool->metapers = md_metapers_get(&md_ddf_pers);
		spare_pool->metadata_private = ddf_dm_get(di->domain);
		md_member_add_child(&spare_arrays, to_member(mddev));
		MD_DDF_DEBUG(("New spare pool %p created\n", spare_pool));
	}
	di->rdev = rdev;
	disk->data_size = (sector_t)le64_to_cpu(di->pde->Configured_Size);
	disk->metadata_private = di;
	disk->state = ddf_to_md_disk_state(NULL, di->pde->PD_State);
	disk->metapers = md_metapers_get(&md_ddf_pers);
	rdev->kdev = di->dev;
	MD_DDF_DEBUG(("Added spare with di->pde == %p with size 0x%llx\n",
		      di->pde, (unsigned long long)disk->data_size));
	md_member_add_child(spare_pool, disk);
	member_put(spare_pool);

	ITERATE_CHILDREN(&toplevel_arrays, array, tmp, flags) {

		if (array->metapers != &md_ddf_pers) 
			continue;

		if (array->type != MDK_MTYPE_MDDEV)
			continue;

		mddev = to_mddev(array);

		MD_DDF_DEBUG(("Announcing spare to %p\n", array));	
		array->pers->raid_op(array, MDK_RAID_OP_SPARE_ADDED, NULL);
	}
	return (0);
}

static int
ddf_get_spares(mddev_t *mddev, mdu_spare_disk_info_t *usi)
{
	usi->count = 0;
	return (0);
}

static int
ddf_delete_spare(mdk_member_t *member)
{
	struct md_ddf_disk_info *di;
	struct md_ddf_domain *dm;
	mdk_member_t *spare_pool;
	mddev_t *spare_pool_mddev;
	mdk_member_t *spare;
	struct list_head *tmp;
	u_long flags;
	int count;
	
	MD_DDF_DEBUG(("ddf_delete_spare called\n"));
	if (member == NULL)
		return -ENODEV;

	if (member->metapers != &md_ddf_pers)
		return (-EINVAL);

	di = member->metadata_private;
	dm = di->domain;
	if (di == NULL || dm == NULL)
		return (-EINVAL);

	spare_pool = member->parent;
	if (spare_pool == NULL)
		return (-EINVAL);

	/*
	 * Use ddf_find_spare_pool to obtain a reference
	 * on our spare pool.  This takes care of holding
	 * the spare_array's member lock while acquiring
	 * the reference.
	 */
	spare_pool_mddev = ddf_find_spare_pool(dm);
	if (to_member(spare_pool_mddev) != spare_pool) {
		if (spare_pool_mddev != NULL)
			member_put(to_member(spare_pool_mddev));
		return (-EINVAL);
	}

	if (spare_pool->subtype != MDK_MSUBTYPE_POOL_SPARE) {
		member_put(spare_pool);
		return (-EINVAL);
	}
	down(&spare_pool_mddev->reconfig_sem);
	if (member->parent != spare_pool) {
		/*
		 * Spare was activated or removed by
		 * someone else.
		 */
		up(&spare_pool_mddev->reconfig_sem);
		member_put(spare_pool);
		return (-ENODEV);
	}
	
	/*
	 * Mark the spare as going away.
	 */
	member->state = MDU_MSTATE_TEARDOWN;

	/*
	 * Count the number of spares in this pool.
	 * If only one, the spare pool must be torn
	 * down too.
	 */
	count = 0;
	ITERATE_CHILDREN(spare_pool, spare, tmp, flags) {
		if (spare->state != MDU_MSTATE_TEARDOWN)
			count++;
	}

	/* Remove the sentinel count on the spare. */
	member_put(member);
	if (count == 0) {
		/* And on the spare pool if empty. */
		member_put(spare_pool);
	}
	up(&spare_pool_mddev->reconfig_sem);
	member_put(spare_pool);
	return (0);
}


static int
ddf_create_spare(mddev_t *mddev, mdu_disk_info_t *ui)
{
	struct md_ddf_disk_info *di;
	struct md_ddf_domain *dm;
	struct ddf_pd_entry *pde;
	struct gendisk *gd;
 	sector_t configured_size;
	u_long dmflags;
	kdev_t dev;
	int entry, type, state, error;

	dev = MKDEV(ui->major, ui->minor);
	di = ddf_find_disk_info(dev);
	if (di != NULL) {
		u_int type;
		u_int state;
		u_int mask;

		MD_DDF_DEBUG(("ddf_add_new_disk: "
			      "Found existing disk %p\n", di));
		mask = DDF_PDE_GLOBAL_SPARE
		     | DDF_PDE_CONFIG_SPARE
		     | DDF_PDE_PARTICIPATING;
		type = le16_to_cpu(di->pde->PD_Type);
		state = le16_to_cpu(di->pde->PD_State);
		if ((type & mask) != 0
		 && (state & (DDF_PDE_FAILED|DDF_PDE_MISSING)) == 0)
			return (-EBUSY);

		/*
		 * Remove from the domain prior to using
		 * this disk for a new array.
		 */
		ddf_free_disk_info(di);
	}

	error = ddf_find_domain(&dm);
	if (error != 0)
		return (-ENOMEM);

	gd = get_gendisk(dev);
	if (gd == NULL)
		return (-ENODEV);
	di = ddf_alloc_disk_info(dev, gd->part[MINOR(dev)].nr_sects);
	if (di == NULL)
		return (-ENOMEM);
	di->crl = dm->crl;

	type = DDF_PDE_GLOBAL_SPARE;
	state = DDF_PDE_ONLINE;

	/* Set up initial records */
	di->hdr_anchor = ddf_init_header(DDF_HEADER_ANCHOR,
					 di->dev_size - 1, di->crl,
					 di->crl * 2);
	di->hdr_pri = ddf_init_header(DDF_HEADER_PRIMARY,
				      di->dev_size - 1, di->crl, di->crl * 2);
	di->cr = ddf_init_cr(di->crl, 2);
	di->vendor = ddf_init_vendor(NULL);

	if (di->hdr_anchor == NULL
	 || di->hdr_pri == NULL
	 || di->cr == NULL
	 || di->vendor == NULL) {
		ddf_free_disk_info(di);
		return (-ENOMEM);
	}

	di->domain = ddf_dm_get(dm);
	di->cd = di->domain->cd;
	di->pdr = di->domain->pdr;
	di->vdr = di->domain->vdr;
	di->sa = (void *)di->cr;
	di->cr_consumers++;
	error = ddf_init_sa(dm, di->sa);
	if (error != 0) {
		ddf_free_disk_info(di);
		return (-ENOMEM);
	}
	di->cd_length = di->domain->cd_length;
	di->pdr_length = di->domain->pdr_length;
	di->vdr_length = di->domain->vdr_length;

	/* Figure out the correct PDE slot for this disk */
	for (entry = 0; entry < le16_to_cpu(di->pdr->Max_PD_Entries); entry++) {
		pde = &di->pdr->entry[entry];
		if (pde->PD_GUID[0] == 0xff)
			break;
	}
	if (entry >= le16_to_cpu(di->pdr->Max_PD_Entries)) {
		printk("DDF: Cannot allocate any more PDEs\n");
		ddf_free_disk_info(di);
		return (-EINVAL);
	}

	/* Allocate the PDD and PDE, and update the PDR (parent of PDE) */
	di->pdd = ddf_init_pdd();
	error = ddf_get_freesize(di->dev, &configured_size);
	if (error != 0) {
		printk("DDF: Cannot determine device freesize\n");
		ddf_free_disk_info(di);
		return (error);
	}
		
	di->pde = ddf_init_pde(di->pdr, entry, type, state, configured_size,
			       di->pdd->PD_GUID, di->pdd->PD_Reference);
	di->pdr->Populated_PDEs =
	    cpu_to_le16(le16_to_cpu(di->pdr->Populated_PDEs) + 1);

	/* Final housekeeping to complete the disk_info */
	di->cr_length = di->crl * 2;
	di->pdd_length = DDF_ADPT_PDD_LENGTH;
	di->vendor_length = DDF_ADPT_VENDOR_LENGTH;
	memcpy(di->vend_id, di->cd->Controller_GUID, 8);

	spin_lock_irqsave(&di->domain->ddf_lock, dmflags);
	list_add(&di->ddf_disk_links, &di->domain->diskqueue);
	di->domain->diskqueue_gencount++;
	ui->cookie = (u_long)di->rdev;

	/* Mark everything as dirty and sync it all out */
	di->ddf_dirty_mask = (DI_DIRTY_HDR_ANCHOR|DI_DIRTY_HDR_PRI|DI_DIRTY_CD|
			      DI_DIRTY_PDR|DI_DIRTY_VDR|DI_DIRTY_VDCR|
			      DI_DIRTY_PDD|DI_DIRTY_VENDOR);
	dm->ddf_dirty_mask |= DM_DIRTY_PDR;
	spin_unlock_irqrestore(&di->domain->ddf_lock, dmflags);
	ddf_schedule_sync(&dm->sync_data, /*wait*/1, /*enable*/1);

	/* Announce to the system */
	error = ddf_add_spare(di);
	if (error != 0)
		ddf_free_disk_info(di);

	return (error);
}

static int
ddf_activate_spare(mddev_t *array_mddev)
{
	mdk_member_t *member;
	struct md_ddf_domain *dm;
	struct md_ddf_array_info *ai;
	struct md_ddf_disk_info *di;
	mddev_t *spare_pool_mddev;
	mdk_member_t *spare_pool;
	mdk_member_t *spare;
	mdk_member_t *best_spare;
	struct list_head *tmp;
	uint64_t needed_size;
	u_long flags;
	u_int num_pdes;
	int num_disks;
	int i;
	
	MD_DDF_DEBUG(("ddf_activate_spare called\n"));

	member = to_member(array_mddev);
	if (member == NULL)
		return -ENODEV;

	if (member->metapers != &md_ddf_pers)
		return (-EINVAL);

	/*
	 * If we need to activate a spare, we should
	 * first update the metadata to account for
	 * missing devices.  Otherwise we may not be
	 * able to find a valid slot in the VDCR to
	 * place the spare.
	 */
	ddf_mark_missing_members(array_mddev);

	ai = member->metadata_private;
	dm = ai->domain;
	if (ai == NULL || ai->vdcr == NULL)
		return (-EINVAL);
	/* XXX How does this work for multi-level arrays. */
	needed_size = le64_to_cpu(ai->vdcr->Size);
	MD_DDF_DEBUG(("ddf_activate_spare: Needed Size is 0x%llx\n",
		     needed_size));
	num_disks = le16_to_cpu(ai->vdcr->Primary_Element_Count);
	switch (ai->vdcr->Primary_RAID_Level) {
	case 1:
		break;
	default:
		return (-EOPNOTSUPP);
	}

	/*
	 * Search through the spare arrays to find
	 * spares allocated to the domain for this
	 * array.
	 */
	spare_pool_mddev = ddf_find_spare_pool(dm);
	spare_pool = to_member(spare_pool_mddev);
	if (spare_pool == NULL)
		return (-ENODEV);

	/*
	 * Prevent modifications to our spare pool.
	 */
	down(&spare_pool_mddev->reconfig_sem);
	best_spare = NULL;
	ITERATE_CHILDREN(spare_pool, spare, tmp, flags) {

		if ((spare->state & MDU_MSTATE_OPTIMAL) == 0)
			continue;

		if (spare->data_size < needed_size) {
			MD_DDF_DEBUG(("ddf_activate_spare: skipping spare "
				      "with size %llx\n",
				      (unsigned long long)spare->data_size));
			continue;
		}

		/*
		 * Found a candidate.  Only switch to using
		 * this device as our spare if it is closer
		 * to the "needed_size" than any previous
		 * candidates.
		 */
		if (best_spare == NULL
		 || best_spare->data_size > spare->data_size)
			best_spare = spare;
	}
	if (best_spare == NULL)
		goto activation_failed;

	/*
	 * Verify that there is a slot for this new
	 * disk in the target array.
	 */
	num_pdes = le16_to_cpu(dm->pdr->Max_PD_Entries);
	for (i = 0; i < num_disks; i++) {
		struct ddf_pd_entry *pde;
		uint32_t pdref;
		int j;

		pdref = ai->pds_array[i];
		/*
		 * An unused slot is certainly valid.
		 */
		if (pdref == 0xFFFFFFFF)
			break;
		for (pde = dm->pdr->entry, j = 0; j < num_pdes; pde++, j++) {
			u_int state;

			state = le16_to_cpu(pde->PD_State);
			if (pde->PD_Reference == pdref
			 && (state & (DDF_PDE_FAILED|DDF_PDE_MISSING)) != 0)
				goto search_done;
		}
	}

search_done:
	if (i >= num_disks)
		goto activation_failed;

	/*
	 * We now have a spare *and* a valid slot to place it
	 * in.  Convert this drive to an active member of the
	 * array.
	 *
	 * Disassociate the spare from the spare pool.
	 */
	md_member_orphan(best_spare);

	/*
	 * Schedule the spare pool for release if
	 * we are the last member of the pool.
	 */
	if (list_empty(&spare_pool->child_set))
		member_put(spare_pool);
	di = best_spare->metadata_private;
	di->parent = ddf_ai_get(ai);
	di->ss_len = ai->ss_len;
	di->pds_len = ai->pds_len;
	ai->pds_array[i] = di->pdd->PD_Reference;
	ai->ss_array[i] = 0;

	/* Convert the Spare Activation record into a VDCR. */
	di->vdcr = (struct ddf_vdc_record *)di->sa;
	di->sa = NULL;

	/*
	 * Update the rdev.
	 */
	best_spare->priority = i;
	best_spare->subtype = MDK_MSUBTYPE_CHILD;
	best_spare->data_size = needed_size;
	best_spare->data_offset = 0;
	best_spare->state = MDU_MSTATE_OPTIMAL|MDU_MSTATE_REBUILD_TARGET;
	di->pde->PD_Type = le16_to_cpu(DDF_PDE_PARTICIPATING);
	di->pde->PD_State =
	    le16_to_cpu(md_to_ddf_disk_state(best_spare->state));
	member->pers->raid_op(member, MDK_RAID_OP_INIT_MEMBER, best_spare);
	md_member_add_child(member, best_spare);

	/* Mark everything as dirty and sync it all out */
	di->ddf_dirty_mask = (DI_DIRTY_HDR_ANCHOR|DI_DIRTY_HDR_PRI|DI_DIRTY_CD|
			      DI_DIRTY_PDR|DI_DIRTY_VDR|DI_DIRTY_VDCR|
			      DI_DIRTY_PDD|DI_DIRTY_VENDOR);
	ai->ddf_dirty_mask |= AI_DIRTY_VDCR;
	ai->domain->ddf_dirty_mask |= (DM_DIRTY_PDR|DM_DIRTY_VDR);
	ddf_schedule_sync(&ai->sync_data, /*wait*/1, /*enable*/1);

	member->pers->raid_op(member, MDK_RAID_OP_ADD_MEMBER, best_spare);
	up(&spare_pool_mddev->reconfig_sem);
	member_put(spare_pool);
	return 0;

activation_failed:
	up(&spare_pool_mddev->reconfig_sem);
	member_put(spare_pool);
	return (-ENODEV);
}

/*
 * Mark all absent devices as missing in the metadata
 * for the passed in array.
 */
static void
ddf_mark_missing_members(mddev_t *mddev)
{
	mdk_member_t *array;
	mdk_member_t *child;
	struct md_ddf_domain *dm;
	struct md_ddf_array_info *ai;
	struct list_head *tmp;
	u_long dmflags;
	u_long flags;
	u_int num_disks;
	u_int num_pdes;
	u_int i;

	array = to_member(mddev);
	ai = array->metadata_private;
	dm = ai->domain;
	spin_lock_irqsave(&dm->ddf_lock, dmflags);
	spin_lock_irqsave(&array->member_lock, flags);
	num_disks = le16_to_cpu(ai->vdcr->Primary_Element_Count);
	num_pdes = le16_to_cpu(dm->pdr->Max_PD_Entries);
	for (i = 0; i < num_disks; i++) {
		struct ddf_pd_entry *pde;
		uint32_t pdref;
		int j;

		pdref = ai->pds_array[i];
		/* Skip unused slots */
		if (pdref == 0xFFFFFFFF)
			continue;
		for (pde = dm->pdr->entry, j = 0; j < num_pdes; pde++, j++) {
			if (pde->PD_Reference == pdref) {
				struct md_ddf_disk_info *di;

				di = NULL;
				ITERATE_CHILDREN_LOCKED(array, child, tmp) {
					struct md_ddf_disk_info *tmp_di;

					tmp_di = child->metadata_private;
					if (tmp_di->pde == pde) {
						di = tmp_di;
						break;
					}
				}
				if (di == NULL) {
					u_int state;

					state = le16_to_cpu(pde->PD_State);
			 		state |= DDF_PDE_MISSING;
					pde->PD_State = cpu_to_le16(state);
					dm->ddf_dirty_mask |= DM_DIRTY_PDR;
				}
			}
		}
	}
	spin_unlock_irqrestore(&array->member_lock, flags);
	spin_unlock_irqrestore(&dm->ddf_lock, dmflags);

	ddf_schedule_sync(&dm->sync_data, /*wait*/0, /*enable*/1);
}

static struct md_ddf_domain *
ddf_alloc_domain()
{
	struct md_ddf_domain *new_dm;

	new_dm = kmalloc(sizeof(*new_dm), GFP_KERNEL);
	if (new_dm != NULL) {
		memset(new_dm, 0, sizeof(*new_dm));
		atomic_set(&new_dm->refcount, 1);
		INIT_LIST_HEAD(&new_dm->ddf_domain_links);
		INIT_LIST_HEAD(&new_dm->diskqueue);
		INIT_LIST_HEAD(&new_dm->arrayqueue);
		spin_lock_init(&new_dm->ddf_lock);
		new_dm->sync_data.sync_dm = new_dm;
		new_dm->sync_data.sync_lock = &new_dm->ddf_lock;
		new_dm->sync_data.sync_flags |= DDF_SYNC_FLAGS_SYNC_ALLOWED;
		md_init_workitem(&new_dm->sync_data.sync_workitem,
				 MD_DDF_WORKOP_SYNC, md_ddf_work_control,
				 md_ddf_sync_handler, &new_dm->sync_data);
	}
	return (new_dm);
}

static struct md_ddf_array_info *
ddf_alloc_array_info(struct md_ddf_domain *dm)
{
	struct md_ddf_array_info *ai;

	ai = kmalloc(sizeof(struct md_ddf_array_info), GFP_KERNEL);
	if (ai != NULL) {
		memset(ai, 0, sizeof(struct md_ddf_array_info));
		atomic_set(&ai->refcount, 1);
		INIT_LIST_HEAD(&ai->ddf_array_links);
		if (dm != NULL)
			ai->domain = ddf_dm_get(dm);
		ai->sync_data.sync_dm = dm;
		ai->sync_data.sync_ai = ai;
		md_init_workitem(&ai->sync_data.sync_workitem,
				 MD_DDF_WORKOP_SYNC, md_ddf_work_control,
				 md_ddf_sync_handler, &ai->sync_data);
	}
	return (ai);
}

static struct md_ddf_disk_info *
ddf_alloc_disk_info(kdev_t dev, uint64_t size)
{
	struct md_ddf_disk_info *di;

	di = kmalloc(sizeof(struct md_ddf_disk_info), GFP_KERNEL);
	if (di != NULL) {
		memset(di, 0, sizeof(struct md_ddf_disk_info));
		INIT_LIST_HEAD(&di->ddf_disk_links);
		di->dev = dev;
		di->dev_size = size;
	}
	return (di);
}

static void
ddf_free_domain(struct md_ddf_domain *dm)
{
	if (dm == NULL)
		return;

	if (dm->cd != NULL)
		kfree(dm->cd);
	if (dm->pdr != NULL)
		kfree(dm->pdr);
	if (dm->vdr != NULL)
		kfree(dm->vdr);
	if (dm->sync_cd != NULL)
		kfree(dm->sync_cd);
	dm->sync_cd = NULL;
	if (dm->sync_pdr != NULL)
		kfree(dm->sync_pdr);
	if (dm->sync_vdr != NULL)
		kfree(dm->sync_vdr);
	list_del_init(&dm->ddf_domain_links);
	ddf_domains_gencount++;
	kfree(dm);
}

static void
ddf_free_array_info(struct md_ddf_array_info *ai)
{
	if (ai == NULL)
		return;

	if (ai->domain != NULL) {
		u_long flags;

		spin_lock_irqsave(&ai->domain->ddf_lock, flags);
		if (ai->vde != NULL) {
			uint16_t pop_vdes;

			pop_vdes = le16_to_cpu(ai->domain->vdr->Populated_VDEs);
			pop_vdes = cpu_to_le16(pop_vdes - 1);
			memset(ai->vde, 0xFF, sizeof(*ai->vde));
			ai->domain->ddf_dirty_mask |= DM_DIRTY_VDR;
			ai->domain->vdr->Populated_VDEs = pop_vdes;
			ai->vde = NULL;
		}
		list_del_init(&ai->ddf_array_links);
		ai->domain->array_gencount++;
		spin_unlock_irqrestore(&ai->domain->ddf_lock, flags);
		ddf_schedule_sync(&ai->domain->sync_data, /*wait*/0,
				  /*enable*/1);
		ddf_dm_put(ai->domain);
		ai->domain = NULL;
	}
	if (ai->vdcr != NULL) {
		kfree(ai->vdcr);
		ai->vdcr = NULL;
	}
	if (ai->vendor != NULL) {
		kfree(ai->vendor);
		ai->vendor = NULL;
	}
	kfree(ai);
}

static void
ddf_free_disk_info(struct md_ddf_disk_info *di)
{
	if (di == NULL)
		return;

	if (di->hdr_anchor != NULL) {
		kfree(di->hdr_anchor);
		di->hdr_anchor = NULL;
	}
	if (di->hdr_pri != NULL) {
		kfree(di->hdr_pri);
		di->hdr_pri = NULL;
	}
	if (di->domain == NULL) {
		if (di->cd != NULL)
			kfree(di->cd);
		if (di->pdr != NULL)
			kfree(di->pdr);
		if (di->vdr != NULL)
			kfree(di->vdr);
	} else {
		u_long flags;

		spin_lock_irqsave(&di->domain->ddf_lock, flags);
		list_del_init(&di->ddf_disk_links);
		di->domain->diskqueue_gencount++;
		spin_unlock_irqrestore(&di->domain->ddf_lock, flags);
		ddf_dm_put(di->domain);
		di->domain = NULL;
	}
	di->cd = NULL;
	di->pdr = NULL;
	di->vdr = NULL;

	if (di->cr != NULL) {
		if (di->vdcr != NULL) {
			di->vdcr = NULL;
			di->cr_consumers--;
		}
		if (di->sa != NULL) {
			di->sa = NULL;
			di->cr_consumers--;
		}
		if (di->vuc != NULL) {
			di->vuc = NULL;
			di->cr_consumers--;
		}
		if (di->cr_consumers == 0)
			kfree(di->cr);
		di->cr = NULL;
	}
	if (di->pdd != NULL) {
		kfree(di->pdd);
		di->pdd = NULL;
	}
	if (di->bbm != NULL) {
		kfree(di->bbm);
		di->bbm = NULL;
	}
	if (di->vendor != NULL) {
		kfree(di->vendor);
		di->vendor = NULL;
	}
	if (di->parent != NULL) {
		ddf_ai_put(di->parent);
		di->parent = NULL;
	}
	kfree(di);
}

static int
ddf_to_md_array_state(u_int vde_state, u_int init_state, mdu_mstate_t *state)
{
	switch (vde_state & DDF_VDE_STATE_MASK) {
	case DDF_VDE_OPTIMAL:
		*state = MDU_MSTATE_OPTIMAL;
		break;
	case DDF_VDE_DEGRADED:
		*state = MDU_MSTATE_DEGRADED;
		break;
	case DDF_VDE_DELETED:
	case DDF_VDE_MISSING:
	case DDF_VDE_FAILED:
	default:
		*state = MDU_MSTATE_FAILED;
		break;
	}

	switch (init_state & DDF_VDE_INIT_MASK) {
	case DDF_VDE_INIT_QUICK:
	case DDF_VDE_INIT_FULL:
		break;
	case DDF_VDE_UNINTIALIZED:
	default:
		MD_DDF_DEBUG(("Setting Initializing State: "
			      "VDEState = 0x%x IState 0x%x\n",
			      vde_state, init_state));
		*state |= MDU_MSTATE_INITIALIZING;
		break;
	}

	return (0);
}

/*
 * Revalidate our array state.   This routine is called whenever
 * a member is added, removed, or changes state.
 */
static int
ddf_update_array_state(mddev_t *mddev)
{
	mdk_set_state_arg_t sstate;
	mdk_member_t *array;
	struct md_ddf_array_info *ai;
	u_long flags;
	u_int optimal_cnt;
	u_int degraded_cnt;
	u_int failed_cnt;
	u_int build_targets;

	array = to_member(mddev);
	ai = array->metadata_private;
	/*
	 * Prevent state from changing during this update.
	 */
	spin_lock_irqsave(&array->member_lock, flags);
	sstate.oldstate = array->state;

	md_get_array_stats(mddev, &optimal_cnt, &degraded_cnt,
			   &failed_cnt, &build_targets);
	MD_DDF_DEBUG(("md_ddf_update_array_state: o %d d %d f %d b %d\n",
		      optimal_cnt, degraded_cnt, failed_cnt, build_targets));

	if (optimal_cnt >= mddev->max_disks) 
		sstate.newstate = MDU_MSTATE_OPTIMAL;
	else if (optimal_cnt + degraded_cnt >= ai->needed_disks)
		sstate.newstate = MDU_MSTATE_DEGRADED;
	else if (sstate.oldstate & MDU_MSTATE_INITIALIZING) {
		/*
		 * If we have sufficent build targets,
		 * update our state.
		 */
		sstate.newstate = MDU_MSTATE_INITIALIZING;
		if (optimal_cnt + build_targets == mddev->max_disks) 
			sstate.newstate |= MDU_MSTATE_OPTIMAL;
		else if (optimal_cnt + build_targets
		      >= ai->needed_disks)
			sstate.newstate |= MDU_MSTATE_DEGRADED;
	} else
		 sstate.newstate = MDU_MSTATE_FAILED;

	if ((sstate.oldstate & MDU_MSTATE_DIRTY) != 0)
		sstate.newstate |= MDU_MSTATE_DIRTY;
	spin_unlock_irqrestore(&array->member_lock, flags);

	if (sstate.newstate != sstate.oldstate) {

		ddf_set_array_state(mddev, &sstate);
		MD_DDF_DEBUG(("Array %p transitions from %#x to %#x\n",
			      array, sstate.oldstate, sstate.newstate));

		return (1);
	}
	return (0);
}

static int ddf_to_md_disk_state(mddev_t *array, uint8_t pd_state)
{

	switch (pd_state) {
	case DDF_PDE_ONLINE:
	{
		mdu_mstate_t mdstate;

		mdstate = MDU_MSTATE_OPTIMAL;
		if (array != NULL
		 && (to_member(array)->state & MDU_MSTATE_INITIALIZING) != 0)
			mdstate |= MDU_MSTATE_REBUILD_TARGET;
		return (mdstate);
	}
	case DDF_PDE_REBUILD:
	case DDF_PDE_TRANSITION:
		return (MDU_MSTATE_REBUILD_TARGET | MDU_MSTATE_OPTIMAL);
	case DDF_PDE_FAILED:
	case DDF_PDE_PFA:
	case DDF_PDE_UNRECOVERED:
	case DDF_PDE_MISSING:
	default:
		return (MDU_MSTATE_FAILED);
	}
}

static int ddf_open_disks(struct md_ddf_domain *dm, int open)
{
	struct ddf_header *hdr;
	struct md_ddf_disk_info *di;
	uint64_t ddf_primary_base;
	uint32_t hdr_seq;
	u_long flags;
	u_int saved_gencount;
	int error;

	error = 0;
	spin_lock_irqsave(&dm->ddf_lock, flags);
rescan:
	saved_gencount = dm->diskqueue_gencount;
	list_for_each_entry(di, &dm->diskqueue, ddf_disk_links) {
		mdk_member_t *disk;

		disk = to_member(di->rdev);
		if (disk != NULL)
			member_get(disk);
		hdr = di->hdr_pri;
		/*
		 * When opening, skip non-dirty devices.
		 * When closing, skip devices that weren't
		 * open.
		 */
		if ((open && di->ddf_dirty_mask == 0)
		 || (!open && hdr->Open_Flag == 0)) {
			MD_DDF_DEBUG(("Skipping clean disk\n"));
			if (disk != NULL)
				member_put(disk);
			continue;
		}

		hdr->Open_Flag = open;
		hdr_seq = le32_to_cpu(hdr->Sequence_Number);
		if (hdr_seq != 0xffffffff)
			hdr_seq++;
		if (hdr_seq == 0xffffffff)
			hdr_seq = 1;
		hdr->Sequence_Number = cpu_to_le32(hdr_seq);
		hdr->CRC = ddf_calc_cksum(hdr, DDF_HEADER_BYTES);
		ddf_primary_base = le64_to_cpu(hdr->Primary_Header_LBA);
		MD_DDF_DEBUG(("open at %llx\n", ddf_primary_base));
		spin_unlock_irqrestore(&dm->ddf_lock, flags);
		error = md_write_blocks(di->dev, ddf_primary_base,
				        DDF_HEADER_LENGTH, (uint8_t *)hdr);
		if (error == -EIO) {
			if (disk)
				md_set_member_state(disk, MDU_MSTATE_FAILED);
			printk("ddf_open_disks: Unable to write metadata. "
			       "Marking the disk (%d,%d) as failed\n",
			       MAJOR(di->dev), MINOR(di->dev));
		}
		if (disk)
			member_put(disk);
		if (error) {
			printk("DDF: Error opening DDF\n");
			break;
		}
		spin_lock_irqsave(&dm->ddf_lock, flags);
		if (dm->diskqueue_gencount != saved_gencount)
			goto rescan;
	}
	spin_unlock_irqrestore(&dm->ddf_lock, flags);

	return (error);
}

static int ddf_sync_disk(struct md_ddf_disk_info *di)
{
	mdk_member_t *member;
	struct ddf_header *hdr;
	uint64_t ddf_primary_base, ddf_anchor_base;
	uint64_t offset;
	u_long flags;
	u_int ddf_dirty_mask;
	int error;

	MD_DDF_DEBUG(("DDF: ddf_sync_disk called, dev_t 0x%x\n", di->dev));
	member = to_member(di->rdev);
	hdr = di->hdr_pri;
	ddf_primary_base = le64_to_cpu(hdr->Primary_Header_LBA);
	error = 0;
	flags = 0;
	if (member)
		spin_lock_irqsave(&member->member_lock, flags);
	ddf_dirty_mask = di->ddf_dirty_mask;
	di->ddf_dirty_mask = 0;
	if (member)
		spin_unlock_irqrestore(&member->member_lock, flags);

	ddf_anchor_base = di->dev_size - 1;
	MD_DDF_DEBUG(("ddf_anchor %llx, pri_hdr %llx, dev_size %llx\n",
		       ddf_anchor_base, ddf_primary_base, di->dev_size));
	if (ddf_dirty_mask & DI_DIRTY_HDR_ANCHOR) {
		offset = ddf_anchor_base;
		di->hdr_anchor->CRC =
		    ddf_calc_cksum(di->hdr_anchor, DDF_HEADER_BYTES);
		MD_DDF_DEBUG(("Syncing HDR_ANCHOR at %llx len %d\n",
			      offset, 1));
		error = md_write_blocks(di->dev, offset, DDF_HEADER_LENGTH,
					(uint8_t *)di->hdr_anchor);
		if (error)
			goto had_error;
		ddf_dirty_mask &= ~DI_DIRTY_HDR_ANCHOR;
	}

	if (ddf_dirty_mask & DI_DIRTY_HDR_PRI) {
		offset = ddf_primary_base;
		MD_DDF_DEBUG(("Syncing HDR_PRI at %llx len %d\n", offset, 1));
		di->hdr_pri->CRC =
		    ddf_calc_cksum(di->hdr_pri, DDF_HEADER_BYTES);
		error = md_write_blocks(di->dev, offset, DDF_HEADER_LENGTH,
					(uint8_t *)di->hdr_pri);
		if (error)
			goto had_error;
		ddf_dirty_mask &= ~DI_DIRTY_HDR_PRI;
	}

	if (ddf_dirty_mask & DI_DIRTY_CD) {
		offset = ddf_primary_base + le32_to_cpu(hdr->cd_section);
		MD_DDF_DEBUG(("Syncing CD at %llx len %d\n",
			      offset, di->cd_length));
		error = md_write_blocks(di->dev, offset, di->cd_length,
					(uint8_t *)di->domain->sync_cd);
		if (error)
			goto had_error;
		ddf_dirty_mask &= ~DI_DIRTY_CD;
	}

	if (ddf_dirty_mask & DI_DIRTY_PDR) {
		offset = ddf_primary_base + le32_to_cpu(hdr->pdr_section);
		MD_DDF_DEBUG(("Syncing PDR at %llx len %d\n",
			      offset, di->pdr_length));
		error = md_write_blocks(di->dev, offset, di->pdr_length,
					(uint8_t *)di->domain->sync_pdr);
		if (error)
			goto had_error;
		ddf_dirty_mask &= ~DI_DIRTY_PDR;
	}

	if (ddf_dirty_mask & DI_DIRTY_VDR) {
		offset = ddf_primary_base + le32_to_cpu(hdr->vdr_section);
		MD_DDF_DEBUG(("Syncing VDR at %llx len %d\n",
			      offset, di->vdr_length));
		error = md_write_blocks(di->dev, offset, di->vdr_length,
					(uint8_t *)di->domain->sync_vdr);
		if (error)
			goto had_error;
		ddf_dirty_mask &= ~DI_DIRTY_VDR;
	}

	if (ddf_dirty_mask & (DI_DIRTY_VDCR|DI_DIRTY_VUC|DI_DIRTY_SA)) {
		offset = ddf_primary_base + le32_to_cpu(hdr->cr_section);
		MD_DDF_DEBUG(("Syncing CR at %llx len %d\n",
			      offset, di->cr_length));
		error = md_write_blocks(di->dev, offset, di->cr_length,
					(uint8_t *)di->cr);
		if (error)
			goto had_error;
		ddf_dirty_mask &= ~(DI_DIRTY_VDCR|DI_DIRTY_VUC|DI_DIRTY_SA);
	}

	if (ddf_dirty_mask & DI_DIRTY_PDD) {
		di->pdd->CRC = ddf_calc_cksum(di->pdd, di->pdd_length * 512);
		offset = ddf_primary_base + le32_to_cpu(hdr->pdd_section);
		MD_DDF_DEBUG(("Syncing PDD at %llx len %d\n",
			      offset, di->pdd_length));
		error = md_write_blocks(di->dev, offset, di->pdd_length,
					(uint8_t *)di->pdd);
		if (error)
			goto had_error;
		ddf_dirty_mask &= ~DI_DIRTY_PDD;
	}

	if (ddf_dirty_mask & DI_DIRTY_VENDOR) {
		offset = ddf_primary_base + le32_to_cpu(hdr->vendor_section);
		MD_DDF_DEBUG(("Syncing vendor at %llx len %d\n",
			      offset, di->vendor_length));
		error = md_write_blocks(di->dev, offset, di->vendor_length,
					(uint8_t *)di->vendor);
		if (error)
			goto had_error;
		ddf_dirty_mask &= ~DI_DIRTY_VENDOR;
	}

	return (0);

had_error:
	if (member)
		spin_lock_irqsave(&member->member_lock, flags);
	di->ddf_dirty_mask |= ddf_dirty_mask;
	if (member)
		spin_unlock_irqrestore(&member->member_lock, flags);
	return (error);
}

static void
ddf_schedule_sync(struct md_ddf_sync_data *sync_data, int wait, int enable_sync)
{
	u_long flags;
	int sched_work;

	sched_work = 0;

	spin_lock_irqsave(sync_data->sync_lock, flags);
	if (enable_sync)
		sync_data->sync_flags |= DDF_SYNC_FLAGS_SYNC_ALLOWED;
	if ((sync_data->sync_flags & DDF_SYNC_FLAGS_SYNC_INPROG) == 0) {
		sync_data->sync_flags |= DDF_SYNC_FLAGS_SYNC_INPROG;
		sched_work = 1;
	}
	spin_unlock_irqrestore(sync_data->sync_lock, flags);

	if (sched_work) {
		/*
		 * The "syncer" holds a reference on the container
		 * of the workitem to prevent it from disappearing
		 * out from under it.
		 */
		ddf_sync_data_get(sync_data);
		md_queue_workitem(&md_ddf_syncer, &sync_data->sync_workitem, 0);
	}

	if (wait == 0)
		return;

	/*
	 * Hold a reference while waiting for the
	 * workitem to complete just in case the container
	 * of the workitem is removed as the result of
	 * a final sync on a failed member.
	 */
	ddf_sync_data_get(sync_data);
	for (;;) {
		int syncing;

		md_wait_workitem(&md_ddf_syncer, &sync_data->sync_workitem);
		spin_lock_irqsave(sync_data->sync_lock, flags);
		syncing = sync_data->sync_flags & DDF_SYNC_FLAGS_SYNC_INPROG;
		spin_unlock_irqrestore(sync_data->sync_lock, flags);
		if (syncing == 0)
			break;
	}
	ddf_sync_data_put(sync_data);
}

/*
 * Handle state change notifications on objects.  In general
 * state changes are only allowed if the old presented to us
 * matches the current state of the object.  The teardown state
 * is special.  We don't bother to check the previous state
 * of the member.  We just change its externally visible state
 * to the teardown state and decrement its reference so it goes
 * away.
 */
static int
ddf_set_disk_state(mdk_rdev_t *disk, mdk_set_state_arg_t *sstate)
{
	struct md_ddf_domain *dm;
	struct md_ddf_disk_info *di;
	mdk_member_t *member;
	u_long mflags;
	u_long dmflags;
	int ddf_state;
	int put_member;
	int put_member_after_sync;

	di = to_member(disk)->metadata_private;
	dm = di->domain;
	member = to_member(disk);
	put_member = 0;
	put_member_after_sync = 0;

	ddf_state = md_to_ddf_disk_state(sstate->newstate);

	if (sstate->newstate == MDU_MSTATE_FAILED
	 && sstate->oldstate != MDU_MSTATE_FAILED)
		put_member_after_sync = 1;

	spin_lock_irqsave(&dm->ddf_lock, dmflags);
	spin_lock_irqsave(&member->member_lock, mflags);
	if (sstate->newstate != MDU_MSTATE_TEARDOWN) {

		if (sstate->oldstate != member->state) {
			spin_unlock_irqrestore(&member->member_lock, mflags);
			spin_unlock_irqrestore(&dm->ddf_lock, dmflags);
			return (-EINVAL);
		}

		/* Update DDF state. */
		di->pde->PD_State = ddf_state;
		MD_DDF_DEBUG(("DDF Setting disk state %x:%x\n",
			      sstate->newstate, di->pde->PD_State));
		di->domain->ddf_dirty_mask |= DM_DIRTY_PDR;
	} else if (member->state != MDU_MSTATE_TEARDOWN) {
		if (member->state != MDU_MSTATE_FAILED)
			put_member = 1;
		MD_DDF_DEBUG(("DDF_TEARDOWN_DISK %p: Refcount %d\n",
			      member, (int)atomic_read(&member->refcount)));
	}
	member->state = sstate->newstate;
	if (put_member_after_sync)
		di->flags |= DI_FLAGS_PUT_AFTER_SYNC;	
	spin_unlock_irqrestore(&member->member_lock, mflags);
	spin_unlock_irqrestore(&dm->ddf_lock, dmflags);

	if (sstate->newstate != MDU_MSTATE_TEARDOWN) {
		mdk_member_t *parent;
		int enable_sync;

		parent = member->parent;
		enable_sync = 0;
		if (sstate->newstate & (MDU_MSTATE_FAILED|MDU_MSTATE_DEGRADED))
			enable_sync = 1;

		if (parent != NULL)
			ddf_update_array_state(to_mddev(parent));

		if (member->pers != NULL) {
			member->pers->raid_op(member,
					      MDK_RAID_OP_MSTATE_CHANGED,
					      member);
		}

		ddf_schedule_sync(&di->domain->sync_data,
				  /*wait*/0, enable_sync);
	}

	if (put_member) {
		mdk_member_t *parent;

		/*
		 * The rules for "sentinel puts" require us to
		 * guarantee that all other threads have seen
		 * a state change to either TEARDOWN or FAILED
		 * prior to the last reference count being released.
		 * This is achieved by briefly acquiring and
		 * releasing the parent's lock so that all threads
		 * iterating on the parent's child set have exited
		 * their loop since the state change and prior to
		 * our put.
		 */
		parent = member->parent;
		if (parent) {
			spin_lock_irqsave(&parent->member_lock, mflags);
			spin_unlock_irqrestore(&parent->member_lock, mflags);
		}

		member_put(member);
	}

	return (0);
}

/*
 * Handle state change notifications on objects.  In general
 * state changes are only allowed if the old presented to us
 * matches the current state of the object.  The teardown state
 * is special.  We don't bother to check the previous state
 * of the member.  We just change its externally visible state
 * to the teardown state and decrement its reference so it goes
 * away.
 */
static int
ddf_set_array_state(mddev_t *mddev, mdk_set_state_arg_t *sstate)
{
	mdk_member_t *member;
	struct md_ddf_array_info *ai;
	struct md_ddf_domain *dm;
	u_long mflags;
	u_long dmflags;
	u_int ddf_state;
	u_int init_state;
	int put_member;
	int put_member_after_sync;
	int error;

	member = to_member(mddev);
	ai = member->metadata_private;
	dm = ai->domain;
	put_member = 0;
	put_member_after_sync = 0;

	error = md_to_ddf_array_state(sstate->newstate,
				      &ddf_state, &init_state);
	if (error != 0)
		return (-EINVAL);

	spin_lock_irqsave(&dm->ddf_lock, dmflags);
	spin_lock_irqsave(&member->member_lock, mflags);
	if (sstate->newstate != MDU_MSTATE_TEARDOWN) {

		if (sstate->oldstate != member->state) {
			spin_unlock_irqrestore(&member->member_lock, mflags);
			spin_unlock_irqrestore(&dm->ddf_lock, dmflags);
			return (-EINVAL);
		}

		/*
		 * Preserve the dirty flag.
		 */
		if ((ai->vde->VD_State & DDF_VDE_DIRTY) != 0)
			ddf_state |= DDF_VDE_DIRTY;
		ai->vde->VD_State = ddf_state;
		ai->vde->Init_State = init_state;
		MD_DDF_DEBUG(("DDF_SET_ARRAY_STATE %p: S 0x%x I 0x%x\n",
			      member, ai->vde->VD_State, ai->vde->Init_State));
		ai->domain->ddf_dirty_mask |= DM_DIRTY_VDR;
	} else if (member->state != MDU_MSTATE_TEARDOWN) {
		put_member = 1;
		MD_DDF_DEBUG(("DDF_TEARDOWN_ARRAY %p: Refcount %d\n",
			      member, (int)atomic_read(&member->refcount)));
	}
	member->state = sstate->newstate;
	spin_unlock_irqrestore(&member->member_lock, mflags);
	spin_unlock_irqrestore(&dm->ddf_lock, dmflags);

	if (sstate->newstate != MDU_MSTATE_TEARDOWN) {
		int enable_sync;

		enable_sync = 0;
		if (member->subtype != MDK_MSUBTYPE_TOPLEVEL)
			ddf_update_array_state(to_mddev(member->parent));

		if (member->pers != NULL) {
			member->pers->raid_op(member,
					      MDK_RAID_OP_MSTATE_CHANGED,
					      member);
		}

		ddf_schedule_sync(&ai->sync_data, /* wait */0, enable_sync);
	}

	if (put_member) {
		mdk_member_t *parent;

		/*
		 * The rules for "sentinel puts" require us to
		 * guarantee that all other threads have seen
		 * a state change to either TEARDOWN or FAILED
		 * prior to the last reference count being released.
		 * This is achieved by briefly acquiring and
		 * releasing the parent's lock so that all threads
		 * iterating on the parent's child set have exited
		 * their loop since the state change and prior to
		 * our put.
		 */
		parent = member->parent;
		if (parent) {
			spin_lock_irqsave(&parent->member_lock, mflags);
			spin_unlock_irqrestore(&parent->member_lock, mflags);
		}

		member_put(member);
	}

	return (0);
}

static uint8_t
md_to_ddf_disk_state(mdu_mstate_t state)
{
	switch (state) {
	case MDU_MSTATE_OPTIMAL:
		return (DDF_PDE_ONLINE);
	case MDU_MSTATE_OPTIMAL|MDU_MSTATE_REBUILD_TARGET:
		return (DDF_PDE_REBUILD);
	case MDU_MSTATE_FAILED:
	default:
		return (DDF_PDE_FAILED);
	}
}

static int
md_to_ddf_array_state(mdu_mstate_t state, u_int *ddf_state, u_int *init_state)
{
	if ((state & MDU_MSTATE_INITIALIZING) != 0)
		*init_state = DDF_VDE_UNINTIALIZED;
	else
		*init_state = DDF_VDE_INIT_FULL;

	switch (state & ~(MDU_MSTATE_DIRTY|MDU_MSTATE_INITIALIZING)) {
	case MDU_MSTATE_OPTIMAL:
		*ddf_state = DDF_VDE_OPTIMAL;
		break;
	case MDU_MSTATE_DEGRADED:
		*ddf_state = DDF_VDE_DEGRADED;
	case MDU_MSTATE_FAILED:
	default:
		*ddf_state = DDF_VDE_FAILED;
	}

	return (0);
}

static int md_ddf_work_control(mdk_member_t *member, int op, void *arg)
{
	struct md_ddf_sync_data *si;
	mdk_workitem_t *work;
	struct list_head *tmp;
	struct md_ddf_array_info *ai;
	struct md_ddf_disk_info *di;
	struct md_ddf_domain *dm;
	u_long flags;
	int error;

	work = (mdk_workitem_t *)arg;
	si = (struct md_ddf_sync_data *)work->work_data;
	ai = si->sync_ai;
	dm = si->sync_dm;
	error = 0;

	switch (op) {
	case MDK_WORKITEM_OP_OPEN:
		
		spin_lock_irqsave(si->sync_lock, flags);
		if ((si->sync_flags & DDF_SYNC_FLAGS_SYNC_ALLOWED) == 0) {
			MD_DDF_DEBUG(("Sync not allowed\n"));
			spin_unlock_irqrestore(si->sync_lock, flags);
			error = -ESTALE;
			break;
		}
		spin_unlock_irqrestore(si->sync_lock, flags);
		/*
		 * Figure out which array records have changed, and copy them
		 * to all of the appropriate members.
		 *
		 * The whole point of copying the records into per-disk storage
		 * is that there might be a lot of disks, and/or they might
		 * take a long time to sync out.  Also, it's unlikely that you
		 * can hold the spinlock while trying to do I/O.
		 */
		spin_lock_irqsave(&dm->ddf_lock, flags);
		dm->pdr->CRC = ddf_calc_cksum(dm->pdr, dm->pdr_length * 512);
		memcpy(dm->sync_pdr, dm->pdr, dm->pdr_length * 512);
		dm->vdr->CRC = ddf_calc_cksum(dm->vdr, dm->vdr_length * 512);
		memcpy(dm->sync_vdr, dm->vdr, dm->vdr_length * 512);
		dm->cd->CRC = ddf_calc_cksum(dm->cd, dm->cd_length * 512);
		memcpy(dm->sync_cd, dm->cd, dm->cd_length * 512);

		list_for_each_entry(di, &dm->diskqueue, ddf_disk_links)
			di->ddf_dirty_mask |= dm->ddf_dirty_mask;
		dm->ddf_dirty_mask = 0;
		spin_unlock_irqrestore(&dm->ddf_lock, flags);
		
		if (ai && ai->ddf_dirty_mask) {
			mdk_member_t *disk;
			uint32_t array_seq;

			member = to_member(ai->mddev);
			spin_lock_irqsave(&member->member_lock, flags);
			/*
			 * Increment the sequence number, don't let it
			 * become -1 or 0
			 */
			array_seq = le32_to_cpu(ai->vdcr->Sequence_Number);
			if (array_seq != 0xffffffff)
				array_seq++;
			if (array_seq == 0xffffffff)
				array_seq = 1;
			ai->vdcr->Sequence_Number = cpu_to_le32(array_seq);
			ai->vdcr->CRC = ddf_calc_cksum(ai->vdcr, 2 * 512);
			if (ai->vendor != NULL) {
				ai->vendor->CRC =
				    ddf_calc_cksum(ai->vendor,
						   ai->vendor_length * 512);
			}
			ITERATE_CHILDREN_LOCKED(member, disk, tmp) {
				di = disk->metadata_private;

				if (di->vdcr != NULL) {
					memcpy(di->vdcr, ai->vdcr,
					       di->crl * 512);
					di->ddf_dirty_mask |= DI_DIRTY_VDCR;
				}
				if (ai->vendor != NULL && di->vendor != NULL) {
					memcpy(di->vendor, ai->vendor,
					       di->vendor_length * 512);
					di->ddf_dirty_mask |= DI_DIRTY_VENDOR;
				}
			}
			ai->ddf_dirty_mask &= ~(AI_DIRTY_VDCR|AI_DIRTY_VENDOR);
			spin_unlock_irqrestore(&member->member_lock, flags);
		}
		/*
		 * Mark the DDF header as open for all of the
		 * dirty disks in the domain.
		 */
		error = ddf_open_disks(dm, 1 /* open */);
		if (error) {
			printk("DDF: error %d while opening disks for update,"
			       " aborting update\n", error);
			break;
		}
		break;
	case MDK_WORKITEM_OP_CLOSE:
		/* Now that the update is complete, it's time to mark the
		 * DDF header as closed.
		 */
		spin_lock_irqsave(si->sync_lock, flags);
		if ((si->sync_flags & DDF_SYNC_FLAGS_SYNC_ALLOWED) == 0) {
			spin_unlock_irqrestore(si->sync_lock, flags);
			error = -ESTALE;
			break;
		}
		spin_unlock_irqrestore(si->sync_lock, flags);
		error = ddf_open_disks(dm, /*open*/0);
		if (error) {
			printk("DDF: error %d while closing disks after "
			       "update, disks might be inconsistent\n", error);
			break;
		}
		break;
	case MDK_WORKITEM_OP_FREE:
	{
		u_long dmflags;
		int requeue;

		spin_lock_irqsave(si->sync_lock, flags);
		si->sync_flags &= ~DDF_SYNC_FLAGS_SYNC_INPROG;
		if ((si->sync_flags & DDF_SYNC_FLAGS_SYNC_ALLOWED) == 0) {
			spin_unlock_irqrestore(si->sync_lock, flags);
			ddf_sync_data_put(si);
			break;
		}
		spin_unlock_irqrestore(si->sync_lock, flags);
		/*
		 * Check the dirty flags to see if state has changed
		 * during the sync.  If so, requeue our workitem.
		 */
		requeue = 0;
		spin_lock_irqsave(&dm->ddf_lock, dmflags);
		if (dm->ddf_dirty_mask != 0)
			requeue++;
		if (ai != NULL) {
			mdk_member_t *disk;
			
			member = to_member(ai->mddev);
			spin_lock_irqsave(&member->member_lock, flags);
			if (ai->ddf_dirty_mask != 0)
				requeue++;
			ITERATE_CHILDREN_LOCKED(member, disk, tmp) {
				di = disk->metadata_private;
				if (requeue || di->ddf_dirty_mask != 0) {
					requeue++;
					break;
				}
			}
			spin_unlock_irqrestore(&member->member_lock, flags);
		}
		spin_unlock_irqrestore(&dm->ddf_lock, dmflags);

		spin_lock_irqsave(si->sync_lock, flags);
		if (requeue
		 && (si->sync_flags & DDF_SYNC_FLAGS_SYNC_INPROG) == 0) {
			si->sync_flags |= DDF_SYNC_FLAGS_SYNC_INPROG;
			spin_unlock_irqrestore(si->sync_lock, flags);
			md_queue_workitem(&md_ddf_syncer,
					  &si->sync_workitem, 0);
		} else {
			spin_unlock_irqrestore(si->sync_lock, flags);
			ddf_sync_data_put(si);
		}
		break;
	}
	default:
		error = -EINVAL;
	}

	return (error);
}

static int
md_ddf_sync_handler(mddev_t *mddev, mdk_workitem_t *workitem)
{
	struct md_ddf_sync_data *sync_data;
	struct md_ddf_domain *dm;
	struct md_ddf_disk_info *di;
	u_long flags;
	int saved_gencount;
	int error;

	sync_data = (struct md_ddf_sync_data *)workitem->work_data;
	dm = sync_data->sync_dm;
	error = 0;

	/* Iterate all children write out the dirty records */
	spin_lock_irqsave(&dm->ddf_lock, flags);
	ddf_dm_get(dm);
rerun:
	saved_gencount = dm->diskqueue_gencount;
	list_for_each_entry(di, &dm->diskqueue, ddf_disk_links) {
		mdk_rdev_t *disk;
		mdk_member_t *member;

		disk = di->rdev;
		member = to_member(disk);
		if (di->ddf_dirty_mask) {
			MD_DDF_DEBUG(("Updating disk %x\n", di->dev));
			if (member)
				member_get(member);
			spin_unlock_irqrestore(&dm->ddf_lock, flags);
			error = ddf_sync_disk(di);
			if (error == -EIO) {
				if (member)
					md_set_member_state(member,
							    MDU_MSTATE_FAILED);
				printk("md_ddf_sync_handler: Unable to sync "
				       "metadata. So marking the disk (%d,%d) "
				       "as failed\n", MAJOR(di->dev), 
				       MINOR(disk->kdev));
			}
			if ((di->flags & DI_FLAGS_PUT_AFTER_SYNC) != 0
			 && member) {
				mdk_member_t *array;

				/*
				 * The rules for "sentinel puts" require us to
				 * guarantee that all other threads have seen
				 * a state change to either TEARDOWN or FAILED
				 * prior to the last reference count being
				 * released.   This is achieved by briefly
				 * acquiring and releasing the parent's lock
				 * so that all threads iterating on the
				 * parent's child set have exited their loop
				 * since the state change and prior to our put.
				 */
				array = member->parent;

				if (array != NULL) {
					spinlock_t *plock;

					plock = &array->member_lock;
					spin_lock_irqsave(plock, flags);
					spin_unlock_irqrestore(plock, flags);
				}
				member_put(member);
				di->flags &= ~DI_FLAGS_PUT_AFTER_SYNC;
			}
			if (member)
				member_put(member);
			spin_lock_irqsave(&dm->ddf_lock, flags);
			if (error)
				break;
			if (dm->diskqueue_gencount != saved_gencount)
				goto rerun;
		}
	}
	ddf_dm_put(dm);
	spin_unlock_irqrestore(&dm->ddf_lock, flags);
	return (error);
}

/*
 * Calculate the meta-data checksum with the checksum field
 * set to 0xFFFFFFFF, using a seed of 0xFFFFFFFF, and a final
 * xor with 0xFFFFFFFF after processing all data.  The CRC
 * field of the passed in data is *not* modified.
 */
static uint32_t
ddf_calc_cksum(void *data, int size)
{
	struct ddf_header *ddf;
	uint32_t old_sum;
	uint32_t new_sum;

	/*
	 * Make it look like a DDF Header so that the CRC field can be
	 * reset.  All DDF records that record a CRC record it in the
	 * same position, so this is safe for now.
	 */
	ddf = (struct ddf_header *)data;
	old_sum = ddf->CRC;
	ddf->CRC = 0xffffffff;

	new_sum = crc32(0xffffffff, data, size) ^ 0xffffffff;
	ddf->CRC = old_sum;
	return (cpu_to_le32(new_sum));
}

static void
ddf_create_controller_guid(char *data, char *vendor)
{
	char buf[25];
	uint32_t ddf_time;
	uint32_t random;

	memcpy(&data[0], vendor, 8);
	ddf_time = ddf_gettimeofday();
	get_random_bytes(&random, 4);
	sprintf(buf, "%-8.8s%08x%08x", vendor, ddf_time, random);
/* XXX BIOS looks for this at the end */
	memcpy(&buf[16], vendor, 4);
/* XXX BIOS only handles 20 byte GUIDs */
	memset(&buf[20], 0xFF, 4);
	MD_DDF_DEBUG(("controller_guid %s\n", buf));
	memcpy(data, buf, 24);
}

static void
ddf_create_pd_guid(char *data)
{
	char buf[25];
	uint32_t random[3];

	get_random_bytes(random, 12);
	sprintf(buf, "HDD %04x%08x%08x",
		random[0] & 0xFFFF,
		random[1], random[2]);
/* XXX BIOS only handles 20 byte GUIDs */
	memset(&buf[20], 0xFF, 4);
	memcpy(data, buf, 24);
	MD_DDF_DEBUG(("pd_guid %s\n", buf));
}

static void
ddf_create_vd_guid(char *data, char *vendor)
{
	char buf[25];
	uint32_t ddf_time;
	uint32_t random;

	ddf_time = ddf_gettimeofday();
	get_random_bytes(&random, 4);
	sprintf(buf, "%-8.8s%08x%08x", vendor,
		ddf_time, random);
/* XXX BIOS only handles 20 byte GUIDs */
	memset(&buf[20], 0xFF, 4);
	MD_DDF_DEBUG(("vd_guid %s\n", buf));
	memcpy(data, buf, 24);
}

static void
ddf_create_hdr_guid(char *data, char *vendor)
{
	ddf_create_vd_guid(data, vendor);
}

static struct ddf_header *
ddf_init_header(int header_type, sector_t size, int crl, int cr_size)
{
	struct ddf_header *header;
	uint64_t ddf_start;

	ddf_start = size - 11;
	header = kmalloc(DDF_HEADER_BYTES, GFP_KERNEL);
	if (header == NULL)
		return (NULL);

	memset(header, 0xff, DDF_HEADER_BYTES);
	header->Signature = cpu_to_le32(DDF_HEADER_SIGNATURE);
	ddf_create_hdr_guid(header->DDF_Header_GUID, "ADPT    ");
	memcpy(header->DDF_rev, DDF_REV, 8);
	header->Sequence_Number = cpu_to_le32(0x1);
	header->TimeStamp = ddf_gettimeofday();
	header->Open_Flag = 0x00;
	header->Foreign_Flag = 0x0;
	header->Diskgrouping = 0x0;
	header->Primary_Header_LBA = cpu_to_le64(ddf_start);
	header->Header_Type = header_type;
	header->WorkSpace_Length = cpu_to_le32(0);
	header->Max_PD_Entries = cpu_to_le16(MD_DDF_MAX_PDE);
	header->Max_VD_Entries = cpu_to_le16(MD_DDF_MAX_VDE);
	header->Max_Partitions = cpu_to_le16(0x1);
	header->Configuration_Record_Length = cpu_to_le16(crl);
	header->cd_section = cpu_to_le32(DDF_ADPT_CD_OFFSET);
	header->cd_length = cpu_to_le32(DDF_ADPT_CD_LENGTH);
	header->pdr_section = cpu_to_le32(DDF_ADPT_PDR_OFFSET);
	header->pdr_length = cpu_to_le32(DDF_ADPT_PDR_LENGTH);
	header->vdr_section = cpu_to_le32(DDF_ADPT_VDR_OFFSET);
	header->vdr_length = cpu_to_le32(DDF_ADPT_VDR_LENGTH);
	header->cr_section = cpu_to_le32(DDF_ADPT_VDCR_OFFSET);
	header->cr_length = cpu_to_le32(cr_size);
	header->pdd_section = cpu_to_le32(DDF_ADPT_PDD_OFFSET);
	header->pdd_length = cpu_to_le32(DDF_ADPT_PDD_LENGTH);
	header->vendor_section = cpu_to_le32(DDF_ADPT_VENDOR_OFFSET);
	header->vendor_length = cpu_to_le32(DDF_ADPT_VENDOR_LENGTH);

	return (header);
}

static void *
ddf_init_cr(uint32_t crl, u_int num_recs)
{
	void *cr;

	cr = kmalloc(crl * num_recs * 512, GFP_KERNEL);
	if (cr)
		memset(cr, 0xFF, crl * num_recs * 512);
	return (cr);
}

static int
ddf_init_cd(struct md_ddf_domain *dm)
{
	struct ddf_cd_record *cd;
	struct ddf_cd_record *sync_cd;

	cd = kmalloc(DDF_ADPT_CD_LENGTH * 512, GFP_KERNEL);
	if (cd == NULL)
		return (-ENOMEM);
	sync_cd = kmalloc(DDF_ADPT_CD_LENGTH * 512, GFP_KERNEL);
	if (sync_cd == NULL) {
		kfree(cd);
		return (-ENOMEM);
	}

	memset(cd, 0xff, DDF_ADPT_CD_LENGTH * 512);
	cd->Signature = cpu_to_le32(DDF_CD_SIGNATURE);
	ddf_create_controller_guid(cd->Controller_GUID, "ADPT");
	/*
	 * XXX We don't know the real Adapter, so try and make the BIOS happy.
	 */
	cd->Adapter_Type.Vendor_ID = cpu_to_le16(0x8086);
	cd->Adapter_Type.Device_ID = cpu_to_le16(0x24DF);
	cd->Adapter_Type.SubVendor_ID = cpu_to_le16(0x8086);
	cd->Adapter_Type.SubDevice_ID = cpu_to_le16(0x24DF);
	dm->cd = cd;
	dm->sync_cd = sync_cd;
	dm->cd_length = DDF_ADPT_CD_LENGTH;
	memcpy(dm->Controller_GUID, cd->Controller_GUID, DDF_GUID_BYTES);
	return (0);
}

static int
ddf_init_pdr(struct md_ddf_domain *dm, int populated_pdes)
{
	struct ddf_pd_record *pdr;
	struct ddf_pd_record *sync_pdr;

	pdr = kmalloc(DDF_ADPT_PDR_LENGTH * 512, GFP_KERNEL);
	if (pdr == NULL)
		return (-ENOMEM);

	sync_pdr = kmalloc(DDF_ADPT_PDR_LENGTH * 512, GFP_KERNEL);
	if (sync_pdr == NULL) {
		kfree(pdr);
		return (-ENOMEM);
	}

	memset(pdr, 0xff, DDF_ADPT_PDR_LENGTH * 512);
	pdr->Signature = cpu_to_le32(DDF_PDR_SIGNATURE);
	pdr->Populated_PDEs = cpu_to_le16(populated_pdes);
	pdr->Max_PD_Entries = cpu_to_le16(MD_DDF_MAX_PDE);

	dm->pdr = pdr;
	dm->sync_pdr = sync_pdr;
	dm->pdr_length = DDF_ADPT_PDR_LENGTH;

	return (0);
}

static struct ddf_pd_entry *
ddf_init_pde(struct ddf_pd_record *pdr, int entry, int type, int state,
	     uint64_t size, char *guid, uint32_t ref)
{
	struct ddf_pd_entry *pde;

	pde = &pdr->entry[entry];
	memcpy(pde->PD_GUID, guid, DDF_GUID_BYTES);
	pde->PD_Reference = ref;
	pde->PD_Type = cpu_to_le16(type);
	pde->PD_State = cpu_to_le16(state);
	pde->Configured_Size = cpu_to_le64(size);

	return (pde);
}

static int
ddf_init_vdr(struct md_ddf_domain *dm, int populated_vdes)
{
	struct ddf_vd_record *vdr;
	struct ddf_vd_record *sync_vdr;

	vdr = kmalloc(DDF_ADPT_VDR_LENGTH * 512, GFP_KERNEL);
	if (vdr == NULL)
		return (-ENOMEM);

	sync_vdr = kmalloc(DDF_ADPT_VDR_LENGTH * 512, GFP_KERNEL);
	if (sync_vdr == NULL) {
		kfree(vdr);
		return (-ENOMEM);
	}

	memset(vdr, 0xff, DDF_ADPT_VDR_LENGTH * 512);
	vdr->Signature = cpu_to_le32(DDF_VD_RECORD_SIGNATURE);
	vdr->Populated_VDEs = cpu_to_le16(populated_vdes);
	vdr->Max_VD_Entries = cpu_to_le16(MD_DDF_MAX_VDE);
	dm->vdr = vdr;
	dm->sync_vdr = sync_vdr;
	dm->vdr_length = DDF_ADPT_VDR_LENGTH;
	return (0);
}

static struct ddf_vd_entry *
ddf_init_vde(struct ddf_vd_record *vdr, int entry, int type, int state,
	     int init_state, char *name, char *guid)
{
	struct ddf_vd_entry *vde;

	vde = &vdr->entry[entry];
	memcpy(vde->VD_GUID, guid, DDF_GUID_BYTES);
	vde->VD_Number = cpu_to_le16(entry);
	vde->VD_Type = cpu_to_le16(type);
	vde->VD_State = cpu_to_le16(state);
	vde->Init_State = cpu_to_le16(init_state);
	memcpy(vde->VD_Name, name, 16);

	return (vde);
}

static struct ddf_vdc_record *
ddf_init_vdcr(int count, int ss, int level, int rlq, uint64_t size, int crl)
{
	struct ddf_vdc_record *vdcr;

	vdcr = kmalloc(crl * 512, GFP_KERNEL);
	if (vdcr == NULL)
		return (NULL);

	memset(vdcr, 0xff, crl * 512);
	vdcr->Signature = cpu_to_le32(DDF_VDCR_SIGNATURE);
	ddf_create_vd_guid(vdcr->VD_GUID, "ADPT");
	vdcr->Timestamp = ddf_gettimeofday();
	vdcr->Sequence_Number = cpu_to_le32(0x0);
	vdcr->Primary_Element_Count = cpu_to_le16(count);
	vdcr->Stripe_Size = ffs(ss >> 9) - 1;
	vdcr->Primary_RAID_Level = level;
	vdcr->RLQ = rlq;
	vdcr->Secondary_Element_Count = 0x0;
	vdcr->Block_Count = cpu_to_le64(size);
	vdcr->Size = cpu_to_le64(size);
	vdcr->Cache_Flags = cpu_to_le64(0x0);
	vdcr->BG_Rate = 0x0;

	return (vdcr);
}

int
ddf_init_sa(struct md_ddf_domain *dm, struct ddf_sa_record *sa)
{
	memset(sa, 0xff, dm->crl * 512);
	sa->Signature = cpu_to_le32(DDF_SA_SIGNATURE);
	sa->Channel_Affinity = 0;
	sa->Spare_Type = 0;
	sa->CRC = ddf_calc_cksum(sa, dm->crl * 512);
	return (0);
}

static struct ddf_pdd_record *
ddf_init_pdd(void)
{
	struct ddf_pdd_record *pdd;

	pdd = kmalloc(DDF_ADPT_PDD_LENGTH * 512, GFP_KERNEL);
	if (pdd == NULL)
		return (NULL);

	memset(pdd, 0xff, DDF_ADPT_PDD_LENGTH * 512);
	pdd->Signature = cpu_to_le32(DDF_PDD_SIGNATURE);
	ddf_create_pd_guid(pdd->PD_GUID);
	get_random_bytes(&pdd->PD_Reference, 4);
	pdd->Forced_Ref_Flag = 0;
	pdd->Forced_PD_GUID_Flag = 0;

	return (pdd);
}

static struct ddf_vendor_log *
ddf_init_vendor(struct md_ddf_domain *dm)
{
	struct ddf_vendor_log *vendor;
	struct adpt_vendor_data *adpt;

	vendor = kmalloc(DDF_ADPT_VENDOR_LENGTH * 512, GFP_KERNEL);
	if (vendor == NULL)
		return (NULL);

	memset(vendor, 0xff, DDF_ADPT_VENDOR_LENGTH * 512);
	vendor->Signature = cpu_to_le32(DDF_VENDOR_LOG_SIGNATURE);
	if (dm != NULL && dm->cd != NULL) 
		memcpy(vendor->Log_Owner, &dm->cd->Adapter_Type,
		       sizeof(vendor->Log_Owner));

	adpt = (struct adpt_vendor_data *)&vendor->data[0];
	memset(adpt, 0, sizeof(struct adpt_vendor_data));

	return (vendor);
}

static int
ddf_set_checkpoint(mddev_t *mddev, mdk_checkpoint_arg_t *cp)
{
	struct md_ddf_array_info *ai;
	int build_complete, sync_md;

	sync_md = 0;
	build_complete = 0;
	ai = to_member(mddev)->metadata_private;

	if (cp->checkpoint_type == MD_SYNCOP_NONE) {
		ai->checkpoint = 0;
		sync_md = 1;
	} else if (cp->checkpoint < to_member(mddev)->data_size) {
		/* update the checkpoint record */
		ai->checkpoint = cp->checkpoint;
	} else {
		MD_DDF_DEBUG(("DDF: build complete, check point 0x%lx\n",
			      cp->checkpoint));
		build_complete = 1;
		sync_md = 1;
		ai->checkpoint = 0;
	}

	/*
	 * Update vendor record if available.
	 */
	if (ai->vendor_type == AI_VENDOR_ADPT) {
		struct adpt_vendor_data *adpt;

		adpt = (struct adpt_vendor_data *)ai->vendor->data;
		adpt->CheckpointLBA = cpu_to_le32(ai->checkpoint);
		adpt->irocFlags &= ~DDF_ADPT_IF_VERIFY_WITH_AUTOFIX;
		switch (cp->checkpoint_type) {
		case MD_SYNCOP_REBUILD:
			adpt->CheckpointActivity = DDF_ADPT_CHKPNT_ACT_REBUILD;
			break;
		case MD_SYNCOP_INIT:
			adpt->CheckpointActivity = DDF_ADPT_CHKPNT_ACT_ZERO;
			break;
		case MD_SYNCOP_VERIFY_FIX:
			adpt->irocFlags |= DDF_ADPT_IF_VERIFY_WITH_AUTOFIX;
			/* FALLTHROUGH */
		case MD_SYNCOP_VERIFY:
			adpt->CheckpointActivity = DDF_ADPT_CHKPNT_ACT_VERIFY;
			break;
		default:
			adpt->CheckpointActivity = DDF_ADPT_CHKPNT_ACT_NONE;
			break;
		}
		if (build_complete)
			adpt->CheckpointActivity = DDF_ADPT_CHKPNT_ACT_NONE;
		ai->ddf_dirty_mask |= AI_DIRTY_VENDOR;
	}

	if (sync_md)
		ddf_schedule_sync(&ai->sync_data, /*wait*/0, /*enable*/1);

	if (build_complete) {
		mdk_member_t *array;
		mdk_member_t *child;
		struct list_head *tmp;
		u_long cflags;

		array = to_member(mddev);
		ITERATE_CHILDREN(array, child, tmp, cflags) {
			u_long cflags;

			spin_lock_irqsave(&child->member_lock, cflags);
			if (MDK_WRITABLE_MEMBER(child)
			 && (child->state & MDU_MSTATE_REBUILD_TARGET) != 0) {
				mdk_set_state_arg_t sstate;
				mdk_metapersonality_t *mpers;

				sstate.oldstate = child->state;
				spin_unlock_irqrestore(&child->member_lock,
						       cflags);
				sstate.newstate = child->state
						& ~MDU_MSTATE_REBUILD_TARGET;
				mpers = child->metapers;
				mpers->member_op(child,
						 MDK_MEMBER_OP_SET_STATE,
						 &sstate);
			} else {
				spin_unlock_irqrestore(&child->member_lock,
						       cflags);
			}
		}

		/*
		 * When verifies complete, clear the dirty flag.
		 */
		if ((cp->checkpoint_type & MD_SYNCOP_VERIFY) != 0) {
			mdk_set_state_arg_t sstate;

			spin_lock_irqsave(&array->member_lock, cflags);
			sstate.oldstate = array->state;
			sstate.newstate = array->state & ~MDU_MSTATE_DIRTY;
			spin_unlock_irqrestore(&array->member_lock, cflags);
			ddf_set_array_state(mddev, &sstate);
		}
	}

	return (0);
}

#define DDF_WIPE_RECORD(section, length)				\
	offset = ddf_primary_base + le32_to_cpu(di->hdr_pri->section);	\
	len = le32_to_cpu(di->hdr_pri->length);				\
	MD_DDF_DEBUG((#section "off %llx len %x\n", offset, len));	\
	if ((len != 0xffffffff) && (offset != 0xffffffff)) {		\
		MD_DDF_DEBUG(("Wiping" #section "\n"));			\
		while (len--) {						\
			error = md_write_blocks(dev, offset, 1,		\
						(uint8_t *)buf);	\
			offset++;					\
			if (error)					\
				goto done;				\
		}							\
	}

static int
ddf_wipe_metadata(kdev_t dev)
{
	struct ddf_header *hdr;
	struct md_ddf_disk_info *di;
	struct md_ddf_disk_info *tmp_info;
	struct md_ddf_disk_info *active_di;
	struct md_ddf_domain *dm;
	struct md_ddf_array_info *ai;
	uint8_t *buf;
	uint64_t dev_size;
	uint64_t offset;
	uint64_t ddf_primary_base;
	uint32_t len;
	u_long dmflags;
	u_long mflags;
	int error;

	buf = kmalloc(DDF_HEADER_BYTES, GFP_KERNEL);
	if (buf == NULL)
		return (-ENOMEM);
	memset(buf, 0xFF, DDF_HEADER_BYTES);
	error = ddf_get_anchor(dev, &hdr, &dev_size);
	if (error) {
		printk("ddf_Get_anchor returned %d\n", error);
		kfree(buf);
		return (error);
	}

	ddf_primary_base = le64_to_cpu(hdr->Primary_Header_LBA);
	if (ddf_primary_base > dev_size - 1)
		ddf_primary_base = dev_size - 1;
	/* Next get the DDF records for the disk_info structure */
	di = NULL;
	error = ddf_get_disk_info(dev, hdr, dev_size, &di);
	if (error != 0
	 || ddf_get_domain(di, &dm) != 0) {
		/*
		 * Since we found an anchor, effect the wipe
		 * by killing it and the primary header.
		 */
		error = md_write_blocks(dev, dev_size - 1, DDF_HEADER_LENGTH,
				        (uint8_t *)buf);
		if (error == 0)
			error = md_write_blocks(dev, ddf_primary_base,
						DDF_HEADER_LENGTH,
						(uint8_t *)buf);
		if (di != NULL)
			ddf_free_disk_info(di);
		else
			kfree(hdr);
		kfree(buf);
		return (error);
	}

	/*
	 * If we already have a disk info for this disk in the domain,
	 * use it instead of putting one together.
	 */
	ai = NULL;
	active_di = NULL;
	spin_lock_irqsave(&dm->ddf_lock, dmflags);
	list_for_each_entry(tmp_info, &dm->diskqueue, ddf_disk_links) {

		if (memcmp(&di->pde->PD_GUID[0],
			   &tmp_info->pde->PD_GUID[0], DDF_GUID_BYTES) == 0) {
			active_di = tmp_info;
			break;
		}
	}
	spin_unlock_irqrestore(&dm->ddf_lock, dmflags);
	if (active_di != NULL) {
		ddf_free_disk_info(di);
		di = active_di;
		ai = di->parent;
	} else if ((le16_to_cpu(di->pde->PD_Type)
		 & DDF_PDE_PARTICIPATING) != 0) {
		error = ddf_get_array_info(di, &ai);
		if (error != 0)
			ai = NULL;
	}

	/* Remove the PDE record for this disk in the domain. */
	spin_lock_irqsave(&dm->ddf_lock, dmflags);
	if (ai != NULL && ai->vdcr != NULL) {
		mdk_member_t *array;
		u_int num_disks;
		u_int remaining;
		u_int i;

		array = to_member(ai->mddev);
		mflags = 0;
		if (ai->mddev != NULL)
			spin_lock_irqsave(&array->member_lock, mflags);

		/* Wipe our entry from the VDCR record. */
		num_disks = le16_to_cpu(ai->vdcr->Primary_Element_Count);
		for (remaining = 0, i = 0; i < num_disks; i++) {
			uint32_t pdref;

			pdref = ai->pds_array[i];
			if (di->pde->PD_Reference == pdref) {
				ai->pds_array[i] = 0xFFFFFFFF;
				ai->ss_array[i] = 0xFFFFFFFF;
				MD_DDF_DEBUG(("Clearing pds entry %i\n", i));
			} else if (pdref != 0xFFFFFFFF)
				remaining++;
		}
		/*
		 * This array no longer has any valid entries.
		 * Remove it from the domain.
		 */
		if (remaining == 0) {
			uint16_t pop_vdes;

			pop_vdes = le16_to_cpu(dm->vdr->Populated_VDEs);
			memset(ai->vde, 0xFF, sizeof(*ai->vde));
			dm->vdr->Populated_VDEs = cpu_to_le16(pop_vdes - 1);
		}
		ai->ddf_dirty_mask |= AI_DIRTY_VDCR;
		dm->ddf_dirty_mask |= DM_DIRTY_VDR;
		if (ai->mddev != NULL)
			spin_unlock_irqrestore(&array->member_lock, mflags);
	}
	if (di->pde != NULL) {
		uint16_t pop_pdes;

		pop_pdes = le16_to_cpu(dm->pdr->Populated_PDEs);
		memset(di->pde, 0xFF, sizeof(*di->pde));
		dm->pdr->Populated_PDEs = cpu_to_le16(pop_pdes - 1);
		dm->ddf_dirty_mask |= DM_DIRTY_PDR;
	}
	spin_unlock_irqrestore(&dm->ddf_lock, dmflags);

	if (ai != NULL && ai->mddev != NULL)
		ddf_schedule_sync(&ai->sync_data, /*wait*/1, /*enable*/1);
	else
		ddf_schedule_sync(&dm->sync_data, /*wait*/1, /*enable*/1);

	/* Kill the Anchor first in case wiping any other records fail.*/
	error = md_write_blocks(dev, dev_size - 1,
				DDF_HEADER_LENGTH, (uint8_t *)buf);
	if (error == 0)
		goto done;

	DDF_WIPE_RECORD(cd_section, cd_length);
	DDF_WIPE_RECORD(pdr_section, pdr_length);
	DDF_WIPE_RECORD(vdr_section, vdr_length);
	DDF_WIPE_RECORD(cr_section, cr_length);
	DDF_WIPE_RECORD(pdd_section, pdd_length);
	DDF_WIPE_RECORD(bbmlog_section, bbmlog_length);
	DDF_WIPE_RECORD(Diagnostic_Space, Diagnostic_Space_Length);
	DDF_WIPE_RECORD(vendor_section, vendor_length);

	/* Kill the primary header */
	error = md_write_blocks(dev, ddf_primary_base,
				DDF_HEADER_LENGTH, (uint8_t *)hdr);
done:
	kfree(buf);
	ddf_free_disk_info(di);
	return (error);
}

static int
ddf_find_domain(struct md_ddf_domain **dm)
{
	struct md_ddf_domain *new_dm;
	u_long flags;

	*dm = NULL;
	spin_lock_irqsave(&ddf_domains_lock, flags);
	/* For now, assume that there is a single global domain. */
	if (!list_empty(&ddf_domains)) { 
		*dm = list_entry(ddf_domains.next, struct md_ddf_domain,
				 ddf_domain_links);
		ddf_dm_get(*dm);
		spin_unlock_irqrestore(&ddf_domains_lock, flags);
		return (0);
	}
	spin_unlock_irqrestore(&ddf_domains_lock, flags);

	new_dm = ddf_alloc_domain();
	if (ddf_init_cd(new_dm) != 0
	 || ddf_init_pdr(new_dm, 0)
	 || ddf_init_vdr(new_dm, 0)) {
		ddf_dm_put(new_dm);
		return (-ENOMEM);
	}
	new_dm->crl = 2;

	spin_lock_irqsave(&ddf_domains_lock, flags);
	list_add_tail(&new_dm->ddf_domain_links, &ddf_domains);
	spin_unlock_irqrestore(&ddf_domains_lock, flags);

	*dm = new_dm;
	return (0);
}

static int
ddf_set_array_info(mdk_member_t *member, mdu_array_info_t *ui)
{
	struct md_ddf_array_info *ai;
	struct md_ddf_domain *dm;
	struct ddf_vd_entry *vde;
	mddev_t *mddev;
	uint64_t size;
	u_long flags;
	int level, rlq;
	int state, init_state;
	int entry, crl;
	int error;

	error = ddf_find_domain(&dm);
	if (error)
		return (error);
	ai = ddf_alloc_array_info(dm);
	if (ai == NULL) {
		ddf_dm_put(dm);
		return (-ENOMEM);
	}

	ai->vendor = ddf_init_vendor(dm);
	if (ai->vendor == NULL) {
		ddf_ai_put(ai);
		return (-ENOMEM);
	}
	ai->vendor_length = DDF_ADPT_VENDOR_LENGTH;
	ai->vendor_type = AI_VENDOR_ADPT;

	/* Figure out array properties */
	size = ui->size;

	switch (ui->level) {
	case 0:
		size *= ui->raid_disks;
		break;
	case 1:
		break;
	case 5:
		size *= ui->raid_disks - 1;
		break;
	default:
		ddf_ai_put(ai);
		return (-EINVAL);
	}

	error = ddf_md_to_ddf_raid_type(ui->level, &level, &rlq);
	if (error) {
		ddf_ai_put(ai);
		return (error);
	}
	error = md_to_ddf_array_state(ui->state, &state, &init_state);
	if (error) {
		ddf_ai_put(ai);
		return (error);
	}
	crl = dm->crl;

	/* Figure out the proper VDE slot to put this new array */
	for (entry = 0;
	     entry < le16_to_cpu(ai->domain->vdr->Max_VD_Entries); entry++) {
		vde = &ai->domain->vdr->entry[entry];
		if (vde->VD_GUID[0] == 0xff)
			break;
	}
	if (entry >= le16_to_cpu(ai->domain->vdr->Max_VD_Entries)) {
		printk("DDF: Cannot allocate any more VDEs: %d:%d\n",
		       entry, le16_to_cpu(ai->domain->vdr->Max_VD_Entries));
		ddf_ai_put(ai);
		return (-EINVAL);
	}

	ai->vdcr = ddf_init_vdcr(ui->raid_disks, ui->chunk_size, level, rlq,
				 size, crl);
	ai->vde = ddf_init_vde(dm->vdr, entry, 0 /*type*/, state, init_state,
			       ui->array_name, ai->vdcr->VD_GUID);
	dm->vdr->Populated_VDEs =
	    cpu_to_le16(le16_to_cpu(dm->vdr->Populated_VDEs) + 1);
	ai->ddf_dirty_mask |= AI_DIRTY_VDCR;
	dm->ddf_dirty_mask |= DM_DIRTY_VDR;

	ai->pds_len = le16_to_cpu(dm->pdr->Max_PD_Entries) * 4;
	ai->ss_len = le16_to_cpu(dm->pdr->Max_PD_Entries) * 8;
	ai->pds_array =
	    (uint32_t *)&((uint8_t *)(ai->vdcr))[DDF_VDCR_PDS_OFFSET];
	ai->ss_array =
	    (uint32_t *)&((uint8_t *)(ai->vdcr))[DDF_VDCR_PDS_OFFSET +
						 ai->pds_len];

	ai->checkpoint = 0;
	spin_lock_irqsave(&dm->ddf_lock, flags);
	list_add(&ai->ddf_array_links, &dm->arrayqueue);
	spin_unlock_irqrestore(&dm->ddf_lock, flags);

	mddev = ddf_attach_array(ai);
	if (mddev == NULL) {
		ddf_ai_put(ai);
		return (-ENOMEM);
	}
	md_assign_unit_number(mddev);
	ui->cookie = (u_long)mddev;
	ddf_ai_put(ai);
	return (0);
}

static int
ddf_add_new_disk(mddev_t *mddev, mdu_disk_info_t *ui)
{
	struct md_ddf_disk_info *di;
	struct md_ddf_array_info *ai;
	struct ddf_pd_entry *pde;
	struct gendisk *gd;
 	sector_t configured_size;
	u_long dmflags;
	kdev_t dev;
	int entry, type, state, error;

	dev = MKDEV(ui->major, ui->minor);
	di = ddf_find_disk_info(dev);
	if (di != NULL) {
		u_int type;
		u_int mask;

		MD_DDF_DEBUG(("ddf_add_new_disk: "
			      "Found existing disk %p\n", di));
		mask = DDF_PDE_GLOBAL_SPARE
		     | DDF_PDE_CONFIG_SPARE
		     | DDF_PDE_PARTICIPATING;
		type = le16_to_cpu(di->pde->PD_Type);
		if ((type & mask) != 0
		 && (di->pde->PD_State & (DDF_PDE_FAILED|DDF_PDE_MISSING)) == 0)
			return (-EBUSY);

		/*
		 * Remove from the domain prior to using
		 * this disk for a new array.
		 */
		ddf_free_disk_info(di);
	}

	gd = get_gendisk(dev);
	if (gd == NULL)
		return (-ENODEV);
	di = ddf_alloc_disk_info(dev, gd->part[MINOR(dev)].nr_sects);
	if (di == NULL)
		return (-ENOMEM);
	ai = to_member(mddev)->metadata_private;
	di->parent = ddf_ai_get(ai);
	di->crl = ai->domain->crl;
	di->cr_length = di->crl * 2;

	type = DDF_PDE_PARTICIPATING;
	state = md_to_ddf_disk_state(ui->state);

	/* Set up initial records */
	di->hdr_anchor = ddf_init_header(DDF_HEADER_ANCHOR,
					 di->dev_size - 1, di->crl,
					 di->crl * 2);
	di->hdr_pri = ddf_init_header(DDF_HEADER_PRIMARY,
				      di->dev_size - 1, di->crl, di->crl * 2);
	di->cr = ddf_init_cr(di->crl, di->cr_length);
	di->vendor = ddf_init_vendor(NULL);

	if (di->hdr_anchor == NULL
	 || di->hdr_pri == NULL
	 || di->cr == NULL
	 || di->vendor == NULL) {
		ddf_free_disk_info(di);
		return (-ENOMEM);
	}

	di->domain = ddf_dm_get(ai->domain);
	di->cd = di->domain->cd;
	di->pdr = di->domain->pdr;
	di->vdr = di->domain->vdr;
	di->vdcr = (void *)di->cr;
	di->cr_consumers++;
	di->cd_length = di->domain->cd_length;
	di->pdr_length = di->domain->pdr_length;
	di->vdr_length = di->domain->vdr_length;

	/* Figure out the correct PDE slot for this disk */
	for (entry = 0; entry < le16_to_cpu(di->pdr->Max_PD_Entries); entry++) {
		pde = &di->pdr->entry[entry];
		if (pde->PD_GUID[0] == 0xff)
			break;
	}
	if (entry >= le16_to_cpu(di->pdr->Max_PD_Entries)) {
		printk("DDF: Cannot allocate any more PDEs\n");
		ddf_free_disk_info(di);
		return (-EINVAL);
	}

	/* Allocate the PDD and PDE, and update the PDR (parent of PDE) */
	di->pdd = ddf_init_pdd();
	error = ddf_get_freesize(di->dev, &configured_size);
	if (error != 0) {
		printk("DDF: Cannot determine device freesize\n");
		ddf_free_disk_info(di);
		return (error);
	}
		
	di->pde = ddf_init_pde(di->pdr, entry, type, state, configured_size,
			       di->pdd->PD_GUID,
			       di->pdd->PD_Reference);
	di->pdr->Populated_PDEs =
	    cpu_to_le16(le16_to_cpu(di->pdr->Populated_PDEs) + 1);

	/* Insert the member into the VDCR array, and update */
	if (ai->pds_array[ui->number] == 0xffffffff) {
		ai->pds_array[ui->number] = di->pdd->PD_Reference;
	} else {
		printk("DDF: pds_array entry already filled!\n");
		ddf_free_disk_info(di);
		return (-EINVAL);
	}
	if (ai->ss_array[ui->number] == 0xffffffff)
		ai->ss_array[ui->number] = 0x0;
	else {
		printk("DDF: ss_array entry already filled!\n");
		ddf_free_disk_info(di);
		return (-EINVAL);
	}

	/* Final housekeeping to complete the disk_info */
	di->cr_length = di->crl * 2;
	di->pdd_length = DDF_ADPT_PDD_LENGTH;
	di->vendor_length = ai->vendor_length;
	di->dev = dev;
	di->ss_len = ai->ss_len;
	di->pds_len = ai->pds_len;
	memcpy(di->vend_id, di->cd->Controller_GUID, 8);

	/* Announce to the system */
	error = ddf_attach_drive(di, ai);
	if (error != 0) {
		ddf_free_disk_info(di);
		return (error);
	}

	spin_lock_irqsave(&di->domain->ddf_lock, dmflags);
	list_add(&di->ddf_disk_links, &di->domain->diskqueue);
	di->domain->diskqueue_gencount++;
	ui->cookie = (u_long)di->rdev;

	/* Mark everything as dirty and sync it all out */
	di->ddf_dirty_mask = (DI_DIRTY_HDR_ANCHOR|DI_DIRTY_HDR_PRI|DI_DIRTY_CD|
			      DI_DIRTY_PDR|DI_DIRTY_VDR|DI_DIRTY_VDCR|
			      DI_DIRTY_PDD|DI_DIRTY_VENDOR);
	ai->ddf_dirty_mask |= AI_DIRTY_VDCR;
	ai->domain->ddf_dirty_mask |= (DM_DIRTY_PDR|DM_DIRTY_VDR);
	spin_unlock_irqrestore(&di->domain->ddf_lock, dmflags);
	ddf_schedule_sync(&ai->sync_data, /*wait*/1, /*enable*/1);

	return (0);
}

/*
 * Handle notification of our array transitioning into
 * Read/Write mode or, in MD parlance, "exiting safemode".
 */
static int
ddf_exit_safemode(mddev_t *mddev)
{
	struct md_ddf_array_info *ai;
	mdk_member_t *member;
	u_long dmflags;
	u_long mflags;
	int sched_write;

	/*
	 * Update the sequence number and mark the
	 * array meta-data as open.  We use this
	 * status to detect improper shutdown of
	 * the array.  Incrementing the sequence number
	 * automaticallly prevents "missing" members
	 * that arrive late from being accepted into
	 * the array.
	 */
	MD_DDF_DEBUG(("DDF Exit Safemode\n"));
	member = to_member(mddev);
	sched_write = 0;
	ai = member->metadata_private;
	spin_lock_irqsave(&ai->domain->ddf_lock, dmflags);
	spin_lock_irqsave(&member->member_lock, mflags);
	if ((ai->vde->VD_State & DDF_VDE_DIRTY) == 0) {
		ai->vde->VD_State |= DDF_VDE_DIRTY;
		sched_write = 1;
		ai->domain->ddf_dirty_mask |= DM_DIRTY_VDR;
	 	if (ai->vendor_type == AI_VENDOR_ADPT) {
 			struct adpt_vendor_data *adpt;

			adpt = (struct adpt_vendor_data *)ai->vendor->data;
 			adpt->Dirty = 1;
			ai->ddf_dirty_mask |= AI_DIRTY_VENDOR;
		}
	}
	spin_unlock_irqrestore(&member->member_lock, mflags);
	spin_unlock_irqrestore(&ai->domain->ddf_lock, dmflags);
	if (sched_write)
		ddf_schedule_sync(&ai->sync_data, /*wait*/1, /*enable*/1);
	return (0);
}

/*
 * Handle notification of our array transitioning to Read-Only
 * mode or, in MD parlance, "entering safemode".
 */
static int
ddf_enter_safemode(mddev_t *mddev)
{
	struct md_ddf_array_info *ai;
	mdk_member_t *member;
	u_long dmflags;
	u_long mflags;
	int sched_write;

	/*
	 * Update the sequence number and mark the
	 * array meta-data as closed.  We increment
	 * the sequence number so that we are not
	 * confused by any members that for some
	 * reason are not written before we lose
	 * power or are incorrectly shutdown.
	 */
	MD_DDF_DEBUG(("DDF Enter Safemode\n"));
	member = to_member(mddev);
	sched_write = 0;
	ai = member->metadata_private;
	spin_lock_irqsave(&ai->domain->ddf_lock, dmflags);
	spin_lock_irqsave(&member->member_lock, mflags);
	if ((ai->vde->VD_State & DDF_VDE_DIRTY) != 0) {
		ai->vde->VD_State &= ~DDF_VDE_DIRTY;
		sched_write = 1;
		ai->domain->ddf_dirty_mask |= DM_DIRTY_VDR;
		MD_DDF_DEBUG(("Clearing DIRTY bit\n"));

	 	if (ai->vendor_type == AI_VENDOR_ADPT) {
			struct adpt_vendor_data *adpt;

			adpt = (struct adpt_vendor_data *)ai->vendor->data;
			adpt->Dirty = 0;
			ai->ddf_dirty_mask |= AI_DIRTY_VENDOR;
		}
	}
	spin_unlock_irqrestore(&member->member_lock, mflags);
	spin_unlock_irqrestore(&ai->domain->ddf_lock, dmflags);
	if (sched_write)
		ddf_schedule_sync(&ai->sync_data, /*wait*/1, /*enable*/1);
	return (0);
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
static int __init ddf_init(void)
#else
int __init ddf_init(void)
#endif
{
	int error;

	error = md_init_workitem_support(NULL, &md_ddf_syncer, "ddf_syncer");
	if (error)
		return (error);
	error = register_md_metapersonality(&md_ddf_pers);
	if (error)
		md_teardown_workitem_support(&md_ddf_syncer);
	return (error);
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
static void ddf_exit(void)
#else
void ddf_exit(void)
#endif
{
	struct md_ddf_domain *dm;
	struct md_ddf_disk_info *di;
	struct md_ddf_array_info *ai;

	unregister_md_metapersonality(&md_ddf_pers);

	while (!list_empty(&ddf_domains)) {
		dm = list_entry(ddf_domains.next, struct md_ddf_domain,
				ddf_domain_links);
		ddf_dm_get(dm);
		while (!list_empty(&dm->diskqueue)) {
			di = list_entry(dm->diskqueue.next,
					struct md_ddf_disk_info,
					ddf_disk_links);
			ddf_free_disk_info(di);
		}
		while (!list_empty(&dm->arrayqueue)) {
			ai = list_entry(dm->arrayqueue.next,
					struct md_ddf_array_info,
					ddf_array_links);
			ddf_free_array_info(ai);
		}
		ddf_free_domain(dm);
	}
	md_teardown_workitem_support(&md_ddf_syncer);
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
module_init(ddf_init);
module_exit(ddf_exit);
#endif
MODULE_LICENSE("Dual BSD/GPL");
