/*
   raid0.c : Multiple Devices driver for Linux
             Copyright (C) 1994-96 Marc ZYNGIER
	     <zyngier@ufr-info-p7.ibp.fr> or
	     <maz@gloups.fdn.fr>
             Copyright (C) 1999, 2000 Ingo Molnar, Red Hat


   RAID-0 management functions.


   Added pluggable metadata support.
   Copyright (C) 2004 Adaptec Inc.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.
   
   You should have received a copy of the GNU General Public License
   (for example /usr/src/linux/COPYING); if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  
*/

#include <linux/module.h>
#include <linux/raid/emd/raid0.h>

/* size in KB */
#define MAX_REQUEST_SIZE_KB (4096)

static int raid0_make_request(mdk_member_t *, int, struct buffer_head *);
static int raid0_raidop(mdk_member_t *, int, void *);
static mdk_personality_t raid0_personality=
{
	.name		= "raid0",
	.owner		= THIS_MODULE,
	.make_request	= raid0_make_request,
	.raid_op	= raid0_raidop,
};


/*
 * Number of guaranteed r0bios in case of extreme VM load:
 */

#define md_min(a,b)	(((a) < (b)) ? (a) : (b))

static void * r0_tbuf_pool_alloc(int gfp_flags, void *data)
{
	struct raid0_transaction *r0_tbuf;

	r0_tbuf = kmalloc(sizeof(struct raid0_transaction), gfp_flags); 

	if (r0_tbuf != NULL) 
		memset(r0_tbuf, 0, sizeof(*r0_tbuf));

	return r0_tbuf;
}

static void  r0_tbuf_pool_free(void *arg, void *data)
{
	struct raid0_transaction *r0_tbuf = (struct raid0_transaction *)arg;

	kfree(r0_tbuf);

	return;
}

static void * r0_bh_pool_alloc(int gfp_flags, void *data)
{
	struct buffer_head *r0_bh;

	r0_bh = kmalloc(sizeof(struct buffer_head), gfp_flags);

	if (r0_bh != NULL) 
		memset(r0_bh, 0, sizeof(*r0_bh));
	
	return r0_bh;
}

static void r0_bh_pool_free(void *bh, void *data)
{
	struct buffer_head *tbh = (struct buffer_head *)bh;

	kfree(tbh);

	return;
}

static struct buffer_head *
raid0_alloc_bh(raid0_conf_t *conf, int cnt)
{
	struct buffer_head *bh;

	bh = NULL;
	while (cnt > 0) {
		struct buffer_head *t;

		t = mempool_alloc(conf->r0_bh_pool, GFP_NOIO);
		if (t != NULL) {
			t->b_next = bh;
			bh = t;
			cnt--;
		}
	}

	return (bh);
}

static void
raid0_free_bh(raid0_conf_t *conf, struct buffer_head *bh)
{
	while (bh != NULL) {
		struct buffer_head *t;

		t = bh;
		bh = bh->b_next;
		if (t->b_pprev != NULL)
			MD_BUG();
		mempool_free(t, conf->r0_bh_pool);
	}
}

static int create_strip_zones (mddev_t *mddev)
{
	int i, c, j;
	sector_t current_offset, curr_zone_offset;
	sector_t min_spacing;
	raid0_conf_t *conf = mddev_to_conf(mddev);
	mdk_member_t *smallest, *rdev1, *rdev2, *rdev;
	struct list_head *tmp1, *tmp2;
	struct strip_zone *zone;
	int cnt;
	u_long flags;
 
	/*
	 * The number of 'same size groups'
	 */
	conf->nr_strip_zones = 0;

	ITERATE_CHILDREN(to_member(mddev), rdev1, tmp1, flags) {
		printk("raid0: looking at %s\n", mdname(rdev1));
		c = 0;
		ITERATE_CHILDREN(to_member(mddev), rdev2, tmp2, flags) {
			printk("raid0:   comparing %s(%llu)",
			       mdname(rdev1),
			       (unsigned long long)rdev1->data_size);
			printk(" with %s(%llu)\n",
			       mdname(rdev2),
			       (unsigned long long)rdev2->data_size);
			if (rdev2 == rdev1) {
				printk("raid0:   END\n");
				member_put(rdev2);
				break;
			}
			if (rdev2->data_size == rdev1->data_size)
			{
				/*
				 * Not unique, don't count it as a new
				 * group
				 */
				printk("raid0:   EQUAL\n");
				c = 1;
				member_put(rdev2);
				break;
			}
			printk("raid0:   NOT EQUAL\n");
		}
	
		if (!c) {
			printk("raid0:   ==> UNIQUE\n");
			conf->nr_strip_zones++;
			printk("raid0: %d zones\n", conf->nr_strip_zones);
		}
	}

	printk("raid0: FINAL %d zones\n", conf->nr_strip_zones);

	conf->strip_zone = kmalloc(sizeof(struct strip_zone)*
				conf->nr_strip_zones, GFP_KERNEL);
	if (!conf->strip_zone)
		return -ENOMEM;
	conf->devlist = kmalloc(sizeof(mdk_rdev_t*)*
				conf->nr_strip_zones*mddev->max_disks,
				GFP_KERNEL);
	if (!conf->devlist) {
		kfree(conf->strip_zone);
		conf->strip_zone = NULL;
		return -ENOMEM;
	}

	memset(conf->strip_zone, 0,sizeof(struct strip_zone)*
				   conf->nr_strip_zones);
	memset(conf->devlist, 0,
	       sizeof(mdk_rdev_t*) * conf->nr_strip_zones * mddev->max_disks);

	/* The first zone must contain all devices, so here we check that
	 * there is a proper alignment of slots to devices and find them all
	 */
	zone = &conf->strip_zone[0];
	cnt = 0;
	smallest = NULL;
	zone->dev = conf->devlist;
	ITERATE_CHILDREN(to_member(mddev), rdev1, tmp1, flags) {
		int j = rdev1->priority;

		if (j < 0 || j >= mddev->max_disks) {
			printk("raid0: bad disk number %d - aborting!\n", j);
			member_put(rdev1);
			goto abort;
		}
		if (zone->dev[j]) {
			printk("raid0: multiple devices for %d - aborting!\n",
				j);
			member_put(rdev1);
			goto abort;
		}
		zone->dev[j] = rdev1;
#if 0
		q = md_member_get_queue(rdev1);
		blk_queue_stack_limits(mddev->queue, q);
#endif
		/* as we don't honour merge_bvec_fn, we must never risk
		 * violating it, so limit ->max_sector to one PAGE, as
		 * a one page request is never in violation.
		 */

		if (!smallest || (rdev1->data_size <smallest->data_size))
			smallest = rdev1;
		cnt++;
	}
	if (cnt != mddev->max_disks) {
		printk("raid0: too few disks (%d of %d) - aborting!\n",
			cnt, mddev->max_disks);
		goto abort;
	}
	zone->nb_dev = cnt;
	zone->size = smallest->data_size * cnt;
	zone->zone_offset = 0;

	current_offset = smallest->data_size;
	curr_zone_offset = zone->size;

	/* now do the other zones */
	for (i = 1; i < conf->nr_strip_zones; i++)
	{
		zone = conf->strip_zone + i;
		zone->dev = conf->strip_zone[i-1].dev + mddev->max_disks;

		printk("raid0: zone %d\n", i);
		zone->dev_offset = current_offset;
		smallest = NULL;
		c = 0;

		for (j=0; j<cnt; j++) {
			rdev = conf->strip_zone[0].dev[j];
			printk("raid0: checking %s ...", mdname(rdev));
			if (rdev->data_size > current_offset)
			{
				printk(" contained as device %d\n", c);
				zone->dev[c] = rdev;
				c++;
				if (!smallest || (rdev->data_size <smallest->data_size)) {
					smallest = rdev;
					printk("  (%llu) is smallest!.\n", 
						(unsigned long long)rdev->data_size);
				}
			} else
				printk(" nope.\n");
		}

		zone->nb_dev = c;
		zone->size = (smallest->data_size - current_offset) * c;
		printk("raid0: zone->nb_dev: %d, size: %llu\n",
			zone->nb_dev, (unsigned long long)zone->size);
		
		zone->zone_offset = curr_zone_offset;
		curr_zone_offset += zone->size;

		current_offset = smallest->data_size;
		printk("raid0: current zone offset: %llu\n",
			(unsigned long long)current_offset);
	}

	/* Now find appropriate hash spacing.
	 * We want a number which causes most hash entries to cover
	 * at most two strips, but the hash table must be at most
	 * 1 PAGE.  We choose the smallest strip, or contiguous collection
	 * of strips, that has big enough size.  We never consider the last
	 * strip though as it's size has no bearing on the efficacy of the hash
	 * table.
	 */
	conf->hash_spacing = curr_zone_offset;
	min_spacing = curr_zone_offset;
	sector_div(min_spacing, PAGE_SIZE/sizeof(struct strip_zone*));
	for (i=0; i < conf->nr_strip_zones-1; i++) {
		sector_t sz = 0;
		for (j=i; j<conf->nr_strip_zones-1 &&
			     sz < min_spacing ; j++)
			sz += conf->strip_zone[j].size;
		if (sz >= min_spacing && sz < conf->hash_spacing)
			conf->hash_spacing = sz;
	}

	printk("raid0: done.\n");
	return 0;
 abort:
	if (conf->strip_zone != NULL)
		kfree(conf->strip_zone);
	conf->strip_zone = NULL;
	if (conf->devlist != NULL)
		kfree(conf->devlist);
	conf->devlist = NULL;
	return -EINVAL;
}

static int raid0_array_init (mddev_t *mddev)
{
	raid0_conf_t *conf;
	int min_bhs = 0, min_tbufs = 0, r0_req_size = 0;

#if 0
	printk("%s: setting max_sectors to %d, segment boundary to %d\n",
	       mdname(to_member(mddev)),
	       mddev->chunk_size >> 9,
	       (mddev->chunk_size>>1)-1);
	blk_queue_max_sectors(mddev->queue, mddev->chunk_size >> 9);
	blk_queue_segment_boundary(mddev->queue, (mddev->chunk_size>>1) - 1);
#endif

	conf = kmalloc(sizeof (raid0_conf_t), GFP_KERNEL);
	if (!conf)
		return (-ENOMEM);
	memset(conf, 0, sizeof(*conf));
	mddev->member.pers_private = (void *)conf;
	mddev->member.pers = md_pers_get(&raid0_personality);
	conf->mddev = mddev;

	r0_req_size = (MAX_REQUEST_SIZE_KB / mddev->chunk_size); 
	r0_req_size = (r0_req_size<=0)?1:r0_req_size;
	
	//we don't know the number of stripe zones yet, use max_disks instead
	min_tbufs = r0_req_size * mddev->max_disks + 1;

	conf->r0_tbuf_pool = mempool_create(min_tbufs, r0_tbuf_pool_alloc,
					  r0_tbuf_pool_free, mddev);
	if (!conf->r0_tbuf_pool) {
		printk(KERN_ERR "raid1: couldn't allocate tbuf pool memory\n");
		kfree(conf);
		mddev->member.pers_private = NULL;
		md_pers_put(mddev->member.pers);
		mddev->member.pers = NULL;
		return -ENOMEM;
	}
	
	min_bhs = min_tbufs;
	conf->r0_bh_pool = mempool_create(min_bhs, r0_bh_pool_alloc,
					  r0_bh_pool_free, mddev);
	if (!conf->r0_bh_pool) {
		printk(KERN_ERR "raid1: couldn't allocate bh pool memory\n");
		mempool_destroy(conf->r0_tbuf_pool);
		kfree(conf);
		mddev->member.pers_private = NULL;
		md_pers_put(mddev->member.pers);
		mddev->member.pers = NULL;
		return -ENOMEM;
	}
	return (0);
}
 
static int raid0_array_release(mddev_t *mddev)
{
	raid0_conf_t *conf = mddev_to_conf(mddev);

	if (conf == NULL)
		return 0;

	if (conf->hash_table)
		kfree (conf->hash_table);
	conf->hash_table = NULL;

	if (conf->strip_zone)
		kfree (conf->strip_zone);
	conf->strip_zone = NULL;

	if (conf->devlist)
		kfree (conf->devlist);
	conf->devlist = NULL;

	kfree (conf);
	mddev->member.pers_private = NULL;
	if (mddev->member.pers != NULL)
		md_pers_put(mddev->member.pers);

	return 0;
}
 
static int raid0_run (mddev_t *mddev)
{
	raid0_conf_t *conf;
	mdk_member_t *member;
	int64_t size;
	u_int cur, i, nb_zone;
	u_long flags;
	int err;

	conf = mddev->member.pers_private;
	member = to_member(mddev);
	spin_lock_irqsave(&member->member_lock, flags);
	if ((mddev->flags & MD_MDDEV_ALLOWED_TO_RUN) == 0) {
		spin_unlock_irqrestore(&member->member_lock, flags);
		return -EPERM;
	}
	if ((mddev->flags
	  & (MD_MDDEV_RUNNING|MD_MDDEV_ATTEMPTING_TO_RUN)) != 0) {
		spin_unlock_irqrestore(&member->member_lock, flags);
		return 0;
	}
	if ((mddev->flags & MD_MDDEV_UNIT_NUMBER_ASSIGNED) == 0) {
		spin_unlock_irqrestore(&member->member_lock, flags);
		md_assign_unit_number(mddev);
		spin_lock_irqsave(&member->member_lock, flags);

		if ((mddev->flags & MD_MDDEV_UNIT_NUMBER_ASSIGNED) == 0) {
			spin_unlock_irqrestore(&member->member_lock, flags);
			return -EPERM;
		}
	}
	if (!MDU_MSTATE_READABLE(member->state)) {
		spin_unlock_irqrestore(&member->member_lock, flags);
		return -ENXIO;
	}
	mddev->flags |= MD_MDDEV_ATTEMPTING_TO_RUN;
	spin_unlock_irqrestore(&member->member_lock, flags);
	conf->strip_zone = NULL;
	conf->devlist = NULL;
	if ((err = create_strip_zones (mddev))) 
		goto out_free_conf;

	if ((err = md_check_chunksize(mddev)))
		goto out_free_conf;

	printk("raid0 : md_size is %llu blocks.\n", 
		(unsigned long long)mddev->member.data_size);
	printk("raid0 : conf->hash_spacing is %llu blocks.\n",
		(unsigned long long)conf->hash_spacing);
	{
#if __GNUC__ < 3
		volatile
#endif
		sector_t s = mddev->member.data_size;
		sector_t space = conf->hash_spacing;
		int round;
		conf->preshift = 0;
		if (sizeof(sector_t) > sizeof(unsigned long)) {
			/*shift down space and s so that sector_div will work */
			while (space > (sector_t) (~(unsigned long)0)) {
				s >>= 1;
				space >>= 1;
				s += 1; /* force round-up */
				conf->preshift++;
			}
		}
		round = sector_div(s, (unsigned long)space) ? 1 : 0;
		nb_zone = s + round;
	}
	printk("raid0 : nb_zone is %d.\n", nb_zone);

	printk("raid0 : Allocating %Zd bytes for hash.\n",
				nb_zone*sizeof(struct strip_zone*));
	conf->hash_table = kmalloc (sizeof (struct strip_zone *)*nb_zone, GFP_KERNEL);
	if (conf->hash_table == NULL) {
		err = -ENOMEM;
		goto out_free_conf;
	}
	size = conf->strip_zone[0].size;

	for (i = 0, cur = 0; i < nb_zone; i++) {
		conf->hash_table[i] = conf->strip_zone + cur;
		while (size <= conf->hash_spacing) {
			cur++;
			size += conf->strip_zone[cur].size;
		}
		size -= conf->hash_spacing;
	}
	if (conf->preshift) {
		conf->hash_spacing >>= conf->preshift;
		/* round hash_spacing up so when we divide by it, we
		 * err on the side of too-low, which is safest
		 */
		conf->hash_spacing++;
	}

	/* calculate the max read-ahead size.
	 * For read-ahead of large files to be effective, we need to
	 * readahead at least a whole stripe. i.e. number of devices
	 * multiplied by chunk size.
	 * If an individual device has an ra_pages greater than the
	 * chunk size, then we will not drive that device as hard as it
	 * wants.  We consider this a configuration error: a larger
	 * chunksize should be used in that case.
	 */
#if 0
	{
		int stripe = mddev->max_disks * mddev->chunk_size / PAGE_CACHE_SIZE;
		int mddev_minor = 0; /* always zerop */
		if (max_readahead[mddev->gendisk->major][0] < stripe)
			max_readahead[mddev->gendisk->major][0] = stripe;
	}
#endif
	spin_lock_irqsave(&member->member_lock, flags);
	mddev->flags |= MD_MDDEV_RUNNING;
	mddev->flags &= ~MD_MDDEV_ATTEMPTING_TO_RUN;
	spin_unlock_irqrestore(&member->member_lock, flags);
	md_export_array(mddev);
	return 0;

out_free_conf:
	if (conf->strip_zone)
		kfree(conf->strip_zone);
	if (conf->devlist)
		kfree (conf->devlist);
	spin_lock_irqsave(&member->member_lock, flags);
	mddev->flags &= ~MD_MDDEV_ATTEMPTING_TO_RUN;
	spin_unlock_irqrestore(&member->member_lock, flags);
	return err;
}

static int raid0_stop (mddev_t *mddev)
{
	return 0;
}

void
raid0_end_request(struct buffer_head *bh, int uptodate)
{
	struct raid0_transaction *tbuf, *tbuf2;

	tbuf2 = (struct raid0_transaction *)bh->b_private;

	tbuf = tbuf2->master_tbuf;

	/*
	 * Unlike RAID-1 and RAID-5, which initiate reconstruction when
	 * they get an error from a disk, we just pass the error back up
	 * the stack to the caller.
	 */
	if (uptodate == 0)
		atomic_inc(&tbuf->error_count);

	if (atomic_dec_and_test(&tbuf->remaining)) {
		struct raid0_transaction *tbuf3;
		raid0_conf_t *conf;
#ifdef RAID0_DEBUG
		unsigned long sector;
#endif /* RAID0_DEBUG */
		int nuptodate;

		nuptodate = 1;

		conf = tbuf->conf;

		if (atomic_read(&tbuf->error_count) > 0) {
			nuptodate = 0;
			printk("%s: %s of raid0 array sector %ld size %d "
			       "failed\n", mdname(to_member(conf->mddev)),
			       (tbuf->optype == RAID0_OP_READ) ? "read" :
			       "write", tbuf->bh->b_rsector, tbuf->bh->b_size);
		}

						
		tbuf->bh->b_end_io(tbuf->bh, nuptodate);
#ifdef RAID0_DEBUG
		sector = tbuf->bh->b_rsector;
		printk("%s: %s at array sector %ld size %d done\n",
		       mdname(to_member(conf->mddev)),
		       (tbuf->optype == RAID0_OP_READ) ? "read" : "write",
		       tbuf->bh->b_rsector, tbuf->bh->b_size);
#endif /* RAID0_DEBUG */
		for (tbuf2 = tbuf->next_tbuf; tbuf2 != NULL; tbuf2 = tbuf3) {
			tbuf3 = tbuf2->next_tbuf;
			tbuf2->bh->b_next = NULL;
			raid0_free_bh(conf, tbuf2->bh);
			mempool_free(tbuf2, conf->r0_tbuf_pool);
		}
		mempool_free(tbuf, conf->r0_tbuf_pool);
#ifdef RAID0_DEBUG
		printk("%s: done freeing tbufs for sector %ld\n",
		       mdname(to_member(conf->mddev)), sector);
#endif /* RAID0_DEBUG */
	}
}

static int raid0_make_request (mdk_member_t *member, int rw, 
				struct buffer_head *bh)	
{
	mddev_t *mddev;
	raid0_conf_t *conf;
	struct strip_zone *zone;
	mdk_member_t *tmp_dev, **dev_list;
	unsigned long *starting_sector, *sector_count, *trans_offset;
	struct buffer_head *new_bh, *bh_list;
	struct raid0_transaction *tbuf, *tbuf2 = NULL, *tbuf3 = NULL;
	sector_t rsect, x;
	unsigned long chunk;
	unsigned int sect_in_chunk, chunksize_bits, chunk_sects;
	int num_pieces, max_pieces, i=0, sectors_assigned, num_buffers_queued;


	if (member == NULL)
		return 0;

	member_get(member);

	mddev  = to_mddev(member);
	conf = mddev_to_conf(mddev);

	tbuf = mempool_alloc(conf->r0_tbuf_pool, GFP_NOIO);

	chunk_sects = mddev->chunk_size >> 9;
	chunksize_bits = ffz(~chunk_sects);

	max_pieces = (bh->b_size / mddev->chunk_size) + 2 +
		conf->nr_strip_zones;

	starting_sector = (unsigned long *)kmalloc(sizeof(unsigned long) *
						   max_pieces, GFP_NOIO);
	sector_count = (unsigned long *)kmalloc(sizeof(unsigned long) *
						max_pieces, GFP_NOIO);
	dev_list = (mdk_member_t **)kmalloc(sizeof(mdk_member_t *) * max_pieces,
					 GFP_NOIO);
	trans_offset = (unsigned long *)kmalloc(sizeof(unsigned long) *
						max_pieces, GFP_NOIO);
	
	if ((starting_sector == NULL)
	 || (sector_count == NULL)
	 || (dev_list == NULL)
	 || (trans_offset == NULL))
		goto outerr;

	for (sectors_assigned = 0, num_pieces = 0;
	     sectors_assigned < (bh->b_size >> 9); num_pieces++) {
		unsigned long cur_sector, cur_count;

		cur_sector = bh->b_rsector + sectors_assigned;

		x = cur_sector >> conf->preshift;
		sector_div(x, (unsigned long)conf->hash_spacing);
		zone = conf->hash_table[x];
		
		while (cur_sector >= (zone->zone_offset + zone->size)) 
			zone++;

		/*
		 * Determine the size of the current chunk of data.
		 * There are three factors that can signal the end of a
		 * piece of data:
		 * - we're at the end of our buffer
		 * - we're at the end of a zone
		 * - we're at the end of a chunk on a disk
		 * 
		 * The idea is to segregate the buffers by the disks they
		 * correspond to.
		 */
		sect_in_chunk = cur_sector & (chunk_sects -1);
		cur_count = md_min((bh->b_size >> 9) - sectors_assigned,
				   (zone->zone_offset + zone->size) - cur_sector);
		cur_count = md_min(cur_count, chunk_sects - sect_in_chunk);

		sector_count[num_pieces] = cur_count;
		trans_offset[num_pieces] = sectors_assigned;
		sectors_assigned += cur_count;

		chunk = (cur_sector - zone->zone_offset) /
			(zone->nb_dev << chunksize_bits);
		tmp_dev = zone->dev[(cur_sector >> chunksize_bits) % zone->nb_dev];
		rsect = ((chunk << chunksize_bits) + zone->dev_offset) +
			sect_in_chunk;
		starting_sector[num_pieces] = rsect;
		dev_list[num_pieces] = tmp_dev;
	}
	

	bh_list = raid0_alloc_bh(conf, num_pieces);

	tbuf->bh = bh;
	tbuf->conf = conf;
	atomic_set(&tbuf->remaining, num_pieces);
	atomic_set(&tbuf->error_count, 0);
	if ((rw == READ) || (rw == READA))
		tbuf->optype = RAID0_OP_READ;
	else
		tbuf->optype = RAID0_OP_WRITE;

	tbuf->next_tbuf = NULL;
	tbuf->master_tbuf = NULL;

	new_bh = bh_list;

	for (i = 0; i < num_pieces; i++) {

		tbuf2 =  mempool_alloc(conf->r0_tbuf_pool, GFP_NOIO);

		tbuf2->master_tbuf = tbuf;
		tbuf2->next_tbuf = tbuf->next_tbuf;
		tbuf->next_tbuf = tbuf2;
		tbuf2->bh = new_bh;
		tbuf2->optype = tbuf->optype;
		tbuf2->conf = conf;
		atomic_set(&tbuf2->remaining, 0);
		atomic_set(&tbuf2->error_count, 0);
		tbuf2->transaction_offset = trans_offset[i];

		new_bh->b_size = sector_count[i] << 9;
		new_bh->b_list = BUF_LOCKED;
		new_bh->b_state = (1 << BH_Req) | (1 << BH_Mapped) |
				  (1 << BH_Lock);
		atomic_set(&new_bh->b_count, 1);

		new_bh->b_page = bh->b_page;
		new_bh->b_data = (trans_offset[i] << 9) + bh->b_data;
		if ((new_bh->b_data < bh->b_data)
		 || (new_bh->b_data + new_bh->b_size >
		     (bh->b_data + 4096))) {
			printk("pointer values out of range\n");
			BUG();
		}

		new_bh->b_blocknr = starting_sector[i];
		new_bh->b_rsector = starting_sector[i];
		new_bh->b_end_io = raid0_end_request;
		new_bh->b_private = tbuf2;
		init_waitqueue_head(&new_bh->b_wait);

		new_bh = new_bh->b_next;
	}

	kfree(starting_sector);
	kfree(sector_count);
	kfree(trans_offset);

	/* This shouldn't happen */
	if (i != num_pieces)
		BUG();

	/* Fire off the requests */
	for (tbuf2 = tbuf->next_tbuf, num_buffers_queued = num_pieces-1; tbuf2 != NULL;
	     tbuf2 = tbuf3, num_buffers_queued--) {
		tbuf3 = tbuf2->next_tbuf;
		tbuf2->bh->b_next = NULL;
		tbuf2->bh->b_reqnext = NULL;
#ifdef RAID0_DEBUG
		printk("raid0: queueing %d byte %s to %s at %ld\n",
		       tbuf2->bh->b_size,
		       (tbuf2->optype == RAID0_OP_READ) ? "read" : "write",
		       partition_name(tbuf2->bh->b_dev),
		       tbuf2->bh->b_rsector);
		printk("raid0: memory address %p\n", tbuf2->bh->b_data);
		printk("raid0: original block address %ld\n", bh->b_rsector);
		printk("raid0: original memory address %p\n", bh->b_data);
#endif /* RAID0_DEBUG */
		if (tbuf2->bh->b_data == NULL)
			BUG();
		md_make_request(dev_list[num_buffers_queued], rw, tbuf2->bh);
	}

	kfree(dev_list);

	if (num_buffers_queued != -1) {
		printk("raid0_make_request: num_buffers_queued %d != "
		       "num_pieces %d\n", num_buffers_queued, num_pieces);
	}
#ifdef RAID0_DEBUG
	 else {
		printk("raid0_make_request: queued %d buffers\n", num_pieces);
	}
#endif /* RAID0_DEBUG */

	member_put(member);
	return (0);

outerr:
	if (starting_sector != NULL)
		kfree(starting_sector);
	if (sector_count != NULL)
		kfree(sector_count);
	if (dev_list != NULL)
		kfree(dev_list);
	if (trans_offset != NULL)
		kfree(trans_offset);

	buffer_IO_error(bh);
	member_put(member);
	
	return 0;
}

static int raid0_state_change(mdk_member_t *member)
{
	switch (member->type) {
	case MDK_MTYPE_MDDEV:
		switch (member->state) {
		case MDU_MSTATE_OPTIMAL:
			raid0_run(to_mddev(member));
			break;
		case MDU_MSTATE_DEGRADED:
		case MDU_MSTATE_FAILED:
		default:
			return (raid0_stop(to_mddev(member)));
		}
	case MDK_MTYPE_RDEV:
	default:
		break;
	}

	return (0);
}

static int raid0_raidop(mdk_member_t *member, int op, void *arg)
{
	switch (op) {
	case MDK_RAID_OP_MSTATE_CHANGED:
		return (raid0_state_change(member));
	case MDK_RAID_OP_INIT:
	case MDK_RAID_OP_RELEASE:
		if (member->type != MDK_MTYPE_MDDEV)
			return (0);
		if (op == MDK_RAID_OP_INIT)
			return (raid0_array_init(to_mddev(member)));
		else
			return (raid0_array_release(to_mddev(member)));
	case MDK_RAID_OP_ADD_MEMBER:
		/*
		 * No action here.  RAID-0 cannot support late arrival of
		 * children on existing arrays, and the strip-zone code
		 * must have all members present when it initializes.
		 */
		break;
	case MDK_RAID_OP_RUN:
		return (raid0_run(to_mddev(member)));
	case MDK_RAID_OP_PAUSE_MEMBER:
	case MDK_RAID_OP_REMOVE_MEMBER:
	case MDK_RAID_OP_REQUEST_VERIFY:
	case MDK_RAID_OP_CANCEL_VERIFY:
		return -EOPNOTSUPP;
	default:
		return (-EINVAL);
	}
	return (0);
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
static int __init raid0_init (void)
#else
int __init raid0_init (void)
#endif
{
	return register_md_personality (RAID0, &raid0_personality);
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
static void raid0_exit (void)
#else
void raid0_exit (void)
#endif
{
	unregister_md_personality (RAID0);
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
module_init(raid0_init);
module_exit(raid0_exit);
#endif
MODULE_LICENSE("GPL");
