/*
 * raid1.c : Multiple Devices driver for Linux
 *
 * Copyright (C) 1999, 2000, 2001 Ingo Molnar, Red Hat
 *
 * Copyright (C) 1996, 1997, 1998 Ingo Molnar, Miguel de Icaza, Gadi Oxman
 *
 * RAID-1 management functions.
 *
 * Better read-balancing code written by Mika Kuoppala <miku@iki.fi>, 2000
 *
 * Fixes to reconstruction by Jakob )Bstergaard" <jakob@ostenfeld.dk>
 * Various fixes by Neil Brown <neilb@cse.unsw.edu.au>
 *
 * Added pluggable metadata support.
 * Copyright (C) 2004 Adaptec Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * You should have received a copy of the GNU General Public License
 * (for example /usr/src/linux/COPYING); if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/raid/emd/raid1.h>

#define R1_READABLE_MEMBER(conf, member, below_sync_mark)	\
    (MDK_READABLE_MEMBER(member)				\
  || (below_sync_mark && MDK_WRITABLE_MEMBER(member)))

#define R1_WRITABLE_MEMBER(member) MDK_WRITABLE_MEMBER(member)
#define R1_REBUILDABLE(array)					\
  (((array)->state						\
  & ~(MDU_MSTATE_DIRTY|MDU_MSTATE_DEGRADED|MDU_MSTATE_REBUILD_TARGET)) == 0)

#define R1_VERIFIABLE(array)					\
   (((array)->state & MDU_MSTATE_DIRTY) != 0			\
 && ((array)->state & ~(MDU_MSTATE_DIRTY|MDU_MSTATE_OPTIMAL)) == 0)

#define R1_INITIALIZABLE(array)					\
   (((array)->state & MDU_MSTATE_INITIALIZING)			\
 && ((array)->state & (MDU_MSTATE_DEGRADED|MDU_MSTATE_OPTIMAL)))

/*
 * Number of guaranteed r1bios in case of extreme VM load:
 */
#define	NR_RAID1_BIOS 64

static int raid1_workop(mdk_member_t *member, int op, void *arg);
static mdk_personality_t raid1_personality;
static spinlock_t retry_list_lock = SPIN_LOCK_UNLOCKED;
static LIST_HEAD(retry_list_head);

typedef enum {
	R1_CHECK_STATE_CHANGE,
	R1_CHECK_WO_OPEN,
	R1_CHECK_WO_FREE,
	R1_CHECK_SPARE_ACTIVATE
} r1_check_context_t;
static int md_raid1_check_recovery(mddev_t *mddev,
				   r1_check_context_t check_context);

static void * r1_bh_pool_alloc(int gfp_flags, void *data)
{
	struct buffer_head *bh;

	bh = kmalloc(sizeof(*bh), gfp_flags);
	if (bh != NULL)
		memset(bh, 0, sizeof(*bh));
	return bh;
}

static void r1_bh_pool_free(void *bh, void *data)
{
	kfree(bh);
}

static void * r1bio_pool_alloc(int gfp_flags, void *data)
{
	mddev_t *mddev = data;
	conf_t	*conf = mddev_to_conf(mddev);
	r1bio_t *r1_bio;
	size_t	alloc_size;
	int	num_bios;
	int	i;

	/*
	 * allocate a r1bio with room to support this
	 * array if all members were optimal.  Should
	 * this module every support adding new mirror
	 * members beyond the initial configuration,
	 * some method for "updating" pre-allocated
	 * r1bios in the mempool to include room for
	 * the new member must be added.
	 */
	num_bios = mddev->max_disks;
	alloc_size = sizeof(r1bio_t) + sizeof(r1bio_holder_t) * num_bios;
	r1_bio = kmalloc(alloc_size, gfp_flags);
	if (r1_bio) {
		memset(r1_bio, 0, alloc_size);
		r1_bio->mddev = mddev;

		r1_bio->num_bios = num_bios;
		for (i = 0; i < num_bios; i++) {
			struct buffer_head *bh;

			r1_bio->bios[i].r1bio = r1_bio;
			bh = mempool_alloc(conf->r1_bh_pool, gfp_flags);
			if (bh != NULL) {
				r1_bio->bios[i].bh = bh;
				continue;
			}
			/*
			 * Teardown previously allocated objects.
			 * and return failure.
			 */
			while (--i > 0) {
				bh = r1_bio->bios[i].bh;
				mempool_free(bh, conf->r1_bh_pool);
			}
			kfree(r1_bio);
			r1_bio = NULL;
			break;
		}
	}
	return r1_bio;
}

static void free_all_bhs(conf_t *conf, r1bio_t *r1_bio)
{
	int i;

	for (i = 0; i < r1_bio->num_bios; i++) {
		r1bio_holder_t *bioh = &r1_bio->bios[i];
		if (bioh->bh)
			mempool_free(bioh->bh, conf->r1_bh_pool);
		bioh->bh = NULL;
	}
}

static void r1bio_pool_free(void *r1_bio, void *data)
{
	mddev_t *mddev = (mddev_t *)data;
	conf_t *conf = mddev_to_conf(mddev);

	free_all_bhs(conf, r1_bio);
	kfree(r1_bio);
}

#define RESYNC_BLOCK_SIZE PAGE_SIZE
#define RESYNC_SECTORS (RESYNC_BLOCK_SIZE >> 9)
#define RESYNC_WINDOW (512*1024)

static void * r1buf_pool_alloc(int gfp_flags, void *data)
{
	conf_t *conf;
	struct page *page;
	r1bio_t *r1_bio;

	conf = data;
	r1_bio = r1bio_pool_alloc(gfp_flags, conf->mddev);
	if (!r1_bio)
		return NULL;

	/*
	 * MD_SYNCOP_VERIFY works with two buffers at a time; the one to
	 * be verified and one to be verified against.
	 */
	if (conf->syncitem.work_opcode != MD_SYNCOP_INIT) {
		if ((conf->syncitem.work_opcode & MD_SYNCOP_VERIFY) != 0) {
			r1_bio->verify_page = alloc_page(gfp_flags);
			if (r1_bio->verify_page == NULL) {
				r1bio_pool_free(r1_bio, conf->mddev);
				return (NULL);
			}
		}

		/*
		 * We don't chain buffers yet, so we only need
		 * a single page to back the first buffer.  This
		 * will be our "read" buffer for a resync or verify.
		 */
		page = alloc_page(gfp_flags);
		if (unlikely(!page)) {
			if (r1_bio->verify_page != NULL)
				__free_page(r1_bio->verify_page);
			r1bio_pool_free(r1_bio, conf->mddev);
			return (NULL);
		}
		r1_bio->bios[0].bh->b_page = page;
		r1_bio->bios[0].bh->b_data = page_address(page);
	}

	r1_bio->master_bh = r1_bio->bios[0].bh;

	return r1_bio;
}

static void r1buf_pool_free(void *__r1_bio, void *data)
{
	conf_t *conf;
	r1bio_t *r1bio = __r1_bio;

	conf = data;
	
	if (conf->syncitem.work_opcode != MD_SYNCOP_INIT) {
		__free_page(r1bio->bios[0].bh->b_page);
		r1bio->bios[0].bh->b_page = NULL;
		if (r1bio->verify_page != NULL) {
			__free_page(r1bio->verify_page);
			r1bio->verify_page = NULL;
		}
	}
	
	r1bio_pool_free(r1bio, conf->mddev);
}

static inline void put_io_r1bio(r1bio_t *r1_bio, int rw)
{
	conf_t *conf = mddev_to_conf(r1_bio->mddev);

	/*
	 * Wake up any possible resync thread that waits for the device
	 * to go idle.
	 */
	syncable_vol_io_complete(&conf->syncinfo, rw, r1_bio->phase);
	mempool_free(r1_bio, conf->r1bio_pool);
}

static inline void put_sync_r1bio(r1bio_t *r1_bio)
{
	conf_t *conf = mddev_to_conf(r1_bio->mddev);
	u_int phase;

	phase = r1_bio->phase;
	mempool_free(r1_bio, conf->r1buf_pool);
	syncable_vol_syncio_complete(&conf->syncinfo, phase);
}

/*
 * Find a readable member for this array.
 */
static int map(mddev_t *mddev, mdk_member_t **member)
{
	mdk_member_t *array = to_member(mddev);
	mdk_member_t *child;
	struct list_head *tmp;

	/*
	 * Later we do read balancing on the read side
	 * now we use the first available disk.
	 */
	spin_lock_irq(&array->member_lock);
	ITERATE_CHILDREN_LOCKED(array, child, tmp) {

		if (MDK_READABLE_MEMBER(child)) {
			*member = child;
			member_get(child);
			spin_unlock_irq(&array->member_lock);
			return 0;
		}
	}
	spin_unlock_irq(&array->member_lock);

	printk(KERN_ERR "raid1_map(): huh, no more operational devices?\n");
	return -1;
}

static void reschedule_retry(r1bio_t *r1_bio)
{
	unsigned long flags;
	mddev_t *mddev = r1_bio->mddev;

	spin_lock_irqsave(&retry_list_lock, flags);
	list_add(&r1_bio->retry_list, &retry_list_head);
	spin_unlock_irqrestore(&retry_list_lock, flags);

	md_wakeup_thread(mddev->thread);
}

/*
 * raid_end_bio_io() is called when we have finished servicing a mirrored
 * operation and are ready to return a success/failure code to the buffer
 * cache layer.
 */
static void raid_end_bio_io(r1bio_t *r1_bio)
{
	struct buffer_head *bh = r1_bio->master_bh;

	bh->b_end_io(bh, test_bit(R1BIO_Uptodate, &r1_bio->state));
	put_io_r1bio(r1_bio, r1_bio->master_rw);
}

/*
 * Update disk head position estimator based on IRQ completion info.
 */
static inline void update_head_pos(mdk_member_t *member, r1bio_t *r1_bio)
{
	mconf_t *mconf;

	mconf = member_to_mconf(member);
	mconf->head_position = r1_bio->sector + r1_bio->sectors;
}

static void raid1_end_read_request(struct buffer_head *bh, int uptodate)
{
	r1bio_holder_t *bioh;
	r1bio_t *r1_bio;
	mdk_member_t *member;

	bioh = (r1bio_holder_t *)bh->b_private;
	r1_bio = bioh->r1bio;
	member = bioh->member;
	/*
	 * this branch is our 'one mirror IO has finished' event handler:
	 */
	if (!uptodate) {

		/*
		 * oops, read error:
		 */
		md_set_member_state(member, MDU_MSTATE_FAILED);
		printk(KERN_ERR "raid1: %s: rescheduling sector %llu\n",
		       mdname(member), (unsigned long long)r1_bio->sector);
		reschedule_retry(r1_bio);
	} else {
		/*
		 * Set R1BIO_Uptodate in our master bio, so that
		 * we will return a good error code for to the higher
		 * levels even if IO on some other mirrored buffer fails.
		 *
		 * The 'master' represents the composite IO operation to
		 * user-side. So if something waits for IO, then it will
		 * wait for the 'master' bio.
		 */
		set_bit(R1BIO_Uptodate, &r1_bio->state);
		raid_end_bio_io(r1_bio);
	}
	member_put(member);
	return;
}

static void raid1_end_write_request(struct buffer_head *bh, int uptodate)
{
	r1bio_holder_t *r1bioh;
	r1bio_t *r1_bio;
	mdk_member_t *member;

	r1bioh = (r1bio_holder_t *)(bh->b_private);
	r1_bio = r1bioh->r1bio;
	member = r1bioh->member;
	/*
	 * this branch is our 'one mirror IO has finished' event handler:
	 */
	if (!uptodate)
		md_set_member_state(member, MDU_MSTATE_FAILED);
	else {
		/*
		 * Set R1BIO_Uptodate in our master bio, so that
		 * we will return a good error code for to the higher
		 * levels even if IO on some other mirrored buffer fails.
		 *
		 * The 'master' represents the composite IO operation to
		 * user-side. So if something waits for IO, then it will
		 * wait for the 'master' bio.
		 */
		set_bit(R1BIO_Uptodate, &r1_bio->state);
	}

	/*
	 *
	 * Let's see if all mirrored write operations have finished
	 * already.
	 */
	if (atomic_dec_and_test(&r1_bio->remaining))
		raid_end_bio_io(r1_bio);

	member_put(member);
	return;
}

/*
 * This routine returns the disk from which the requested read should
 * be done. There is a per-array 'next expected sequential IO' sector
 * number - if this matches on the next IO then we use the last disk.
 * There is also a per-disk 'last know head position' sector that is
 * maintained from IRQ contexts, both the normal and the resync IO
 * completion handlers update this position correctly. If there is no
 * perfect sequential match then we pick the disk whose head is closest.
 *
 * If there are 2 mirrors in the same 2 devices, performance degrades
 * because position is mirror, not device based.
 *
 * The rdev for the device selected will have nr_pending incremented.
 */
static mdk_member_t *read_balance(mddev_t *mddev, struct buffer_head *bh,
				  r1bio_t *r1_bio, int below_sync_mark)
{
	mdk_member_t *array = to_member(mddev);
	struct list_head *list_item;
	conf_t *conf = mddev_to_conf(mddev);
	mdk_member_t *new_disk;
	mdk_member_t *disk;
	mconf_t *mconf;
	sector_t new_distance, current_distance;

	/*
	 * Traverse the list of members looking for the one who's
	 * head position is closest to this request.  Optimal or
	 * degraded members (meaning the member is another array
	 * that is degraded) can handle reads to any sector.  Rebuild
	 * targets can only handle requests below the last rebuild
	 * checkpoint.
	 */
	if (list_empty(&array->child_set))
		return (NULL);

	new_disk = conf->last_used;
	if (new_disk == NULL) {
		new_disk = list_entry(array->child_set.next,
				      mdk_member_t, peer_links);
	}
		
	/*
	 * If our last disk is not optimal, find some disk that is.
	 */
	if (!R1_READABLE_MEMBER(conf, new_disk, below_sync_mark)) {
		new_disk = NULL;
		ITERATE_CHILDREN_LOCKED(array, disk, list_item) {
			if (R1_READABLE_MEMBER(conf, disk, below_sync_mark)) {
				new_disk = disk;
				break;
			}
		}
	}
	if (new_disk == NULL)
		return NULL;

	/*
	 * Don't change to another disk for sequential reads:
	 */
	mconf = member_to_mconf(new_disk);
	if (mconf->head_position == r1_bio->sector)
		return (new_disk);

	/*
	 * Find the disk whose head is closest.
	 *
	 * Update last_used so we will always terminate the
	 * loop on a member in optimal state even if no other
	 * valid member is found.  new_disk is guaranteed to
	 * by optimal by the loops above.
	 */
	conf->last_used = new_disk;
	disk = new_disk;
	mconf = member_to_mconf(new_disk);
	current_distance = abs(r1_bio->sector - mconf->head_position);

	do {
		struct list_head *next;

		next = disk->peer_links.next;
		if (next == &array->child_set)
			next = array->child_set.next;
		disk = list_entry(next, mdk_member_t, peer_links);
		if (!R1_READABLE_MEMBER(conf, disk, below_sync_mark))
			continue;

		mconf = member_to_mconf(disk);
		new_distance = abs(r1_bio->sector - mconf->head_position);
		if (new_distance < current_distance) {
			current_distance = new_distance;
			new_disk = disk;
		}
	} while (disk != conf->last_used);

	update_head_pos(disk, r1_bio);
	conf->last_used = new_disk;

	return new_disk;
}

static int make_request (mdk_member_t *array, int rw, struct buffer_head *bh)
{
	mddev_t *mddev = to_mddev(array);
	conf_t *conf = mddev_to_conf(mddev);
	mdk_member_t *member;
	struct list_head *list_item;
	r1bio_t *r1_bio;
	u_long flags;
	u_int num_bios;
	u_int i;
	int cur_phase;
	int below_sync_mark;

	/* Block on the sync window is necessary. */            
	cur_phase = syncable_vol_io_start(&conf->syncinfo, rw, bh,  
					  &below_sync_mark);
	/*
	 * make_request() can abort the operation when READA is being
	 * used and no empty request is available.
	 */
	r1_bio = mempool_alloc(conf->r1bio_pool, GFP_NOIO);

	r1_bio->master_rw = rw;
	r1_bio->master_bh = bh;
	r1_bio->sectors = bh->b_size;
	r1_bio->sector = bh->b_rsector;
	r1_bio->phase = cur_phase;

	if (rw == READ) {
		struct buffer_head *read_bh;

		/*
		 * read balancing logic:
		 */
		spin_lock_irqsave(&array->member_lock, flags);
		member = read_balance(mddev, bh, r1_bio, below_sync_mark);
		if (member == NULL) {
/* XXX abort request??? */
			spin_unlock_irqrestore(&array->member_lock, flags);
			return 0;
		}
		member_get(member);
		spin_unlock_irqrestore(&array->member_lock, flags);
		r1_bio->bios[0].member = member;
		r1_bio->bios[0].rw = READ;
		read_bh = r1_bio->bios[0].bh;
		memcpy(read_bh, bh, sizeof(*bh));
		read_bh->b_rsector = bh->b_rsector + member->data_offset;
		read_bh->b_blocknr = bh->b_rsector + member->data_offset;
		read_bh->b_end_io = raid1_end_read_request;
		read_bh->b_private = &r1_bio->bios[0];
		md_make_request(member, rw, read_bh);
		return 0;
	}

	/*
	 * WRITE:
	 *
	 * First select target members under spinlock and
	 * inc their refcount so that they cannot be deleted.
	 */
	num_bios = 0;
	spin_lock_irqsave(&array->member_lock, flags);
	ITERATE_CHILDREN_LOCKED(array, member, list_item) {

		if (R1_WRITABLE_MEMBER(member)) {

			member_get(member);
			r1_bio->bios[num_bios].member = member;
			num_bios++;
		}
	}
	spin_unlock_irqrestore(&array->member_lock, flags);

	atomic_set(&r1_bio->remaining, num_bios);
	for (i = 0; i < num_bios; i++) {
		struct buffer_head *wbh;

		member = r1_bio->bios[i].member;
		wbh = r1_bio->bios[i].bh;
		wbh->b_blocknr	= bh->b_rsector + member->data_offset;
		wbh->b_rsector	= bh->b_rsector + member->data_offset;
		wbh->b_state	= (1<<BH_Req) | (1<<BH_Dirty) |
				  (1<<BH_Mapped) | (1<<BH_Lock);

		atomic_set(&wbh->b_count, 1);
		wbh->b_size       = bh->b_size;
		wbh->b_page       = bh->b_page;
		wbh->b_data       = bh->b_data;
		wbh->b_list       = BUF_LOCKED;
		wbh->b_end_io	  = raid1_end_write_request;
		wbh->b_private    = &r1_bio->bios[i];

		r1_bio->bios[i].rw = WRITE;
		md_make_request(member, rw, wbh);
	}
	
	return 0;
}

struct r1_status_info {
	u_int	member_num;
	char	state_code;
};

static void print_conf(conf_t *conf)
{
#if 0
	int i;
	mirror_info_t *tmp;

	printk("RAID1 conf printout:\n");
	if (!conf) {
		printk("(!conf)\n");
		return;
	}
	printk(" --- wd:%d rd:%d\n", conf->working_disks,
		conf->raid_disks);

	for (i = 0; i < conf->raid_disks; i++) {
		char b[BDEVNAME_SIZE];
		tmp = conf->mirrors + i;
		if (tmp->rdev)
			printk(" disk %d, wo:%d, o:%d, dev:%s\n",
				i, !tmp->rdev->in_sync, !tmp->rdev->faulty,
				bdevname(tmp->rdev->bdev,b));
	}
#endif
}

static void end_sync_read(struct buffer_head *bh, int uptodate)
{
	r1bio_holder_t *bioh = (r1bio_holder_t *)(bh->b_private);
	r1bio_t *r1_bio = bioh->r1bio;
	mdk_member_t *member = bioh->member;

	if (bioh->bh != bh)
		BUG();
	/*
	 * we have read a block, now it needs to be re-written,
	 * or re-read if the read failed.
	 * We don't do much here, just schedule handling by raid1d
	 */
	if (!uptodate)
		md_set_member_state(member, MDU_MSTATE_FAILED);
	else
		set_bit(R1BIO_Uptodate, &r1_bio->state);
	member_put(member);
	reschedule_retry(r1_bio);
}

static void end_verify_fix(struct buffer_head *bh, int uptodate)
{
	r1bio_holder_t *bioh = (r1bio_holder_t *)(bh->b_private);
	r1bio_t *r1_bio = bioh->r1bio;
	mdk_member_t *member = bioh->member;

	if (!uptodate)
		md_set_member_state(member, MDU_MSTATE_FAILED);
	/*printk("fixed miscompare in sector %llu in %s\n",
	       (unsigned long long)r1_bio->sector, mdname(member));*/
	if (atomic_dec_and_test(&r1_bio->remaining)) {
		/*
		 * No more fixes or reads pending.
		 */
		put_sync_r1bio(r1_bio);
	}
	member_put(member);
}

static void end_sync_verify_read(struct buffer_head *bh, int uptodate)
{
	r1bio_holder_t *bioh;
	r1bio_t *r1_bio;
	mdk_member_t *member;
	char *data;
	int bio_index;

	bioh = (r1bio_holder_t *)(bh->b_private);
	r1_bio = bioh->r1bio;
	member = bioh->member;
	if (!uptodate) {
		md_set_member_state(member, MDU_MSTATE_FAILED);
		printk(KERN_ERR "member %s failed during verify read %llu\n",
		       mdname(member), (unsigned long long)r1_bio->sector);
		put_sync_r1bio(r1_bio);
		member_put(member);
		return;
	}
	/*
	 * r1_bio->bios[0].bh->b_data is the master buffer to be verified
	 * against. This buffer is preserved until the completion all
	 * members have been verified (and potentially fixed) for this
	 * sector range.
	 */
	data = r1_bio->bios[0].bh->b_data;
	if (data != bh->b_data) {
		/*
		 * Compare with the master copy 
		 */
		if (memcmp(bh->b_data, data, bh->b_size)) {
			conf_t *conf;
			int sync_op;

			/*
			 * Miscompare encountered, do fix if it is verify
			 * with fix.
			 */
			printk("miscompare in sector %llu in %s\n",
			       (unsigned long long)r1_bio->sector,
			       mdname(member));
			conf = mddev_to_conf(r1_bio->mddev);
			sync_op = conf->syncitem.work_opcode ;
			if (sync_op == MD_SYNCOP_VERIFY_FIX) {

				bioh->rw = WRITE;
				bioh->bh->b_end_io = end_verify_fix;
				bioh->bh->b_state |= (1<<BH_Dirty);
				/*
				 * write the contents of the master buffer
				 */
				bioh->bh->b_page = r1_bio->bios[0].bh->b_page;
				bioh->bh->b_data = r1_bio->bios[0].bh->b_data;
				atomic_set(&bioh->bh->b_count, 1);
				bioh->bh->b_list = BUF_LOCKED;
				init_waitqueue_head(&bioh->bh->b_wait);
				md_sync_acct(bioh->member,
					     bioh->bh->b_size >> 9);
				md_make_request(bioh->member, WRITE, bioh->bh);
				/*
				 * Gain extra references for fix operation.
				 */
				member_get(member);
				atomic_inc(&r1_bio->remaining);	
			}
		} 
	} 

	/* Done with the member */	
	member_put(member);
		
	if (atomic_dec_and_test(&r1_bio->remaining)) { 
		/*
		 * No more fixes or reads pending.
		 */
		put_sync_r1bio(r1_bio);
	} else {
		/*
		 * Schedule next read. The next read will use the other
		 * buffer. This buffer is over written by subsequent reads.
		 */
		bio_index = r1_bio->num_bios - atomic_read(&r1_bio->remaining);
		bh = r1_bio->bios[bio_index].bh;
		bioh = (r1bio_holder_t *)(bh->b_private);
		md_sync_acct(bioh->member, bh->b_size >> 9);
		md_make_request(bioh->member, READ, bh);
	}
}

static void end_sync_write(struct buffer_head *bh, int uptodate)
{
	r1bio_holder_t *bioh = (r1bio_holder_t *)(bh->b_private);
	r1bio_t *r1_bio = bioh->r1bio;
	mdk_member_t *member = bioh->member;

	if (!uptodate)
		md_set_member_state(member, MDU_MSTATE_FAILED);

	if (atomic_dec_and_test(&r1_bio->remaining))
		put_sync_r1bio(r1_bio);
	member_put(member);
}

static int raid1_sync_write(mddev_t *mddev, r1bio_t *r1_bio)
{
	r1bio_holder_t *bioh;
	struct buffer_head *wbh;
	int writes;
	int i;

	/*
	 * The write data is always placed in
	 * the first bio.
	 */
	bioh = &r1_bio->bios[0];

	/*
	 * schedule writes
	 */
	if (!test_bit(R1BIO_Uptodate, &r1_bio->state)) {
		/*
		 * There is no point trying a read-for-reconstruct as
		 * reconstruct is about to be aborted
		 */
		printk(KERN_ALERT "raid1: %s: unrecoverable I/O read error"
			" for block %llu\n", mdname(bioh->member), 
			(unsigned long long)r1_bio->sector);
		put_sync_r1bio(r1_bio);
		return 0;
	}
	for (i = 0, writes = 0; i < r1_bio->num_bios; i++) {
		if (r1_bio->bios[i].member == NULL)
			break;
		wbh = r1_bio->bios[i].bh;
		if (wbh->b_end_io != end_sync_write)
			continue;
		md_sync_acct(r1_bio->bios[i].member, wbh->b_size >> 9);
		md_make_request(r1_bio->bios[i].member,
				r1_bio->bios[i].rw, wbh);
		writes++;
	}
	if (writes == 0) {
		printk(KERN_ALERT "raid1: %s: Sync request with no writers!\n",
		       mdname(bioh->member));
		put_sync_r1bio(r1_bio);
	}
	return 0;
}

/*
 * perform a "sync" on one "block"
 *
 * We need to make sure that no normal I/O request - particularly write
 * requests - conflict with active sync requests.
 *
 * This is achieved by tracking pending requests and a 'barrier' concept
 * that can be installed to exclude normal IO requests.
 */
static int raid1_sync_request(mddev_t *mddev, mdk_workitem_t *workitem)
{
	mdk_member_t *array;
	mdk_member_t *child;
	struct list_head *tmp;
	conf_t *conf;
	mdk_sync_support_t *syncinfo;
	md_sector_range_t *syncrange;
	r1bio_t *r1_bio;
	r1bio_holder_t *bioh;
	struct buffer_head *first_bh;
	struct buffer_head *bh;
	u_long flags;
	u_int bsize;
	int num_sectors;
	int num_bhs;
	int sync_op;
	int i;

	array = to_member(mddev);
	conf = mddev_to_conf(mddev);
	syncinfo = &conf->syncinfo;
	syncrange = &syncinfo->sync_range[syncinfo->phase];
	sync_op = workitem->work_opcode;
	
	if (syncrange->start >= array->data_size)
		return -EINVAL;

	if (syncrange->end > array->data_size)
		num_sectors = array->data_size - syncrange->start;
	else
		num_sectors = syncrange->end - syncrange->start;

	if (num_sectors > (PAGE_SIZE >> 9))
		num_sectors = PAGE_SIZE >> 9;
	/*
	 * Allocate an r1_bio for the sync operation.
	 */
	r1_bio = mempool_alloc(conf->r1buf_pool, GFP_NOIO);
	r1_bio->mddev = mddev;
	r1_bio->sector = syncrange->start;
	set_bit(R1BIO_IsSync, &r1_bio->state);
	/*
	 * Loop through our disks assigning bios to
	 * them.  If we are initializing the array, no
	 * read is necessary.  Just schedule a write to
	 * all writable members.
	 */
	num_bhs = 0;
	if (sync_op == MD_SYNCOP_INIT) 
		bioh = &r1_bio->bios[0];
	else
		bioh = &r1_bio->bios[1];
	first_bh = r1_bio->bios[0].bh;
	r1_bio->bios[0].member = NULL;
	spin_lock_irqsave(&array->member_lock, flags);
	ITERATE_CHILDREN_LOCKED(array, child, tmp) {
		if (MDK_READABLE_MEMBER(child)) {
			int bio_index;
			
			/* 
			 * In case of rebuild only first readable member is
			 * taken as master copy.
			 */
			if (workitem->work_opcode == MD_SYNCOP_REBUILD) {
				if (r1_bio->bios[0].member != NULL) 
					continue;
				bio_index = 0;
			} else
				bio_index = num_bhs;
			member_get(child);
			r1_bio->bios[bio_index].member = child;
			r1_bio->bios[bio_index].rw = READ;
			if (sync_op & MD_SYNCOP_VERIFY) {
				r1_bio->bios[bio_index].bh->b_end_io =
				    end_sync_verify_read;
			} else {
				r1_bio->bios[bio_index].bh->b_end_io =
				    end_sync_read;
			}
			num_bhs++;
			/*
			 * No need to write an optimal or
			 * degraded member...
			 */
		} else if (MDK_WRITABLE_MEMBER(child)) {
			member_get(child);
			bioh->member = child;
			bioh->rw = WRITE;
			bioh->bh->b_end_io = end_sync_write;
			bioh++;
			num_bhs++;
		}
	}
	spin_unlock_irqrestore(&array->member_lock, flags);
	if (r1_bio->bios[0].member == NULL
	 || (sync_op != MD_SYNCOP_INIT && num_bhs <= 1)){
		/*
		 * No valid member to read from.
		 */
		printk("raid1_sync_request: No Valid Sync Devices\n");
		for (i = 0; i < num_bhs; i++) {
			if (r1_bio->bios[i].member != NULL)
				member_put(r1_bio->bios[i].member);
		}
		mempool_free(r1_bio, conf->r1buf_pool);
		return -EIO;
	}
	bsize = num_sectors << 9;
	r1_bio->phase = syncable_vol_syncio_start(syncinfo);
	for (i = 0; i < num_bhs; i++) {
		bioh = &r1_bio->bios[i];
		bh = bioh->bh;

		bh->b_size = bsize;
		if (sync_op == MD_SYNCOP_INIT) {
			bh->b_page = conf->init_page;
			bh->b_data = page_address(conf->init_page);
		} else if (i != 0) {
			/*
			 * For both verify and rebuild, the buffer in
			 * bios[0] is the master buffer.  For a rebuild,
			 * all other I/O references the master.  In the
			 * verify case, all other I/Os will use the shared
			 * verify buffer with the members verified serially.
			 */ 
			if (sync_op & MD_SYNCOP_VERIFY) {
				bh->b_page = r1_bio->verify_page;
				bh->b_data = page_address(r1_bio->verify_page);
			} else {
				bh->b_page = r1_bio->bios->bh->b_page;
				bh->b_data = r1_bio->bios->bh->b_data;
			}
		}
		bh->b_list = BUF_LOCKED;
		bh->b_state = (1<<BH_Req) | (1<<BH_Mapped) | (1<<BH_Lock);
		if (bioh->rw == WRITE) {
			bh->b_state |= (1<<BH_Dirty);
			atomic_set(&bh->b_count, 1);
			bh->b_list = BUF_LOCKED;
		}
		bh->b_private = bioh;
	        bh->b_blocknr = syncrange->start + bioh->member->data_offset;
		bh->b_rsector = syncrange->start + bioh->member->data_offset;
		init_waitqueue_head(&bh->b_wait);
	}
        syncrange->start += num_sectors;
	r1_bio->sectors = num_sectors;

	/*
	 * Set a marker so our write routine knows where to stop.
	 */
	if (num_bhs < r1_bio->num_bios)
		r1_bio->bios[num_bhs].member = NULL;

	if (sync_op == MD_SYNCOP_INIT) {
		atomic_set(&r1_bio->remaining, num_bhs);
		set_bit(R1BIO_Uptodate, &r1_bio->state);
		raid1_sync_write(mddev, r1_bio);
		return 0;
	}

	/*
	 * Don't count the read in the writes
	 * remaining count.
	 */
	if (sync_op & MD_SYNCOP_VERIFY) {
		atomic_set(&r1_bio->remaining, num_bhs);
	} else
		atomic_set(&r1_bio->remaining, num_bhs - 1);

	md_sync_acct(r1_bio->bios[0].member, num_sectors);

	md_make_request(r1_bio->bios[0].member, READ, first_bh);

	return 0;
}

/*
 * This is a kernel thread which:
 *
 *	1.	Retries failed read operations on working mirrors.
 *	2.	Updates the raid superblock when problems encounter.
 *	3.	Performs writes following reads for array syncronising.
 */
static void raid1d(mddev_t *mddev, void *arg)
{
	struct list_head *head = &retry_list_head;
	r1bio_t *r1_bio;
	struct buffer_head *bh;
	unsigned long flags;
	conf_t *conf = mddev_to_conf(mddev);
	mdk_member_t *member;

	for (;;) {
		spin_lock_irqsave(&retry_list_lock, flags);
		if (list_empty(head))
			break;
		r1_bio = list_entry(head->prev, r1bio_t, retry_list);
		list_del(head->prev);
		spin_unlock_irqrestore(&retry_list_lock, flags);

		mddev = r1_bio->mddev;
		conf = mddev_to_conf(mddev);
		bh = r1_bio->master_bh;
		if (test_bit(R1BIO_IsSync, &r1_bio->state)) {
			raid1_sync_write(mddev, r1_bio);
		} else {
			if (map(mddev, &member) == -1) {
				printk(KERN_ALERT "raid1: %s: unrecoverable I/O"
				       " read error for block %llu\n",
				       bdevname(bh->b_dev),
				       (unsigned long long)r1_bio->sector);
				raid_end_bio_io(r1_bio);
			} else {

				printk(KERN_ERR "raid1: %s: redirecting sector "
				       "%llu to another mirror\n",
				       mdname(member),
				       (unsigned long long)r1_bio->sector);
				memcpy(r1_bio->bios[0].bh, bh, sizeof(*bh));
				r1_bio->bios[0].member = member;
				r1_bio->bios[0].bh->b_rsector =
				    r1_bio->sector + member->data_offset;
				r1_bio->bios[0].bh->b_blocknr =
				    r1_bio->sector + member->data_offset;
				r1_bio->bios[0].bh->b_end_io =
				    raid1_end_read_request;
				r1_bio->bios[0].bh->b_private =
				    &r1_bio->bios[0];
				md_make_request(member, READ,
						r1_bio->bios[0].bh);
			}
		}
	}
	spin_unlock_irqrestore(&retry_list_lock, flags);
}

static int raid1_stop(mddev_t *mddev, int ro)
{
	conf_t *conf;
	mdk_member_t *member;
	mdk_metapersonality_t *metapers;
	mdk_sync_support_t *syncinfo;
	u_long flags;
	int lastphase;

	conf = mddev_to_conf(mddev);
	member = to_member(mddev);
	metapers = member->metapers;
	syncinfo = &conf->syncinfo;

	/*
	 * Cancel any rebuild operations, wait for all writes to
	 * complete, and enter safe mode.  The normal I/O path is
	 * optimized to only do wake-ups when writes cease for the
	 * an old phase.  This avoids wakeups except when doing a
	 * reconstruction.  So, we must artificially toggle the
	 * phase to ensure that both phases are clear of writes.
	 */
	spin_lock_irqsave(&member->member_lock, flags);
	conf->flags |= R1_CONF_STOPPING;
	spin_unlock_irqrestore(&member->member_lock, flags);
	md_cancel_syncops(syncinfo);
	lastphase = syncinfo->phase ^ 1;
	spin_lock_irq(&syncinfo->sync_lock);
	wait_event_lock_irq(syncinfo->sync_wait,
			    (syncinfo->writes_pending[lastphase] == 0),
			    syncinfo->sync_lock);
	lastphase = syncinfo->phase;
	syncinfo->phase ^= 1;
	wait_event_lock_irq(syncinfo->sync_wait,
			    (syncinfo->writes_pending[lastphase] == 0),
			    syncinfo->sync_lock);
	spin_unlock_irq(&syncinfo->sync_lock);
	metapers->member_op(member, MDK_MEMBER_OP_ENTER_SAFEMODE, NULL);
	return 0;
}

/*
 * Initialize an mddev that will soon become
 * a RAID1 managed object.
 */
static int
raid1_raid_init(mddev_t *mddev)
{
	mdk_member_t *member;
	conf_t *conf;
	int error;
	int min_bhs;
	
	member = to_member(mddev);
	conf = kmalloc(sizeof(conf_t), GFP_KERNEL);
	if (!conf) {
		printk(KERN_ERR "raid1: couldn't allocate conf memory\n");
		return -ENOMEM;
	}
	memset(conf, 0, sizeof(*conf));
	member->pers_private = conf;
	member->pers = md_pers_get(&raid1_personality);
	conf->mddev = mddev;
	conf->last_used = NULL;
	conf->raid_disks = mddev->max_disks;

	min_bhs = NR_RAID1_BIOS * conf->raid_disks;
	conf->r1_bh_pool = mempool_create(min_bhs, r1_bh_pool_alloc,
					  r1_bh_pool_free, mddev);
	if (!conf->r1_bh_pool) {
		printk(KERN_ERR "raid1: couldn't allocate bh pool memory\n");
		goto cleanup;
	}
	conf->r1bio_pool = mempool_create(NR_RAID1_BIOS, r1bio_pool_alloc,
					  r1bio_pool_free, mddev);
	if (!conf->r1bio_pool) {
		printk(KERN_ERR "raid1: couldn't allocate bio pool memory\n");
		goto cleanup;
	}

	error = md_init_sync_support(mddev, &conf->syncinfo, NULL);
	if (error != 0) {
		printk(KERN_ERR "raid1: couldn't initialize sync support\n");
		goto cleanup;
	}
	conf->syncinfo.window_size = RESYNC_WINDOW >> 9;
	mddev->thread = md_register_thread(raid1d, mddev, NULL, "%s_raid1");
	if (!mddev->thread) {
		printk(KERN_ERR 
			"raid1: couldn't allocate thread for %s\n", 
			mdname(to_member(mddev)));
		goto cleanup;
	}

	return 0;

cleanup:
	md_teardown_sync_support(mddev, &conf->syncinfo);
	if (conf->r1bio_pool != NULL)
		mempool_destroy(conf->r1bio_pool);
	if (conf->r1_bh_pool)
		mempool_destroy(conf->r1_bh_pool);
	kfree(conf);
	member->pers_private = NULL;
	if (member->pers) {
		md_pers_put(member->pers);
		member->pers = NULL;
	}
	return -ENOMEM;
}

static int raid1_raid_release(mddev_t *mddev)
{
	conf_t *conf;

	conf = mddev_to_conf(mddev);
	if (conf == NULL) {
		BUG();
		return 0;
	}
	md_teardown_sync_support(mddev, &conf->syncinfo);
	if (mddev->thread != NULL) {
		md_unregister_thread(mddev->thread);
		mddev->thread = NULL;
	}
	if (conf->r1buf_pool)
		BUG();
	if (conf->r1_bh_pool)
		mempool_destroy(conf->r1_bh_pool);
	if (conf->r1bio_pool)
		mempool_destroy(conf->r1bio_pool);
	kfree(conf);
	mddev->member.pers_private = NULL;
	if (mddev->member.pers != NULL) {
		md_pers_put(mddev->member.pers);
		mddev->member.pers = NULL;
	}
	return 0;
}

static int raid1_member_init(mddev_t *mddev, mdk_member_t *member)
{
	mconf_t *mconf;

	mconf = kmalloc(sizeof(*mconf), GFP_KERNEL);
	if (mconf == NULL)
		return (-ENOMEM);
	
	mconf->head_position = 0;
	member->parent_pers_private = mconf;
	print_conf(mddev_to_conf(mddev));
/* XXX Maxsector propogation issues??? */
	return 0;
}

static int raid1_member_release(mddev_t *mddev, mdk_member_t *member)
{

	if (member->parent_pers_private == NULL) {
		BUG();
		return 0;
	}
	kfree(member->parent_pers_private);
	member->parent_pers_private = NULL;
	return 0;
}

/*
 * Returns the number of items added or removed
 * from the sync set.
 */
static int raid1_update_sync_set(mddev_t *mddev, int *padded, int *ptargets)
{
	mdk_member_t *array;
	mdk_member_t *member;
	struct list_head *tmp;
	int changed;
	int added;
	int targets;

	array = to_member(mddev);
	changed = 0;
	added = 0;
	targets = 0;
	ITERATE_CHILDREN_LOCKED(array, member, tmp) {
		mconf_t *mconf;
		int	old_sync;

		mconf = member_to_mconf(member);
		old_sync = mconf->sync_member;
		/*
		 * All writable members can participate
		 * in a sync operation.  We don't care
		 * here about which one will be the source
		 * for a rebuild.
		 */
		if (R1_WRITABLE_MEMBER(member)) {
			mconf->sync_member = 1;
			if (member->state & MDU_MSTATE_REBUILD_TARGET)
				targets++;
		} else
			mconf->sync_member = 0;
		if (mconf->sync_member != old_sync) {
			changed++;
			if (mconf->sync_member == 1)
				added++;
		}
	}

	if (padded != NULL)
		*padded = added;
	if (ptargets != NULL)
		*ptargets = targets;
	return (changed);
}

static int raid1_run(mddev_t *mddev)
{
	conf_t *conf;
	mdk_member_t *member;
	u_long flags;
	int exported;

	conf = mddev_to_conf(mddev);
	member = to_member(mddev);
	spin_lock_irqsave(&member->member_lock, flags);
	if ((mddev->flags & MD_MDDEV_ALLOWED_TO_RUN) == 0) {
		spin_unlock_irqrestore(&member->member_lock, flags);
		return -EPERM;
	}

	if ((mddev->flags & MD_MDDEV_RUNNING) != 0)
		goto check_recovery;

	if ((mddev->flags & MD_MDDEV_ATTEMPTING_TO_RUN) != 0) {
		spin_unlock_irqrestore(&member->member_lock, flags);
		return -EBUSY;
	}
	mddev->flags |= MD_MDDEV_ATTEMPTING_TO_RUN;

	if ((mddev->flags & MD_MDDEV_UNIT_NUMBER_ASSIGNED) == 0) {
		spin_unlock_irqrestore(&member->member_lock, flags);
		md_assign_unit_number(mddev);
		spin_lock_irqsave(&member->member_lock, flags);

		if ((mddev->flags & MD_MDDEV_UNIT_NUMBER_ASSIGNED) == 0) {
			mddev->flags &= ~MD_MDDEV_ATTEMPTING_TO_RUN;
			spin_unlock_irqrestore(&member->member_lock, flags);
			return -EPERM;
		}
	}
	spin_unlock_irqrestore(&member->member_lock, flags);
	md_update_thread_string(mddev->thread, "%s_raid1");
	md_update_sync_support_string(&conf->syncinfo, NULL);
	spin_lock_irqsave(&member->member_lock, flags);
	mddev->flags |= MD_MDDEV_RUNNING;
	mddev->flags &= ~MD_MDDEV_ATTEMPTING_TO_RUN;
check_recovery:
	exported = mddev->flags & MD_MDDEV_EXPORTED_TO_OS;
	spin_unlock_irqrestore(&member->member_lock, flags);
	md_raid1_check_recovery(mddev, R1_CHECK_STATE_CHANGE);
	if (MDU_MSTATE_READABLE(member->state))
		md_export_array(mddev);
	return (exported ? 0 : -ENXIO);
}

static int raid1_state_change(mddev_t *mddev, mdk_member_t *member)
{
	conf_t *conf;
	mdk_member_t *array;
	mdu_mstate_t new_state;
	u_long flags;

	conf = mddev_to_conf(mddev);
	array = to_member(mddev);

	spin_lock_irqsave(&member->member_lock, flags);
	new_state = member->state;
	spin_unlock_irqrestore(&member->member_lock, flags);
	if (member == array) {
		/*
		 * Handle status changes that have propogated
		 * to our array object.
		 */
		switch (new_state & ~MDU_MSTATE_MODIFIER_MASK) {
		case MDU_MSTATE_OPTIMAL:
		case MDU_MSTATE_DEGRADED:
			raid1_run(mddev);
			break;
		case MDU_MSTATE_FAILED:
			break;
		default:
			break;
		}
	} else if ((new_state & MDU_MSTATE_REBUILD_TARGET) == 0) {
		/*
		 * For member state changes, no recovery action changes
		 * are required unless the member happens to be a newly
		 * discovered rebuild target, or a member we have rebuilt
		 * to in the past.
		 */
		return 0;
	}
	md_raid1_check_recovery(mddev, R1_CHECK_STATE_CHANGE);
	return 0;
}

int raid1_announce_spare(mddev_t *mddev)
{
	md_raid1_check_recovery(mddev, R1_CHECK_SPARE_ACTIVATE);
	return 0;
}

/*
 * md_raid1_check_recovery manages all the workitems run for raid1.
 * It runs only one workitem per array at a time. This routine is 
 * called from following three contexts.
 * 	- array state changes, R1_CHECK_STATE_CHANGE
 * 	- workitem completes and has to be freed, R1_CHECK_WO_FREE
 * 	- workitem has to be started, R1_CHECK_WO_OPEN
 */
static int
md_raid1_check_recovery(mddev_t *mddev, r1_check_context_t check_context)
{
	mdk_checkpoint_arg_t checkpoint_arg;
	mdk_member_t *array;
	conf_t *conf;
	u_long flags;
	u_int sync_op;
	u_int cur_op;
	int added;
	int targets;
	int changed;
	int err;
	int sched_work;

	/* Default to no action. */
	err = 0;
	sched_work = 0;
	sync_op = 0;
	array = to_member(mddev);
	conf = mddev_to_conf(mddev);

	/*
	 * We can't schedule any workitems if we are
	 * trying to teardown the array.
	 */
	spin_lock_irqsave(&array->member_lock, flags);
	if (((conf->flags & R1_CONF_STOPPING) != 0)
	|| (mddev->flags & MD_MDDEV_RUNNING) == 0) {
		spin_unlock_irqrestore(&array->member_lock, flags);
		return 0;
	}
	changed = raid1_update_sync_set(mddev, &added, &targets);
	if (R1_INITIALIZABLE(array))
		sync_op = MD_SYNCOP_INIT;
	else if (R1_VERIFIABLE(array)) 
		sync_op = mddev->verify_op;
	cur_op = conf->syncitem.work_opcode;
	switch (check_context) {
	case R1_CHECK_WO_OPEN:
		/*
		 * Check if the queued workitem is still valid. If yes 
		 * do the allocation else return that queued workitem
		 * is no more valid.
		 */ 
		if (R1_REBUILDABLE(array) && targets != 0)
			sync_op = MD_SYNCOP_REBUILD;

		if (sync_op == cur_op) {
			int buffs;

			buffs = 2 * RESYNC_WINDOW / RESYNC_BLOCK_SIZE;
			if (conf->r1buf_pool)
				BUG();
			spin_unlock_irqrestore(&array->member_lock, flags);
			conf->r1buf_pool = mempool_create(buffs,
							  r1buf_pool_alloc,
							  r1buf_pool_free,
							  conf);
			if (!conf->r1buf_pool)
				err = -ENOMEM;
			if (sync_op == MD_SYNCOP_INIT) {
				conf->init_page = alloc_page(GFP_KERNEL);
				if (conf->init_page == NULL) {
					mempool_destroy(conf->r1buf_pool);
					err = -ENOMEM;
				}
				memset(page_address(conf->init_page), 0,
				       PAGE_SIZE);
			}
			spin_lock_irqsave(&array->member_lock, flags);
		} else {
			err = -ESTALE;
		}
		break;
	case R1_CHECK_STATE_CHANGE:
		
		if (R1_REBUILDABLE(array))
			sync_op = MD_SYNCOP_REBUILD;
		/*
		 * Check if there is a syncop that has to be run. Schedule it
		 * to run if no other syncop is running. If another syncop
		 * is running wake up the thread so that the latest recovery
		 * action will take effect.  R1_CHECK_WO_FREE will re-schedule
		 * the workitem if needed. 
		 */
		if (((conf->flags & R1_CONF_INPROG) != 0) && sync_op) {
			spin_unlock_irqrestore(&array->member_lock, flags);
			md_interrupt_thread(mddev->thread);
		} else if (sync_op) {
			sched_work = 1;
			conf->flags |= R1_CONF_INPROG;
		}
		break;
	case R1_CHECK_WO_FREE:
		conf->syncitem.work_opcode = 0;
		conf->flags &= ~R1_CONF_INPROG;
		/* FALLTHROUGH */
	case R1_CHECK_SPARE_ACTIVATE:
		/*
		 * If we have no rebuild targets, attempt to allocate
		 * a spare.  We can only do this safely from a thread
		 * context since the meta-data may have to block, syncing
		 * meta-data to array members, in order to effect the
		 * activation.  One other complication is that activating
		 * the spare will result in a member state change
		 * notification.  Our notification handler will call back
		 * into this routine, notice that the sync set has changed,
		 * and interrupt any currently active workitem so that
		 * this change is noted.  This means that after a successful
		 * spare activation, there will be a signal pending for the
		 * current workitem.  So, we only allocate spares from our
		 * Free context after we have released the current workitem
		 * or from the context of a spare offering.  The actual
		 * rebuild will occur the next time the workitem is executed.
		 */
		if (R1_REBUILDABLE(array) && targets == 0) {
			mdk_metapersonality_t *metapers;
			
			spin_unlock_irqrestore(&array->member_lock, flags);
			metapers = array->metapers;
			metapers->member_op(array,
					    MDK_MEMBER_OP_ACTIVATE_SPARE,
					    NULL);
			/* Reset the checkpoint to 0. */
			checkpoint_arg.checkpoint = 0;
			checkpoint_arg.checkpoint_type = MD_SYNCOP_REBUILD;
			metapers->member_op(array, 
					    MDK_MEMBER_OP_SET_CHECKPOINT,
					    &checkpoint_arg);
			spin_lock_irqsave(&array->member_lock, flags);
			break;
		} else if (check_context == R1_CHECK_WO_FREE && sync_op) {

			sched_work = 1;	
			conf->flags |= R1_CONF_INPROG;
		}
		break;
	}
	spin_unlock_irqrestore(&array->member_lock, flags);
	if (sched_work) {
		md_init_workitem(&conf->syncitem, sync_op, raid1_workop,
				 md_sync_handler, &conf->syncinfo);
		err = md_queue_workitem(&conf->syncinfo.work_state,
			  		  &conf->syncitem, /*head_of_queue*/1);
		if (err != 0) {
			spin_lock_irqsave(&array->member_lock, flags);
			conf->flags &= ~R1_CONF_INPROG;
			spin_unlock_irqrestore(&array->member_lock, flags);
		}
	} 
	return err;
}

static int raid1_cancel_verify(mddev_t *mddev)
{
	mdk_checkpoint_arg_t checkpoint_arg;
	conf_t *conf;
	mdk_sync_support_t *syncinfo;
	mdk_member_t *member;
	mdk_set_state_arg_t sstate;
	int lastphase;
	int err;
	
	conf = mddev_to_conf(mddev);
	member = to_member(mddev);
	if (conf->syncitem.work_opcode & MD_SYNCOP_VERIFY) {
		/* Cancel verify syncop */
		syncinfo = &conf->syncinfo;
		md_cancel_syncops(syncinfo);
		lastphase = syncinfo->phase ^ 1;
		spin_lock_irq(&syncinfo->sync_lock);
		wait_event_lock_irq(syncinfo->sync_wait,
		 		    (syncinfo->writes_pending[lastphase] == 0),
				    syncinfo->sync_lock);
		lastphase = syncinfo->phase;
		syncinfo->phase ^= 1;
		wait_event_lock_irq(syncinfo->sync_wait,
				    (syncinfo->writes_pending[lastphase] == 0),
				    syncinfo->sync_lock);
		spin_unlock_irq(&syncinfo->sync_lock);
	}
	/*
	 * Unmark dirty flag
	 */
	sstate.oldstate = member->state;
	sstate.newstate = member->state & ~MDU_MSTATE_DIRTY;
	err = member->metapers->member_op(member,
					  MDK_MEMBER_OP_SET_STATE,
					  &sstate);
	/*
	 * Reset the checkpoint to 0.
	 */
	checkpoint_arg.checkpoint = 0;
	checkpoint_arg.checkpoint_type = MD_SYNCOP_NONE;
	member->metapers->member_op(member, MDK_MEMBER_OP_SET_CHECKPOINT,
				    &checkpoint_arg);
	
	return err;
}

static int raid1_verify(mddev_t *mddev, int fix)
{
	mdk_set_state_arg_t sstate;
	mdk_member_t *array;
	conf_t *conf;
	u_long flags;
	int err;

	array = to_member(mddev);
	conf = mddev_to_conf(mddev);
	spin_lock_irqsave(&array->member_lock, flags);
	sstate.oldstate = array->state;
	if (fix)
		mddev->verify_op = MD_SYNCOP_VERIFY_FIX;
	else
		mddev->verify_op = MD_SYNCOP_VERIFY;
	spin_unlock_irqrestore(&array->member_lock, flags);
	sstate.newstate = array->state|MDU_MSTATE_DIRTY;
	err = array->metapers->member_op(array, MDK_MEMBER_OP_SET_STATE,
					 &sstate);
	return err;
}

static int raid1_close_workitem(mddev_t *mddev, mdk_workitem_t *workitem)
{
	conf_t *conf;

	conf = mddev_to_conf(mddev);
	if (conf->r1buf_pool != NULL) {
		mempool_destroy(conf->r1buf_pool);
		conf->r1buf_pool = NULL;
	}
	if (conf->init_page != NULL) {
		__free_page(conf->init_page);
		conf->init_page = NULL;
	}
	return 0;
}

static int raid1_workop(mdk_member_t *member, int op, void *arg)
{
	switch (op) {
	case MDK_WORKITEM_OP_OPEN:
		return md_raid1_check_recovery(to_mddev(member),
			       		       R1_CHECK_WO_OPEN);
		break;
	case MDK_WORKITEM_OP_CLOSE:
		return raid1_close_workitem(to_mddev(member),
					    (mdk_workitem_t *)arg);
		break;
	case MDK_WORKITEM_OP_FREE:
		return md_raid1_check_recovery(to_mddev(member), 
					       R1_CHECK_WO_FREE);
	default:
		return (-EINVAL);
	}
}

static int raid1_raidop(mdk_member_t *member, int op, void *arg)
{

	switch (op) {
	case MDK_RAID_OP_INIT:
	case MDK_RAID_OP_RELEASE:
		if (member->type != MDK_MTYPE_MDDEV)
			return (0);
		if (op == MDK_RAID_OP_INIT)
			return raid1_raid_init(to_mddev(member));
		else
			return raid1_raid_release(to_mddev(member));
	case MDK_RAID_OP_INIT_MEMBER:
	case MDK_RAID_OP_RELEASE_MEMBER:
	{
		mdk_member_t *child;

		child = (mdk_member_t *)arg;
		if (op == MDK_RAID_OP_INIT_MEMBER)
			return raid1_member_init(to_mddev(member), child);
		else
			return raid1_member_release(to_mddev(member), child);
	}
	case MDK_RAID_OP_MSTATE_CHANGED:
		return raid1_state_change(to_mddev(member),
					  (mdk_member_t *)arg);
	case MDK_RAID_OP_SPARE_ADDED:
		return raid1_announce_spare(to_mddev(member));
	case MDK_RAID_OP_SYNC_REQUEST:
		return raid1_sync_request(to_mddev(member),
					  (mdk_workitem_t *)arg);
	case MDK_RAID_OP_RUN: 
		return (raid1_run(to_mddev(member)));
	case MDK_RAID_OP_STOP:
		return raid1_stop(to_mddev(member), (int)arg);
	case MDK_RAID_OP_ENTER_SAFEMODE:
	{
		mdk_metapersonality_t *metapers;

		metapers = member->metapers;
		return (metapers->member_op(member,
					    MDK_MEMBER_OP_ENTER_SAFEMODE,
					    NULL));
	}
	case MDK_RAID_OP_EXIT_SAFEMODE:
	{
		mdk_metapersonality_t *metapers;

		metapers = member->metapers;
		return (metapers->member_op(member,
					    MDK_MEMBER_OP_EXIT_SAFEMODE,
					    NULL));
	}
	case MDK_RAID_OP_ADD_MEMBER:
	{
		mdu_mstate_t  parent_state;
		u_long flags;

		spin_lock_irqsave(&member->member_lock, flags);
		parent_state = member->state;
		spin_unlock_irqrestore(&member->member_lock, flags);
		md_raid1_check_recovery(to_mddev(member),
					R1_CHECK_STATE_CHANGE);
		return 0;
	}
	case MDK_RAID_OP_SYNC_STATS:
	{
		conf_t *conf;
		
		if(!member->pers_private)
			return -ENODEV;
		conf = member->pers_private;
		return md_get_sync_op_stat(&conf->syncinfo, arg);
	}
	case MDK_RAID_OP_REQUEST_VERIFY:
		return raid1_verify(to_mddev(member), (int)arg);
	case MDK_RAID_OP_CANCEL_VERIFY:
		return raid1_cancel_verify(to_mddev(member));
	case MDK_RAID_OP_REQUEST_RESYNC:
	case MDK_RAID_OP_PAUSE_MEMBER:
	case MDK_RAID_OP_REMOVE_MEMBER:
		return -EOPNOTSUPP;
	default:
		return (-EINVAL);
	}

	return 0;
}

static mdk_personality_t raid1_personality =
{
	.name		= "raid1",
	.owner		= THIS_MODULE,
	.make_request	= make_request,
	.raid_op	= raid1_raidop,
};

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
static int __init raid1_init(void)
#else
int __init raid1_init(void)
#endif
{
	return register_md_personality(RAID1, &raid1_personality);
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
static void raid1_exit(void)
#else
void raid1_exit(void)
#endif
{
	unregister_md_personality(RAID1);
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
module_init(raid1_init);
module_exit(raid1_exit);
#endif
MODULE_LICENSE("GPL");
